Package: nc
Authors@R: c(
    person("Toby", "Hocking",
     email="toby.hocking@r-project.org",
     role=c("aut", "cre")))
Version: 2026.4.20
License: GPL-3
Title: Named Capture to Data Tables
Description: User-friendly functions for extracting a data
 table (row for each match, column for each group)
 from non-tabular text data using regular expressions,
 and for melting columns that match a regular expression.
 Patterns are defined using a readable syntax
 that makes it easy to build complex patterns
 in terms of simpler, re-usable sub-patterns.
 Named R arguments are translated to column names
 in the output; capture groups without names are used
 internally in order to provide a standard interface
 to three regular expression 'C' libraries
 ('PCRE', 'RE2', 'ICU').
 Output can also include numeric columns via
 user-specified type conversion functions.
Depends: R (>= 2.14)
Imports: data.table (>= 1.15.0)
Suggests: testthat, re2, stringi, ggplot2, tidyr (>= 1.0.0), cdata,
        reshape2, knitr, markdown, rmarkdown, R.utils, covr, arrow
VignetteBuilder: knitr
URL: https://github.com/tdhock/nc
BugReports: https://github.com/tdhock/nc/issues
NeedsCompilation: no
Packaged: 2026-04-20 19:35:21 UTC; hoct2726
Author: Toby Hocking [aut, cre]
Maintainer: Toby Hocking <toby.hocking@r-project.org>
Repository: CRAN
Date/Publication: 2026-04-21 05:10:12 UTC
Built: R 4.6.0; ; 2026-04-25 20:43:37 UTC; unix

export(alevels)
export(before_match)
export(altlist)
export(alternatives_with_shared_groups)
export(measure)
export(quantifier)
export(alternatives)
export(field)
export(group)
export(var_args_list)
export(capture_all_str)
export(capture_melt_multiple)
export(capture_melt_single)
export(capture_first_vec)
export(capture_first_glob)
export(capture_first_df)
export(capture_longer_spec)
import(data.table)
importFrom("stats","na.omit")
importFrom("utils","type.convert")

nc
===

![Build status](https://github.com/xushaohua/nc/actions/workflows/rust.yml/badge.svg)
[![Latest version](https://img.shields.io/crates/v/nc.svg)](https://crates.io/crates/nc)
[![Documentation](https://docs.rs/nc/badge.svg)](https://docs.rs/nc)
![Minimum rustc version](https://img.shields.io/badge/rustc-1.56+-yellow.svg)
![License](https://img.shields.io/crates/l/nc.svg)

Access system calls directly without `std` or `libc`.

- [Documentation](https://docs.rs/nc)
- [Release notes](https://github.com/xushaohua/nc/releases)

Features:
- No glibc or musl required
- Access syscalls directly, via assembly
- No global errno variable, every function returns an errno instead
- Support latest kernel APIs, like io-uring and pidfd, introduced in linux 5.0+

## Usage
Add this to `Cargo.toml`:
```toml
[dependencies]
nc = "0.8"
```

## Examples
Get file stat:
```rust
let mut statbuf = nc::stat_t::default();
match unsafe { nc::stat("/etc/passwd", &mut statbuf) } {
    Ok(_) => println!("s: {:?}", statbuf),
    Err(errno) => eprintln!("Failed to get file status, got errno: {}", errno),
}
```

Get human-readable error string:
```rust
let errno = nc::EPERM;
println!("err: {:?}", nc::strerror(errno);
```

Fork process:
```rust
let pid = unsafe { nc::fork() };
match pid {
    Ok(pid) => {
        if pid == 0 {
            println!("child process: {}", pid);
            let args = [""];
            let env = [""];
            match unsafe { nc::execve("/bin/ls", &args, &env) } {
                Ok(_) => {},
                Err(errno) => eprintln!("`ls` got err: {}", errno),
            }
        } else if pid < 0 {
            eprintln!("fork() error!");
        } else {
            println!("parent process!");
        }
    },
    Err(errno) => eprintln!("errno: {}", errno),
}
```

Kill self:
```rust
let pid = unsafe { nc::getpid() };
let ret = unsafe { nc::kill(pid, nc::SIGTERM) };
// Never reach here.
println!("ret: {:?}", ret);
```

Or handle signals:
```rust
fn handle_alarm(signum: i32) {
    assert_eq!(signum, nc::SIGALRM);
}

fn main() {
    let ret = unsafe { nc::signal(nc::SIGALRM, handle_alarm as nc::sighandler_t) };
    assert!(ret.is_ok());
    let remaining = unsafe {nc::alarm(1) };
    let ret = unsafe { nc::pause() };
    assert!(ret.is_err());
    assert_eq!(ret, Err(nc::EINTR));
    assert_eq!(remaining, 0);
}
```

## Stable version
For stable version of rustc, please install a C compiler (`gcc` or `clang`) first.
As `llvm_asm!` feature is unavailable in stable version.


## Supported Operating Systems and Architectures
- linux
  - aarch64
  - arm
  - loongarch64
  - mips
  - mips64
  - mips64el
  - mipsel
  - powerpc64
  - riscv64
  - s390x
  - x86
  - x86-64
- android
  - aarch64
- freebsd
  - x86-64
- netbsd
  - x86-64
- mac os
  - x86-64

## Related projects
- [nix][nix]
- [syscall][syscall]
- [relibc][relibc]
- [Linux Syscall Support][lss]
- [syscall pkg in golang][go-syscall]

[syscall]: https://github.com/kmcallister/syscall.rs
[relibc]: https://gitlab.redox-os.org/redox-os/relibc.git
[nix]: https://github.com/nix-rust/nix
[lss]: https://chromium.googlesource.com/linux-syscall-support
[go-syscall]: https://github.com/golang/go/tree/master/src/syscall

## License
This library is release in [Apache License](LICENSE).

Netcat 1.10
===========							   /\_/\
								  / 0 0 \
Netcat is a simple Unix utility which reads and writes data	 ====v====
across network connections, using TCP or UDP protocol.		  \  W  /
It is designed to be a reliable "back-end" tool that can	  |     |     _
be used directly or easily driven by other programs and		  / ___ \    /
scripts.  At the same time, it is a feature-rich network	 / /   \ \  |
debugging and exploration tool, since it can create almost	(((-----)))-'
any kind of connection you would need and has several		 /
interesting built-in capabilities.  Netcat, or "nc" as the	(      ___
actual program is named, should have been supplied long ago	 \__.=|___E
as another one of those cryptic but standard Unix tools.	        /

In the simplest usage, "nc host port" creates a TCP connection to the given
port on the given target host.  Your standard input is then sent to the host,
and anything that comes back across the connection is sent to your standard
output.  This continues indefinitely, until the network side of the connection
shuts down.  Note that this behavior is different from most other applications
which shut everything down and exit after an end-of-file on the standard input.

Netcat can also function as a server, by listening for inbound connections
on arbitrary ports and then doing the same reading and writing.  With minor
limitations, netcat doesn't really care if it runs in "client" or "server"
mode -- it still shovels data back and forth until there isn't any more left.
In either mode, shutdown can be forced after a configurable time of inactivity
on the network side.

And it can do this via UDP too, so netcat is possibly the "udp telnet-like"
application you always wanted for testing your UDP-mode servers.  UDP, as the
"U" implies, gives less reliable data transmission than TCP connections and
some systems may have trouble sending large amounts of data that way, but it's
still a useful capability to have.

You may be asking "why not just use telnet to connect to arbitrary ports?"
Valid question, and here are some reasons.  Telnet has the "standard input
EOF" problem, so one must introduce calculated delays in driving scripts to
allow network output to finish.  This is the main reason netcat stays running
until the *network* side closes.  Telnet also will not transfer arbitrary
binary data, because certain characters are interpreted as telnet options and
are thus removed from the data stream.  Telnet also emits some of its
diagnostic messages to standard output, where netcat keeps such things
religiously separated from its *output* and will never modify any of the real
data in transit unless you *really* want it to.  And of course telnet is
incapable of listening for inbound connections, or using UDP instead.  Netcat
doesn't have any of these limitations, is much smaller and faster than telnet,
and has many other advantages.

Some of netcat's major features are:

	Outbound or inbound connections, TCP or UDP, to or from any ports
	Full DNS forward/reverse checking, with appropriate warnings
	Ability to use any local source port
	Ability to use any locally-configured network source address
	Built-in port-scanning capabilities, with randomizer
	Built-in loose source-routing capability
	Can read command line arguments from standard input
	Slow-send mode, one line every N seconds
	Hex dump of transmitted and received data
	Optional ability to let another program service established connections
	Optional telnet-options responder

Efforts have been made to have netcat "do the right thing" in all its various
modes.  If you believe that it is doing the wrong thing under whatever
circumstances, please notify me and tell me how you think it should behave.
If netcat is not able to do some task you think up, minor tweaks to the code
will probably fix that.  It provides a basic and easily-modified template for
writing other network applications, and I certainly encourage people to make
custom mods and send in any improvements they make to it.  This is the second
release; the overall differences from 1.00 are relatively minor and have mostly
to do with portability and bugfixes.  Many people provided greatly appreciated
fixes and comments on the 1.00 release.  Continued feedback from the Internet
community is always welcome!

Netcat is entirely my own creation, although plenty of other code was used as
examples.  It is freely given away to the Internet community in the hope that
it will be useful, with no restrictions except giving credit where it is due.
No GPLs, Berkeley copyrights or any of that nonsense.  The author assumes NO
responsibility for how anyone uses it.  If netcat makes you rich somehow and
you're feeling generous, mail me a check.  If you are affiliated in any way
with Microsoft Network, get a life.  Always ski in control.  Comments,
questions, and patches to hobbit@avian.org.

Building
========

Compiling is fairly straightforward.  Examine the Makefile for a SYSTYPE that
matches yours, and do "make <systype>".  The executable "nc" should appear.
If there is no relevant SYSTYPE section, try "generic".  If you create new
sections for generic.h and Makefile to support another platform, please follow
the given format and mail back the diffs.

There are a couple of other settable #defines in netcat.c, which you can
include as DFLAGS="-DTHIS -DTHAT" to your "make" invocation without having to
edit the Makefile.  See the following discussions for what they are and do.

If you want to link against the resolver library on SunOS [recommended] and
you have BIND 4.9.x, you may need to change XLIBS=-lresolv in the Makefile to
XLIBS="-lresolv -l44bsd".

Linux sys/time.h does not really support presetting of FD_SETSIZE; a harmless
warning is issued.

Some systems may warn about pointer types for signal().  No problem, though.

Exploration of features
=======================

Where to begin?  Netcat is at the same time so simple and versatile, it's like
trying to describe everything you can do with your Swiss Army knife.  This will
go over the basics; you should also read the usage examples and notes later on
which may give you even more ideas about what this sort of tool is good for.

If no command arguments are given at all, netcat asks for them, reads a line
from standard input, and breaks it up into arguments internally.  This can be
useful when driving netcat from certain types of scripts, with the side effect
of hiding your command line arguments from "ps" displays.

The host argument can be a name or IP address.  If -n is specified, netcat
will only accept numeric IP addresses and do no DNS lookups for anything.  If
-n is not given and -v is turned on, netcat will do a full forward and reverse
name and address lookup for the host, and warn you about the all-too-common
problem of mismatched names in the DNS.  This often takes a little longer for
connection setup, but is useful to know about.  There are circumstances under
which this can *save* time, such as when you want to know the name for some IP
address and also connect there.  Netcat will just tell you all about it, saving
the manual steps of looking up the hostname yourself.  Normally mismatch-
checking is case-insensitive per the DNS spec, but you can define ANAL at
compile time to make it case-sensitive -- sometimes useful for uncovering minor
errors in your own DNS files while poking around your networks.

A port argument is required for outbound connections, and can be numeric or a
name as listed in /etc/services.  If -n is specified, only numeric arguments
are valid.  Special syntax and/or more than one port argument cause different
behavior -- see details below about port-scanning.

The -v switch controls the verbosity level of messages sent to standard error.
You will probably want to run netcat most of the time with -v turned on, so you
can see info about the connections it is trying to make.  You will probably
also want to give a smallish -w argument, which limits the time spent trying to
make a connection.  I usually alias "nc" to "nc -v -w 3", which makes it
function just about the same for things I would otherwise use telnet to do.
The timeout is easily changed by a subsequent -w argument which overrides the
earlier one.  Specifying -v more than once makes diagnostic output MORE
verbose.  If -v is not specified at all, netcat silently does its work unless
some error happens, whereupon it describes the error and exits with a nonzero
status.  Refused network connections are generally NOT considered to be errors,
unless you only asked for a single TCP port and it was refused.

Note that -w also sets the network inactivity timeout.  This does not have any
effect until standard input closes, but then if nothing further arrives from
the network in the next <timeout> seconds, netcat tries to read the net once
more for good measure, and then closes and exits.  There are a lot of network
services now that accept a small amount of input and return a large amount of
output, such as Gopher and Web servers, which is the main reason netcat was
written to "block" on the network staying open rather than standard input.
Handling the timeout this way gives uniform behavior with network servers that
*don't* close by themselves until told to.

UDP connections are opened instead of TCP when -u is specified.  These aren't
really "connections" per se since UDP is a connectionless protocol, although
netcat does internally use the "connected UDP socket" mechanism that most
kernels support.  Although netcat claims that an outgoing UDP connection is
"open" immediately, no data is sent until something is read from standard
input.  Only thereafter is it possible to determine whether there really is a
UDP server on the other end, and often you just can't tell.  Most UDP protocols
use timeouts and retries to do their thing and in many cases won't bother
answering at all, so you should specify a timeout and hope for the best.  You
will get more out of UDP connections if standard input is fed from a source
of data that looks like various kinds of server requests.

To obtain a hex dump file of the data sent either way, use "-o logfile".  The
dump lines begin with "<" or ">" to respectively indicate "from the net" or
"to the net", and contain the total count per direction, and hex and ascii
representations of the traffic.  Capturing a hex dump naturally slows netcat
down a bit, so don't use it where speed is critical.

Netcat can bind to any local port, subject to privilege restrictions and ports
that are already in use.  It is also possible to use a specific local network
source address if it is that of a network interface on your machine.  [Note:
this does not work correctly on all platforms.]  Use "-p portarg" to grab a
specific local port, and "-s ip-addr" or "-s name" to have that be your source
IP address.  This is often referred to as "anchoring the socket".  Root users
can grab any unused source port including the "reserved" ones less than 1024.
Absence of -p will bind to whatever unused port the system gives you, just like
any other normal client connection, unless you use -r [see below].

Listen mode will cause netcat to wait for an inbound connection, and then the
same data transfer happens.  Thus, you can do "nc -l -p 1234 < filename" and
when someone else connects to your port 1234, the file is sent to them whether
they wanted it or not.  Listen mode is generally used along with a local port
argument -- this is required for UDP mode, while TCP mode can have the system
assign one and tell you what it is if -v is turned on.  If you specify a target
host and optional port in listen mode, netcat will accept an inbound connection
only from that host and if you specify one, only from that foreign source port.
In verbose mode you'll be informed about the inbound connection, including what
address and port it came from, and since listening on "any" applies to several
possibilities, which address it came *to* on your end.  If the system supports
IP socket options, netcat will attempt to retrieve any such options from an
inbound connection and print them out in hex.

If netcat is compiled with -DGAPING_SECURITY_HOLE, the -e argument specifies
a program to exec after making or receiving a successful connection.  In the
listening mode, this works similarly to "inetd" but only for a single instance.
Use with GREAT CARE.  This piece of the code is normally not enabled; if you
know what you're doing, have fun.  This hack also works in UDP mode.  Note that
you can only supply -e with the name of the program, but no arguments.  If you
want to launch something with an argument list, write a two-line wrapper script
or just use inetd like always.

If netcat is compiled with -DTELNET, the -t argument enables it to respond
to telnet option negotiation [always in the negative, i.e. DONT or WONT].
This allows it to connect to a telnetd and get past the initial negotiation
far enough to get a login prompt from the server.  Since this feature has
the potential to modify the data stream, it is not enabled by default.  You
have to understand why you might need this and turn on the #define yourself.

Data from the network connection is always delivered to standard output as
efficiently as possible, using large 8K reads and writes.  Standard input is
normally sent to the net the same way, but the -i switch specifies an "interval
time" which slows this down considerably.  Standard input is still read in
large batches, but netcat then tries to find where line breaks exist and sends
one line every interval time.  Note that if standard input is a terminal, data
is already read line by line, so unless you make the -i interval rather long,
what you type will go out at a fairly normal rate.  -i is really designed
for use when you want to "measure out" what is read from files or pipes.

Port-scanning is a popular method for exploring what's out there.  Netcat
accepts its commands with options first, then the target host, and everything
thereafter is interpreted as port names or numbers, or ranges of ports in M-N
syntax.  CAVEAT: some port names in /etc/services contain hyphens -- netcat
currently will not correctly parse those, so specify ranges using numbers if
you can.  If more than one port is thus specified, netcat connects to *all* of
them, sending the same batch of data from standard input [up to 8K worth] to
each one that is successfully connected to.  Specifying multiple ports also
suppresses diagnostic messages about refused connections, unless -v is
specified twice for "more verbosity".  This way you normally get notified only
about genuinely open connections.  Example: "nc -v -w 2 -z target 20-30" will
try connecting to every port between 20 and 30 [inclusive] at the target, and
will likely inform you about an FTP server, telnet server, and mailer along the
way.  The -z switch prevents sending any data to a TCP connection and very
limited probe data to a UDP connection, and is thus useful as a fast scanning
mode just to see what ports the target is listening on.  To limit scanning
speed if desired, -i will insert a delay between each port probe.  There are
some pitfalls with regard to UDP scanning, described later, but in general it
works well.

For each range of ports specified, scanning is normally done downward within
that range.  If the -r switch is used, scanning hops randomly around within
that range and reports open ports as it finds them.  [If you want them listed
in order regardless, pipe standard error through "sort"...]  In addition, if
random mode is in effect, the local source ports are also randomized.  This
prevents netcat from exhibiting any kind of regular pattern in its scanning.
You can exert fairly fine control over your scan by judicious use of -r and
selected port ranges to cover.  If you use -r for a single connection, the
source port will have a random value above 8192, rather than the next one the
kernel would have assigned you.  Note that selecting a specific local port
with -p overrides any local-port randomization.

Many people are interested in testing network connectivity using IP source
routing, even if it's only to make sure their own firewalls are blocking
source-routed packets.  On systems that support it, the -g switch can be used
multiple times [up to 8] to construct a loose-source-routed path for your
connection, and the -G argument positions the "hop pointer" within the list.
If your network allows source-routed traffic in and out, you can test
connectivity to your own services via remote points in the internet.  Note that
although newer BSD-flavor telnets also have source-routing capability, it isn't
clearly documented and the command syntax is somewhat clumsy.  Netcat's
handling of "-g" is modeled after "traceroute".

Netcat tries its best to behave just like "cat".  It currently does nothing to
terminal input modes, and does no end-of-line conversion.  Standard input from
a terminal is read line by line with normal editing characters in effect.  You
can freely suspend out of an interactive connection and resume.  ^C or whatever
your interrupt character is will make netcat close the network connection and
exit.  A switch to place the terminal in raw mode has been considered, but so
far has not been necessary.  You can send raw binary data by reading it out of
a file or piping from another program, so more meaningful effort would be spent
writing an appropriate front-end driver.

Netcat is not an "arbitrary packet generator", but the ability to talk to raw
sockets and/or nit/bpf/dlpi may appear at some point.  Such things are clearly
useful; I refer you to Darren Reed's excellent ip_filter package, which now
includes a tool to construct and send raw packets with any contents you want.

Example uses -- the light side
==============================

Again, this is a very partial list of possibilities, but it may get you to
think up more applications for netcat.  Driving netcat with simple shell or
expect scripts is an easy and flexible way to do fairly complex tasks,
especially if you're not into coding network tools in C.  My coding isn't
particularly strong either [although undoubtedly better after writing this
thing!], so I tend to construct bare-metal tools like this that I can trivially
plug into other applications.  Netcat doubles as a teaching tool -- one can
learn a great deal about more complex network protocols by trying to simulate
them through raw connections!

An example of netcat as a backend for something else is the shell-script
Web browser, which simply asks for the relevant parts of a URL and pipes
"GET /what/ever" into a netcat connection to the server.  I used to do this
with telnet, and had to use calculated sleep times and other stupidity to
kludge around telnet's limitations.  Netcat guarantees that I get the whole
page, and since it transfers all the data unmodified, I can even pull down
binary image files and display them elsewhere later.  Some folks may find the
idea of a shell-script web browser silly and strange, but it starts up and
gets me my info a hell of a lot faster than a GUI browser and doesn't hide
any contents of links and forms and such.  This is included, as scripts/web,
along with several other web-related examples.

Netcat is an obvious replacement for telnet as a tool for talking to daemons.
For example, it is easier to type "nc host 25", talk to someone's mailer, and
just ^C out than having to type ^]c or QUIT as telnet would require you to do.
You can quickly catalog the services on your network by telling netcat to
connect to well-known services and collect greetings, or at least scan for open
ports.  You'll probably want to collect netcat's diagnostic messages in your
output files, so be sure to include standard error in the output using
`>& file' in *csh or `> file 2>&1' in bourne shell.

A scanning example: "echo QUIT | nc -v -w 5 target 20-250 500-600 5990-7000"
will inform you about a target's various well-known TCP servers, including
r-services, X, IRC, and maybe a few you didn't expect.  Sending in QUIT and
using the timeout will almost guarantee that you see some kind of greeting or
error from each service, which usually indicates what it is and what version.
[Beware of the "chargen" port, though...]  SATAN uses exactly this technique to
collect host information, and indeed some of the ideas herein were taken from
the SATAN backend tools.  If you script this up to try every host in your
subnet space and just let it run, you will not only see all the services,
you'll find out about hosts that aren't correctly listed in your DNS.  Then you
can compare new snapshots against old snapshots to see changes.  For going
after particular services, a more intrusive example is in scripts/probe.

Netcat can be used as a simple data transfer agent, and it doesn't really
matter which end is the listener and which end is the client -- input at one
side arrives at the other side as output.  It is helpful to start the listener
at the receiving side with no timeout specified, and then give the sending side
a small timeout.  That way the listener stays listening until you contact it,
and after data stops flowing the client will time out, shut down, and take the
listener with it.  Unless the intervening network is fraught with problems,
this should be completely reliable, and you can always increase the timeout.  A
typical example of something "rsh" is often used for: on one side,

	nc -l -p 1234 | uncompress -c | tar xvfp -

and then on the other side

	tar cfp - /some/dir | compress -c | nc -w 3 othermachine 1234

will transfer the contents of a directory from one machine to another, without
having to worry about .rhosts files, user accounts, or inetd configurations
at either end.  Again, it matters not which is the listener or receiver; the
"tarring" machine could just as easily be running the listener instead.  One
could conceivably use a scheme like this for backups, by having cron-jobs fire
up listeners and backup handlers [which can be restricted to specific addresses
and ports between each other] and pipe "dump" or "tar" on one machine to "dd
of=/dev/tapedrive" on another as usual.  Since netcat returns a nonzero exit
status for a denied listener connection, scripts to handle such tasks could
easily log and reject connect attempts from third parties, and then retry.

Another simple data-transfer example: shipping things to a PC that doesn't have
any network applications yet except a TCP stack and a web browser.  Point the
browser at an arbitrary port on a Unix server by telling it to download
something like http://unixbox:4444/foo, and have a listener on the Unix side
ready to ship out a file when the connect comes in.  The browser may pervert
binary data when told to save the URL, but you can dig the raw data out of
the on-disk cache.

If you build netcat with GAPING_SECURITY_HOLE defined, you can use it as an
"inetd" substitute to test experimental network servers that would otherwise
run under "inetd".  A script or program will have its input and output hooked
to the network the same way, perhaps sans some fancier signal handling.  Given
that most network services do not bind to a particular local address, whether
they are under "inetd" or not, it is possible for netcat avoid the "address
already in use" error by binding to a specific address.  This lets you [as
root, for low ports] place netcat "in the way" of a standard service, since
inbound connections are generally sent to such specifically-bound listeners
first and fall back to the ones bound to "any".  This allows for a one-off
experimental simulation of some service, without having to screw around with
inetd.conf.  Running with -v turned on and collecting a connection log from
standard error is recommended.

Netcat as well can make an outbound connection and then run a program or script
on the originating end, with input and output connected to the same network
port.  This "inverse inetd" capability could enhance the backup-server concept
described above or help facilitate things such as a "network dialback" concept.
The possibilities are many and varied here; if such things are intended as
security mechanisms, it may be best to modify netcat specifically for the
purpose instead of wrapping such functions in scripts.

Speaking of inetd, netcat will function perfectly well *under* inetd as a TCP
connection redirector for inbound services, like a "plug-gw" without the
authentication step.  This is very useful for doing stuff like redirecting
traffic through your firewall out to other places like web servers and mail
hubs, while posing no risk to the firewall machine itself.  Put netcat behind
inetd and tcp_wrappers, perhaps thusly:

	www stream tcp nowait nobody /etc/tcpd /bin/nc -w 3 realwww 80

and you have a simple and effective "application relay" with access control
and logging.  Note use of the wait time as a "safety" in case realwww isn't
reachable or the calling user aborts the connection -- otherwise the relay may
hang there forever.

You can use netcat to generate huge amounts of useless network data for
various performance testing.  For example, doing

	yes AAAAAAAAAAAAAAAAAAAAAA | nc -v -v -l -p 2222 > /dev/null

on one side and then hitting it with

	yes BBBBBBBBBBBBBBBBBBBBBB | nc othermachine 2222 > /dev/null

from another host will saturate your wires with A's and B's.  The "very
verbose" switch usage will tell you how many of each were sent and received
after you interrupt either side.  Using UDP mode produces tremendously MORE
trash per unit time in the form of fragmented 8 Kbyte mobygrams -- enough to
stress-test kernels and network interfaces.  Firing random binary data into
various network servers may help expose bugs in their input handling, which
nowadays is a popular thing to explore.  A simple example data-generator is
given in data/data.c included in this package, along with a small collection
of canned input files to generate various packet contents.  This program is
documented in its beginning comments, but of interest here is using "%r" to
generate random bytes at well-chosen points in a data stream.  If you can
crash your daemon, you likely have a security problem.

The hex dump feature may be useful for debugging odd network protocols,
especially if you don't have any network monitoring equipment handy or aren't
root where you'd need to run "tcpdump" or something.  Bind a listening netcat
to a local port, and have it run a script which in turn runs another netcat
to the real service and captures the hex dump to a log file.  This sets up a
transparent relay between your local port and wherever the real service is.
Be sure that the script-run netcat does *not* use -v, or the extra info it
sends to standard error may confuse the protocol.  Note also that you cannot
have the "listen/exec" netcat do the data capture, since once the connection
arrives it is no longer netcat that is running.

Binding to an arbitrary local port allows you to simulate things like r-service
clients, if you are root locally.  For example, feeding "^@root^@joe^@pwd^@"
[where ^@ is a null, and root/joe could be any other local/remote username
pair] into a "rsh" or "rlogin" server, FROM your port 1023 for example,
duplicates what the server expects to receive.  Thus, you can test for insecure
.rhosts files around your network without having to create new user accounts on
your client machine.  The program data/rservice.c can aid this process by
constructing the "rcmd" protocol bytes.  Doing this also prevents "rshd" from
trying to create that separate standard-error socket and still gives you an
input path, as opposed to the usual action of "rsh -n".  Using netcat for
things like this can be really useful sometimes, because rsh and rlogin
generally want a host *name* as an argument and won't accept IP addresses.  If
your client-end DNS is hosed, as may be true when you're trying to extract
backup sets on to a dumb client, "netcat -n" wins where normal rsh/rlogin is
useless.

If you are unsure that a remote syslogger is working, test it with netcat.
Make a UDP connection to port 514 and type in "<0>message", which should
correspond to "kern.emerg" and cause syslogd to scream into every file it has
open [and possibly all over users' terminals].  You can tame this down by
using a different number and use netcat inside routine scripts to send syslog
messages to places that aren't configured in syslog.conf.  For example,
"echo '<38>message' | nc -w 1 -u loggerhost 514" should send to auth.notice
on loggerhost.  The exact number may vary; check against your syslog.h first.

Netcat provides several ways for you to test your own packet filters.  If you
bind to a port normally protected against outside access and make a connection
to somewhere outside your own network, the return traffic will be coming to
your chosen port from the "outside" and should be blocked.  TCP may get through
if your filter passes all "ack syn", but it shouldn't be even doing that to low
ports on your network.  Remember to test with UDP traffic as well!  If your
filter passes at least outbound source-routed IP packets, bouncing a connection
back to yourself via some gateway outside your network will create "incoming"
traffic with your source address, which should get dropped by a correctly
configured anti-spoofing filter.  This is a "non-test" if you're also dropping
source-routing, but it's good to be able to test for that too.  Any packet
filter worth its salt will be blocking source-routed packets in both
directions, but you never know what interesting quirks you might turn up by
playing around with source ports and addresses and watching the wires with a
network monitor.

You can use netcat to protect your own workstation's X server against outside
access.  X is stupid enough to listen for connections on "any" and never tell
you when new connections arrive, which is one reason it is so vulnerable.  Once
you have all your various X windows up and running you can use netcat to bind
just to your ethernet address and listen to port 6000.  Any new connections
from outside the machine will hit netcat instead your X server, and you get a
log of who's trying.  You can either tell netcat to drop the connection, or
perhaps run another copy of itself to relay to your actual X server on
"localhost".  This may not work for dedicated X terminals, but it may be
possible to authorize your X terminal only for its boot server, and run a relay
netcat over on the server that will in turn talk to your X terminal.  Since
netcat only handles one listening connection per run, make sure that whatever
way you rig it causes another one to run and listen on 6000 soon afterward, or
your real X server will be reachable once again.  A very minimal script just
to protect yourself could be

	while true ; do
	  nc -v -l -s <your-addr> -p 6000 localhost 2
	done

which causes netcat to accept and then close any inbound connection to your
workstation's normal ethernet address, and another copy is immediately run by
the script.  Send standard error to a file for a log of connection attempts.
If your system can't do the "specific bind" thing all is not lost; run your
X server on display ":1" or port 6001, and netcat can still function as a probe
alarm by listening on 6000.

Does your shell-account provider allow personal Web pages, but not CGI scripts?
You can have netcat listen on a particular port to execute a program or script
of your choosing, and then just point to the port with a URL in your homepage.
The listener could even exist on a completely different machine, avoiding the
potential ire of the homepage-host administrators.  Since the script will get
the raw browser query as input it won't look like a typical CGI script, and
since it's running under your UID you need to write it carefully.  You may want
to write a netcat-based script as a wrapper that reads a query and sets up
environment variables for a regular CGI script.  The possibilities for using
netcat and scripts to handle Web stuff are almost endless.  Again, see the
examples under scripts/.

Example uses -- the dark side
=============================

Equal time is deserved here, since a versatile tool like this can be useful
to any Shade of Hat.  I could use my Victorinox to either fix your car or
disassemble it, right?  You can clearly use something like netcat to attack
or defend -- I don't try to govern anyone's social outlook, I just build tools.
Regardless of your intentions, you should still be aware of these threats to
your own systems.

The first obvious thing is scanning someone *else's* network for vulnerable
services.  Files containing preconstructed data, be it exploratory or
exploitive, can be fed in as standard input, including command-line arguments
to netcat itself to keep "ps" ignorant of your doings.  The more random the
scanning, the less likelihood of detection by humans, scan-detectors, or
dynamic filtering, and with -i you'll wait longer but avoid loading down the
target's network.  Some examples for crafting various standard UDP probes are
given in data/*.d.

Some configurations of packet filters attempt to solve the FTP-data problem by
just allowing such connections from the outside.  These come FROM port 20, TO
high TCP ports inside -- if you locally bind to port 20, you may find yourself
able to bypass filtering in some cases.  Maybe not to low ports "inside", but
perhaps to TCP NFS servers, X servers, Prospero, ciscos that listen on 200x
and 400x...  Similar bypassing may be possible for UDP [and maybe TCP too] if a
connection comes from port 53; a filter may assume it's a nameserver response.

Using -e in conjunction with binding to a specific address can enable "server
takeover" by getting in ahead of the real ones, whereupon you can snarf data
sent in and feed your own back out.  At the very least you can log a hex dump
of someone else's session.  If you are root, you can certainly use -s and -e to
run various hacked daemons without having to touch inetd.conf or the real
daemons themselves.  You may not always have the root access to deal with low
ports, but what if you are on a machine that also happens to be an NFS server?
You might be able to collect some interesting things from port 2049, including
local file handles.  There are several other servers that run on high ports
that are likely candidates for takeover, including many of the RPC services on
some platforms [yppasswdd, anyone?].  Kerberos tickets, X cookies, and IRC
traffic also come to mind.  RADIUS-based terminal servers connect incoming
users to shell-account machines on a high port, usually 1642 or thereabouts.
SOCKS servers run on 1080.  Do "netstat -a" and get creative.

There are some daemons that are well-written enough to bind separately to all
the local interfaces, possibly with an eye toward heading off this sort of
problem.  Named from recent BIND releases, and NTP, are two that come to mind.
Netstat will show these listening on address.53 instead of *.53.  You won't
be able to get in front of these on any of the real interface addresses, which
of course is especially interesting in the case of named, but these servers
sometimes forget about things like "alias" interface addresses or interfaces
that appear later on such as dynamic PPP links.  There are some hacked web
servers and versions of "inetd" floating around that specifically bind as well,
based on a configuration file -- these generally *are* bound to alias addresses
to offer several different address-based services from one machine.

Using -e to start a remote backdoor shell is another obvious sort of thing,
easier than constructing a file for inetd to listen on "ingreslock" or
something, and you can access-control it against other people by specifying a
client host and port.  Experience with this truly demonstrates how fragile the
barrier between being "logged in" or not really is, and is further expressed by
scripts/bsh.  If you're already behind a firewall, it may be easier to make an
*outbound* connection and then run a shell; a small wrapper script can
periodically try connecting to a known place and port, you can later listen
there until the inbound connection arrives, and there's your shell.  Running
a shell via UDP has several interesting features, although be aware that once
"connected", the UDP stub sockets tend to show up in "netstat" just like TCP
connections and may not be quite as subtle as you wanted.  Packets may also be
lost, so use TCP if you need reliable connections.  But since UDP is
connectionless, a hookup of this sort will stick around almost forever, even if
you ^C out of netcat or do a reboot on your side, and you only need to remember
the ports you used on both ends to reestablish.  And outbound UDP-plus-exec
connection creates the connected socket and starts the program immediately.  On
a listening UDP connection, the socket is created once a first packet is
received.  In either case, though, such a "connection" has the interesting side
effect that only your client-side IP address and [chosen?] source port will
thereafter be able to talk to it.  Instant access control!  A non-local third
party would have to do ALL of the following to take over such a session:

	forge UDP with your source address [trivial to do; see below]
	guess the port numbers of BOTH ends, or sniff the wire for them
	arrange to block ICMP or UDP return traffic between it and your real
	  source, so the session doesn't die with a network write error.

The companion program data/rservice.c is helpful in scripting up any sort of
r-service username or password guessing attack.  The arguments to "rservice"
are simply the strings that get null-terminated and passed over an "rcmd"-style
connection, with the assumption that the client does not need a separate
standard-error port.  Brute-force password banging is best done via "rexec" if
it is available since it is less likely to log failed attempts.  Thus, doing
"rservice joe joespass pwd | nc target exec" should return joe's home dir if
the password is right, or "Permission denied."  Plug in a dictionary and go to
town.  If you're attacking rsh/rlogin, remember to be root and bind to a port
between 512 and 1023 on your end, and pipe in "rservice joe joe pwd" and such.

Netcat can prevent inadvertently sending extra information over a telnet
connection.  Use "nc -t" in place of telnet, and daemons that try to ask for
things like USER and TERM environment variables will get no useful answers, as
they otherwise would from a more recent telnet program.  Some telnetds actually
try to collect this stuff and then plug the USER variable into "login" so that
the caller is then just asked for a password!  This mechanism could cause a
login attempt as YOUR real username to be logged over there if you use a
Borman-based telnet instead of "nc -t".

Got an unused network interface configured in your kernel [e.g. SLIP], or
support for alias addresses?  Ifconfig one to be any address you like, and bind
to it with -s to enable all sorts of shenanigans with bogus source addresses.
The interface probably has to be UP before this works; some SLIP versions
need a far-end address before this is true.  Hammering on UDP services is then
a no-brainer.  What you can do to an unfiltered syslog daemon should be fairly
obvious; trimming the conf file can help protect against it.  Many routers out
there still blindly believe what they receive via RIP and other routing
protocols.  Although most UDP echo and chargen servers check if an incoming
packet was sent from *another* "internal" UDP server, there are many that still
do not, any two of which [or many, for that matter] could keep each other
entertained for hours at the expense of bandwidth.  And you can always make
someone wonder why she's being probed by nsa.gov.

Your TCP spoofing possibilities are mostly limited to destinations you can
source-route to while locally bound to your phony address.  Many sites block
source-routed packets these days for precisely this reason.  If your kernel
does oddball things when sending source-routed packets, try moving the pointer
around with -G.  You may also have to fiddle with the routing on your own
machine before you start receiving packets back.  Warning: some machines still
send out traffic using the source address of the outbound interface, regardless
of your binding, especially in the case of localhost.  Check first.  If you can
open a connection but then get no data back from it, the target host is
probably killing the IP options on its end [this is an option inside TCP
wrappers and several other packages], which happens after the 3-way handshake
is completed.  If you send some data and observe the "send-q" side of "netstat"
for that connection increasing but never getting sent, that's another symptom.
Beware: if Sendmail 8.7.x detects a source-routed SMTP connection, it extracts
the hop list and sticks it in the Received: header!

SYN bombing [sometimes called "hosing"] can disable many TCP servers, and if
you hit one often enough, you can keep it unreachable for days.  As is true of
many other denial-of-service attacks, there is currently no defense against it
except maybe at the human level.  Making kernel SOMAXCONN considerably larger
than the default and the half-open timeout smaller can help, and indeed some
people running large high-performance web servers have *had* to do that just to
handle normal traffic.  Taking out mailers and web servers is sociopathic, but
on the other hand it is sometimes useful to be able to, say, disable a site's
identd daemon for a few minutes.  If someone realizes what is going on,
backtracing will still be difficult since the packets have a phony source
address, but calls to enough ISP NOCs might eventually pinpoint the source.
It is also trivial for a clueful ISP to watch for or even block outgoing
packets with obviously fake source addresses, but as we know many of them are
not clueful or willing to get involved in such hassles.  Besides, outbound
packets with an [otherwise unreachable] source address in one of their net
blocks would look fairly legitimate.

Notes
=====

A discussion of various caveats, subtleties, and the design of the innards.

As of version 1.07 you can construct a single file containing command arguments
and then some data to transfer.  Netcat is now smart enough to pick out the
first line and build the argument list, and send any remaining data across the
net to one or multiple ports.  The first release of netcat had trouble with
this -- it called fgets() for the command line argument, which behind the
scenes does a large read() from standard input, perhaps 4096 bytes or so, and
feeds that out to the fgets() library routine.  By the time netcat 1.00 started
directly read()ing stdin for more data, 4096 bytes of it were gone.  It now
uses raw read() everywhere and does the right thing whether reading from files,
pipes, or ttys.  If you use this for multiple-port connections, the single
block of data will now be a maximum of 8K minus the first line.  Improvements
have been made to the logic in sending the saved chunk to each new port.  Note
that any command-line arguments hidden using this mechanism could still be
extracted from a core dump.

When netcat receives an inbound UDP connection, it creates a "connected socket"
back to the source of the connection so that it can also send out data using
normal write().  Using this mechanism instead of recvfrom/sendto has several
advantages -- the read/write select loop is simplified, and ICMP errors can in
effect be received by non-root users.  However, it has the subtle side effect
that if further UDP packets arrive from the caller but from different source
ports, the listener will not receive them.  UDP listen mode on a multihomed
machine may have similar quirks unless you specifically bind to one of its
addresses.  It is not clear that kernel support for UDP connected sockets
and/or my understanding of it is entirely complete here, so experiment...

You should be aware of some subtleties concerning UDP scanning.  If -z is on,
netcat attempts to send a single null byte to the target port, twice, with a
small time in between.  You can either use the -w timeout, or netcat will try
to make a "sideline" TCP connection to the target to introduce a small time
delay equal to the round-trip time between you and the target.  Note that if
you have a -w timeout and -i timeout set, BOTH take effect and you wait twice
as long.  The TCP connection is to a normally refused port to minimize traffic,
but if you notice a UDP fast-scan taking somewhat longer than it should, it
could be that the target is actually listening on the TCP port.  Either way,
any ICMP port-unreachable messages from the target should have arrived in the
meantime.  The second single-byte UDP probe is then sent.  Under BSD kernels,
the ICMP error is delivered to the "connected socket" and the second write
returns an error, which tells netcat that there is NOT a UDP service there.
While Linux seems to be a fortunate exception, under many SYSV derived kernels
the ICMP is not delivered, and netcat starts reporting that *all* the ports are
"open" -- clearly wrong.  [Some systems may not even *have* the "udp connected
socket" concept, and netcat in its current form will not work for UDP at all.]
If -z is specified and only one UDP port is probed, netcat's exit status
reflects whether the connection was "open" or "refused" as with TCP.

It may also be that UDP packets are being blocked by filters with no ICMP error
returns, in which case everything will time out and return "open".  This all
sounds backwards, but that's how UDP works.  If you're not sure, try "echo
w00gumz | nc -u -w 2 target 7" to see if you can reach its UDP echo port at
all.  You should have no trouble using a BSD-flavor system to scan for UDP
around your own network, although flooding a target with the high activity that
-z generates will cause it to occasionally drop packets and indicate false
"opens".  A more "correct" way to do this is collect and analyze the ICMP
errors, as does SATAN's "udp_scan" backend, but then again there's no guarantee
that the ICMP gets back to you either.  Udp_scan also does the zero-byte
probes but is excruciatingly careful to calculate its own round-trip timing
average and dynamically set its own response timeouts along with decoding any
ICMP received.  Netcat uses a much sleazier method which is nonetheless quite
effective.  Cisco routers are known to have a "dead time" in between ICMP
responses about unreachable UDP ports, so a fast scan of a cisco will show
almost everything "open".  If you are looking for a specific UDP service, you
can construct a file containing the right bytes to trigger a response from the
other end and send that as standard input.  Netcat will read up to 8K of the
file and send the same data to every UDP port given.  Note that you must use a
timeout in this case [as would any other UDP client application] since the
two-write probe only happens if -z is specified.

Many telnet servers insist on a specific set of option negotiations before
presenting a login banner.  On a raw connection you will see this as small
amount of binary gook.  My attempts to create fixed input bytes to make a
telnetd happy worked some places but failed against newer BSD-flavor ones,
possibly due to timing problems, but there are a couple of much better
workarounds.  First, compile with -DTELNET and use -t if you just want to get
past the option negotiation and talk to something on a telnet port.  You will
still see the binary gook -- in fact you'll see a lot more of it as the options
are responded to behind the scenes.  The telnet responder does NOT update the
total byte count, or show up in the hex dump -- it just responds negatively to
any options read from the incoming data stream.  If you want to use a normal
full-blown telnet to get to something but also want some of netcat's features
involved like settable ports or timeouts, construct a tiny "foo" script:

	#! /bin/sh
	exec nc -otheroptions targethost 23

and then do

	nc -l -p someport -e foo localhost &
	telnet localhost someport

and your telnet should connect transparently through the exec'ed netcat to
the target, using whatever options you supplied in the "foo" script.  Don't
use -t inside the script, or you'll wind up sending *two* option responses.

I've observed inconsistent behavior under some Linuxes [perhaps just older
ones?] when binding in listen mode.  Sometimes netcat binds only to "localhost"
if invoked with no address or port arguments, and sometimes it is unable to
bind to a specific address for listening if something else is already listening
on "any".  The former problem can be worked around by specifying "-s 0.0.0.0",
which will do the right thing despite netcat claiming that it's listening on
[127.0.0.1].  This is a known problem -- for example, there's a mention of it
in the makefile for SOCKS.  On the flip side, binding to localhost and sending
packets to some other machine doesn't work as you'd expect -- they go out with
the source address of the sending interface instead.  The Linux kernel contains
a specific check to ensure that packets from 127.0.0.1 are never sent to the
wire; other kernels may contain similar code.  Linux, of course, *still*
doesn't support source-routing, but they claim that it and many other network
improvements are at least breathing hard.

There are several possible errors associated with making TCP connections, but
to specifically see anything other than "refused", one must wait the full
kernel-defined timeout for a connection to fail.  Netcat's mechanism of
wrapping an alarm timer around the connect prevents the *real* network error
from being returned -- "errno" at that point indicates "interrupted system
call" since the connect attempt was interrupted.  Some old 4.3 BSD kernels
would actually return things like "host unreachable" immediately if that was
the case, but most newer kernels seem to wait the full timeout and *then* pass
back the real error.  Go figure.  In this case, I'd argue that the old way was
better, despite those same kernels generally being the ones that tear down
*established* TCP connections when ICMP-bombed.

Incoming socket options are passed to applications by the kernel in the
kernel's own internal format.  The socket-options structure for source-routing
contains the "first-hop" IP address first, followed by the rest of the real
options list.  The kernel uses this as is when sending reply packets -- the
structure is therefore designed to be more useful to the kernel than to humans,
but the hex dump of it that netcat produces is still useful to have.

Kernels treat source-routing options somewhat oddly, but it sort of makes sense
once one understands what's going on internally.  The options list of addresses
must contain hop1, hop2, ..., destination.  When a source-routed packet is sent
by the kernel [at least BSD], the actual destination address becomes irrelevant
because it is replaced with "hop1", "hop1" is removed from the options list,
and all the other addresses in the list are shifted up to fill the hole.  Thus
the outbound packet is sent from your chosen source address to the first
*gateway*, and the options list now contains hop2, ..., destination.  During
all this address shuffling, the kernel does NOT change the pointer value, which
is why it is useful to be able to set the pointer yourself -- you can construct
some really bizarre return paths, and send your traffic fairly directly to the
target but around some larger loop on the way back.  Some Sun kernels seem to
never flip the source-route around if it contains less than three hops, never
reset the pointer anyway, and tries to send the packet [with options containing
a "completed" source route!!] directly back to the source.  This is way broken,
of course.  [Maybe ipforwarding has to be on?  I haven't had an opportunity to
beat on it thoroughly yet.]

"Credits" section: The original idea for netcat fell out of a long-standing
desire and fruitless search for a tool resembling it and having the same
features.  After reading some other network code and realizing just how many
cool things about sockets could be controlled by the calling user, I started
on the basics and the rest fell together pretty quickly.  Some port-scanning
ideas were taken from Venema/Farmer's SATAN tool kit, and Pluvius' "pscan"
utility.  Healthy amounts of BSD kernel source were perused in an attempt to
dope out socket options and source-route handling; additional help was obtained
from Dave Borman's telnet sources.  The select loop is loosely based on fairly
well-known code from "rsh" and Richard Stevens' "sock" program [which itself is
sort of a "netcat" with more obscure features], with some more paranoid
sanity-checking thrown in to guard against the distinct likelihood that there
are subtleties about such things I still don't understand.  I found the
argument-hiding method cleanly implemented in Barrett's "deslogin"; reading the
line as input allows greater versatility and is much less prone to cause
bizarre problems than the more common trick of overwriting the argv array.
After the first release, several people contributed portability fixes; they are
credited in generic.h and the Makefile.  Lauren Burka inspired the ascii art
for this revised document.  Dean Gaudet at Wired supplied a precursor to
the hex-dump code, and mudge@l0pht.com originally experimented with and
supplied code for the telnet-options responder.  Outbound "-e <prog>" resulted
from a need to quietly bypass a firewall installation.  Other suggestions and
patches have rolled in for which I am always grateful, but there are only 26
hours per day and a discussion of feature creep near the end of this document.

Netcat was written with the Russian railroad in mind -- conservatively built
and solid, but it *will* get you there.  While the coding style is fairly
"tight", I have attempted to present it cleanly [keeping *my* lines under 80
characters, dammit] and put in plenty of comments as to why certain things
are done.  Items I know to be questionable are clearly marked with "XXX".
Source code was made to be modified, but determining where to start is
difficult with some of the tangles of spaghetti code that are out there.
Here are some of the major points I feel are worth mentioning about netcat's
internal design, whether or not you agree with my approach.

Except for generic.h, which changes to adapt more platforms, netcat is a single
source file.  This has the distinct advantage of only having to include headers
once and not having to re-declare all my functions in a billion different
places.  I have attempted to contain all the gross who's-got-what-.h-file
things in one small dumping ground.  Functions are placed "dependencies-first",
such that when the compiler runs into the calls later, it already knows the
type and arguments and won't complain.  No function prototyping -- not even the
__P(()) crock -- is used, since it is more portable and a file of this size is
easy enough to check manually.  Each function has a standard-format comment
ahead of it, which is easily found using the regexp " :$".  I freely use gotos.
Loops and if-clauses are made as small and non-nested as possible, and the ends
of same *marked* for clarity [I wish everyone would do this!!].

Large structures and buffers are all malloc()ed up on the fly, slightly larger
than the size asked for and zeroed out.  This reduces the chances of damage
from those "end of the buffer" fencepost errors or runaway pointers escaping
off the end.  These things are permanent per run, so nothing needs to be freed
until the program exits.

File descriptor zero is always expected to be standard input, even if it is
closed.  If a new network descriptor winds up being zero, a different one is
asked for which will be nonzero, and fd zero is simply left kicking around
for the rest of the run.  Why?  Because everything else assumes that stdin is
always zero and "netfd" is always positive.  This may seem silly, but it was a
lot easier to code.  The new fd is obtained directly as a new socket, because
trying to simply dup() a new fd broke subsequent socket-style use of the new fd
under Solaris' stupid streams handling in the socket library.

The catch-all message and error handlers are implemented with an ample list of
phoney arguments to get around various problems with varargs.  Varargs seems
like deliberate obfuscation in the first place, and using it would also
require use of vfprintf() which not all platforms support.  The trailing
sleep in bail() is to allow output to flush, which is sometimes needed if
netcat is already on the other end of a network connection.

The reader may notice that the section that does DNS lookups seems much
gnarlier and more confusing than other parts.  This is NOT MY FAULT.  The
sockaddr and hostent abstractions are an abortion that forces the coder to
deal with it.  Then again, a lot of BSD kernel code looks like similar
struct-pointer hell.  I try to straighten it out somewhat by defining my own
HINF structure, containing names, ascii-format IP addresses, and binary IP
addresses.  I fill this structure exactly once per host argument, and squirrel
everything safely away and handy for whatever wants to reference it later.

Where many other network apps use the FIONBIO ioctl to set non-blocking I/O
on network sockets, netcat uses straightforward blocking I/O everywhere.
This makes everything very lock-step, relying on the network and filesystem
layers to feed in data when needed.  Data read in is completely written out
before any more is fetched.  This may not be quite the right thing to do under
some OSes that don't do timed select() right, but this remains to be seen.

The hexdump routine is written to be as fast as possible, which is why it does
so much work itself instead of just sprintf()ing everything together.  Each
dump line is built into a single buffer and atomically written out using the
lowest level I/O calls.  Further improvements could undoubtedly be made by
using writev() and eliminating all sprintf()s, but it seems to fly right along
as is.  If both exec-a-prog mode and a hexdump file is asked for, the hexdump
flag is deliberately turned off to avoid creating random zero-length files.
Files are opened in "truncate" mode; if you want "append" mode instead, change
the open flags in main().

main() may look a bit hairy, but that's only because it has to go down the
argv list and handle multiple ports, random mode, and exit status.  Efforts
have been made to place a minimum of code inside the getopt() loop.  Any real
work is sent off to functions in what is hopefully a straightforward way.

Obligatory vendor-bash: If "nc" had become a standard utility years ago,
the commercial vendors would have likely packaged it setuid root and with
-DGAPING_SECURITY_HOLE turned on but not documented.  It is hoped that netcat
will aid people in finding and fixing the no-brainer holes of this sort that
keep appearing, by allowing easier experimentation with the "bare metal" of
the network layer.

It could be argued that netcat already has too many features.  I have tried
to avoid "feature creep" by limiting netcat's base functionality only to those
things which are truly relevant to making network connections and the everyday
associated DNS lossage we're used to.  Option switches already have slightly
overloaded functionality.  Random port mode is sort of pushing it.  The
hex-dump feature went in later because it *is* genuinely useful.  The
telnet-responder code *almost* verges on the gratuitous, especially since it
mucks with the data stream, and is left as an optional piece.  Many people have
asked for example "how 'bout adding encryption?" and my response is that such
things should be separate entities that could pipe their data *through* netcat
instead of having their own networking code.  I am therefore not completely
enthusiastic about adding any more features to this thing, although you are
still free to send along any mods you think are useful.

Nonetheless, at this point I think of netcat as my tcp/ip swiss army knife,
and the numerous companion programs and scripts to go with it as duct tape.
Duct tape of course has a light side and a dark side and binds the universe
together, and if I wrap enough of it around what I'm trying to accomplish,
it *will* work.  Alternatively, if netcat is a large hammer, there are many
network protocols that are increasingly looking like nails by now...

_H* 960320 v1.10 RELEASE -- happy spring!

## How to get syscall asm instructions?
musl project contains syscall asm instructioins of main architectures,
We can use this command to generate readable asm code:

```bash
$ gcc -E ${arch}/syscall_arch.h
```

## How to install debian sysroot?
First install deps:
```bash
$ sudo apt install qemu-user-static debootstrap
```

Then create sysroot of specific architecture:
```bash
$ sudo debootstrap --arch aarch64 buster aarch64-root http://ftp.cn.debian.org/debian
```
Here:
* `--arch aarch64`, use `aarch64`
* `buster`, use latest debian stable image
* `aarch64-root`, rootfs folder name
* `http://ftp.cn.debian.org/debian`, debian mirror url. Most of unofficial debian
mirrors do not support non-x86 architectures

## TODO
- [x] support freebsd
- [ ] support windows
- [x] sighandler types
- [x] implement more syscalls
- [x] test arm
- [x] test mips
- [x] test mipsel
- [x] test ppc64
- [x] test ppc64le
- [ ] test s390x
- [x] add linux uapi types
- [x] remove little endians
- [ ] linux socket example
- [ ] big endian/litten endian bitwise integers
- [x] cstring issue
- [x] support stable version
- [x] derive Clone, Default traits
- [x] fix stat_t type error
- [ ] simplify syscalls

For now, read the header comments inside each of these for documentation.
The programs are simple enough that they don't really need a Makefile any more
complex than the one given; ymmv.  Data and xor may also be useful on DOS,
which is why there are hooks for it in the code.

data.c		a primitive atob / btoa byte generator
*.d		example input to "data -g"
rservice.c	a utility for scripting up rsh/rexec attacks
xor.c		generic xor handler

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## -----------------------------------------------------------------------------

library(data.table)
data.table(iris)


## -----------------------------------------------------------------------------

iris.parts <- list(
  nc=nc::capture_melt_multiple(
    iris,
    column=".*?",
    "[.]",
    dim=".*"),
  tidyr=if(requireNamespace("tidyr"))tidyr::pivot_longer(
    iris, 
    cols=1:4, 
    names_to=c(".value", "dim"),
    names_sep="[.]"),
  stats=stats::reshape(
    iris,
    direction="long",
    timevar="dim",
    varying=1:4,
    sep="."),
  "data.table::melt"=melt(
    data.table(iris),
    measure.vars=patterns(
      Sepal="^Sepal",
      Petal="^Petal")
  )[data.table(
    variable=factor(1:2), dim=c("Length", "Width")
  ), on=.(variable)],
  if(requireNamespace("cdata"))cdata::rowrecs_to_blocks(
    iris,
    controlTable=data.frame(
      dim=c("Length", "Width"),
      Petal=c("Petal.Length", "Petal.Width"),
      Sepal=c("Sepal.Length", "Sepal.Width"),
      stringsAsFactors=FALSE),
    columnsToCopy="Species"))
iris.parts$nc


## -----------------------------------------------------------------------------

if(require(ggplot2)){
  ggplot()+
    theme_bw()+
    theme(panel.spacing=grid::unit(0, "lines"))+
    facet_grid(dim ~ Species)+
    coord_equal()+
    geom_abline(slope=1, intercept=0, color="grey")+
    geom_point(aes(
      Petal, Sepal),
      data=iris.parts$nc)
}


## -----------------------------------------------------------------------------

iris.dims <- list(
  nc=nc::capture_melt_multiple(
    iris,
    part=".*?",
    "[.]",
    column=".*"),
  stats=stats::reshape(
    structure(iris, names=sub("(.*?)[.](.*)", "\\2.\\1", names(iris))),
    direction="long",
    timevar="part",
    varying=1:4,
    sep="."))
iris.dims$nc


## -----------------------------------------------------------------------------

if(require(ggplot2)){
  ggplot()+
    theme_bw()+
    theme(panel.spacing=grid::unit(0, "lines"))+
    facet_grid(part ~ Species)+
    coord_equal()+
    geom_abline(slope=1, intercept=0, color="grey")+
    geom_point(aes(
      Length, Width),
      data=iris.dims$nc)
}


## -----------------------------------------------------------------------------

TC <- data.table::data.table(
  age.treatment=c(1, 5),
  sex.control=c("M", "M"),
  sex.treatment=c("F", "F"),
  age.control=c(10, 50))


## -----------------------------------------------------------------------------

input.list <- list(
  "nc"=nc::capture_melt_multiple(
    TC,
    column=".*?",
    "[.]",
    group=".*"),
  "cdata"=if(requireNamespace("cdata"))cdata::rowrecs_to_blocks(
    TC,
    controlTable=data.frame(
      group=c("treatment", "control"),
      age=c("age.treatment", "age.control"),
      sex=c("sex.treatment", "sex.control"),
      stringsAsFactors=FALSE)),
  "data.table"=data.table::melt(TC, measure.vars=patterns(
    age="age",
    sex="sex")),
  "stats"=stats::reshape(
    TC,
    varying=1:4,
    direction="long"),
  "tidyr"=if(requireNamespace("tidyr"))tidyr::pivot_longer(
    TC, 1:4,
    names_to=c(".value", "group"),
    names_sep="[.]"))
output.list <- list()
for(pkg in names(input.list)){
  df.or.null <- input.list[[pkg]]
  if(is.data.frame(df.or.null)){
    output.list[[pkg]] <- data.table::data.table(df.or.null)[order(age)]
  }
}
output.list
sapply(output.list, function(DT)identical(DT$sex, c("F", "F", "M", "M")))


## -----------------------------------------------------------------------------

if(requireNamespace("tidyr")){
  data(who, package="tidyr")
}else{
  who <- data.frame(id=1, new_sp_m5564=2, newrel_f65=3)
}
names(who)


## -----------------------------------------------------------------------------

who.chr.list <- list(
  nc=nc::capture_melt_single(
    who,
    "new_?",
    diagnosis=".*",
    "_",
    gender=".",
    ages=".*"),
  tidyr=if(requireNamespace("tidyr"))tidyr::pivot_longer(
    who,
    new_sp_m014:newrel_f65,
    names_to=c("diagnosis", "gender", "ages"),
    names_pattern="new_?(.*)_(.)(.*)"))


## -----------------------------------------------------------------------------

who.pattern <- "new_?(.*)_(.)((0|[0-9]{2})([0-9]{0,2}))"
as.numeric.Inf <- function(y)ifelse(y=="", Inf, as.numeric(y))
who.typed.list <- list(
  nc=nc::capture_melt_single(
    who,
    "new_?",
    diagnosis=".*",
    "_",
    gender=".",
    ages=list(
      ymin.num="0|[0-9]{2}", as.numeric,
      ymax.num="[0-9]{0,2}", as.numeric.Inf),
    value.name="count",
    na.rm=TRUE),
  tidyr=if(requireNamespace("tidyr"))try(tidyr::pivot_longer(
    who,
    cols=grep(who.pattern, names(who)),
    names_transform=list(
      ymin.num=as.numeric,
      ymax.num=as.numeric.Inf),
    names_to=c("diagnosis", "gender", "ages", "ymin.num", "ymax.num"),
    names_pattern=who.pattern,
    values_drop_na=TRUE,
    values_to="count")))
str(who.typed.list)


## -----------------------------------------------------------------------------

if(requireNamespace("tidyr")){
  gather.result <- tidyr::gather(
    who,
    "variable",
    "count",
    grep(who.pattern, names(who)),
    na.rm=TRUE)
  extract.result <- tidyr::extract(
    gather.result,
    "variable",
    c("diagnosis", "gender", "ages", "ymin.int", "ymax.int"),
    who.pattern,
    convert=TRUE)
  transform.result <- base::transform(
    extract.result,
    ymin.num=as.numeric(ymin.int),
    ymax.num=ifelse(is.na(ymax.int), Inf, as.numeric(ymax.int)))
  str(transform.result)
}


## -----------------------------------------------------------------------------

reshape2.result <- if(requireNamespace("reshape2")){
  reshape2:::melt.data.frame(
    who,
    measure.vars=grep(who.pattern, names(who)),
    na.rm=TRUE,
    value.name="count")
}


## -----------------------------------------------------------------------------

dt.result <- data.table::melt.data.table(
  data.table(who),
  measure.vars=patterns(who.pattern),
  na.rm=TRUE,
  value.name="count")


## -----------------------------------------------------------------------------

who.df <- data.frame(who)
is.varying <- grepl(who.pattern, names(who))
names(who.df)[is.varying] <- paste0("count.", names(who)[is.varying])
stats.result <- stats::reshape(
  who.df,
  direction="long",
  timevar="variable",
  varying=is.varying)


## -----------------------------------------------------------------------------

if(requireNamespace("cdata")){
  cdata.result <- cdata::rowrecs_to_blocks(
    who, 
    cdata::build_unpivot_control(
      "variable",
      "count",
      grep(who.pattern, names(who), value=TRUE)),
    columnsToCopy=grep(who.pattern, names(who), value=TRUE, invert=TRUE))
}


## -----------------------------------------------------------------------------

## Example 1: melting a wider iris data back to original.
library(data.table)
iris.dt <- data.table(
  i=1:nrow(iris),
  iris[,1:4],
  Species=paste(iris$Species))
print(iris.dt)

## what if we had two observations on each row?
set.seed(1)
iris.rand <- iris.dt[sample(.N)]
iris.wide <- cbind(treatment=iris.rand[1:75], control=iris.rand[76:150])
print(iris.wide, topn=2, nrows=10)

## This is the usual data.table syntax for getting the original iris back.
iris.melted <- melt(iris.wide, value.factor=TRUE, measure.vars = patterns(
  i="i$",
  Sepal.Length="Sepal.Length$",
  Sepal.Width="Sepal.Width$",
  Petal.Length="Petal.Length$",
  Petal.Width="Petal.Width$",
  Species="Species$"))
identical(iris.melted[order(i), names(iris.dt), with=FALSE], iris.dt)

## nc can do the same thing -- you must define an R argument named
## column, and another named argument which identifies each group.
(nc.melted <- nc::capture_melt_multiple(
  iris.wide,
  group="[^.]+",
  "[.]",
  column=".*"))
identical(nc.melted[order(i), names(iris.dt), with=FALSE], iris.dt)

## This is how we do it using stats::reshape.
iris.wide.df <- data.frame(iris.wide)
names(iris.wide.df) <- sub("(.*?)[.](.*)", "\\2_\\1", names(iris.wide))
iris.reshaped <- stats::reshape(
  iris.wide.df,
  direction="long",
  timevar="group",
  varying=names(iris.wide.df),
  sep="_")
identical(data.table(iris.reshaped[, names(iris.dt)])[order(i)], iris.dt)

## get the parts columns and groups -- is there any difference
## between groups? of course not!
parts.wide <- nc::capture_melt_multiple(
  iris.wide,
  group=".*?",
  "[.]",
  column=".*?",
  "[.]",
  dim=".*")
if(require("ggplot2")){
  ggplot()+
    theme_bw()+
    theme(panel.spacing=grid::unit(0, "lines"))+
    facet_grid(dim ~ group)+
    coord_equal()+
    geom_abline(slope=1, intercept=0, color="grey")+
    geom_point(aes(
      Petal, Sepal),
      data=parts.wide)
}


## -----------------------------------------------------------------------------

## Example 2. Lots of column types, from example(melt.data.table).
DT <- data.table(
  i_1 = c(1:5, NA),
  i_2 = c(NA,6:10),
  f_1 = factor(sample(c(letters[1:3], NA), 6, TRUE)),
  f_2 = factor(c("z", "a", "x", "c", "x", "x"), ordered=TRUE),
  c_1 = sample(c(letters[1:3], NA), 6, TRUE),
  d_1 = as.Date(c(1:3,NA,4:5), origin="2013-09-01"),
  d_2 = as.Date(6:1, origin="2012-01-01"))
## add a couple of list cols
DT[, l_1 := DT[, list(c=list(rep(i_1, sample(5,1)))), by = i_1]$c]
DT[, l_2 := DT[, list(c=list(rep(c_1, sample(5,1)))), by = i_1]$c]

## original DT syntax is quite repetitive.
melt(DT, measure=patterns(
  i="^i",
  f="^f",
  d="^d",
  l="^l"
))

## nc syntax uses a single regex rather than four.
nc::capture_melt_multiple(
  DT,
  column="^[^c]",
  "_",
  number="[12]")

## id.vars can be specified using original DT syntax.
melt(DT, id=1:2, measure=patterns(
  f="^f",
  l="^l"
))

nc::capture_melt_multiple(
  DT,
  column="^[fl]",
  "_",
  number="[12]")

## reshape does not support list columns.
reshape(
  DT,
  varying=grep("^[fid]", names(DT)),
  sep="_",
  direction="long",
  timevar="number")

## tidyr does, but errors for combining ordered and un-ordered factors.
if(requireNamespace("tidyr")){
  tidyr::pivot_longer(
    DT, grep("[cf]", names(DT), invert=TRUE),
    names_pattern="(.)_(.)",
    names_to=c(".value", "number"))
}


## -----------------------------------------------------------------------------

## Example 3, three children, one family per row, from data.table
## vignette.
family.dt <- fread(text="
family_id age_mother dob_child1 dob_child2 dob_child3 gender_child1 gender_child2 gender_child3
1         30 1998-11-26 2000-01-29         NA             1             2            NA
2         27 1996-06-22         NA         NA             2            NA            NA
3         26 2002-07-11 2004-04-05 2007-09-02             2             2             1
4         32 2004-10-10 2009-08-27 2012-07-21             1             1             1
5         29 2000-12-05 2005-02-28         NA             2             1            NA")
(children.melt <- melt(family.dt, measure = patterns(
  dob="^dob", gender="^gender"
), na.rm=TRUE, variable.factor=FALSE))

## nc::field can be used to define group name and pattern at the
## same time, to avoid repetitive code.
(children.nc <- nc::capture_melt_multiple(
  family.dt,
  column="[^_]+",
  "_",
  nc::field("child", "", "[1-3]"),
  na.rm=TRUE))

## reshape works too.
stats::reshape(
  family.dt,
  varying=grep("child", names(family.dt)),
  direction="long",
  sep="_",
  timevar="child.str")


## -----------------------------------------------------------------------------

## Comparison with base R. 1. mfrow means parts on rows, mfcol means
## parts on columns. 2. same number of lines of code. 3. nc/ggplot2
## code has more names and fewer numbers.
imat <- as.matrix(iris[, 1:4])
ylim <- range(table(imat))
xlim <- range(imat)
old.par <- par(mfcol=c(2,2), mar=c(2,2,1,1))
for(col.i in 1:ncol(imat)){
  hist(
    imat[, col.i],
    breaks=seq(xlim[1], xlim[2], by=0.1),
    ylim=ylim,
    main=colnames(imat)[col.i])
}


## -----------------------------------------------------------------------------

pen.peaks.wide <- data.table::data.table(
  data.set=c("foo", "bar"),
  "10.1"=c(5L, 10L),
  "0.3"=c(26L, 39L))
pen.peaks.gather <- if(requireNamespace("tidyr"))tidyr::gather(
  pen.peaks.wide,
  "penalty",
  "peaks",
  -1,
  convert=TRUE)
str(pen.peaks.gather)

pen.peaks.nc <- nc::capture_melt_single(
  pen.peaks.wide,
  penalty="^[0-9.]+", as.numeric,
  value.name="peaks")
str(pen.peaks.nc)

pen.peaks.pivot <- if(requireNamespace("tidyr"))try(tidyr::pivot_longer(
  pen.peaks.wide,
  -1,
  names_to="penalty",
  names_transform=list(penalty=as.numeric),
  values_to="peaks"))
str(pen.peaks.pivot)

varying <- 2:3
pen.peaks.reshape.times <- stats::reshape(
  pen.peaks.wide,
  direction="long",
  varying=varying,
  times=as.numeric(names(pen.peaks.wide)[varying]),
  v.names="peaks",
  timevar="penalty")
str(pen.peaks.reshape.times)

pen.peaks.renamed <- pen.peaks.wide
names(pen.peaks.renamed) <- paste0(ifelse(
  grepl("^[0-9]", names(pen.peaks.wide)),
  "peaks_", ""),
  names(pen.peaks.wide))
pen.peaks.reshape.sep <- stats::reshape(
  pen.peaks.renamed,
  direction="long",
  varying=varying,
  sep="_",
  timevar="penalty")
str(pen.peaks.reshape.sep)


## -----------------------------------------------------------------------------
peaks.csv <- system.file(
  "extdata", "RD12-0002_PP16HS_5sec_GM_F_1P.csv",
  package="nc", mustWork=TRUE)
peaks.wide <- data.table::fread(peaks.csv)
tidyr.long <- tidyr::pivot_longer(
  peaks.wide,
  grep(" [0-9]", names(peaks.wide)),
  names_pattern = "(.*) ([0-9]+)",
  names_to = c(".value", "peak"),
  names_transform = list(peak=as.integer))
peaks.tall <- nc::capture_melt_multiple(
  peaks.wide,
  column=".*",
  " ",
  peak="[0-9]+", as.integer)
options(old.par)

bg.vec <- c("H3K4me3_XJ_immune/2/counts/McGill0024.bedGraph.gz","H3K4me3_TDH_immune/9/counts/McGill0024.bedGraph.gz","H3K36me3_AM_immune/9/counts/McGill0101.bedGraph.gz","H3K36me3_TDH_other/1/counts/McGill0019.bedGraph.gz")
prefix <- "https://rcdata.nau.edu/genomic-ml/chip-seq-chunk-db/"
for(suffix in bg.vec){
  local.dir <- "~/R/nc/inst/extdata/chip-seq-chunk-db"
  local.file <- file.path(local.dir, suffix)
  if(!file.exists(local.file)){
    dir.create(dirname(local.file), showWarnings = FALSE, recursive = TRUE)
    u <- paste0(prefix, suffix)
    download.file(u, local.file)
  }
}

capture_first_vec <- structure(function # Capture first match in each character vector element
### Use a regular expression (regex) with capture groups to extract
### the first matching text from each of several subject strings. For
### all matches in one multi-line text file or string use
### capture_all_str. For the first match in every row of a data.frame,
### using a different regex for each column, use capture_first_df. For
### reading regularly named files, use capture_first_glob. For
### matching column names in a wide data frame and then
### melting/reshaping those columns to a taller/longer data frame, see
### capture_melt_single and capture_melt_multiple. To simplify the
### definition of the regex you can use field, quantifier, and
### alternatives.
(...,
### subject, name1=pattern1, fun1, etc. The first argument must be a
### character vector of length>0 (subject strings to parse with a
### regex). Arguments after the first specify the regex/conversion and
### must be string/list/function. All character strings are pasted
### together to obtain the final regex used for matching. Each
### string/list with a named argument in R becomes a capture group in
### the regex, and the name is used for the corresponding column of
### the output data table. Each function must be un-named, and is used
### to convert the previous capture group. Each un-named list becomes
### a non-capturing group. Elements in each list are parsed
### recursively using these rules.
  nomatch.error=getOption("nc.nomatch.error", TRUE),
### if TRUE (default), stop with an error if any subject does not
### match; otherwise subjects that do not match are reported as
### missing/NA rows of the result.
  engine=getOption("nc.engine", "PCRE"),
### character string, one of PCRE, ICU, RE2
  type.convert=getOption("nc.type.convert", FALSE)
### Default conversion function, which will be used on each capture
### group, unless a specific conversion is specified for that
### group. If TRUE, use utils::type.convert; if FALSE, use
### base::identity; otherwise must be a function of at least one
### argument (character), returning an atomic vector of the same
### length.
){
  L <- subject_var_args(..., type.convert=type.convert)
  subject.vec <- L[["subject"]]
  stop_for_engine(engine)
  ##alias<< nc
  stop_for_na <- function(no.match){
    if(isTRUE(nomatch.error) && any(no.match)){
      no.match.i <- which(no.match)
      stop(domain=NA, gettextf("subject(s) %s (%d total) did not match regex below; to output missing rows use nomatch.error=FALSE
%s", collapse_some(no.match.i), length(no.match.i), L[["pattern"]]))
    }
  }
  m <- if(engine=="PCRE"){
    vec.with.attrs <- try_or_stop_print_pattern({
      regexpr(L[["pattern"]], subject.vec, perl=TRUE)
    }, L[["pattern"]], engine)
    make.na <- vec.with.attrs == -1 | is.na(subject.vec)
    stop_for_na(make.na)
    first <- attr(vec.with.attrs, "capture.start")
    first[make.na] <- NA
    last <- attr(vec.with.attrs, "capture.length")-1+first
    last[make.na] <- NA
    subs <- substring(subject.vec, first, last)
    matrix(subs, length(subject.vec))
  }else{
    match.fun <- if(engine=="ICU"){
      stringi::stri_match_first_regex
    }else{
      re2::re2_match
    }
    match.mat <- try_or_stop_print_pattern({
      match.fun(subject.vec, L[["pattern"]])
    }, L[["pattern"]], engine)
    only_captures(match.mat, stop_for_na)
  }
  apply_type_funs(m, L[["fun.list"]])
### data.table with one row for each subject, and one column for each
### capture group.
}, ex=function(){

  chr.pos.vec <- c(
    "chr10:213,054,000-213,055,000",
    "chrM:111,000",
    "chr1:110-111 chr2:220-222") # two possible matches.
  ## Find the first match in each element of the subject character
  ## vector. Named argument values are used to create capture groups
  ## in the generated regex, and argument names become column names in
  ## the result.
  (dt.chr.cols <- nc::capture_first_vec(
    chr.pos.vec,
    chrom="chr.*?",
    ":",
    chromStart="[0-9,]+"))

  ## Even when no type conversion functions are specified, the result
  ## is always a data.table:
  str(dt.chr.cols)

  ## Conversion functions are used to convert the previously named
  ## group, and patterns may be saved in lists for re-use.
  keep.digits <- function(x)as.integer(gsub("[^0-9]", "", x))
  int.pattern <- list("[0-9,]+", keep.digits)
  range.pattern <- list(
    chrom="chr.*?",
    ":",
    chromStart=int.pattern,
    list( # un-named list becomes non-capturing group.
      "-",
      chromEnd=int.pattern
    ), "?") # chromEnd is optional.
  (dt.int.cols <- nc::capture_first_vec(
    chr.pos.vec, range.pattern))

  ## Conversion functions used to create non-char columns.
  str(dt.int.cols)

  ## NA used to indicate no match or missing subject.
  na.vec <- c(
    "this will not match",
    NA, # neither will this.
    chr.pos.vec)
  nc::capture_first_vec(na.vec, range.pattern, nomatch.error=FALSE)

  ## another subject from https://adventofcode.com/2024/day/14
  ## type.convert=TRUE means to use utils::type.convert as default
  ## conversion function
  pvxy.subject <- c("p=0,4 v=3,-3","p=6,3 v=-1,-3")
  nc::capture_first_vec(
    pvxy.subject,
    "p=",
    px="[0-9]",
    ",",
    py="[0-9]",
    " v=",
    vx="[-0-9]+",
    ",",
    vy="[-0-9]+",
    type.convert=TRUE)

  ## to do the same as above but with less repetition:
  g <- function(prefix,suffix)nc::group(
    name=paste0(prefix,suffix),
    "[-0-9]+")
  xy <- function(prefix)list(
    prefix,
    "=",
    g(prefix,"x"),
    ",",
    g(prefix,"y"))
  nc::capture_first_vec(
    pvxy.subject,
    xy("p"),
    " ",
    xy("v"),
    type.convert=TRUE)

  ## or use a sub-pattern list without type.convert arg:
  ipat <- list("[-0-9]+", as.integer)
  nc::capture_first_vec(
    pvxy.subject,
    "p=",
    px=ipat,
    ",",
    py=ipat,
    " v=",
    vx=ipat,
    ",",
    vy=ipat)

})

check_df_names <- function
### Check that first argument is a data frame and then call
### check_names on its names.
(...
### data frame, regex pattern args.
){
  all.args <- list(...)
  subject.df <- all.args[[1]]
  var.args <- all.args[-1]
  if(!is.data.frame(subject.df)){
    stop("first argument (subject) must be a data.frame")
  }
  subject.vec <- names(subject.df)
  ans <- check_names(subject.vec, var.args)
  ans$data <- if(is.data.table(subject.df)){
    subject.df
  } else {
    data.table(subject.df)
  }
  ans
}

check_names <- function
### Check that subject is a vector of unique names and then call
### capture_first_vec.
(subject,
### character vector, data frame column names.
  var.args
### regex pattern list.
){
  names.tab <- table(subject)
  names.rep <- names.tab[1 < names.tab]
  if(length(names.rep)){
    stop(domain=NA, gettextf("input must have columns with unique names, problems: %s", paste(names(names.rep), collapse = ", ")))
  }
  capture.args <- function(L){
    c(list(subject), L, nomatch.error=FALSE)
  }
  match.dt <- do.call(capture_first_vec, capture.args(var.args))
  missing.vec <- apply(is.na(match.dt), 1, all)
  if(all(missing.vec)){
    no.funs.args <- Filter(function(x)!is.function(x), unlist(var.args))
    no.funs.dt <- do.call(capture_first_vec, capture.args(no.funs.args))
    no.funs.missing <- apply(is.na(no.funs.dt), 1, all)
    if(all(no.funs.missing)){
      stop(domain=NA, gettextf("no column names match regex below, please change regex or column names
%s", var_args_list(var.args)[["pattern"]]))
    }else{
      no.funs.names <- subject[which(!no.funs.missing)]
      stop(domain=NA, gettextf("need to change type conversion function(s), which should return at least one non-NA, but are always returning NA, even though regex matched %s column(s): %s", length(no.funs.names), collapse_some(no.funs.names)))
    }
  }
  list(
    match.dt=match.dt,
    no.match=missing.vec)
}

collapse_some <- function
### Create character string with some or all items.
(all.vec,
### Vector of all items.
  max.first.last=5,
### Max number of items to show.
  collapse=","
### Passed to paste.
){
  all.n <- length(all.vec)
  some.vec <- if(all.n > max.first.last*2) c(
    all.vec[1:max.first.last],
    "...",
    all.vec[seq(all.n-max.first.last+1, all.n)]
  ) else all.vec
  paste(some.vec, collapse=collapse) 
### Character string formed by paste with collapse on some items of
### all.vec (first/last few items used if length is greater than
### max.first.last*2, otherwise all items).
}  

capture_first_df <- structure(function # Capture first match in columns of a data frame
### Capture first matching text from one or more character columns of
### a data frame, using a different regular expression for each
### column.
(# can NOT have subject arg outside of dots (column named subject)
  ...,
### subject data frame, colName1=list(groupName1=pattern1, fun1, etc),
### colName2=list(etc), etc. First argument must be a data frame with
### one or more character columns of subjects for matching. If the
### first argument is a data table then it will be modified using
### data.table::set (for memory efficiency, to avoid copying the whole
### data table); otherwise the input data frame will be copied to a
### new data table. Each other argument must be named using a column
### name of the subject data frame, e.g. colName1, colName2. Each
### other argument value must be a list that specifies the
### regex/conversion to use (in string/function/list format as
### documented in capture_first_vec, which is used on each named
### column), and possibly a column-specific engine to use.
  nomatch.error=getOption("nc.nomatch.error", TRUE),
### if TRUE (default), stop with an error if any subject does not
### match; otherwise subjects that do not match are reported as
### missing/NA rows of the result.
  existing.error=getOption("nc.existing.error", TRUE),
### if TRUE (default to avoid data loss), stop with an error if any
### capture groups have the same name as an existing column of
### subject.
  engine=getOption("nc.engine", "PCRE"),
### character string, one of PCRE, ICU, RE2. This engine will be used
### for each column, unless another engine is specified for that
### column in ...
  type.convert=getOption("nc.type.convert", FALSE)
### Default conversion function, which will be used on each capture
### group, unless a specific conversion is specified for that
### group. If TRUE, use utils::type.convert; if FALSE, use
### base::identity; otherwise must be a function of at least one
### argument (character), returning an atomic vector of the same
### length.
){
  all.arg.list <- list(...)
  subject <- all.arg.list[[1]]
  if(!is.data.table(subject)){
    if(!is.data.frame(subject)){
      stop("subject must be a data.frame with character columns to match")
    }
    subject <- as.data.table(subject)
  }
  col.pattern.list <- all.arg.list[-1]
  if(length(col.pattern.list)==0){
    stop(domain=NA, gettext("no patterns specified in ...; must specify subjectColName=list(groupName=pattern, etc), etc"))
  }
  valid.name <- names(col.pattern.list) %in% names(subject)
  invalid.vec <- names(col.pattern.list)[!valid.name]
  if(is.null(names(col.pattern.list)) || length(invalid.vec)){
    stop(domain=NA, gettextf("named args (%s) not found in subject column names (%s); each pattern in ... must be named using a column name of subject", paste(invalid.vec, collapse = ", "), paste(names(subject), collapse = ", ")))
  }
  if(names(all.arg.list)[[1]] != ""){
    stop("first argument (subject data.frame) should not be named")
  }
  name.tab <- table(names(col.pattern.list))
  if(any(bad <- 1 < name.tab)){
    stop(domain=NA, gettextf("each argument / subject column name should be unique, problems: %s", paste(names(name.tab)[bad], collapse = ", ")))
  }
  name.group.used <- FALSE
  for(col.name in names(col.pattern.list)){
    subject.vec <- subject[[col.name]]
    col.arg.list <- c(list(subject.vec), col.pattern.list[[col.name]])
    maybe.rep <- c("engine", "nomatch.error", "type.convert")
    to.rep <- maybe.rep[!maybe.rep %in% names(col.arg.list)]
    col.arg.list[to.rep] <- lapply(to.rep, get, environment())
    tryCatch({
      m <- do.call(capture_first_vec, col.arg.list)
    }, error=function(e){
      stop(domain=NA, gettextf("problem for subject column %s: %s", col.name, e))
    })
    new.bad <- names(m) %in% names(subject)
    if(isTRUE(existing.error) && any(new.bad)){
      stop(domain=NA, gettextf("capture group names (%s) must not conflict with existing column names (%s); fix by changing capture group names or use existing.error=FALSE to overwrite existing column names", paste(names(m), collapse = ", "), paste(names(subject), collapse = ", ")))
    }
    set(subject, j=names(m), value=m)
  }
  subject
### data.table with same number of rows as subject, with an additional
### column for each named capture group specified in ...
}, ex=function(){

  ## The JobID column can be match with a complicated regular
  ## expression, that we will build up from small sub-pattern list
  ## variables that are easy to understand independently.
  (sacct.df <- data.frame(
    JobID = c(
      "13937810_25", "13937810_25.batch",
      "13937810_25.extern", "14022192_[1-3]", "14022204_[4]"),
    Elapsed = c(
      "07:04:42", "07:04:42", "07:04:49",
      "00:00:00", "00:00:00"),
    stringsAsFactors=FALSE))

  ## Just match the end of the range.
  int.pattern <- list("[0-9]+", as.integer)
  end.pattern <- list(
    "-",
    task.end=int.pattern)
  nc::capture_first_df(sacct.df, JobID=list(
    end.pattern, nomatch.error=FALSE))

  ## Match the whole range inside square brackets.
  range.pattern <- list(
    "[[]",
    task.start=int.pattern,
    end.pattern, "?", #end is optional.
    "[]]")
  nc::capture_first_df(sacct.df, JobID=list(
    range.pattern, nomatch.error=FALSE))

  ## Match either a single task ID or a range, after an underscore.
  task.pattern <- list(
    "_",
    list(
      task.id=int.pattern,
      "|",#either one task(above) or range(below)
      range.pattern))
  nc::capture_first_df(sacct.df, JobID=task.pattern)

  ## Match type suffix alone.
  type.pattern <- list(
    "[.]",
    type=".*")
  nc::capture_first_df(sacct.df, JobID=list(
    type.pattern, nomatch.error=FALSE))

  ## Match task and optional type suffix.
  task.type.pattern <- list(
    task.pattern,
    type.pattern, "?")
  nc::capture_first_df(sacct.df, JobID=task.type.pattern)

  ## Match full JobID and Elapsed columns.
  nc::capture_first_df(
    sacct.df,
    JobID=list(
      job=int.pattern,
      task.type.pattern),
    Elapsed=list(
      hours=int.pattern,
      ":",
      minutes=int.pattern,
      ":",
      seconds=int.pattern))

  ## If input is data table then it is modified for memory efficiency,
  ## to avoid copying entire table.
  sacct.DT <- data.table::as.data.table(sacct.df)
  nc::capture_first_df(sacct.df, JobID=task.pattern)
  sacct.df #not modified.
  nc::capture_first_df(sacct.DT, JobID=task.pattern)
  sacct.DT #modified!

})

alternatives <- structure(function
### Make a pattern that matches one of the specified alternatives. The
### altlist function can be helpful for defining named sub-patterns
### that are used in several alternatives.
(...
### Each argument is a different alternative pattern.
){
  in.list <- list(...)
  if(length(in.list) < 2){
    stop("alternatives should have at least two arguments")
  }
  out.list <- in.list[1]
  for(i in 2:length(in.list)){
    out.list <- c(out.list, c("|", in.list[i]))
  }
  out.list
### Pattern list.
}, ex=function(){

  ## simple example.
  subject <- c("foooo1", "barrr2")
  str(foo.or.bar <- nc::alternatives(bar="bar+", foo="fo+"))
  str(foo.or.bar <- list(bar="bar+", "|", foo="fo+"))#same
  nc::capture_first_vec(subject, foo.or.bar, number="[12]")

  ## More complicated regular expression for matching the JobID column
  ## of SLURM sacct output.
  JobID <- c(
    "13937810_25", "13937810_25.batch",
    "13937810_25.extern", "14022192_[1-3]", "14022204_[4]")
  int.pattern <- list("[0-9]+", as.integer)
  ## Match the whole range inside square brackets.
  range.pattern <- list(
    "[[]",
    task.start=int.pattern,
    nc::quantifier("-", task.end=int.pattern, "?"),
    "[]]")
  nc::capture_first_vec(JobID, range.pattern, nomatch.error=FALSE)

  ## Match either a single task ID or a range, after an underscore.
  task.pattern <- list(job="[0-9]+", "_", nc::alternatives(
    task.id=int.pattern,
    range.pattern))
  nc::capture_first_vec(JobID, task.pattern)

})

altlist <- structure(function
### Create a named list containing named patterns, useful for creating
### a named list of named sub-pattern groups to be used with
### alternatives. This function is used to implement the more
### user-friendly alternatives_with_shared_groups, which should be
### preferred.
(...
### Named patterns which will be used as sub-pattern groups in alternatives.
){
  L <- list(...)
  if(is.null(names(L)) || any(names(L)=="")){
    stop("all arguments to altlist must be named")
  }
  for(i in seq_along(L)){
    L[[i]] <- structure(list(L[[i]]), names=names(L)[[i]])
  }
  L
### Named list of patterns to be used for constructing alternatives
### using base::with, see examples.
}, ex=function(){

  ## Example 1: matching dates in different formats, but always same
  ## type in each alternative.
  subject.vec <- c("mar 17, 1983", "26 sep 2017", "17 mar 1984")
  pat.list <- nc::altlist(month="[a-z]{3}", day="[0-9]{2}", year="[0-9]{4}")
  pattern <- with(pat.list, nc::alternatives(
    american=list(month, " ", day, ", ", year),
    european=list(day, " ", month, " ", year)))
  match.dt <- nc::capture_first_vec(subject.vec, pattern)
  print(match.dt, class=TRUE)
  match.dt[, data.table::as.IDate(paste0(month, day, year), format="%b%d%Y")]

  ## Example 2: matching dates in different formats, but with
  ## different types in different alternatives.
  subject.vec <- c("3/17/1983", "26 sep 2017")
  month2int <- c(#this approach is locale-indepdendent.
    jan=1L, feb=2L, mar=3L, apr=4L,  may=5L,  jun=6L,
    jul=7L, aug=8L, sep=9L, oct=10L, nov=11L, dec=12L)
  pat.list <- nc::altlist(
    day=list("[0-9]{2}", as.integer),
    year=list("[0-9]{4}", as.integer))
  pattern <- with(pat.list, nc::alternatives(
    american=list(month="[0-9]", as.integer, "/", day, "/", year),
    european=list(
      day, " ", month="[a-z]{3}", function(m)month2int[m], " ", year)))
  match.dt <- nc::capture_first_vec(subject.vec, pattern)
  print(match.dt, class=TRUE)

})

alternatives_with_shared_groups <- structure(function
### Create a pattern which matches alternatives with common
### sub-pattern groups.
(...
### named arguments are sub-pattern groups, un-named arguments are
### alternative patterns which can refer to argument names.
){
  mcall <- match.call()
  args.list <- as.list(mcall[-1])
  is.named <- names(args.list)!=""
  if(all(!is.named)){
    stop("alternatives_with_shared_groups must have at least one named argument; each named argument should define a sub-pattern/group")
  }
  if(all(is.named)){
    stop("alternatives_with_shared_groups must have at least one un-named argument; each un-named argument should define an alternative pattern")
  }
  altlist.result <- do.call(altlist, args.list[is.named])
  with(altlist.result, do.call(alternatives, args.list[!is.named]))
### Pattern created by first using altlist on named arguments to
### create a list of sub-patterns, then using alternatives on un-named
### arguments (evaluated using the names defined in the sub-pattern
### list).
}, ex=function(){

  ## Example 0: matching family and given names.
  nc::capture_first_vec(
    c("Toby Dylan Hocking","Hocking, Toby Dylan"),
    nc::alternatives_with_shared_groups(
      family="[A-Z][a-z]+",
      given="[^,]+",
      list(given, " ", family),
      list(family, ", ", given)
    )
  )

  ## Example 1: matching dates in different formats, but always same
  ## type in each alternative.
  subject.vec <- c("mar 17, 1983", "26 sep 2017", "17 mar 1984")
  ## Another way is with alternative regular expressions.
  pattern <- nc::alternatives_with_shared_groups(
    month="[a-z]{3}",
    day="[0-9]{2}",
    year="[0-9]{4}",
    list(american=list(month, " ", day, ", ", year)),
    list(european=list(day, " ", month, " ", year)))
  (match.dt <- nc::capture_first_vec(subject.vec, pattern))
  match.dt[, data.table::as.IDate(paste0(month, day, year), format="%b%d%Y")]
  ## american and european columns can be searched to see which
  ## alternative matched.
  match.dt[european!=""]

  ## Example 2: matching dates in different formats, but with
  ## different types in different alternatives.
  subject.vec <- c("3/17/1983", "26 sep 2017")
  month2int <- c(#this approach is locale-indepdendent.
    jan=1L, feb=2L, mar=3L, apr=4L,  may=5L,  jun=6L,
    jul=7L, aug=8L, sep=9L, oct=10L, nov=11L, dec=12L)
  pattern <- nc::alternatives_with_shared_groups(
    day=list("[0-9]{2}", as.integer),
    year=list("[0-9]{4}", as.integer),
    list(month="[0-9]", as.integer, "/", day, "/", year),
    list(day, " ", month="[a-z]{3}", function(m)month2int[m], " ", year))
  match.dt <- nc::capture_first_vec(subject.vec, pattern)
  print(match.dt, class=TRUE)

  ## Example 3: three alternatives with four groups each.
  subject.vec <- c(
    "EUR 5.00 Theft in delivery inserted in wire transfer 11/02/2021",
    "EUR 50.00 - Refund for theft in delivery - 30/07/2020",
    "EUR68.50 - Refund for theft in delivery 02/07/2020",
    "45.00 EUR 29/10/2020 Refund for theft in delivery",
    "53.00\u20ac Refund for theft in delivery 24/09/2020")
  sep <- function(x, y, ...){
    if(missing(y)){
      list("^", x, "$")
    }else{
      sep(list(x, list(" - | |"), y), ...)
    }
  }
  pattern <- nc::alternatives_with_shared_groups(
    currency="EUR|\u20ac",
    amount=list("[0-9.]+", as.numeric),
    reason="[A-Za-z ]+?",
    date=list(
      "[0-9]{2}/[0-9]{2}/[0-9]{4}",
      function(d)data.table::as.IDate(d, format="%d/%m/%Y")),
    sep(currency, amount, reason, date),
    sep(amount, currency, date, reason),
    sep(amount, currency, reason, date))
  match.dt <- nc::capture_first_vec(subject.vec, pattern)
  print(match.dt, class=TRUE)

})

subject_var_args <- function
### Parse the complete argument list including subject.
(...,
### subject, regex/conversion.
  type.convert=getOption("nc.type.convert", FALSE)
### passed to var_args_list.
){
  all.arg.list <- list(...)
  first.name <- names(all.arg.list[1])
  no.name <- identical(first.name, "") || identical(first.name, NULL)
  if(!no.name){
    stop(domain=NA, gettextf("first argument is named %s but must NOT be named; please include the subject to match as the first argument, with no name", first.name))
  }
  subject <- all.arg.list[[1]]
  stop_for_subject(subject)
  out.list <- var_args_list(all.arg.list[-1], type.convert=type.convert)
  out.list$subject <- subject
  out.list
### Result of var_args_list plus subject.
}

var_args_list <- structure(function
### Parse the variable-length argument list used in capture_first_vec,
### capture_first_df, and capture_all_str. This function is mostly
### intended for internal use, but is useful if you want to see the
### regex pattern generated by the variable argument syntax.
(...,
### character vectors (for regex patterns) or functions (which specify
### how to convert extracted character vectors to other types). All
### patterns must be character vectors of length 1. If the pattern is
### a named argument in R, it becomes a capture group in the regex
### pattern. All patterns are pasted together to obtain the final
### pattern used for matching. Each named pattern may be followed by
### at most one function which is used to convert the previous named
### pattern. Patterns may also be lists, which are parsed recursively
### for convenience.
  type.convert=getOption("nc.type.convert", FALSE)
### Default conversion function, which will be used on each capture
### group, unless a specific conversion is specified for that
### group. If TRUE, use utils::type.convert; if FALSE, use
### base::identity; otherwise must be a function of at least one
### argument (character), returning an atomic vector of the same
### length.
){
  var.arg.list <- list(...)
  fun.list <- list()
  pattern.list <- list()
  has.name <- FALSE
  group.i <- NULL
  while(length(var.arg.list)){
    var.arg <- var.arg.list[[1]]
    pattern.name <- names(var.arg.list)[1]
    valid.name <- if(is.character(pattern.name)){
      if(is.na(pattern.name))stop("group name must not be missing")
      0 < nchar(pattern.name)
    }else FALSE
    group.start <- if(valid.name){
      if(is.function(var.arg)){
        stop(domain=NA, gettextf("functions must not be named, problem: %s", pattern.name))
      }
      group.i <- length(fun.list)+1L
      fun.list[[group.i]] <- if(isTRUE(type.convert)){
        function(x)utils::type.convert(x,as.is=TRUE)
      }else if(isFALSE(type.convert)){
        identity
      }else if(is.function(type.convert)){
        type.convert
      }else stop("type.convert should be either TRUE or FALSE or a function")
      names(fun.list)[[group.i]] <- pattern.name
      has.name <- TRUE
      "("
    }else{
      "(?:"
    }
    var.arg.list <- var.arg.list[-1]
    if(is.character(var.arg)){
      if(length(var.arg) != 1){
        stop(
          "patterns must be character vectors of length 1")
      }
      if(is.na(var.arg)){
        stop("patterns must not be missing/NA")
      }
      if(is.character(names(var.arg))){
        dquote <- function(chr){
          out <- NULL
          con <- textConnection("out", "w", local=TRUE)
          dput(chr, con)
          close(con)
          out
        }
        list.code <- paste0(
          "list(",
          dquote(names(var.arg)),
          "=",
          dquote(unname(var.arg)),
          ")")
        stop(domain=NA, gettextf("pattern string must not be named; did you mean %s", list.code))
      }
      pattern.list[[length(pattern.list)+1L]] <- if(valid.name){
        paste0(group.start, var.arg, ")")
      }else{
        var.arg
      }
    }else if(is.function(var.arg)){
      if(is.null(group.i)){
        stop(domain=NA, gettext("too many functions; up to one function may follow each named pattern"))
      }
      fun.list[[group.i]] <- var.arg
      group.i <- NULL
    }else if(is.list(var.arg)){
      var.arg.list <- c(group.start, var.arg, ")", var.arg.list)
    }else{
      stop(domain=NA, gettext("arguments must be character (subject/patterns), functions (for converting extracted character vectors to other types), or list (parsed recursively)"))
    }
  }
  if(!has.name){
    stop("must have at least one named argument (capture group)")
  }
  ##value<< a list with two named elements
  list(
    fun.list=##<< list of conversion functions with names corresponding to capture group(s)
      if(length(fun.list))fun.list,
    pattern=##<< regular expression string with capture group(s)
      paste(pattern.list, collapse="")
  )
  ##end<<
}, ex=function(){

  pos.pattern <- list("[0-9]+", as.integer)
  nc::var_args_list(
    chrom="chr.*?",
    ":",
    chromStart=pos.pattern,
    list(
      "-",
      chromEnd=pos.pattern
    ), "?")

})

apply_type_funs <- function
### Convert columns of match.mat using corresponding functions from
### fun.list, then handle any duplicate capture group names.
(match.mat,
### Character matrix (matches X groups).
  fun.list
### Named list of functions to apply to captured groups. If there are
### any duplicate names, they must be in alternatives (only one match
### per unique group name, otherwise error).
){
  stopifnot(is.character(match.mat))
  stopifnot(is.matrix(match.mat))
  if(length(fun.list) < ncol(match.mat)){
    stop(domain=NA, gettext("regex contains more groups than names; please remove literal groups (parentheses) from the regex pattern, and use named arguments in R code instead"))
  }
  colnames(match.mat) <- names(fun.list)
  dt <- data.table(match.mat)
  for(col.i in seq_along(fun.list)){
    type.fun <- fun.list[[col.i]]
    if(!identical(type.fun, identity)){
      col.for.err <- paste0(col.i, "(", names(fun.list)[[col.i]], ")")
      match.vec <- match.mat[, col.i]
      tryCatch({
        fun.result <- type.fun(match.vec)
      }, error=function(e){
        stop(domain=NA, gettextf("type conversion functions should take one argument (character vector of captured text) and return an atomic vector of the same size; function for group %s raised an error: %s", col.for.err, e$message))
      })
      if(!is.atomic(fun.result)){
        stop(domain=NA, gettextf("each type conversion function must return an atomic vector, but function for group %s did not", col.for.err))
      }
      if(length(fun.result) != nrow(match.mat)){
        stop(domain=NA, gettextf("type conversion function for group %s returned vector of length %s but expected length %s", col.for.err, length(fun.result), nrow(match.mat)))
      }
      if(any(is.na(match.vec) & !is.na(fun.result))){
        stop(domain=NA, gettextf("a non-match(NA) was converted to a match(non-NA) by the conversion function in group %s; please fix conversion function", col.for.err))
      }
      set(dt, j=col.i, value=fun.result)
    }
  }
  ## handle duplicates (error or delete columns).
  is.match <- !is.na(match.mat) & match.mat!=""
  dt.type.vec <- sapply(dt, typeof)
  name.tab <- table(names(fun.list))
  dup.name.vec <- names(name.tab)[1 < name.tab]
  remove.col.list <- list()
  for(dup.name in dup.name.vec){
    dup.col.indices <- which(names(fun.list)==dup.name)
    dup.type.tab <- table(dt.type.vec[dup.col.indices])
    if(1 < length(dup.type.tab)){
      stop(domain=NA, gettextf("capture groups with identical names should have conversion functions that all return the same type; problem group name=%s has types %s", dup.name, paste(names(dup.type.tab), collapse = ",")))
    }
    is.match.name <- is.match[, dup.col.indices, drop=FALSE]
    if(any(1 < rowSums(is.match.name))){
      stop(domain=NA, gettextf("duplicate capture group names are only allowed in alternatives, problem: %s", dup.name))
    }
    alt.i.vec <- as.integer(apply(is.match.name, 1, which))
    orig.i.vec <- dup.col.indices[alt.i.vec]
    ## Columns for alternatives other than the first will be removed.
    remove.col.vec <- dup.col.indices[-1]
    for(remove.col in remove.col.vec){
      remove.i <- which(remove.col == orig.i.vec)
      ## Copy values captured in this alternative to the first
      ## alternative.
      set(
        dt, i=remove.i, j=dup.col.indices[[1]],
        value=dt[[remove.col]][remove.i])
    }
    remove.col.list[[dup.name]] <- remove.col.vec
  }
  ## Remove all alternatives other than the first.
  if(length(remove.col.list))set(dt, j=unlist(remove.col.list), value=NULL)
  dt
### data.table with columns defined by calling the functions in
### fun.list on the corresponding column of match.mat. Even if
### fun.list has duplicated names, the output data.table will have
### unique column names (identically named capture groups in
### alternatives will be combined into a single output column).
}

before_match <- structure(function
### Augment pattern so that it can be used to match an entire string. Useful for complex find and replace operations, when used with capture_all_str.
(...
### pattern as in capture_first_vec.
){
  nc::alternatives(
    ## (?s) single line (dotall) makes . match anything (even a newline).
    ## Changes of these options within a group are automatically cancelled at the end of the group.
    list(before="(?s).*?", match=list(...)),
    ## from https://www.pcre.org/current/doc/html/pcre2syntax.html#TOC1
    ##   $           end of subject
    ##                 also before newline at end of subject
    ##                 also before internal newline in multiline mode
    ##   \Z          end of subject
    ##                 also before newline at end of subject
    ##   \z          end of subject
    list(before="(?s).+?", "\\z")) #because $ does not work.
### Pattern with two new groups, before (everything before ...) and match (text matching ... or blank for end of string).
}, ex=function(){

  markdown_link <- list(
    "\\[",
    title=".*?",
    "\\]\\(",
    url="http.*?",
    "\\)")
  markdown_subject <- "before [foo](http) between [bar text](http) after\n"
  nc::capture_all_str(markdown_subject, markdown_link)
  
  before_link <- nc::before_match(markdown_link)
  (all_dt <- nc::capture_all_str(markdown_subject, before_link))

  ## before + match = full subject.
  identical(all_dt[, paste(paste0(before, match), collapse="")], markdown_subject)

  ## replace with org link.
  all_dt[, paste(paste0(before, ifelse(
    match=="", "", sprintf("[[%s][%s]]", url, title)
  )), collapse="")]

  ## also works with no extra text before/after match.
  nc::capture_all_str("[foo](http) between [bar text](http)", before_link)

})

field <- structure(function # Capture a field
### Capture a field with a pattern of the form
### list("field.name", between.pattern,
### field.name=list(...)) -- see examples.
(field.name,
### Field name, used as a pattern and as a capture group (output
### column) name.
  between.pattern,
### Pattern to match after field.name but before the field value.
  ...
### Pattern(s) for matching field value.
){
  if(!(is.character(field.name) && length(field.name)==1))stop(
    "first argument of field must be character string (field name)")
  list(
    field.name,
    between.pattern,
    group(field.name, ...))
### Pattern list which can be used in capture_first_vec,
### capture_first_df, or capture_all_str.
}, ex=function(){

  ## Two ways to create the same pattern.
  str(list("Alignment ", Alignment="[0-9]+"))
  ## To avoid typing Alignment twice use:
  str(nc::field("Alignment", " ", "[0-9]+"))

  ## An example with lots of different fields.
  info.txt.gz <- system.file(
    "extdata", "SweeD_Info.txt.gz", package="nc")
  info.vec <- readLines(info.txt.gz)
  info.vec[24:40]
  ## For each Alignment there are many fields which have a similar
  ## pattern, and occur in the same order. One way to capture these
  ## fields is by coding a pattern that says to look for all of those
  ## fields in that order. Each field is coded using this helper
  ## function.
  g <- function(name, fun=identity, suffix=list()){
    list(
      "\t+",
      nc::field(name, ":\t+", ".*"),
      fun,
      suffix,
      "\n+")
  }
  nc::capture_all_str(
    info.vec,
    nc::field("Alignment", " ", "[0-9]+"),
    "\n+",
    g("Chromosome"),
    g("Sequences", as.integer),
    g("Sites", as.integer),
    g("Discarded sites", as.integer),
    g("Processing", as.integer, " seconds"),
    g("Position", as.integer),
    g("Likelihood", as.numeric),
    g("Alpha", as.numeric))

  ## Another example where field is useful.
  trackDb.txt.gz <- system.file(
    "extdata", "trackDb.txt.gz", package="nc")
  trackDb.vec <- readLines(trackDb.txt.gz)
  cat(trackDb.vec[101:115], sep="\n")
  int.pattern <- list("[0-9]+", as.integer)
   cell.sample.type <- list(
    cellType="[^ ]*?",
    "_",
    sampleName=list(
      "McGill",
      sampleID=int.pattern),
    dataType="Coverage|Peaks")
  ## Each block in the trackDb file begins with track, followed by a
  ## space, followed by the track name. That pattern is coded below,
  ## using field:
  track.pattern <- nc::field(
    "track",
    " ",
    cell.sample.type,
    "|",
    "[^\n]+")
  nc::capture_all_str(trackDb.vec, track.pattern)

  ## Each line in a block has the same structure (field name, space,
  ## field value). Below we use the field function to extract the
  ## color line, along with columns for each of the three channels
  ## (red, green, blue).
  any.lines.pattern <- "(?:\n[^\n]+)*"
  nc::capture_all_str(
    trackDb.vec,
    track.pattern,
    any.lines.pattern,
    "\\s+",
    nc::field(
      "color", " ",
      red=int.pattern, ",",
      green=int.pattern, ",",
      blue=int.pattern))

})

only_captures <- function
### Extract capture group columns from match.mat and assign optional
### groups to "".
(match.mat,
### character matrix.
  stop.fun
### function to call on logical matrix of missing indicators.
){
  group.mat <- match.mat[, -1, drop=FALSE]
  missing.match <- is.na(match.mat[,1])
  stop.fun(missing.match)
  group.mat[is.na(group.mat) & !missing.match] <- "" #optional groups
  group.mat
}

capture_longer_spec <- structure(function
### Create a spec data table for input to tidyr::pivot_longer_spec.
(data,
### Data table to reshape (actually the data are ignored, and only the
### column names are used).
  ...,
### Regex and conversion as described in capture_first_vec. This is
### processed by measure so if "column" is used as an argument name
### then there will be multiple output columns in the reshaped data
### (analogous to names_to=".value" in tidyr::pivot_longer).
  values_to="value"
### string to use for name of value column in output (only used if
### there is a single output column in the reshaped data).
){
  .name <- NULL
  ## above to avoid CRAN check NOTE.
  cols <- names(data)
  vec.or.list <- nc::measure(..., cols=cols)
  vtab <- attr(vec.or.list, "variable_table")
  if(is.integer(vec.or.list)){
    vec.or.list <- structure(list(vec.or.list), names=values_to)
  }
  spec.list <- list()
  for(.value in names(vec.or.list)){
    index.vec <- vec.or.list[[.value]]
    spec.list[[.value]] <- data.table(.name=cols[index.vec], vtab, .value)
  }
  do.call(rbind, spec.list)[!is.na(.name)]
### data table describing a reshape longer operation.
}, ex=function(){

  (one.iris <- iris[1,])
  (single.spec <- nc::capture_longer_spec(iris, part=".*", "[.]", dim=".*", values_to="cm"))
  (multiple.spec <- nc::capture_longer_spec(iris, part=".*", "[.]", column=".*"))
  if(requireNamespace("tidyr")){
    tidyr::pivot_longer_spec(one.iris, single.spec)
    tidyr::pivot_longer_spec(one.iris, multiple.spec)
  }
  
})

group <- structure(function # Capture group
### Create a capture group (named column in output). In the vast
### majority of patterns R arguments can/should be used to specify
### names, e.g. list(name=pattern). This is a helper function which is
### useful for programmatically creating group names (see example for
### a typical use case).
(name,
### Column name in output.
  ...
### Regex pattern(s).
){
  if(!(is.character(name) && length(name)==1))stop(
    "first argument of group must be a character string (group name)")
  pattern <- list(...)
  structure(list(pattern), names=name)
### Named list.
}, ex=function(){

  ## Three ways to create a group named data which matches zero or
  ## more non-newline characters.
  str(list(data=".*"))
  str(nc::group("data", ".*"))
  g.name <- "data"
  str(nc::group(g.name, ".*"))

  ## Data downloaded from
  ## https://en.wikipedia.org/wiki/Hindu%E2%80%93Arabic_numeral_system
  numerals <- system.file(
    "extdata", "Hindu-Arabic-numerals.txt.gz", package="nc")

  ## Use engine="ICU" for unicode character classes
  ## http://userguide.icu-project.org/strings/regexp e.g. match any
  ## character with a numeric value of 2 (including japanese etc).
  if(requireNamespace("stringi"))
    nc::capture_all_str(
    numerals,
    " ",
    two="[\\p{numeric_value=2}]",
    " ",
    engine="ICU")

  ## Create a table of numerals with script names.
  digits.pattern <- list()
  for(digit in 0:9){
    digits.pattern[[length(digits.pattern)+1]] <- list(
      "[|]",
      nc::group(paste(digit), "[^{|]+"),
      "[|]")
  }
  nc::capture_all_str(
    numerals,
    "\n",
    digits.pattern,
    "[|]",
    " *",
    "\\[\\[",
    name="[^\\]|]+")

})

stop_for_capture_same_as_id <- function
### Error if capture names same as id.vars.
(capture.vars,
### character vector of capture column names.
  id.vars
### character vector of id column names.
){
  id.captures <- id.vars[id.vars %in% capture.vars]
  if(length(id.captures)){
    stop(domain=NA, gettextf("some capture group names (%s) are the same as input column names that did not match the pattern; please change either the pattern or the capture group names so that all output column names will be unique", paste(id.captures, collapse = ", ")))
  }
}

stop_for_subject <- function
### Error if subject incorrect type.
(subject
### character vector.
){
  if(!(
    is.character(subject) &&
    0 < length(subject)
  )){
    stop(domain=NA, gettextf("subject has class=%s and length=%s but should be a character vector with length>0", paste(class(subject), collapse = ","), length(subject)))
  }
}

stop_for_engine <- function
### Stop if specified engine is not available.
(engine
### character string: PCRE, RE2, or ICU.
){
  pkg.vec <- c(
    PCRE="base",
    RE2="re2",
    ICU="stringi")
  if(!(
    is.character(engine) &&
    length(engine)==1 &&
    !is.na(engine) &&
    engine %in% names(pkg.vec)
  )){
    stop(domain=NA, gettextf("engine must be character string, one of: %s", paste(names(pkg.vec), collapse = ", ")))
  }
  pkg <- pkg.vec[[engine]]
  if(!requireNamespace(pkg, quietly=TRUE)){
    install.code <- sprintf("install.packages('%s')", pkg)
    stop(domain=NA, gettextf("%s engine not available; try %s", engine, install.code))
  }
### character string.
}

available.engines <- character()
for(e in c("RE2", "ICU", "PCRE")){
  is.available <- tryCatch({
    stop_for_engine(e)
    TRUE
  }, error=function(msg){
    FALSE
  })
  if(is.available){
    available.engines[[e]] <- e
  }
}
test_engines <- function(desc, ...){
  for(e in available.engines){
    old.opt <- options(nc.engine=e)
    test_that(paste(e, desc), ...)
    options(old.opt)
  }
}

library(nc)
library(testthat)
context("df")
source(system.file("test_engines.R", package="nc", mustWork=TRUE), local=TRUE)

if(requireNamespace("stringi"))test_engines("default/specified engine respected", {
  subject.df <- data.frame(f="foo", b="bar")
  match.dt <- capture_first_df(
    subject.df,
    f=list(icu="[\\p{Letter}]"),#only works with ICU
    b=list(re2=".(?R)?", engine="PCRE"),#only works with PCRE
    engine="ICU")
  expect_identical(match.dt[["icu"]], "f")
  expect_identical(match.dt[["re2"]], "bar")
})

subject.df <- data.frame(
  JobID=c(
    "13937810_25",
    "13937810_25.batch",
    "13937810_25.extern",
    "14022192_[1-3]",
    "14022204_[4]"),
  subject=c(
    ten="chr10:213,054,000-213,055,000",
    chrNA="chrNA:111,000-222,000",
    no.match="foo bar",
    missing=NA,
    two="chr1:110-111 chr2:220-222"),
  stringsAsFactors=FALSE)
range.pattern <- list(
  "\\[",
  task1="[0-9]+", as.integer,
  "(?:-",#begin optional end of range.
  taskN="[0-9]+", as.integer,
  ")?", #end is optional.
  "\\]")
test_engines("capture_first_df returns data.table", {
  match.dt <- capture_first_df(
    subject.df,
    JobID=list(
      job="[0-9]+", as.integer,
      "_",
      "(?:",#begin alternate
      task="[0-9]+", as.integer,
      "|",#either one task(above) or range(below)
      range.pattern,
      ")",#end alternate
      "(?:[.]",
      type=".*",
      ")?"),
    subject=list(
      nomatch.error=FALSE,
      chrom="chr.*?",
      ":",
      chromStart=".*?",
      "-",
      chromEnd="[0-9,]*"))
  expect_is(match.dt, "data.table")
  expect_identical(names(match.dt), c(
    "JobID", "subject",
    "job", "task", "task1", "taskN", "type",
    "chrom", "chromStart", "chromEnd"))
  expect_identical(match.dt$job, as.integer(c(
    13937810, 13937810, 13937810, 14022192, 14022204)))
  expect_identical(match.dt$task, as.integer(c(
    25, 25, 25, NA, NA)))
  expect_identical(match.dt$task1, as.integer(c(
    NA, NA, NA, 1, 4)))
  expect_identical(match.dt$taskN, as.integer(c(
    NA, NA, NA, 3, NA)))
  expect_identical(match.dt$type, c(
    "", "batch", "extern", "", ""))
  expect_identical(match.dt$chrom, c(
    "chr10", "chrNA", NA, NA, "chr1"))
  expect_identical(match.dt$chromStart, c(
    "213,054,000", "111,000", NA, NA, "110"))
  expect_identical(match.dt$chromEnd, c(
    "213,055,000", "222,000", NA, NA, "111"))
})

range.square <- list(
  "[[]",
  task1="[0-9]+", as.integer,
  "(?:-",#begin optional end of range.
  taskN="[0-9]+", as.integer,
  ")?", #end is optional.
  "[]]")
test_engines("capture_first_df square brackets pattern", {
  full.square <- list(
    job="[0-9]+", as.integer,
    "_",
    "(?:",#begin alternate
    task="[0-9]+", as.integer,
    "|",#either one task(above) or range(below)
    range.square,
    ")",#end alternate
    "(?:[.]",
    type=".*",
    ")?")
  if(identical(getOption("nc.engine"), "ICU")){
    expect_error({
      capture_first_df(subject.df, JobID=full.square)
    }, "when matching pattern above with ICU engine")
  }else{
    match.dt <- capture_first_df(subject.df, JobID=full.square)
    expect_identical(names(match.dt), c(
      "JobID", "subject",
      "job", "task", "task1", "taskN", "type"))
    expect_identical(match.dt$job, as.integer(c(
      13937810, 13937810, 13937810, 14022192, 14022204)))
    expect_identical(match.dt$task, as.integer(c(
      25, 25, 25, NA, NA)))
    expect_identical(match.dt$task1, as.integer(c(
      NA, NA, NA, 1, 4)))
    expect_identical(match.dt$taskN, as.integer(c(
      NA, NA, NA, 3, NA)))
    expect_identical(match.dt$type, c(
      "", "batch", "extern", "", ""))
  }
})

no.rownames <- data.frame(
  JobID=c(
    "13937810_25",
    "13937810_25.batch",
    "13937810_25.extern",
    "14022192_[1-3]",
    "14022204_[4]"),
  position=c(
    "chr10:213,054,000-213,055,000",
    "chrNA:111,000-222,000",
    "foo bar",
    NA,
    "chr1:110-111 chr2:220-222"),
  stringsAsFactors=FALSE)
test_engines("capture_first_df returns data.frame with default rownames", {
  match.df <- capture_first_df(
    no.rownames,
    JobID=list(
      job="[0-9]+", as.integer,
      "_",
      "(?:",#begin alternate
      task="[0-9]+", as.integer,
      "|",#either one task(above) or range(below)
      range.pattern,
      ")",#end alternate
      "(?:[.]",
      type=".*",
      ")?"),
    position=list(
      nomatch.error=FALSE,
      chrom="chr.*?",
      ":",
      chromStart=".*?",
      "-",
      chromEnd="[0-9,]*"))
  expect_identical(names(match.df), c(
    "JobID", "position",
    "job", "task", "task1", "taskN", "type",
    "chrom", "chromStart", "chromEnd"))
  expect_identical(match.df$job, as.integer(c(
    13937810, 13937810, 13937810, 14022192, 14022204)))
  expect_identical(match.df$task, as.integer(c(
    25, 25, 25, NA, NA)))
  expect_identical(match.df$task1, as.integer(c(
    NA, NA, NA, 1, 4)))
  expect_identical(match.df$taskN, as.integer(c(
    NA, NA, NA, 3, NA)))
  expect_identical(match.df$type, c(
    "", "batch", "extern", "", ""))
  expect_identical(match.df$chrom, c(
    "chr10", "chrNA", NA, NA, "chr1"))
  expect_identical(match.df$chromStart, c(
    "213,054,000", "111,000", NA, NA, "110"))
  expect_identical(match.df$chromEnd, c(
    "213,055,000", "222,000", NA, NA, "111"))
  expect_identical(rownames(match.df), paste(1:5))
})

named.uniq.chr <- data.frame(
  JobID=c(
    foo="13937810_25",
    bar="13937810_25.batch",
    baz="13937810_25.extern",
    sars="14022192_[1-3]",
    last="14022204_[4]"),
  position=c(
    "chr10:213,054,000-213,055,000",
    "chrNA:111,000-222,000",
    "chr2:1-2",
    "chr3:4-5",
    "chr1:110-111 chr2:220-222"),
  stringsAsFactors=FALSE)
keep.digits <- function(x)as.integer(gsub("[^0-9]", "", x))
test_engines("two name groups not OK with named subject", {
  expect_error({
    capture_first_df(
      named.uniq.chr,
      JobID=list(
        name="[^.]+[.].|[0-9]+",
        rest=".*"),
      position=list(
        name="chr.*?",
        ":",
        chromStart=".*?", keep.digits,
        "-",
        chromEnd="[0-9,]*", keep.digits))
  }, "must not conflict with existing column names")
})

test_engines("type.convert OK inside capture_first_df list", {
  type.conv.result <- capture_first_df(
    named.uniq.chr,
    JobID=list(
      job="[0-9]+", 
      "_",
      "(?:",#begin alternate
      task="[0-9]+", 
      "|",#either one task(above) or range(below)
      range.pattern,
      ")",#end alternate
      "(?:[.]",
      type=".*", identity,
      ")?",
      type.convert=as.numeric),
    position=list(
      name="chr.*?",
      ":",
      chromStart=".*?", keep.digits,
      "-",
      chromEnd="[0-9,]*", keep.digits))
  computed.cls <- sapply(type.conv.result, class)
  expected.cls <- c(
    JobID = "character",
    position = "character",
    job = "numeric", 
    task = "numeric",
    task1 = "integer",
    taskN = "integer",
    type = "character", 
    name = "character",
    chromStart = "integer",
    chromEnd = "integer")
  expect_identical(computed.cls, expected.cls)
})

test_engines("type.convert OK as capture_first_df arg", {
  type.conv.result <- capture_first_df(
    named.uniq.chr,
    JobID=list(
      job="[0-9]+", 
      "_",
      "(?:",#begin alternate
      task="[0-9]+", 
      "|",#either one task(above) or range(below)
      range.pattern,
      ")",#end alternate
      "(?:[.]",
      type=".*", identity,
      ")?"),
    position=list(
      name="chr.*?",
      ":",
      chromStart=".*?", keep.digits,
      "-",
      chromEnd="[0-9,]*", keep.digits),
    type.convert=as.factor)
  computed.cls <- sapply(type.conv.result, class)
  expected.cls <- c(
    JobID = "character",
    position = "character",
    job = "factor", 
    task = "factor",
    task1 = "integer",
    taskN = "integer",
    type = "character", 
    name = "factor",
    chromStart = "integer",
    chromEnd = "integer")
  expect_identical(computed.cls, expected.cls)
})

test_engines("error for no pattern", {
  expect_error({
    capture_first_df(named.uniq.chr)
  }, "no patterns specified in ...")
})

test_engines("error for un-named list", {
  expect_error({
    capture_first_df(named.uniq.chr, list())
  }, "each pattern in ... must be named using a column name of subject")
})

test_engines("error for un-named list with name", {
  expect_error({
    capture_first_df(named.uniq.chr, list(foo="bar"))
  }, "each pattern in ... must be named using a column name of subject")
})

test_engines("error for un-recognized name", {
  expect_error({
    capture_first_df(named.uniq.chr, foo="bar")
  }, "each pattern in ... must be named using a column name of subject")
})

test_engines("error for non-df subject", {
  expect_error({
    capture_first_df(c("foo", "bar"), list(foo="bar"))
  }, "subject must be a data.frame with character columns to match")
})

test_engines("error for non-df subject", {
  expect_error({
    capture_first_df(c("foo", "bar"), foo="bar")
  }, "subject must be a data.frame with character columns to match")
})

test_engines("error for factor column", {
  fac.df <- data.frame(foo="bar", stringsAsFactors=TRUE)
  expect_error({
    capture_first_df(fac.df, foo=list(baz="sars"))
  }, "problem for subject column foo: Error in stop_for_subject(subject): subject has class=factor and length=1 but should be a character vector with length>0", fixed=TRUE)
})

test_engines("error for same column name twice", {
  expect_error({
    capture_first_df(
      named.uniq.chr,
      JobID=list(f="baz"),
      JobID="foo")
  }, "each argument / subject column name should be unique")
})

test_engines("error for named subject", {
  expect_error({
    capture_first_df(
      JobID=named.uniq.chr,
      JobID=list(f="[0-9]+"))
  },
  "first argument (subject data.frame) should not be named",
  fixed=TRUE)
})

in.df <- data.frame(bar="foobar", stringsAsFactors=FALSE)
test_engines("df only one group = name", {
  out.dt <- capture_first_df(
    in.df,
    bar=list(
      name="foo"))
  expect_identical(out.dt$name, "foo")
})

matching.subjects <- c(
  "chr10:213,054,000-213,055,000",
  "chrM:111,000",
  "chr1:110-111 chr2:220-222") # two possible matches.
test_engines("df subject no error if nomatch.error=TRUE and all matches", {
  subject.df <- data.frame(
    subject.col=matching.subjects, stringsAsFactors=FALSE)
  match.dt <- capture_first_df(
    subject.df,
    subject.col=list(
      nomatch.error=TRUE,
      chrom="chr.*?",
      ":",
      chromStart="[0-9,]+", keep.digits,
      list(
        "-",
        chromEnd="[0-9,]+", keep.digits
      ), "?"))
  expect_identical(
    match.dt$chromEnd,
    as.integer(c(213055000, NA, 111)))
})
chr.pos.nomatch.vec <- c(
  "chr10:213,054,000-213,055,000",
  "chrM:111,000",
  "this will not match",
  NA, # neither will this.
  "chr1:110-111 chr2:220-222") # two possible matches.
test_engines("df subject stop if nomatch.error=TRUE and no match", {
  subject.df <- data.frame(chr.pos.nomatch.vec, stringsAsFactors=FALSE)
  expect_error({
    capture_first_df(
      subject.df,
      chr.pos.nomatch.vec=list(
        nomatch.error=TRUE,
        chrom="chr.*?",
        ":",
        chromStart="[0-9,]+", keep.digits,
        list(
          "-",
          chromEnd="[0-9,]+", keep.digits
        ), "?"))
  }, "subject(s) 3,4 (2 total) did not match regex below", fixed=TRUE)
})

capture_first_glob <- structure(function
### Glob files, then use capture_first_vec to get meta-data from each
### file name, and combine with contents of each file.
(glob, 
### string: glob specifying files to read.
  ..., 
### passed to capture_first_vec, should include pattern used on vector
### of file names to get meta-data.
  READ=fread
### function of one argument (file name) which returns a data table,
### default data.table::fread.
){
  file.vec <- Sys.glob(glob)
  meta.dt <- capture_first_vec(file.vec, ...)
  result.list <- list()
  for(file.i in seq_along(file.vec)){
    meta.i <- meta.dt[file.i]
    result.i <- READ(file.vec[[file.i]])
    result.list[[file.i]] <- data.table(meta.i, result.i)
  }
  rbindlist(result.list)
### Data table with columns of meta-data specified by pattern, plus
### contents of all files specified by glob.
}, ex=function(){

  data.table::setDTthreads(1)

  ## Example 0: iris data, one file per species.
  library(data.table)
  dir.create(iris.dir <- tempfile())
  icsv <- function(sp)file.path(iris.dir, paste0(sp, ".csv"))
  data.table(iris)[, fwrite(.SD, icsv(Species)), by=Species]
  dir(iris.dir)
  data.table::fread(file.path(iris.dir,"setosa.csv"), nrows=2)
  (iglob <- file.path(iris.dir,"*.csv"))
  nc::capture_first_glob(iglob, Species="[^/]+", "[.]csv")

  ## Example 1: four files, two capture groups, custom read function.
  db <- system.file("extdata/chip-seq-chunk-db", package="nc", mustWork=TRUE)
  suffix <- if(interactive())"gz" else "head"
  (glob <- paste0(db, "/*/*/counts/*", suffix))
  Sys.glob(glob)
  read.bedGraph <- function(f)data.table::fread(
    f, skip=1, col.names = c("chrom","start", "end", "count"))
  data.chunk.pattern <- list(
    data="H.*?",
    "/",
    chunk="[0-9]+", as.integer)
  (data.chunk.dt <- nc::capture_first_glob(glob, data.chunk.pattern, READ=read.bedGraph))

  ## Write same data set in Hive partition, then re-read.
  if(requireNamespace("arrow") && arrow::arrow_with_dataset()){
    path <- tempfile()
    max_rows_per_file <- if(interactive())3 else 1000
    arrow::write_dataset(
      dataset=data.chunk.dt,
      path=path,
      format="csv",
      partitioning=c("data","chunk"),
      max_rows_per_file=max_rows_per_file)
    hive.glob <- file.path(path, "*", "*", "*.csv")
    hive.pattern <- list(
      nc::field("data","=",".*?"),
      "/",
      nc::field("chunk","=",".*?", as.integer),
      "/",
      nc::field("part","-","[0-9]+", as.integer))
    hive.dt <- nc::capture_first_glob(hive.glob, hive.pattern)
    hive.dt[, .(rows=.N), by=.(data,chunk,part)]
  }

  ## Example 2: more complex pattern.
  count.dt <- nc::capture_first_glob(
    glob,
    data.chunk.pattern,
    "/counts/", 
    name=list("McGill", id="[0-9]+", as.integer),
    READ=read.bedGraph)
  count.dt[, .(count=.N), by=.(data, chunk, name, chrom)]

  if(require(ggplot2)){
    ggplot()+
      facet_wrap(~data+chunk+name+chrom, labeller=label_both, scales="free")+
      geom_step(aes(
        start, count),
        data=count.dt)
  }

  ## Example 3: parsing non-CSV data.
  vignettes <- system.file("extdata/vignettes", package="nc", mustWork=TRUE)
  non.greedy.lines <- list(
    list(".*\n"), "*?")
  optional.name <- list(
    list(" ", chunk_name="[^,}]+"), "?")
  chunk.pattern <- list(
    before=non.greedy.lines,
    "```\\{r",
    optional.name,
    parameters=".*",
    "\\}\n",
    code=non.greedy.lines,
    "```")
  chunk.dt <- nc::capture_first_glob(
    paste0(vignettes, "/*.Rmd"),
    "/v",
    vignette_number="[0-9]", as.integer,
    "-",
    vignette_name=".*?",
    ".Rmd",
    READ=function(f)nc::capture_all_str(f, chunk.pattern))
  chunk.dt[, chunk_number := seq_along(chunk_name), by=vignette_number]
  chunk.dt[, .(
    vignette_number, vignette_name, chunk_number, chunk_name, 
    lines=nchar(code))]
  cat(chunk.dt$code[2])
    
})

capture_melt_multiple <- structure(function # Capture and melt into multiple columns
### Match a regex to column names of a wide data frame (many
### columns/few rows), then melt/reshape the matching columns into
### multiple result columns in a taller/longer data table (fewer
### columns/more rows). Input should be a data frame with four or more
### regularly named columns of possibly different types to reshape,
### and output is a data table with at least two columns of reshaped
### data. For melting into a single result column, see
### capture_melt_single.
(...,
### First argument must be a data frame to melt/reshape; column names
### of this data frame will be used as the subjects for regex
### matching. Other arguments (regex/conversion/engine) are passed to
### capture_first_vec along with nomatch.error=FALSE. The regex must
### define a group named "column" -- each unique value captured in
### this group becomes a column name for the reshaped data in the
### output. There must also be at least one other group, and the
### output will contain a column for each other group -- see
### examples.
  fill=FALSE,
### If TRUE, fill missing input reshape columns with runs of rows with
### missing values in the output reshape columns. Otherwise stop with
### an error (default).
  na.rm=FALSE,
### Remove missing values from melted data? (passed to
### data.table::melt.data.table)
  verbose=getOption("datatable.verbose")
### Print verbose output messages? (passed to
### data.table::melt.data.table)
){
  L <- melt_list(measure_multiple, list(...), fill=fill)
  melt(
    L[["data"]],
    measure.vars=L[["measure.vars"]],
    na.rm=na.rm,
    value.factor=FALSE,
    verbose=verbose)
### Data table of reshaped/melted/tall/long data, with a new column
### for each unique value of the capture group named "column", and a
### new column for each other capture group.
}, ex=function(){

  data.table::setDTthreads(1)

  ## Example 1: melt iris columns to compare Sepal and Petal dims, as
  ## in cdata package, https://winvector.github.io/cdata/
  (iris.part.cols <- nc::capture_melt_multiple(
    iris,
    column=".*?",
    "[.]",
    dim=".*"))
  iris.part.cols[Sepal<Petal] #Sepals are never smaller than Petals.
  if(require("ggplot2")){
    ggplot()+
      theme_bw()+
      theme(panel.spacing=grid::unit(0, "lines"))+
      facet_grid(dim ~ Species)+
      coord_equal()+
      geom_abline(slope=1, intercept=0, color="grey")+
      geom_point(aes(
        Petal, Sepal),
        shape=1,
        data=iris.part.cols)
  }

  ## Example 2. melt iris to Length and Width columns.
  (iris.dim.cols <- nc::capture_melt_multiple(
    iris,
    part=".*?",
    "[.]",
    column=".*"))
  iris.dim.cols[Length<Width] #Length is never less than Width.

  ## Example 3. Lots of column types, from example(melt.data.table).
  set.seed(1)
  DT <- data.table::data.table(
    i_1 = c(1:5, NA),
    i_2 = c(NA,6:10),
    f_1 = factor(sample(c(letters[1:3], NA), 6, TRUE)),
    f_2 = factor(c("z", "a", "x", "c", "x", "x"), ordered=TRUE),
    c_1 = sample(c(letters[1:3], NA), 6, TRUE),
    l_2 = list(NULL, NA, c(NA,NA), logical(), 1:2, TRUE),
    d_1 = as.Date(c(1:3,NA,4:5), origin="2013-09-01"),
    d_2 = as.Date(6:1, origin="2012-01-01"))
  ## nc syntax melts to three output columns of different types using
  ## a single regex (na.rm=FALSE by default in order to avoid losing
  ## information).
  nc::capture_melt_multiple(
    DT,
    column="[dfi]",
    "_",
    number="[12]", as.integer)

  ## fill=TRUE means to output NA in positions that correspond to
  ## missing input columns (in this case, there is no l_1 nor c_2).
  nc::capture_melt_multiple(
    DT,
    column=".*",
    "_",
    number="[12]", as.integer,
    fill=TRUE)

  ## Example 4, three children, one family per row, from data.table
  ## vignette.
  family.dt <- data.table::fread(text="
family_id age_mother dob_child1 dob_child2 dob_child3 gender_child1 gender_child2 gender_child3
1         30 1998-11-26 2000-01-29         NA             1             2            NA
2         27 1996-06-22         NA         NA             2            NA            NA
3         26 2002-07-11 2004-04-05 2007-09-02             2             2             1
4         32 2004-10-10 2009-08-27 2012-07-21             1             1             1
5         29 2000-12-05 2005-02-28         NA             2             1            NA")
  ## nc::field can be used to define group name and pattern at the
  ## same time, to avoid repetitive code.
  (children.nc <- nc::capture_melt_multiple(
    family.dt,
    column=".+",
    "_",
    nc::field("child", "", "[1-3]", as.integer), 
    na.rm=TRUE))

  ## Example 5: wide data CSV with 100 possible peaks per row, each
  ## peak has three attributes (Allele, Height, Size) from
  ## https://lftdi.camden.rutgers.edu/repository/PROVEDIt_1-5-Person%20CSVs%20Filtered.zip
  PROVEDIt.csv <- system.file(
    "extdata", "RD12-0002_PP16HS_5sec_GM_F_1P.csv",
    package="nc", mustWork=TRUE)
  PROVEDIt.wide <- data.table::fread(PROVEDIt.csv)
  names(PROVEDIt.wide)
  PROVEDIt.tall <- nc::capture_melt_multiple(
    PROVEDIt.wide,
    column=".*",
    " ",
    peak="[0-9]+", as.integer,
    na.rm=TRUE)
  head(PROVEDIt.tall)

  ## plot number of peaks per row.
  peaks.per.sample.marker <- PROVEDIt.tall[, .(
    peaks=.N
  ), by=.(`Sample File`, Marker)][order(peaks)]
  if(require(ggplot2)){
    ggplot()+
      geom_histogram(aes(
        peaks),
        data=peaks.per.sample.marker,
        binwidth=1)
  }

  ## which row has the most peaks?
  (most <- PROVEDIt.tall[which.max(peak), .(`Sample File`, Marker, Dye)])
  PROVEDIt.tall[most, on=names(most)]
  PROVEDIt.wide[most, on=names(most)]

})

alevels <- structure(function # Alternative levels
### Create a pattern and a conversion function based on alternative
### string literals.
(...
### Optional names are string literals to match; values are
### corresponding factor levels.
){
  old2new <- c(...)
  if(is.null(names(old2new))){
    names(old2new) <- old2new
  }
  to.rep <- names(old2new)==""
  names(old2new)[to.rep] <- old2new[to.rep]
  list(
    paste(names(old2new), collapse="|"),
    function(x)factor(old2new[x], old2new))
### List of pattern and conversion function that returns factor.
}, ex=function(){

  ## Example 0: melt iris data with literal alternatives -> chr columns.
  ichr <- nc::capture_melt_single(
    iris[1,],
    part="Sepal|Petal",
    "[.]",
    dim="Length|Width")
  factor(ichr$part)#default factor levels are alphabetical.

  ## Example 1: melt iris data with alevels() -> factor columns.
  (ifac <- nc::capture_melt_single(
    iris[1,],
    part=nc::alevels("Sepal","Petal"),
    "[.]",
    dim=nc::alevels("Length","Width")))
  ifac$part #factor with levels in same order as given in alevels().

  ## Example 2: alevels(literals_to_match="levels_to_use_in_output").
  tv_wide <- data.frame(
    id=0,
    train.classif.logloss = 1, train.classif.ce = 2, 
    valid.classif.logloss = 3, valid.classif.ce = 4)
  nc::capture_melt_single(
    tv_wide, 
    set=nc::alevels(valid="validation", train="subtrain"),
    "[.]classif[.]",
    measure=nc::alevels(ce="error_prop", auc="AUC", "logloss"))

  ## Example 3: additional groups which output character columns.
  nc::capture_melt_single(
    tv_wide, 
    set_chr=list(set_fac=nc::alevels(valid="validation", train="subtrain")),
    "[.]classif[.]",
    measure_chr=list(measure_fac=nc::alevels(ce="error_prop", auc="AUC", "logloss")))

})

capture_all_str <- structure(function # Capture all matches in a single subject string
### Capture each match of a regex pattern from one multi-line subject
### string or text file. It can be used to convert any regular text
### file (web page, log, etc) to a data table, see examples.
(...,
### subject, name1=pattern1, fun1, etc. The first argument must be a
### subject character vector (or file name which is read via
### base::readLines to get a subject). After removing missing values,
### we use base::paste to collapse the subject (by default using
### newline) and treat it as single character string to
### search. Arguments after the first specify the regex/conversion and
### must be string/function/list, as documented in capture_first_vec.
  engine=getOption("nc.engine", "PCRE"),
### character string, one of PCRE, ICU, RE2
  collapse="\n",
### separator string for combining elements of subject into a single
### string, used as collapse argument of base::paste.
  type.convert=getOption("nc.type.convert", FALSE)
### Default conversion function, which will be used on each capture
### group, unless a specific conversion is specified for that
### group. If TRUE, use utils::type.convert; if FALSE, use
### base::identity; otherwise must be a function of at least one
### argument (character), returning an atomic vector of the same
### length.
){
  stop_for_engine(engine)
  L <- subject_var_args(..., type.convert=type.convert)
  ## instead of explicitly using file.exists here (which may error for
  ## some strange/emoji subjects) we can just use readLines, which
  ## will error if arg has more than one element (invalid
  ## 'description' argument) or single element file name does not
  ## exist (cannot open the connection).
  subject.vec <- tryCatch({
    suppressWarnings(readLines(L[["subject"]]))
  }, error=function(e){
    L[["subject"]]
  })
  subject <- paste(
    subject.vec[!is.na(subject.vec)],
    collapse=collapse)
  group.mat <- if(engine=="PCRE"){
    try_or_stop_print_pattern({
      vec.with.attrs <- gregexpr(L[["pattern"]], subject, perl=TRUE)[[1]]
    }, L[["pattern"]], engine)
    first <- attr(vec.with.attrs, "capture.start")
    if(vec.with.attrs[1] == -1){
      matrix(NA_character_, 0, length(first))
    }else{
      last <- attr(vec.with.attrs, "capture.length")-1+first
      subs <- substring(subject, first, last)
      matrix(subs, nrow=nrow(first))
    }
  }else{
    match.fun <- if(engine=="ICU"){
      stringi::stri_match_all_regex
    }else{
      re2::re2_match_all
    }
    match.mat <- try_or_stop_print_pattern({
      match.fun(subject, L[["pattern"]])[[1]]
    }, L[["pattern"]], engine)
    never.error <- function(...)NULL
    not.na <- !is.na(match.mat[,1])#ICU returns no match as one NA.
    ##RE2 returns a character matrix with 0 rows, which is good.
    only_captures(match.mat[not.na,,drop=FALSE], never.error)
  }
  apply_type_funs(group.mat, L[["fun.list"]])
### data.table with one row for each match, and one column for each
### capture group.
}, ex=function(){

  data.table::setDTthreads(1)

  chr.pos.vec <- c(
    "chr10:213,054,000-213,055,000",
    "chrM:111,000-222,000",
    "this will not match",
    NA, # neither will this.
    "chr1:110-111 chr2:220-222") # two possible matches.
  keep.digits <- function(x)as.integer(gsub("[^0-9]", "", x))
  ## By default elements of subject are treated as separate lines (and
  ## NAs are removed). Named arguments are used to create capture
  ## groups, and conversion functions such as keep.digits are used to
  ## convert the previously named group.
  int.pattern <- list("[0-9,]+", keep.digits)
  (match.dt <- nc::capture_all_str(
    chr.pos.vec,
    chrom="chr.*?",
    ":",
    chromStart=int.pattern,
    "-",
    chromEnd=int.pattern))
  str(match.dt)

  ## Extract all fields from each alignment block, using two regex
  ## patterns, then dcast.
  info.txt.gz <- system.file(
    "extdata", "SweeD_Info.txt.gz", package="nc")
  info.vec <- readLines(info.txt.gz)
  info.vec[24:40]
  info.dt <- nc::capture_all_str(
    sub("Alignment ", "//", info.vec),
    "//",
    alignment="[0-9]+",
    fields="[^/]+")
  (fields.dt <- info.dt[, nc::capture_all_str(
    fields,
    "\t+",
    variable="[^:]+",
    ":\t*",
    value=".*"),
    by=alignment])
  (fields.wide <- data.table::dcast(fields.dt, alignment ~ variable))

  ## Capture all csv tables in report -- the file name can be given as
  ## the subject to nc::capture_all_str, which calls readLines to get
  ## data to parse.
  (report.txt.gz <- system.file(
    "extdata", "SweeD_Report.txt.gz", package="nc"))
  (report.dt <- nc::capture_all_str(
    report.txt.gz,
    "//",
    alignment="[0-9]+",
    "\n",
    csv="[^/]+"
  )[, {
    data.table::fread(text=csv)
  }, by=alignment])

  ## Join report with info fields.
  report.dt[fields.wide, on=.(alignment)]

  ## parsing nbib citation file.
  (pmc.nbib <- system.file(
    "extdata", "PMC3045577.nbib", package="nc"))
  blank <- "\n      "
  pmc.dt <- nc::capture_all_str(
    pmc.nbib,
    Abbreviation="[A-Z]+",
    " *- ",
    value=list(
      ".*",
      list(blank, ".*"), "*"),
    function(x)sub(blank, "", x))
  str(pmc.dt)

  ## What do the variable fields mean? It is explained on
  ## https://www.nlm.nih.gov/bsd/mms/medlineelements.html which has a
  ## local copy in this package (downloaded 18 Sep 2019).
  fields.html <- system.file(
    "extdata", "MEDLINE_Fields.html", package="nc")
  if(interactive())browseURL(fields.html)
  fields.vec <- readLines(fields.html)

  ## It is pretty easy to capture fields and abbreviations if gsub
  ## used to remove some tags first.
  no.strong <- gsub("</?strong>", "", fields.vec)
  no.comments <- gsub("<!--.*?-->", "", no.strong)
  ## grep then capture_first_vec can be used if each desired row in
  ## the output comes from a single line of the input file.
  (h3.vec <- grep("<h3", no.comments, value=TRUE))
  h3.pattern <- list(
    nc::field("name", '="', '[^"]+'),
    '"></a>',
    fields.abbrevs="[^<]+")
  first.fields.dt <- nc::capture_first_vec(
    h3.vec, h3.pattern)
  field.abbrev.pattern <- list(
    Field=".*?",
    " \\(",
    Abbreviation="[^)]+",
    "\\)",
    "(?: and |$)?")
  (first.each.field <- first.fields.dt[, nc::capture_all_str(
    fields.abbrevs, field.abbrev.pattern),
    by=fields.abbrevs])

  ## If we want to capture the information after the initial h3 line
  ## of the input, e.g. the rest column below which contains a
  ## description/example for each field, then capture_all_str can be
  ## used on the full input file.
  h3.fields.dt <- nc::capture_all_str(
    no.comments,
    h3.pattern,
    '</h3>\n',
    rest="(?:.*\n)+?", #exercise: get the examples.
    "<hr />\n")
  (h3.each.field <- h3.fields.dt[, nc::capture_all_str(
    fields.abbrevs, field.abbrev.pattern),
    by=fields.abbrevs])

  ## Either method of capturing abbreviations gives the same result.
  identical(first.each.field, h3.each.field)

  ## but the capture_all_str method returns the additional rest column
  ## which contains data after the initial h3 line.
  names(first.fields.dt)
  names(h3.fields.dt)
  cat(h3.fields.dt[fields.abbrevs=="Volume (VI)", rest])

  ## There are 66 Field rows across three tables.
  a.href <- list('<a href=[^>]+>')
  (td.vec <- fields.vec[240:280])
  fields.pattern <- list(
    "<td.*?>",
    a.href,
    Fields="[^()<]+",
    "</a></td>\n")
  (td.only.Fields <- nc::capture_all_str(fields.vec, fields.pattern))

  ## Extract Fields and Abbreviations. Careful: most fields have one
  ## abbreviation, but one field has none, and two fields have two.
  (td.fields.dt <- nc::capture_all_str(
    fields.vec,
    fields.pattern,
    "<td[^>]*>",
    "(?:\n<div>)?",
    a.href, "?",
    abbrevs=".*?",
    "</"))

  ## Get each individual abbreviation from the previously captured td
  ## data.
  td.each.field <- td.fields.dt[, {
    f <- nc::capture_all_str(
      Fields,
      Field=".*?",
      "(?:$| and )")
    a <- nc::capture_all_str(
      abbrevs,
      "\\(",
      Abbreviation="[^)]+",
      "\\)")
    if(nrow(a)==0)list() else cbind(f, a)
  }, by=Fields]
  str(td.each.field)
  td.each.field[td.fields.dt, .(
    count=.N
  ), on=.(Fields), by=.EACHI][order(count)]

  ## There is a typo in the data captured from the h3 headings.
  td.each.field[!Field %in% h3.each.field$Field]
  h3.each.field[!Field %in% td.each.field$Field]

  ## Abbreviations are consistent.
  td.each.field[!Abbreviation %in% h3.each.field$Abbreviation]
  h3.each.field[!Abbreviation %in% td.each.field$Abbreviation]

  ## There is a a table that provides a description of each comment
  ## type.
  (comment.vec <- fields.vec[840:860])
  comment.dt <- nc::capture_all_str(
    fields.vec,
    "<td><strong>",
    Field="[^<]+",
    "</strong></td>\n",
    "<td><strong>\\(",
    Abbreviation="[^)]+",
    "\\)</strong></td>\n",
    "<td>",
    description=".*",
    "</td>\n")
  str(comment.dt)

  ## Join to original PMC citation file in order to see what the
  ## abbreviations used in that file mean.
  all.abbrevs <- rbind(
    td.each.field[, .(Field, Abbreviation)],
    comment.dt[, .(Field, Abbreviation)])
  all.abbrevs[pmc.dt, .(
    Abbreviation,
    Field,
    value=substr(value, 1, 20)
  ), on=.(Abbreviation)]

  ## There is a listing of examples for each comment type.
  (comment.ex.dt <- nc::capture_all_str(
    fields.vec[938],
    "br />\\s*",
    Abbreviation="[A-Z]+",
    "\\s*-\\s*",
    citation="[^<]+?",
    list(
      "[.] ",
      nc::field("PMID", ": ", "[0-9]+")
    ), "?",
    "<"))

  ## Join abbreviations to see what kind of comments.
  all.abbrevs[comment.ex.dt, on=.(Abbreviation)]

  ## parsing bibtex file.
  refs.bib <- system.file(
    "extdata", "namedCapture-refs.bib", package="nc")
  refs.vec <- readLines(refs.bib)
  at.lines <- grep("@", refs.vec, value=TRUE)
  str(at.lines)
  refs.dt <- nc::capture_all_str(
    refs.vec,
    "@",
    type="[^{]+",
    "[{]",
    ref="[^,]+",
    ",\n",
    fields="(?:.*\n)+?.*",
    "[}]\\s*(?:$|\n)")
  str(refs.dt)

  ## parsing each field of each entry.
  eq.lines <- grep("=", refs.vec, value=TRUE)
  str(eq.lines)
  strip <- function(x)sub("^\\s*\\{*", "", sub("\\}*,?$", "", x))
  refs.fields <- refs.dt[, nc::capture_all_str(
    fields,
    "\\s+",
    variable="\\S+",
    "\\s+=",
    value=".*", strip),
    by=.(type, ref)]
  str(refs.fields)
  with(refs.fields[ref=="HockingUseR2011"], structure(
    as.list(value), names=variable))
  ## the URL of my talk is now
  ## https://user2011.r-project.org/TalkSlides/Lightening/2-StatisticsAndProg_3-Hocking.pdf

  if(!grepl("solaris", R.version$platform)){#To avoid CRAN check error on solaris
    ## Parsing wikimedia tables: each begins with {| and ends with |}.
    emoji.txt.gz <- system.file(
      "extdata", "wikipedia-emoji-text.txt.gz", package="nc")
    tables <- nc::capture_all_str(
      emoji.txt.gz,
      "\n[{][|]",
      first=".*",
      '\n[|][+] style="',
      nc::field("font-size", ":", '.*?'),
      '" [|] ',
      title=".*",
      lines="(?:\n.*)*?",
      "\n[|][}]")
    str(tables)
    ## Rows are separated by |-
    rows.dt <- tables[, {
      row.vec <- strsplit(lines, "|-", fixed=TRUE)[[1]][-1]
      .(row.i=seq_along(row.vec), row=row.vec)
    }, by=title]
    str(rows.dt)
    ## Try to parse columns from each row. Doesn't work for second table
    ## https://en.wikipedia.org/w/index.php?title=Emoji&oldid=920745513#Skin_color
    ## because some entries have rowspan=2.
    contents.dt <- rows.dt[, nc::capture_all_str(
      row,
      "[|] ",
      content=".*?",
      "(?: [|]|\n|$)"),
      by=.(title, row.i)]
    contents.dt[, .(cols=.N), by=.(title, row.i)]
    ## Make data table from
    ## https://en.wikipedia.org/w/index.php?title=Emoji&oldid=920745513#Emoji_versus_text_presentation
    contents.dt[, col.i := 1:.N, by=.(title, row.i)]
    data.table::dcast(
      contents.dt[title=="Sample emoji variation sequences"],
      row.i ~ col.i,
      value.var="content")
  }

  ## Simple way to extract code chunks from Rmd.
  vignette.Rmd <- system.file(
    "extdata", "vignette.Rmd", package="nc")
  non.greedy.lines <- list(
    list(".*\n"), "*?")
  optional.name <- list(
    list(" ", name="[^,}]+"), "?")
  Rmd.dt <- nc::capture_all_str(
    vignette.Rmd,
    before=non.greedy.lines,
    "```\\{r",
    optional.name,
    parameters=".*",
    "\\}\n",
    code=non.greedy.lines,
    "```")
  Rmd.dt[, chunk := 1:.N]
  Rmd.dt[, .(chunk, name, parameters, some.code=substr(code, 1, 20))]

  ## Extract individual parameter names and values.
  Rmd.dt[, nc::capture_all_str(
    parameters,
    ", *",
    variable="[^= ]+",
    " *= *",
    value="[^ ,]+"),
    by=chunk]

  ## Simple way to extract code chunks from Rnw.
  vignette.Rnw <- system.file(
    "extdata", "vignette.Rnw", package="nc")
  Rnw.dt <- nc::capture_all_str(
    vignette.Rnw,
    before=non.greedy.lines,
    "<<",
    name="[^,>]*",
    parameters=".*",
    ">>=\n",
    code=non.greedy.lines,
    "@")
  Rnw.dt[, .(name, parameters, some.code=substr(code, 1, 20))]

  ## The next example involves timing some compression programs that
  ## were run on a 159 megabyte input/uncompressed text file. Here is
  ## how to get a data table from the time command line output.
  times.out <- system.file(
    "extdata", "compress-times.out", package="nc", mustWork=TRUE)
  times.dt <- nc::capture_all_str(
    times.out,
    "coverage.bedGraph ",
    program=".*?",
    " coverage.bedGraph.",
    suffix=".*",
    "\n\nreal\t",
    minutes.only="[0-9]+", as.numeric,
    "m",
    seconds.only="[0-9.]+", as.numeric)
  times.dt[, seconds := minutes.only*60+seconds.only]
  times.dt

  ## join with output from du command line program.
  sizes.out <- system.file(
    "extdata", "compress-sizes.out", package="nc", mustWork=TRUE)
  sizes.dt <- data.table::fread(
    file=sizes.out,
    col.names=c("megabytes", "file"))
  sizes.dt[, suffix := sub("coverage.bedGraph.?", "", file)]
  join.dt <- times.dt[sizes.dt, on="suffix"][order(megabytes)]
  join.dt[file=="coverage.bedGraph", seconds := 0]
  join.dt

  ## visualize with ggplot2.
  if(require(ggplot2)){
    ggplot(join.dt, aes(
      seconds, megabytes, label=suffix))+
      geom_text(vjust=-0.5)+
      geom_point()+
      scale_x_log10()+
      scale_y_log10()
  }

})

capture_melt_single <- structure(function # Capture and melt into a single column
### Match a regex to column names of a wide data frame (many
### columns/few rows), then melt/reshape the matching columns into a
### single result column in a taller/longer data table (fewer columns/more
### rows). It is for the common case of melting several columns of
### the same type in a "wide" input data table which has several
### distinct pieces of information encoded in each column name. For
### melting into several result columns of possibly different types,
### see capture_melt_multiple.
(...,
### First argument must be a data frame to melt/reshape; column names
### of this data frame will be used as the subjects for regex
### matching. Other arguments (regex/conversion/engine) are passed to
### capture_first_vec along with nomatch.error=FALSE.
  value.name="value",
### Name of the column in output which has values taken from
### melted/reshaped column values of input (passed to
### data.table::melt.data.table).
  na.rm=TRUE,
### remove missing values from melted data? (passed to
### data.table::melt.data.table)
  verbose=getOption("datatable.verbose")
### Print verbose output messages? (passed to
### data.table::melt.data.table)
){
  L <- melt_list(measure_single, list(...), value.name=value.name)
  melt(
    L[["data"]],
    measure.vars=L[["measure.vars"]],
    value.name=value.name,
    na.rm=na.rm,
    value.factor=FALSE,
    verbose=verbose)
### Data table of reshaped/melted/tall/long data, with a new column
### for each named argument in the pattern, and additionally
### variable/value columns.
}, ex=function(){

  data.table::setDTthreads(1)

  ## Example 1: melt iris data and barplot for each numeric variable.
  (iris.tall <- nc::capture_melt_single(
    iris,
    part=".*",
    "[.]",
    dim=".*",
    value.name="cm"))
  ## Histogram of cm for each variable.
  if(require("ggplot2")){
    ggplot()+
      theme_bw()+
      theme(panel.spacing=grid::unit(0, "lines"))+
      facet_grid(part ~ dim)+
      geom_bar(aes(cm), data=iris.tall)
  }

  ## Example 2: melt who data and use type conversion functions for
  ## year limits (e.g. for censored regression).
  if(requireNamespace("tidyr")){
    data(who, package="tidyr", envir=environment())
    ##2.1 just extract diagnosis and gender to chr columns.
    new.diag.gender <- list(#save pattern as list for re-use later.
      "new_?",
      diagnosis=".*",
      "_",
      gender=".")
    who.tall.chr <- nc::capture_melt_single(who, new.diag.gender, na.rm=TRUE)
    print(head(who.tall.chr))
    str(who.tall.chr)
    ##2.2 also extract ages and convert to numeric output columns.
    who.tall.num <- nc::capture_melt_single(
      who,
      new.diag.gender,#previous pattern for matching diagnosis and gender.
      ages=list(#new pattern for matching age range.
        min.years="0|[0-9]{2}", as.numeric,#in-line type conversion functions.
        max.years="[0-9]{0,2}", function(x)ifelse(x=="", Inf, as.numeric(x))),
      value.name="count",
      na.rm=TRUE)
    print(head(who.tall.num))
    str(who.tall.num)
    ##2.3 compute total count for each age range then display the
    ##subset with max.years lower than a threshold.
    who.age.counts <- who.tall.num[, .(
      total=sum(count)
    ), by=.(min.years, max.years)]
    print(who.age.counts[max.years < 50])
  }

  ## Example 3: pepseq data.
  if(requireNamespace("R.utils")){#for reading gz files with data.table
    pepseq.dt <- data.table::fread(
      system.file("extdata", "pepseq.txt", package="nc", mustWork=TRUE))
    u.pepseq <- pepseq.dt[, unique(names(pepseq.dt)), with=FALSE]
    nc::capture_melt_single(
      u.pepseq,
      "^",
      prefix=".*?",
      nc::field("D", "", ".*?"),
      "[.]",
      middle=".*?",
      "[.]",
      "[0-9]+",
      suffix=".*",
      "$")
  }

})

quantifier <- structure(function
### Create a group with a quantifier.
(...
### Pattern(s) to be enclosed in a group, and a quantifier (last
### argument). A quantifier is character string: "?" for zero or one,
### "*?"  for non-greedy zero or more, "+" for greedy one or more,
### etc.
){
  L <- list(...)
  N <- length(L)
  if(N < 2)stop(
    "quantifier needs at least two arguments (patterns, quantifier)")
  no.names <- is.null(names(L))
  no.last.name <- identical(names(L)[[N]], "")
  has.last.name <- !(no.names||no.last.name)
  if(has.last.name)stop(
    "last argument to quantifier must be un-named")
  last.arg <- L[[N]]
  if(!(is.character(last.arg) && length(last.arg)==1))stop(domain=NA, gettext("last argument to quantifier must be character string (quantifier such as ?, *, or {0,2})"))
  list(L[-N], last.arg)
### A pattern list.
}, ex=function(){

  ## No need to use nc::quantifier when the pattern to be quantified
  ## is just a string literal.
  digits <- "[0-9]+"

  ## nc::quantifier is useful when there is a sequence of patterns to
  ## be quantified, here an optional group with a dash (not captured)
  ## followed by some digits (captured in the chromEnd group).
  str(optional.end <- nc::quantifier("-", chromEnd=digits, "?"))
  str(optional.end <- list(list("-", chromEnd=digits), "?"))#same

  ## Use it as a sub-pattern for capturing genomic coordinates.
  chr.pos.vec <- c(
    "chr10:213054000-213055000",
    "chrM:111000",
    "chr1:110-111 chr2:220-222") # two possible matches.
  nc::capture_first_vec(
    chr.pos.vec,
    chrom="chr.*?",
    ":",
    chromStart=digits,
    optional.end)

  ## Another example which uses quantifier twice, for extracting code
  ## chunks from Rmd files.
  vignette.Rmd <- system.file(
    "extdata", "vignette.Rmd", package="nc")
  non.greedy.lines <- nc::quantifier(".*\n", "*?")
  optional.name <- nc::quantifier(" ", name="[^,}]+", "?")
  Rmd.dt <- nc::capture_all_str(
    vignette.Rmd,
    before=non.greedy.lines,
    "```\\{r",
    optional.name,
    parameters=".*",
    "\\}\n",
    code=non.greedy.lines,
    "```")
  Rmd.dt[, chunk := 1:.N]
  Rmd.dt[, .(chunk, name, parameters, some.code=substr(code, 1, 20))]

})

measure <- structure(function
### Computes a value to be used as measure.vars argument to
### data.table::melt.data.table. NOTE: only works on newer versions of
### data.table that include the measure function.
(...,
### Regular expression pattern list, passed to capture_first_vec with
### cols as subject.
 cols
### Character vector, column names to match with regex.
){
  var.args <- list(...)
  match.info <- check_names(cols, var.args)
  match.info[["subject.names"]] <- cols
  ##details<<
  ##measure_multiple is called if there is a capture group
  ##named "column" and measure_single is called otherwise.
  m.fun <- if("column" %in% names(match.info[["match.dt"]])){
    measure_multiple
  }else{
    measure_single
  }
  do.call(m.fun, match.info)
### List or vector to use as measure.vars argument to
### data.table::melt.data.table.
}, ex=function(){

  library(data.table)
  iris.dt <- data.table(datasets::iris[c(1,150),])
  melt(iris.dt, measure=nc::measure(part  =".*", "[.]", dim   =".*"))
  melt(iris.dt, measure=nc::measure(column=".*", "[.]", dim   =".*"))
  melt(iris.dt, measure=nc::measure(part  =".*", "[.]", column=".*"))

})

melt_list <- function
### Compute a list of arguments to pass to
### data.table::melt.data.table.
(measure.fun,
### measure_single or measure_multiple.
  dot.args,
### list of arguments for check_df_names.
  ...
### passed to measure.fun.
){
  match.info <- do.call(check_df_names, dot.args)
  data.i <- which(names(match.info) == "data")
  subject.dt <- match.info[[data.i]]
  measure.args <- c(
    list(..., subject.names=names(subject.dt)),
    match.info[-data.i])
  measure.vars <- do.call(measure.fun, measure.args)
  list(
    data=subject.dt,
    measure.vars=measure.vars)
}

measure_single <- function
### Compute a measure.vars vector (indicating a single output column)
### with variable_table attribute to pass to
### data.table::melt.data.table.
(subject.names,
### character vector of data frame column names.
  match.dt,
### data table of matches.
  no.match,
### logical vector.
  value.name=NULL
### string.
){
  id.vars <- subject.names[no.match]
  stop_for_capture_same_as_id(names(match.dt), id.vars)
  if(!is.null(value.name)){
    check.list <- list(
      "an input column name that did not match the pattern"=subject.names,
      "a capture group name"=names(match.dt))
    for(check.name in names(check.list)){
      check.values <- check.list[[check.name]]
      if(value.name %in% check.values){
        stop(domain=NA, gettextf("value.name (%s) is the same as %s; please change one so that all output column names will be unique", value.name, check.name))
      }
    }
  }
  structure(
    which(!no.match),
    variable_table=match.dt[!no.match])
}

measure_multiple <- function
### Compute a measure.vars list (indicating multiple output columns)
### with variable_table attribute to pass to
### data.table::melt.data.table.
(subject.names,
### character vector of data frame column names.
  match.dt,
### data table of matches.
  no.match,
### logical vector.
  fill=TRUE
### logical.
){
  column <- . <- NULL
  ## Above to avoid CRAN NOTE.
  if(is.null(match.dt[["column"]])){
    stop("pattern must define group named column")
  }
  if(!is.character(match.dt[["column"]])){
    stop(domain=NA, gettextf("column group must be character, but conversion function returned %s", class(match.dt[["column"]])[[1]]))
  }
  not.col <- names(match.dt)[names(match.dt) != "column"]
  if(length(not.col)==0){
    stop("need at least one group other than column")
  }
  id.vars <- subject.names[no.match]
  stop_for_capture_same_as_id(not.col, id.vars)
  by.list <- list(
    group=not.col,
    column="column")
  by.result <- list()
  i.name <- paste(names(match.dt), collapse="")
  paste.collapse <- function(x.vec)paste(x.vec, collapse=",")
  for(by.name in names(by.list)){
    by.vec <- by.list[[by.name]]
    by.counts <- match.dt[!is.na(column), {
      structure(list(.N), names=i.name)
    }, keyby=by.vec]#need keyby so variable.name order consistent later.
    count <- by.counts[[i.name]]
    if(!isTRUE(fill) && any(count != max(count))){
      count.vec <- sprintf(
        "%s=%d",
        apply(by.counts[, by.vec, with=FALSE], 1, paste.collapse),
        by.counts[[i.name]])
      stop(domain=NA, gettextf("need %s=same count for each value, but have: %s; please change pattern, edit input column names, or use fill=TRUE to output missing values", paste.collapse(by.vec), paste(count.vec, collapse = " ")))
    }
    by.result[[by.name]] <- by.counts
  }
  by.column <- by.result[["column"]]
  if(all(by.column[[i.name]] == 1)){
    stop(domain=NA, gettext("only one input variable for each value captured in column group; typically this happens when the column group matches the entire input column name; fix by changing regex so that column group matches a strict substring (not the entire input column names)"))
  }
  if(nrow(by.column)==1){
    stop(domain=NA, gettextf("need multiple output columns, but only one value (%s) captured in column group; either provide a different regex that captures more than one value in column group, or use capture_melt_single if you really want only one output column", by.column[["column"]]))
  }
  i.name <- paste(names(match.dt), collapse="")
  i.dt <- data.table(match.dt)
  set(i.dt, j=i.name, value=1:nrow(i.dt))
  ##need to sort by not.col for irregular col ord.
  all.list <- lapply(match.dt, function(x)sort(unique(x[!is.na(x)])))
  value.name <- all.list[["column"]]
  check.list <- list(
    "input column names which do not match the pattern"=id.vars,
    "other regex group names"=not.col)
  for(check.name in names(check.list)){
    check.values <- check.list[[check.name]]
    bad.values <- value.name[value.name %in% check.values]
    if(length(bad.values)){
      stop(domain=NA, gettextf("unable to create unique output column names; some values (%s) captured by the regex group named column are the same as %s; please change either the pattern or the %s so that output column names will be unique", paste(bad.values, collapse = ", "), check.name, check.name))
    }
  }
  all.dt <- do.call(CJ, all.list)
  i.all.dt <- i.dt[all.dt, on=names(all.dt)]
  setkeyv(i.all.dt, c("column", not.col))
  var.tab <- by.result[["group"]][, not.col, with=FALSE]
  measure.vars <- structure(
    list(), variable_table=var.tab)
  for(col.value in value.name){
    measure.vars[[col.value]] <- i.all.dt[col.value, .SD[[i.name]] ]
  }
  measure.vars
}

try_or_stop_print_pattern <- function
### Try to run a capture function. If it fails we wrap the error
### message with a more informative message that also includes the
### generated pattern.
(expr,
### expression to try.  
  pat,
### regex pattern string.
  engine
### string: regex engine.
){
  tryCatch(expr, error=function(e){
    stop(domain=NA, gettextf("%s
when matching pattern above with %s engine, an error occured: %s", pat, engine, e$message))
  })
}

library(nc)
library(testthat)
context("errors")
source(system.file("test_engines.R", package="nc", mustWork=TRUE), local=TRUE)

foo.bar <- c("foo", "bar")
test_engines("no capture groups is an error", {
  expect_error({
    capture_first_vec(foo.bar, "o")
  }, "must have at least one named argument", fixed=TRUE)
})

named.chr.vec <- c("v\"name\\"="[a\"o]+")
expected.obj <- as.list(named.chr.vec)
expect_error_with_code <- function(expr, expected.obj){
  msg <- tryCatch({
    expr
  }, error=function(e){
    e$message
  })
  code <- capture_first_vec(
    msg,
    "did you mean ",
    code=".*")$code
  computed.obj <- eval(parse(text=code))
  expect_identical(computed.obj, expected.obj)
}
test_engines("named character vector is an error with group name", {
  expect_error_with_code({
    capture_first_vec(foo.bar, gname=named.chr.vec)
  }, expected.obj)
})
test_engines("named character vector is an error without group name", {
  expect_error_with_code({
    capture_first_vec(foo.bar, named.chr.vec)
  }, expected.obj)
})

named.pat.list <- as.list(named.chr.vec)
exp.vec <- c("oo", "a")
test_engines("named pattern list in named arg makes two groups", {
  (result.df <- capture_first_vec(foo.bar, gname=named.pat.list))
  expect_identical(names(result.df), c("gname", names(named.pat.list)))
  expect_identical(result.df$gname, exp.vec)
  expect_identical(result.df[[names(named.pat.list)]], exp.vec)
})
test_engines("named pattern list in un-named arg makes one group", {
  (result.df <- capture_first_vec(foo.bar, named.pat.list))
  expect_identical(names(result.df), names(named.pat.list))
  expect_identical(result.df[[names(named.pat.list)]], exp.vec)
})

test_engines("any capture group without a name is an error", {
  expect_error({
    capture_first_vec(foo.bar, "(o)(?P<name>o)")
  }, "must have at least one named argument", fixed=TRUE)
})

test_engines("NA pattern is an error", {
  expect_error({
    capture_first_vec(foo.bar, NA_character_)
  }, "patterns must not be missing/NA", fixed=TRUE)
})

test_engines("factor pattern is an error", {
  expect_error({
    capture_first_vec(foo.bar, factor("(?P<regex>foo)"))
  }, "arguments must be character")
})

test_engines("multiple patterns is an error", {
  expect_error({
    capture_first_vec(foo.bar, c("(?P<name>.)", "(?P<name>.)"))
  }, "patterns must be character vectors of length 1")
})

test_engines("subject of length 0 is an error", {
  expect_error({
    capture_first_vec(character(), "(?P<name>.)")
  }, "subject has class=character and length=0 but should be a character vector with length>0")
})

test_engines("capture all works with only one 'name' group", {
  subject <- c(missing=NA, nomatch="", match="foobar")
  result.dt <- capture_all_str(subject, name="foo")
  expect_equal(dim(result.dt), c(1, 1))
  expect_identical(result.dt$name, "foo")
})

test_engines("informative error when converter fun has zero args", {
  expect_error({
    capture_first_vec(
      "chr2:300-400",
      chrom="chr", function()y)
  }, "atomic vector",
  fixed=TRUE)
})

test_engines("informative error when converter returns wrong length", {
  expect_error({
    capture_first_vec(
      c("chr2:300-400", "chr2:300-400"),
      chrom="chr", function(x)"foo")
  }, "type conversion function for group 1(chrom) returned vector of length 1 but expected length 2", fixed=TRUE)
})

test_engines("informative error when converter returns non-atomic", {
  expect_error({
    capture_first_vec(
      c("chr2:300-400", "chr2:300-400"),
      chrom="chr", function(x)list(foo=200))
  }, "each type conversion function must return an atomic vector, but function for group 1(chrom) did not", fixed=TRUE)
})

test_engines("error for name group, missing subject, nomatch.error=TRUE", {
  expect_error({
    capture_first_vec(
      c("chr1:20-40", NA, "chr2:300-400"),
      name="[^:]+",
      ":",
      chromStart="[0-9]+",
      nomatch.error=TRUE)
  }, "subject(s) 2 (1 total) did not match regex below", fixed=TRUE)
})

test_engines("error for name group, no match, nomatch.error=TRUE", {
  expect_error({
    capture_first_vec(
      c("chr1:20-40", "foobar", "chr2:300-400"),
      name="[^:]+",
      ":",
      chromStart="[0-9]+",
      nomatch.error=TRUE)
  }, "subject(s) 2 (1 total) did not match regex below", fixed=TRUE)
})

test_that("informative error when subject is named", {
  expect_error({
    nc::capture_all_str(chrom="chr.*?", ":", chromStart="[0-9,]+")
  }, "first argument is named chrom but must NOT be named; please include the subject to match as the first argument, with no name")
  expect_error({
    nc::capture_first_vec(chrom="chr.*?", ":", chromStart="[0-9,]+")
  }, "first argument is named chrom but must NOT be named; please include the subject to match as the first argument, with no name")
})

library(testthat)
library(nc)
context("group")

test_that("group returns named list", {
  computed <- group("sample", "McGill", id="[0-9]+", as.integer)
  pat <- list("McGill", id="[0-9]+", as.integer)
  expected <- structure(list(pat), names="sample")
  expect_identical(computed, expected)
})

test_that("group errors for non-character name", {
  expect_error({
    group(factor('foo'), "bar", 'BAZ')
  },
  "first argument of group must be a character string (group name)",
  fixed=TRUE)
})

test_that("group errors for NA character name", {
  l <- group(NA_character_, "bar", 'BAZ')
  ## missing names could also be manually constructed so no need to
  ## stop with an error in group, but we should stop with an error
  ## when the NA is parsed.
  expect_error({
    var_args_list(l)
  }, "group name must not be missing", fixed=TRUE)
})

library(nc)
library(data.table)
library(testthat)
context("alternatives")

test_that("error for no arguments to alternatives", {
  expect_error({
    alternatives()
  }, "alternatives should have at least two arguments")
})

test_that("error for one argument to alternatives", {
  expect_error({
    alternatives("pattern")
  }, "alternatives should have at least two arguments")
})

test_that("two un-named alternatives => two non-capturing groups", {
  subject <- c("foo1", "bar2")
  a.pat <- alternatives("foo", "bar")
  match.dt <- nc::capture_first_vec(subject, a.pat, number="[12]")
  expect_identical(match.dt, data.table(number=paste(1:2)))
})

subject <- c("foooo1", "bar2")
exp.dt <- data.table(foo=c("foooo", ""), number=paste(1:2))
test_that("first named alternative => one more capturing group", {
  a.pat <- alternatives(foo="fo+", "bar")
  match.dt <- nc::capture_first_vec(subject, a.pat, number="[12]")
  expect_identical(match.dt, exp.dt)
})

test_that("second named alternative => one more capturing group", {
  a.pat <- alternatives("bar", foo="fo+")
  match.dt <- nc::capture_first_vec(subject, a.pat, number="[12]")
  expect_identical(match.dt, exp.dt)
})

test_that("both named alternatives => two more capturing groups", {
  subject <- c("foooo1", "barrr2")
  exp.dt <- data.table(
    bar=c("", "barrr"),
    foo=c("foooo", ""),
    number=paste(1:2))
  a.pat <- alternatives(bar="bar+", foo="fo+")
  match.dt <- nc::capture_first_vec(subject, a.pat, number="[12]")
  expect_identical(match.dt, exp.dt)
})

pat.list <- nc::altlist(month="[a-z]{3}", day="[0-9]{2}", year="[0-9]{4}")
pattern <- with(pat.list, nc::alternatives(
  list(month, " ", day, ", ", year),
  list(day, " ", month, " ", year)))
test_that("alternatives with dup names ok", {
  subject.vec <- c("mar 17, 1983", "26 sep 2017")
  computed <- nc::capture_first_vec(subject.vec, pattern)
  expected <- data.table(
    month=c("mar", "sep"), day=c("17", "26"), year=c("1983", "2017"))
  expect_identical(computed, expected)
})
test_that("alternatives with dup names and one subject with no match", {
  subject.vec <- c("mar 17, 1983", "this will not match", "26 sep 2017")
  computed <- nc::capture_first_vec(subject.vec, pattern, nomatch.error=FALSE)
  expected <- data.table(
    month=c("mar", NA, "sep"),
    day=c("17", NA, "26"),
    year=c("1983", NA, "2017"))
  expect_identical(computed, expected)
})

test_that("alternatives with dup names error for diff types", {
  subject.vec <- c("mar 17, 1983", "26 sep 2017")
  pat.list <- nc::altlist(month="[a-z]{3}", day="[0-9]{2}", year="[0-9]{4}")
  pattern <- with(pat.list, nc::alternatives(
    list(month, " ", day, ", ", year, as.integer),
    list(day, " ", month, " ", year)))
  expect_error({
    nc::capture_first_vec(subject.vec, pattern)
  }, "capture groups with identical names should have conversion functions that all return the same type; problem group name=year has types character,integer")
})

test_that("IDate conversion OK", {
  subject.vec <- c(
    "02/07/2020 EUR68.50",
    "EUR45.00 29/10/2020")
  pat.list <- nc::altlist(
    money=list("EUR", amount="[0-9.]+", as.numeric),
    date=list(
      "[0-9]{2}/[0-9]{2}/[0-9]{4}",
      function(d)data.table::as.IDate(d, format="%d/%m/%Y")))
  pattern <- with(pat.list, nc::alternatives(
    list(money, " ", date),
    list(date, " ", money)))
  computed <- nc::capture_first_vec(subject.vec, pattern)
  expected <- data.table(
    money=c("EUR68.50", "EUR45.00"),
    amount=c(68.5, 45),
    date=as.IDate(c("2020-07-02", "2020-10-29")))
  expect_identical(computed, expected)
})

test_that("altlist fails for no names", {
  expect_error({
    nc::altlist("foo")
  }, "all arguments to altlist must be named")
})

test_that("altlist fails for no args", {
  expect_error({
    nc::altlist()
  }, "all arguments to altlist must be named")
})

test_that("altlist fails for one un-named arg", {
  expect_error({
    nc::altlist(foo="bar", "baz")
  }, "all arguments to altlist must be named")
})

test_that("alternatives_with_shared_groups errors for no named args", {
  expect_error({
    nc::alternatives_with_shared_groups("bar")
  }, "alternatives_with_shared_groups must have at least one named argument; each named argument should define a sub-pattern/group")
})

test_that("alternatives_with_shared_groups errors for no un-named args", {
  expect_error({
    nc::alternatives_with_shared_groups(foo="bar")
  }, "alternatives_with_shared_groups must have at least one un-named argument; each un-named argument should define an alternative pattern")
})

test_that("altlist + alternatives with names ok", {
  subject.vec <- c("mar 17, 1983", "26 sep 2017", "17 mar 1984")
  pat.list <- nc::altlist(month="[a-z]{3}", day="[0-9]{2}", year="[0-9]{4}")
  pattern <- with(pat.list, nc::alternatives(
    american=list(month, " ", day, ", ", year),
    european=list(day, " ", month, " ", year)))
  match.dt <- nc::capture_first_vec(subject.vec, pattern)
  sorted.names <- c("american", "day", "european", "month", "year")
  expect_identical(sort(names(match.dt)), sorted.names)
  expect_identical(match.dt[["american"]], c(subject.vec[1], "", ""))
  expect_identical(match.dt[["european"]], c("", subject.vec[2:3]))
  expect_identical(match.dt[["month"]], c("mar", "sep", "mar"))
  expect_identical(match.dt[["day"]], c("17", "26", "17"))
  expect_identical(match.dt[["year"]], c("1983", "2017", "1984"))
})

test_that("alternatives_with_shared_groups ok with 3 subjects", {
  subject.vec <- c("mar 17, 1983", "26 sep 2017", "17 mar 1984")
  pattern <- nc::alternatives_with_shared_groups(
    month="[a-z]{3}", day="[0-9]{2}", year="[0-9]{4}",
    list(month, " ", day, ", ", year),
    list(day, " ", month, " ", year))
  match.dt <- nc::capture_first_vec(subject.vec, pattern)
  sorted.names <- c("day", "month", "year")
  expect_identical(sort(names(match.dt)), sorted.names)
  expect_identical(match.dt[["month"]], c("mar", "sep", "mar"))
  expect_identical(match.dt[["day"]], c("17", "26", "17"))
  expect_identical(match.dt[["year"]], c("1983", "2017", "1984"))
})

test_that("alternatives_with_shared_groups ok with 1 subject", {
  subject.vec <- "mar 17, 1983"
  pattern <- nc::alternatives_with_shared_groups(
    month="[a-z]{3}", day="[0-9]{2}", year="[0-9]{4}",
    list(month, " ", day, ", ", year),
    list(day, " ", month, " ", year))
  match.dt <- nc::capture_first_vec(subject.vec, pattern)
  sorted.names <- c("day", "month", "year")
  expect_identical(sort(names(match.dt)), sorted.names)
  expect_identical(match.dt[["month"]], "mar")
  expect_identical(match.dt[["day"]], "17")
  expect_identical(match.dt[["year"]], "1983")
})

test_that("alevels ok with no names", {
  ifac <- nc::capture_melt_single(
    iris[1,],
    part=nc::alevels("Sepal","Petal"),
    "[.]",
    dim=nc::alevels("Length","Width"))
  expect_identical(levels(ifac$part), c("Sepal","Petal"))
  expect_identical(levels(ifac$dim), c("Length","Width"))
})

test_that("alevels() ok with all and some names", {
  tv_wide <- data.frame(
    id=0,
    train.classif.logloss = 1, train.classif.ce = 2,
    valid.classif.logloss = 3, valid.classif.ce = 4)
  tv_long <- nc::capture_melt_single(
    tv_wide,
    set_chr=list(set_fac=nc::alevels(valid="validation", train="subtrain")),
    "[.]classif[.]",
    measure_chr=list(measure_fac=nc::alevels(ce="error_prop", auc="AUC", "logloss")))
  expect_is(tv_long$set_chr, "character")
  expect_is(tv_long$measure_chr, "character")
  expect_identical(levels(tv_long$set_fac), c("validation","subtrain"))
  expect_identical(levels(tv_long$measure_fac), c("error_prop","AUC","logloss"))
})

library(testthat)
library(nc)
context("field")

test_that("field returns list", {
  int.pattern <- list("[0-9]+", as.integer)
  cell.sample.type <- list(
    cellType="[^ ]*?",
    "_",
    sampleName=list(
      "McGill",
      sampleID=int.pattern),
    dataType="Coverage|Peaks")
  computed <- nc::field(
    "track",
    " ",
    cell.sample.type,
    "|",
    "[^\n]+")
  expected <- list(
    "track",
    " ",
    list(track=list(
      cell.sample.type,
      "|",
      "[^\n]+")))
  expect_identical(computed, expected)
})

test_that("field errors for non-character first arg", {
  expect_error({
    field(identity, "bar", "sars")
  },
  "first argument of field must be character string (field name)",
  fixed=TRUE)
})

library(nc)
library(testthat)
context("quantifier")
source(system.file("test_engines.R", package="nc", mustWork=TRUE), local=TRUE)

test_that("error for no args to quantifier", {
  expect_error({
    quantifier()
  },
  "quantifier needs at least two arguments (patterns, quantifier)",
  fixed=TRUE)
})

test_that("error for one arg to quantifier", {
  expect_error({
    quantifier("foo")
  },
  "quantifier needs at least two arguments (patterns, quantifier)",
  fixed=TRUE)
})

test_that("error for named last arg", {
  expect_error({
    quantifier("a", "b", foo="?")
  }, "last argument to quantifier must be un-named")
})

test_that("error for non-string last arg", {
  expect_error({
    quantifier("a", "b", identity)
  },
  "last argument to quantifier must be character string (quantifier such as ?, *, or {0,2})",
  fixed=TRUE)
})

## Above we use test_that (does not do any matching) and below we do
## test_engines (matches using all regex engines).

test_engines("user-defined optional is ok", {
  job.vec <- readLines(textConnection("26534569
26534569.extern
26534569.0
26534608_63
26534608_63.batch
26534608_63.extern
26534685_[1-373]
26534686_[1]"))
  ## zero_or_one zero_or_more like rex etc?
  optional <- function(...){
    quantifier(..., "?")
  }
  optional.end <- optional("-", taskN="[0-9]+", as.integer)
  range.pattern <- list(
    "\\[",
    task1="[0-9]+", as.integer,
    optional.end,
    "\\]")
  task.pattern <- list(
    task="[0-9]+", as.integer,
    "|",#either one task(above) or range(below)
    range.pattern)
  optional.type <- optional("[.]", type=".*")
  optional.task <- optional("_", task.pattern)
  match.dt <- capture_first_vec(
    job.vec,
    job="[0-9]+", as.integer,
    optional.task,
    optional.type)
  expect_equal(nrow(match.dt), length(job.vec))
  expect_false(any(is.na(match.dt$job)))
})

test_engines("quantifier works for non-greedy lines", {
  vignette.Rmd <- system.file(
    "extdata", "vignette.Rmd", package="nc")
  non.greedy.lines <- quantifier(".*\n", "*?")
  Rmd.dt <- capture_all_str(
    vignette.Rmd,
    before=non.greedy.lines,
    "```\\{r",
    quantifier(" ", name="[^,}]+", "?"),
    parameters=".*",
    "\\}\n",
    code=non.greedy.lines,
    "```")
  expect_equal(nrow(Rmd.dt), 14)
})

library(nc)
library(testthat)
library(data.table)
context("melt")
source(system.file("test_engines.R", package="nc", mustWork=TRUE), local=TRUE)

iris.dt <- data.table(observation=1:nrow(iris), iris)
test_engines("error for regex that matches no column names", {
  expect_error({
    capture_melt_single(iris.dt, part="foo")
  }, "no column names match regex")
})

test_engines("error for fun always returning NA", {
  expect_error(suppressWarnings({
    capture_melt_single(iris.dt, part=".*", as.numeric)
  }),
  "need to change type conversion function(s), which should return at least one non-NA, but are always returning NA, even though regex matched 6 column(s): observation,Sepal.Length,Sepal.Width,Petal.Length,Petal.Width,Species",
  fixed=TRUE)
})

test_engines("error for fun always returning NA abbrev cols", {
  subject <- data.frame(t(1:99))
  expect_error(suppressWarnings({
    capture_melt_single(subject, part=".*", as.numeric)
  }),
  "need to change type conversion function(s), which should return at least one non-NA, but are always returning NA, even though regex matched 99 column(s): X1,X2,X3,X4,X5,...,X95,X96,X97,X98,X99",
  fixed=TRUE)
})

test_engines("possessive (.*+) error(RE2) or OK(others)", {
  posmatch <- function(){
    capture_melt_single(
      iris.dt,
      part=".*",
      "[.]",
      dim=".*+")
  }
  if(identical(getOption("nc.engine"), "RE2")){
    expect_error({
      posmatch()
    }, "bad repetition operator")
  }else{
    iris.tall <- posmatch()
    exp.names <- c(
      "observation", "Species", "part", "dim", "value")
    expect_identical(names(iris.tall), exp.names)
  }
})

test_engines("error if first arg not df", {
  expect_error({
    capture_melt_single("foo", bar="baz")
  }, "first argument (subject) must be a data.frame", fixed=TRUE)
})

## what if an input column is named .variable?
DV <- data.table(
  .variable=c("foo", "bar"), "p10.5"=c(3L, 5L), "p1.1"=c(0L, 1L))
test_engines(".variable input column ok", {
  tall.dt <- capture_melt_single(DV, "p", penalty=".*", as.numeric)
  expect_identical(tall.dt$penalty, c(10.5, 10.5, 1.1, 1.1))
})

## what if a capture group has the same name as variable.name?
test_engines("capture group with funny name is ok", {
  tall.dt <- capture_melt_single(
    DV, "p", .variable.p10.5.p1.1=".*", as.numeric)
  expect_identical(tall.dt$.variable.p10.5.p1.1, c(10.5, 10.5, 1.1, 1.1))
})

## what if input df has repeated names?
bad.df <- data.frame(1, 2)
names(bad.df) <- c("foo", "foo")
test_engines("melting df with same col names is an error", {
  expect_error({
    capture_melt_single(bad.df, o="o+")
  }, "input must have columns with unique names, problems: foo")
})

## what if there are two groups with the same name?
test_engines("groups with the same name is an error", {
  expect_error({
    capture_melt_single(DV, foo="p", foo="1")
  }, "duplicate capture group names are only allowed in alternatives, problem: foo")
})

## what if a capture group has the same name as an input column?
test_engines("err change capture group if same as input col", {
  expect_error({
    capture_melt_single(DV, .variable="p")
  },
  "some capture group names (.variable) are the same as input column names that did not match the pattern; please change either the pattern or the capture group names so that all output column names will be unique",
  fixed=TRUE)
})

## what if value.name is the same as input col?
test_engines("err change value.name if same as input col", {
  expect_error({
    capture_melt_single(DV, "p", num=".*", value.name=".variable")
  },
  "value.name (.variable) is the same as an input column name that did not match the pattern; please change one so that all output column names will be unique",
  fixed=TRUE)
})

test_engines("err change value.name or group names", {
  expect_error({
    capture_melt_single(DV, "p", num=".*", value.name="num")
  },
  "value.name (num) is the same as a capture group name; please change one so that all output column names will be unique",
  fixed=TRUE)
})

i.vec <- 1:10000
one.row <- data.frame(t(i.vec))
test_engines("melting lots of columns is OK", {
  out <- capture_melt_single(one.row, "X", col="[0-9]+", as.integer)
  expect_identical(out$col, i.vec)
  expect_identical(out$value, i.vec)
})

test_engines("melting lots of columns type.convert=TRUE is OK", {
  out <- capture_melt_single(one.row, "X", col="[0-9]+", type.convert=TRUE)
  expect_identical(out$col, i.vec)
  expect_identical(out$value, i.vec)
})

DT.wide <- data.table(id=0, num_ref=1, name_ref="foo", num=2, name="bar")
test_engines("converting NA to non-NA is an error", {
  expect_error({
    nc::capture_melt_multiple(DT.wide, column="name|num", type=".*", function(x)fcase(x=="", "other", default="reference"))
  }, "a non-match(NA) was converted to a match(non-NA) by the conversion function in group 2(type); please fix conversion function", fixed=TRUE)
})

library(nc)
library(data.table)
library(testthat)
source(system.file("test_engines.R", package="nc", mustWork=TRUE), local=TRUE)

test_engines("single spec works", {
  (single.spec <- nc::capture_longer_spec(iris, part=".*", "[.]", dim=".*", values_to="cm"))
  expect_identical(sort(names(single.spec)), c(".name", ".value", "dim", "part"))
})

test_engines("multiple spec works", {
  (multiple.spec <- nc::capture_longer_spec(iris, part=".*", "[.]", column=".*"))
  expect_identical(sort(names(multiple.spec)), c(".name", ".value", "part"))
})

if(requireNamespace("tidyr"))test_engines("missing spec works", {
  missing.col <- data.table(Sepal.Length=1, Sepal.Width=2, Petal.Length=3)
  (missing.spec <- nc::capture_longer_spec(missing.col, part=".*", "[.]", column=".*"))
  computed.df <- tidyr::pivot_longer_spec(missing.col, missing.spec)
  sorted.dt <- data.table(computed.df)[order(part)]
  expected.dt <- data.table(part=c("Petal", "Sepal"), Length=c(3, 1), Width=c(NA, 2))
  expect_identical(sorted.dt, expected.dt)
})

library(testthat)
library(data.table)
context('subject')

test_that("capture_first_vec does not use s arg for subject", {
  dt <- nc::capture_first_vec(
    "author: toby, subject: ml",
    s="\\S+$")
  expect_identical(dt, data.table(s="ml"))
})

test_that("capture_all_str does not use s arg for subject", {
  dt <- nc::capture_all_str(
    "author: toby, subject: ml",
    s="\\S+$")
  expect_identical(dt, data.table(s="ml"))
})

test_that("capture_first_df does not use s arg for subject", {
  subject <- data.frame(x="author: toby, subject: ml")
  dt <- nc::capture_first_df(subject, x=list(s="\\S+$"))
  expected.dt <- data.table(subject, s="ml")
  expect_identical(dt, expected.dt)
})

if(requireNamespace("stringi"))test_that("capture_melt_single does not use s arg for subject", {
  subject <- data.frame(family=1, s2=2, s3=3)
  ## also make sure that engine is passed thru (not interpreted as a
  ## capture group).
  dt <- nc::capture_melt_single(subject, s="[0-9]", engine="ICU")
  expected.dt <- data.table(
    family=1,
    s=c("2", "3"),
    value=c(2, 3))
  expect_identical(dt, expected.dt)
})

if(requireNamespace("stringi"))test_that("capture_melt_multiple does not use s arg for subject", {
  subject <- data.frame(family=1, d2=2, d3=3, c2="a", c3="b")
  ## also make sure that engine is passed thru (not interpreted as a
  ## capture group).
  dt <- nc::capture_melt_multiple(
    subject, column="[dc]", s="[0-9]", engine="ICU")
  expect_identical(dt[["family"]], c(1, 1))
  expect_identical(dt[["d"]], c(2, 3))
  expect_identical(dt[["c"]], c("a", "b"))
  expect_identical(dt[["s"]], c("2", "3"))
})

library(testthat)
library(nc)
library(data.table)
context("multiple")
source(system.file("test_engines.R", package="nc", mustWork=TRUE), local=TRUE)

test_engines("only one input variable per column value is an error", {
  expect_error({
    nc::capture_melt_multiple(iris, column=".*[.].*", other=".*")
  }, "only one input variable for each value captured in column group; typically this happens when the column group matches the entire input column name; fix by changing regex so that column group matches a strict substring (not the entire input column names)",
  fixed=TRUE)
})

iris.dt <- data.table(i=1:nrow(iris), iris)
iris.dims <- capture_melt_multiple(
  iris.dt,
  part=".*?",
  "[.]",
  column=".*")
iris.dims.wide <- dcast(
  iris.dims,
  i + Species ~ part,
  value.var=c("Length", "Width"))
names(iris.dims.wide) <- sub("(.*?)_(.*)", "\\2.\\1", names(iris.dims.wide))
should.be.orig <- iris.dims.wide[, names(iris.dt), with=FALSE]
setkeyv(should.be.orig, key(iris.dt))
test_engines("orig iris.dt and recast data are the same", {
  expect_equal(should.be.orig, iris.dt)
})

iris.dt[, chr := paste(Species)]
compare.cols <- c(
  "Sepal.Length", "Petal.Length",
  "Sepal.Width", "Petal.Width",
  "chr")
set.seed(1)
iris.rand <- iris.dt[sample(.N)]
iris.wide <- cbind(treatment=iris.rand[1:75], control=iris.rand[76:150])
test_engines("iris multiple column matches", {
  nc.melted <- capture_melt_multiple(
    iris.wide,
    group="[^.]+",
    "[.]",
    column=".*")
  expect_identical(nc.melted$group, rep(c("control", "treatment"), each=75))
  nc.melted.orig <- nc.melted[order(i), ..compare.cols]
  expect_identical(nc.melted.orig, iris.dt[, ..compare.cols])
})

set.seed(1)
iris.wide.rand <- iris.wide[, sample(names(iris.wide)), with=FALSE]
test_engines("iris multiple rand column matches", {
  rand.melted <- capture_melt_multiple(
    iris.wide.rand,
    group="[^.]+",
    "[.]",
    column=".*"
  )[order(i)]
  expect_identical(rand.melted[, ..compare.cols], iris.dt[, ..compare.cols])
})

iris.wide.bad <- data.table(iris.wide)
names(iris.wide.bad)[2] <- "treatment.Sepal.bad"
test_engines("iris multiple rand column matches", {
  expect_error({
    capture_melt_multiple(
      iris.wide.bad,
      group="[^.]+",
      "[.]",
      column=".*")
  }, "need column=same count for each value")
})

set.seed(45)
DT <- data.table(
  i_1 = c(1:5, NA),
  i_2 = c(NA,6:10),
  f_1 = factor(sample(c(letters[1:3], NA), 6, TRUE)),
  f_2 = factor(c("z", "a", "x", "c", "x", "x"), ordered=TRUE),
  c_1 = sample(c(letters[1:3], NA), 6, TRUE),
  d_1 = as.Date(c(1:3,NA,4:5), origin="2013-09-01"),
  d_2 = as.Date(6:1, origin="2012-01-01"))
## add a couple of list cols
DT[, l_1 := DT[, list(c=list(rep(i_1, sample(5,1)))), by = i_1]$c]
DT[, l_2 := DT[, list(c=list(rep(c_1, sample(5,1)))), by = i_1]$c]
test_engines("melt multiple column types error if no other group", {
  expect_error({
    capture_melt_multiple(DT, column="^[^c]")
  }, "need at least one group other than column")
})

test_engines("melt multiple column types", {
  result <- capture_melt_multiple(
    DT,
    column="^[^c]",
    "_",
    .number="[0-9]", as.integer)
  expect_identical(result$.number, rep(1:2, each=nrow(DT)))
  expect_is(result$c_1, "character")
  expect_is(result$i, "integer")
  expect_is(result$f, "character")
  expect_is(result$d, "Date")
  expect_is(result$l, "list")
  expect_identical(nrow(result), 12L)
})

family.dt <- fread(text="family_id age_mother dob_child1 dob_child2 dob_child3 gender_child1 gender_child2 gender_child3
1         30 1998-11-26 2000-01-29         NA             1             2            NA
2         27 1996-06-22         NA         NA             2            NA            NA
3         26 2002-07-11 2004-04-05 2007-09-02             2             2             1
4         32 2004-10-10 2009-08-27 2012-07-21             1             1             1
5         29 2000-12-05 2005-02-28         NA             2             1            NA")
test_engines("gender dob example", {
  na.children <- capture_melt_multiple(
    family.dt,
    column="[^_]+",
    between="_child",
    number="[1-3]",
    na.rm=FALSE)
  expect_is(na.children$family_id, "integer")
  expect_is(na.children$age_mother, "integer")
  expect_is(na.children$dob, "IDate")
  expect_is(na.children$gender, "integer")
  expect_equal(sum(is.na(na.children$dob)), 4)
  expect_equal(nrow(na.children), 15)
})

test_engines("error for unequal number of children", {
  bad.groups <- data.table(family.dt, dob_child0="1999-01-01", gender_child4=2)
  expect_error({
    capture_melt_multiple(
      bad.groups,
      column="[^_]+",
      between="_child",
      number="[0-9]")
  }, "need between,number=same count for each value")
})

test_engines("error for column conversion to factor", {
  expect_error({
    capture_melt_multiple(
      family.dt,
      column="[^_]+", as.factor,
      between="_child",
      number="[0-9]")
  },
  "column group must be character, but conversion function returned factor")
})

test_engines("error for column conversion to integer", {
  expect_error({
    expect_warning({
      capture_melt_multiple(
        family.dt,
        column="[^_]+", as.integer,
        between="_child",
        number="[0-9]")
    }, "NAs introduced by coercion")
  },
  "column group must be character, but conversion function returned integer")
})

test_engines("error for unequal number of columns", {
  bad.cols <- data.table(family.dt, bar_child0="1999-01-01", foo_child0=2)
  expect_error({
    capture_melt_multiple(
      bad.cols,
      column="[^_]+",
      between="_child",
      number="[0-9]")
  }, "need column=same count for each value")
})

test_engines("gender dob example na.rm=TRUE", {
  children <- capture_melt_multiple(
    family.dt,
    column="[^_]+",
    between="_child",
    number="[1-3]",
    na.rm=TRUE)
  expect_is(children$family_id, "integer")
  expect_is(children$age_mother, "integer")
  expect_is(children$dob, "IDate")
  expect_is(children$gender, "integer")
  expect_equal(sum(is.na(children$dob)), 0)
  expect_equal(nrow(children), 11)
})

test_engines("variable group ok in melt_multiple", {
  result <- capture_melt_multiple(
    family.dt,
    column="[^_]+",
    variable="_child",
    number="[1-3]")
  exp.names <- c(
    "family_id", "age_mother",
    "variable", "number",
    "dob", "gender")
  expect_identical(names(result), exp.names)
})

test_engines("multiple error if subject not df", {
  expect_error({
    capture_melt_multiple("foobar")
  }, "first argument (subject) must be a data.frame", fixed=TRUE)
})

test_engines("multiple error if no arg named variable", {
  expect_error({
    capture_melt_multiple(family.dt, baz="family")
  }, "pattern must define group named column")
})

test_engines("multiple error if no matching column names", {
  expect_error({
    capture_melt_multiple(family.dt, column="foobar")
  }, "no column names match regex")
})

## what if input df has repeated names?
bad.dt <- data.table(family_id=LETTERS[1:5], family.dt)
test_engines("multiple df with same col names is an error", {
  expect_error({
    capture_melt_multiple(
      bad.dt, column=".*", "_", field("child", "", "[0-9]"))
  }, "input must have columns with unique names, problems: family_id")
})

## what if there are two groups with the same name?
test_engines("multiple groups with the same name is an error", {
  expect_error({
    capture_melt_multiple(
      family.dt, column=".*", child="_", field("child", "", "[0-9]"))
  }, "duplicate capture group names are only allowed in alternatives, problem: child")
})

## what if a capture group has the same name as an input column?
test_engines("err mult capture group same as input col", {
  expect_error({
    capture_melt_multiple(
      family.dt, column=".*", family_id="_", field("child", "", "[0-9]"))
  },
  "some capture group names (family_id) are the same as input column names that did not match the pattern; please change either the pattern or the capture group names so that all output column names will be unique",
  fixed=TRUE)
})

## what if value.name is the same as input col?
bad2 <- data.table(family.dt)
names(bad2)[1] <- "dob"
test_engines("err change value.name if same as input col", {
  expect_error({
    capture_melt_multiple(
      bad2, column=".*", family_id="_", field("child", "", "[0-9]"))
  },
  "unable to create unique output column names; some values (dob) captured by the regex group named column are the same as input column names which do not match the pattern; please change either the pattern or the input column names which do not match the pattern so that output column names will be unique",
  fixed=TRUE)
})

test_engines("err mult value.name same as group names", {
  expect_error({
    capture_melt_multiple(
      family.dt, column=".*", dob="_", field("child", "", "[0-9]"))
  },
  "unable to create unique output column names; some values (dob) captured by the regex group named column are the same as other regex group names; please change either the pattern or the other regex group names so that output column names will be unique",
  fixed=TRUE)
})

i.vec <- 1:10000
c.vec <- paste(i.vec)
one.row <- data.frame(I=t(i.vec), C=t(c.vec), stringsAsFactors=FALSE)
test_engines("multiple melting lots of columns is OK", {
  out <- capture_melt_multiple(
    one.row, column=".", "[.]", int="[0-9]+", as.integer)
  expect_identical(out$int, i.vec)
  expect_identical(out$I, i.vec)
  expect_identical(out$C, c.vec)
})

test_engines("multiple melting lots of columns type.convert=TRUE is OK", {
  out <- capture_melt_multiple(
    one.row, column=".", "[.]", int="[0-9]+", type.convert=TRUE)
  expect_identical(out$int, i.vec)
  expect_identical(out$I, i.vec)
  expect_identical(out$C, c.vec)
})

wide.metrics <- data.table(
  FP.possible=8202, FN.possible=1835,
  FP.count=0, FN.count=1835)
test_engines("count is either 0 or 1835", {
  tall.metrics <- capture_melt_multiple(
    wide.metrics,
    metric=".*?",
    "[.]",
    column=".*")[order(metric)]
  expect_identical(tall.metrics$metric, c("FN", "FP"))
  expect_identical(tall.metrics$count, c(1835, 0))
  expect_identical(tall.metrics$possible, c(1835, 8202))
})

test_engines("melt multiple with count group ok", {
  iris.count <- nc::capture_melt_multiple(
    iris,
    column=".*?",
    "[.]",
    count=".*")
  expect_identical(names(iris.count), c("Species", "count", "Petal", "Sepal"))
})

## apparently there are some data with missing columns
## https://github.com/Rdatatable/data.table/issues/4027
wide.input <- data.table(
  id = 1, a.1 = 1, a.3 = 3, b.1 = 1, b.2 = 2, b.3 = 3)
ix.pattern <- list(column="[a-z]", "[.]", ix="[0-9]", as.integer)
test_engines("error by default for missing column", {
  expect_error({
    capture_melt_multiple(wide.input, ix.pattern)
  }, "need ix=same count for each value, but have: 1=2 2=1 3=2; please change pattern, edit input column names, or use fill=TRUE to output missing values")
})

##But what I really want is the ix of a.3 to be 3, not 2.
##    id ix  a b
## 1:  1  1  1 1
## 2:  1  2 NA 2
## 3:  1  3  3 3
tall.expected <- data.table(
  id=1,
  ix=1:3,
  a=c(1, NA, 3),
  b=1:3)
test_engines("NA for missing column when fill=TRUE", {
  tall.output <- capture_melt_multiple(
    wide.input, ix.pattern, fill=TRUE)
  expect_equal(data.frame(tall.output), data.frame(tall.expected))
})

test_engines("NA for missing columns as in data table example", {
  DT.missing.cols <- DT[, .(d_1, d_2, c_1, f_2)]
  computed <- capture_melt_multiple(
    DT.missing.cols,
    column="[a-z]",
    "_",
    int="[0-9]", as.integer,
    fill=TRUE)
  expected <- DT.missing.cols[, data.table(
    int=rep(1:2, each=.N),
    c=c(c_1, rep(NA, .N)),
    d=c(d_1, d_2),
    f=c(rep(NA, .N), paste(f_2)))]
  expect_identical(data.frame(computed), data.frame(expected))
})

family4.dt <- data.table(family.dt)
family4.dt[, `:=`(dob_child4=NA_character_, gender_child4=NA_integer_)]
test_engines("no families with 4 children", {
  child.pattern <- list(
    column="[^_]+",
    between="_child",
    number="[1-4]")
  na.rm.T <- suppressWarnings({
    capture_melt_multiple(
      family4.dt,
      child.pattern,
      na.rm=TRUE)
  })
  expect_equal(sum(is.na(na.rm.T$dob)), 0)
})

test_engines("lots of missing columns ok with chr capture col small", {
  tall.dt <- data.table(
    family=1,
    child=10:1,
    num=c(10, 9, rep(NA, 8)),
    chr=c("ten", "nine", rep(NA, 8)))
  wide.dt <- dcast(
    tall.dt,
    family ~ child,
    value.var=c("num", "chr"))
  tall.again <- capture_melt_multiple(
    wide.dt,
    column=".*",
    "_",
    child="[0-9]+",
    na.rm=TRUE)[order(num)]
  expect_identical(tall.again$child, paste(9:10))
})

test_engines("lots of missing columns ok with chr capture col big", {
  PROVEDIt.csv <- system.file(
    "extdata", "RD12-0002_PP16HS_5sec_GM_F_1P.csv",
    package="nc", mustWork=TRUE)
  PROVEDIt.wide <- data.table::fread(PROVEDIt.csv)
  PROVEDIt.tall <- suppressWarnings({
    nc::capture_melt_multiple(
      PROVEDIt.wide,
      column=".*",
      " ",
      peak="[0-9]+",
      na.rm=TRUE)
  })
  peak.int <- sort(as.integer(unique(PROVEDIt.tall$peak)))
  expect_identical(peak.int, seq_along(peak.int))
})

test_engines("lots of missing columns chr col big na.rm=FALSE", {
  PROVEDIt.csv <- system.file(
    "extdata", "RD12-0002_PP16HS_5sec_GM_F_1P.csv",
    package="nc", mustWork=TRUE)
  PROVEDIt.wide <- data.table::fread(PROVEDIt.csv)
  PROVEDIt.tall <- suppressWarnings({
    nc::capture_melt_multiple(
      PROVEDIt.wide,
      column=".*",
      " ",
      peak="[0-9]+",
      na.rm=FALSE)[!is.na(Size)]
  })
  ## Why does this test pass? No missing values are removed so data
  ## table assigns peak numbers 1-100 correctly and then we filter the
  ## NAs.
  peak.int <- sort(as.integer(unique(PROVEDIt.tall$peak)))
  expect_identical(peak.int, seq_along(peak.int))
})

test_engines("lots of missing columns int col big na.rm=TRUE", {
  PROVEDIt.csv <- system.file(
    "extdata", "RD12-0002_PP16HS_5sec_GM_F_1P.csv",
    package="nc", mustWork=TRUE)
  PROVEDIt.wide <- data.table::fread(PROVEDIt.csv)
  PROVEDIt.tall <- suppressWarnings({
    nc::capture_melt_multiple(
      PROVEDIt.wide,
      column=".*",
      " ",
      peak="[0-9]+", as.integer,
      na.rm=TRUE)
  })
  ## Why does this test pass? With buggy data.table the missing values
  ## are removed and then the peak IDs are assigned. Because the
  ## missing values are at the end (for peak>36) and the data are
  ## sorted numerically by integers, the assigned IDs match the peak
  ## numbers exactly.
  peak.int <- sort(as.integer(unique(PROVEDIt.tall$peak)))
  expect_identical(peak.int, seq_along(peak.int))
})

test_engines("missing first child", {
  wide.dt <- data.table(
    family=1,
    age_childA=NA_real_, sex_childA=NA_character_,
    age_childB=5, sex_childB="m")
  child.pattern <- list(
    column=".*",
    "_",
    nc::field("child", "", "[A-Z]"))
  tall.dt <- nc::capture_melt_multiple(wide.dt, child.pattern, na.rm=TRUE)
  expect_identical(tall.dt$child, "B")
  expect_identical(tall.dt$age, 5)
  expect_identical(tall.dt$sex, "m")
})

library(nc)
library(testthat)
library(data.table)
source(system.file("test_engines.R", package="nc", mustWork=TRUE), local=TRUE)

db <- system.file("extdata/chip-seq-chunk-db", package="nc", mustWork=TRUE)
glob <- paste0(db, "/*/*/counts/*")
read.bedGraph <- function(f)data.table::fread(
  f, skip=1, col.names = c("chrom","start", "end", "count"))

if(requireNamespace("R.utils")){

  test_engines("capture_first_glob returns expected columns", {
    count.dt <- nc::capture_first_glob(
      glob,
      data="H.*?",
      "/",
      chunk="[0-9]+", as.integer,
      READ=read.bedGraph)
    computed.cls <- sapply(count.dt, class)
    expected.cls <- c(
      data = "character",
      chunk = "integer",
      chrom = "character",
      start = "integer",
      end = "integer",
      count = "integer")
    expect_identical(computed.cls, expected.cls)
  })

  test_engines("capture_first_glob(type.convert) works", {
    count.dt <- nc::capture_first_glob(
      glob,
      data="H.*?",
      "/",
      chunk="[0-9]+",
      READ=read.bedGraph,
      type.convert=TRUE)
    computed.cls <- sapply(count.dt, class)
    expected.cls <- c(
      data = "character",
      chunk = "integer",
      chrom = "character",
      start = "integer",
      end = "integer",
      count = "integer")
    expect_identical(computed.cls, expected.cls)
  })

}

library(nc)
library(testthat)
library(data.table)
context("variable args syntax")
source(system.file("test_engines.R", package="nc", mustWork=TRUE), local=TRUE)

## In stringi there is a difference between matching an empty
## string, and an optional match that fails.
inames <- c("Species", "Petal.Length", "Sepal.Width")
sapply(c("[^.]*(.*)", "[^.]*([.].*)?"), function(pat){
  if(requireNamespace("stringi"))stringi::stri_match_first_regex(inames, pat)
})

test_engines("match nothing in required group is empty string", {
  empty.dt <- capture_first_vec(
    inames,
    "[^.]*",
    rest=".*")
  expect_identical(empty.dt, data.table(rest=c("", ".Length", ".Width")))
})

test_engines("no match for optional group is empty string", {
  opt.empty.dt <- capture_first_vec(
    inames,
    "[^.]*",
    rest="[.].*", "?")
  expect_identical(opt.empty.dt, data.table(rest=c("", ".Length", ".Width")))
})

test_engines("no match for required group is NA", {
  na.dt <- capture_first_vec(
    inames,
    "[^.]*",
    rest="[.].*",
    nomatch.error=FALSE)
  expect_identical(na.dt, data.table(rest=c(NA, ".Length", ".Width")))
})

test_engines("error for capture first regex with literal groups", {
  expect_error({
    capture_first_vec(
      c("chr1:100-200", "chr2:5-6"),
      chrom="chr.",
      ":",
      "([0-9]+)")
  }, "regex contains more groups than names; please remove literal groups (parentheses) from the regex pattern, and use named arguments in R code instead", fixed=TRUE)
})

subject <- c(
  ten="chr10:213,054,000-213,055,000",
  chrNA="chrNA:111,000-222,000",
  no.match="foo bar",
  missing=NA,
  two="chr1:110-111 chr2:220-222")
test_engines("capture_first_vec returns data.table with chr columns", {
  computed <- capture_first_vec(
    subject,
    chrom="chr.*?",
    ":",
    chromStart=".*?",
    "-",
    chromEnd="[0-9,]*",
    nomatch.error=FALSE)
  expected <- data.table(
    chrom=c("chr10", "chrNA", NA, NA, "chr1"),
    chromStart=c("213,054,000", "111,000", NA, NA, "110"),
    chromEnd=c("213,055,000", "222,000", NA, NA, "111"))
  expect_identical(computed, expected)
})

keep.digits <- function(x)as.integer(gsub("[^0-9]", "", x))
test_engines("capture_first_vec returns data.table with int columns", {
  computed <- capture_first_vec(
    subject,
    chrom="chr.*?",
    ":",
    chromStart=".*?", keep.digits,
    "-",
    chromEnd="[0-9,]*", keep.digits,
    nomatch.error=FALSE)
  expected <- data.table(
    chrom=c("chr10", "chrNA", NA, NA, "chr1"),
    chromStart=as.integer(c(213054000, 111000, NA, NA, 110)),
    chromEnd=as.integer(c(213055000, 222000, NA, NA, 111)))
  expect_identical(computed, expected)
})

test_engines("capture_first_vec(type.convert=TRUE) returns one int column", {
  computed <- capture_first_vec(
    "chr1:2-3,000",
    chrom="chr.*?",
    ":",
    chromStart=".*?", 
    "-",
    chromEnd="[0-9,]*",
    type.convert=TRUE)
  expected <- data.table(
    chrom="chr1",
    chromStart=2L,
    chromEnd="3,000")
  expect_identical(computed, expected)
})

test_engines("capture_first_vec(type.convert=TRUE) returns two int columns", {
  computed <- capture_first_vec(
    "chr1:2-3,000",
    chrom="chr.*?",
    ":",
    chromStart=".*?", 
    "-",
    chromEnd="[0-9,]*", keep.digits,
    type.convert=TRUE)
  expected <- data.table(
    chrom="chr1",
    chromStart=2L,
    chromEnd=3000L)
  expect_identical(computed, expected)
})

test_engines("capture_first_vec(type.convert=constant fun) returns all num columns", {
  computed <- capture_first_vec(
    "chr1:2-3,000",
    chrom="chr.*?",
    ":",
    chromStart=".*?",
    "-",
    chromEnd="[0-9,]*",
    type.convert=function(...)5)
  expected <- data.table(
    chrom=5,
    chromStart=5,
    chromEnd=5)
  expect_identical(computed, expected)
})

test_engines("capture_first_vec(type.convert=as.integer) returns all int columns", {
  computed <- suppressWarnings(capture_first_vec(
    "chr1:2-3,000",
    chrom="chr.*?",
    ":",
    chromStart=".*?",
    "-",
    chromEnd="[0-9,]*",
    type.convert=as.integer))
  expected <- data.table(
    chrom=NA_integer_,
    chromStart=2L,
    chromEnd=NA_integer_)
  expect_identical(computed, expected)
})

test_engines("capture_first_vec(type.convert=invalid) errors", {
  expect_error({
    capture_first_vec(
      "chr1:2-3,000",
      chrom="chr.*?",
      ":",
      chromStart=".*?",
      "-",
      chromEnd="[0-9,]*",
      type.convert=function()5)
  }, "type conversion functions should take one argument (character vector of captured text) and return an atomic vector of the same size; function for group 1(chrom) raised an error: unused argument (match.vec)", fixed=TRUE)
})

test_engines("capture_first_vec(type.convert='foo') errors", {
  expect_error({
    capture_first_vec(
      "chr1:2-3,000",
      chrom="chr.*?",
      ":",
      chromStart=".*?",
      "-",
      chromEnd="[0-9,]*",
      type.convert='foo')
  }, "type.convert should be either TRUE or FALSE or a function", fixed=TRUE)
})

test_engines("named function is an error", {
  expect_error({
    capture_first_vec(
      subject,
      chrom="chr.*?",
      ":",
      chromStart=".*?", fun=keep.digits,
      "-",
      chromEnd="[0-9,]*", keep.digits)
  }, "functions must not be named, problem: fun")
})

test_engines("capture_first_vec errors for one argument", {
  expect_error({
    capture_first_vec("foo")
  }, "must have at least one named argument")
})


test_engines("capture_first_vec errors for multi-dim patterns", {
  expect_error({
    capture_first_vec("foo", bar=c("bar", "baz"))
  }, "patterns must be character vectors of length 1")
})

test_engines("capture_first_vec errors for 0-length patterns", {
  expect_error({
    capture_first_vec("foo", bar=character())
  }, "patterns must be character vectors of length 1")
})

test_engines("capture_first_vec errors for non char/fun args", {
  expect_error({
    capture_first_vec("foo", baz="bar", 1)
  }, "arguments must be", fixed=TRUE)
})

test_engines("capture_first_vec errors for two funs in a row", {
  expect_error({
    capture_first_vec(
      "foo", g="bar", as.integer, as.numeric)
  },
  "too many functions; up to one function may follow each named pattern")
})

test_engines("capture_first_vec errors for fun at start", {
  expect_error({
    capture_first_vec("foo", as.numeric, bar="baz")
  },
  "too many functions; up to one function may follow each named pattern")
})

test_engines("capture_first_vec errors for NA pattern", {
  expect_error({
    capture_first_vec("foo", g="bar", NA_character_, "baz")
  }, "patterns must not be missing/NA")
})

range.pattern <- list(
  "\\[",
  task1="[0-9]+", as.integer,
  "(?:-",#begin optional end of range.
  taskN="[0-9]+", as.integer,
  ")?", #end is optional.
  "\\]")
full.pattern <- list(
  job="[0-9]+", as.integer,
  "_",
  "(?:",#begin alternate
  task="[0-9]+", as.integer,
  "|",#either one task(above) or range(below)
  range.pattern,
  ")",#end alternate
  "(?:[.]",
  type=".*",
  ")?")
task.vec <- c(
  "13937810_25",
  "13937810_25.batch",
  "13937810_25.extern",
  "14022192_[1-3]",
  "14022204_[4]")
all.args <- list(task.vec, full.pattern)
test_engines("nested lists are OK", {
  task.df <- do.call(capture_first_vec, all.args)
  expect_identical(
    names(task.df),
    c("job", "task", "task1", "taskN", "type"))
  expect_identical(task.df$job, as.integer(c(
    13937810, 13937810, 13937810, 14022192, 14022204)))
  expect_identical(task.df$task, as.integer(c(
    25, 25, 25, NA, NA)))
  expect_identical(task.df$task1, as.integer(c(
    NA, NA, NA, 1, 4)))
  expect_identical(task.df$taskN, as.integer(c(
    NA, NA, NA, 3, NA)))
  expect_identical(task.df$type, c(
    "", "batch", "extern", "", ""))
})

range.square <- list(
  "[[]",
  task1="[0-9]+", as.integer,
  "(?:-",#begin optional end of range.
  taskN="[0-9]+", as.integer,
  ")?", #end is optional.
  "[]]")
full.square <- list(
  job="[0-9]+", as.integer,
  "_",
  "(?:",#begin alternate
  task="[0-9]+", as.integer,
  "|",#either one task(above) or range(below)
  range.square,
  ")",#end alternate
  "(?:[.]",
  type=".*",
  ")?")
test_engines("vec square brackets pattern", {
  if(identical(getOption("nc.engine"), "ICU")){
    expect_error({
      capture_first_vec(task.vec, full.square)
    }, "when matching pattern above with ICU engine")
  }else{
    task.df <- capture_first_vec(task.vec, full.square)
    expect_identical(
      names(task.df),
      c("job", "task", "task1", "taskN", "type"))
    expect_identical(task.df$job, as.integer(c(
      13937810, 13937810, 13937810, 14022192, 14022204)))
    expect_identical(task.df$task, as.integer(c(
      25, 25, 25, NA, NA)))
    expect_identical(task.df$task1, as.integer(c(
      NA, NA, NA, 1, 4)))
    expect_identical(task.df$taskN, as.integer(c(
      NA, NA, NA, 3, NA)))
    expect_identical(task.df$type, c(
      "", "batch", "extern", "", ""))
  }
})

chr.pos.nomatch.vec <- c(
  "chr10:213,054,000-213,055,000",
  "chrM:111,000",
  "this will not match",
  NA, # neither will this.
  "chr1:110-111 chr2:220-222") # two possible matches.
chr.pos.df <- capture_first_vec(
  chr.pos.nomatch.vec,
  chrom="chr.*?",
  ":",
  chromStart="[0-9,]+", keep.digits,
  list(
    "-",
    chromEnd="[0-9,]+", keep.digits
  ), "?",
  nomatch.error=FALSE)
test_engines("un-named list interpreted as non-capturing group", {
  expect_identical(
    chr.pos.df$chromStart,
    as.integer(c(213054000, 111000, NA, NA, 110)))
  expect_identical(
    chr.pos.df$chromEnd,
    as.integer(c(213055000, NA, NA, NA, 111)))
})

matching.subjects <- c(
  "chr10:213,054,000-213,055,000",
  "chrM:111,000",
  "chr1:110-111 chr2:220-222") # two possible matches.
test_engines("str subject no error if nomatch.error=TRUE and all matches", {
  match.df <- capture_first_vec(
    matching.subjects, nomatch.error=TRUE,
    chrom="chr.*?",
    ":",
    chromStart="[0-9,]+", keep.digits,
    list(
      "-",
      chromEnd="[0-9,]+", keep.digits
    ), "?")
  expect_identical(
    match.df$chromEnd,
    as.integer(c(213055000, NA, 111)))
})
test_engines("str subject stop if nomatch.error=TRUE and no match", {
  expect_error({
    capture_first_vec(
      chr.pos.nomatch.vec, nomatch.error=TRUE,
      chrom="chr.*?",
      ":",
      chromStart="[0-9,]+", keep.digits,
      list(
        "-",
        chromEnd="[0-9,]+", keep.digits
      ), "?")
  }, "subject(s) 3,4 (2 total) did not match regex below", fixed=TRUE)
})

test_that("only a few subject IDs in error", {
  expect_error({
    capture_first_vec(paste(1:1000), foo="bar")
  }, "subject(s) 1,2,3,4,5,...,996,997,998,999,1000 (1000 total) did not match regex below",
  fixed=TRUE)
})

(foo.mat <- capture_first_vec(
  c("foo", "foobar", "fooba"),
  first="foo",
  list("b", second="ar"), "?"))
test_engines("un-named list interpreted as non-capturing group foo subject", {
  expect_identical(foo.mat$first, c("foo", "foo", "foo"))
  expect_identical(foo.mat$second, c("", "ar", ""))
})

subject <- "foo55bar"
test_engines("capture_first_vec returns dt with only one group = name", {
  out.dt <- capture_first_vec(
    subject,
    name="[0-9]+", as.integer)
  expect_equal(dim(out.dt), c(1, 1))
  expect_identical(names(out.dt), "name")
})

library(testthat)
library(nc)
library(data.table)
context("dt")
source(system.file("test_engines.R", package="nc", mustWork=TRUE), local=TRUE)

test_engines("default error for existing column, otherwise modify", {
  pos <- "chr1:100-200"
  pos.dt <- data.table(pos)
  int.pattern <- list("[0-9]+", as.integer)
  chrom.start.pattern <- list(
    chrom=".*?",
    ":",
    start=int.pattern)
  nc::capture_first_df(pos.dt, pos=chrom.start.pattern)
  expect_identical(pos.dt, data.table(pos, chrom="chr1", start=100L))
  full.pattern <- list(chrom.start.pattern, "-", end=int.pattern)
  expect_error({
    nc::capture_first_df(pos.dt, pos=full.pattern)
  }, "capture group names (chrom, start, end) must not conflict with existing column names (pos, chrom, start); fix by changing capture group names or use existing.error=FALSE to overwrite existing column names", fixed=TRUE)
  nc::capture_first_df(pos.dt, pos=full.pattern, existing.error=FALSE)
  expect_identical(pos.dt, data.table(pos, chrom="chr1", start=100L, end=200L))
})

library(testthat)
library(nc)
library(data.table)
context("measure")
source(system.file("test_engines.R", package="nc", mustWork=TRUE), local=TRUE)

DT <- data.table(a_1=10, b_2=21, a_2=20)
test_engines("measure single value", {
  out <- melt(DT, measure.vars=nc::measure(
    letter="[ab]", "_", number="[12]", as.integer))
  expected <- data.table(
    letter=c("a","b","a"),
    number=as.integer(c(1,2,2)),
    value=c(10,21,20))
  expect_identical(out, expected)
})

test_engines("measure single value type.convert=TRUE", {
  out <- melt(DT, measure.vars=nc::measure(
    letter="[ab]", "_", number="[12]", type.convert=TRUE))
  expected <- data.table(
    letter=c("a","b","a"),
    number=as.integer(c(1,2,2)),
    value=c(10,21,20))
  expect_identical(out, expected)
})

test_engines("measure multiple values", {
  out <- melt(DT, measure.vars=nc::measure(
    column="[ab]", "_", number="[12]", as.integer))
  expected <- data.table(
    number=as.integer(c(1,2)),
    a=c(10,20),
    b=c(NA,21))
  expect_identical(out, expected)
})

test_engines("measure multiple values type.convert=TRUE", {
  out <- melt(DT, measure.vars=nc::measure(
    column="[ab]", "_", number="[12]", type.convert=TRUE))
  expected <- data.table(
    number=as.integer(c(1,2)),
    a=c(10,20),
    b=c(NA,21))
  expect_identical(out, expected)
})

library(nc)
library(testthat)
library(data.table)
context("all")
source(system.file("test_engines.R", package="nc", mustWork=TRUE), local=TRUE)

test_engines("capture_all_str has no warning about no such file", {
  expect_warning({
    capture_all_str("chr1:100", baz="sars")
  }, regexp=NA)
})

subject <- c("foobar", "FOOBAR")
test_engines("capture_all_str returns data.frame with 0 rows, 1 chr col", {
  computed <- capture_all_str(subject, baz="sars")
  expect_identical(computed$baz, character())
})

test_engines("capture_all_str returns data.frame with 0 rows, 1 int col", {
  subject <- c("foobar", "FOOBAR")
  computed <- capture_all_str(subject, baz="sars", as.integer)
  expect_identical(computed$baz, integer())
})

keep.digits <- function(x)as.integer(gsub("[^0-9]", "", x))
test_engines("capture_all_str returns data.table", {
  chr.pos.vec <- c(
    "chr10:213,054,000-213,055,000",
    "chrM:111,000-222,000",
    "this will not match",
    NA, # neither will this.
    "chr1:110-111 chr2:220-222") # two possible matches.
  computed <- capture_all_str(
    chr.pos.vec,
    chrom="chr.*?",
    ":",
    chromStart=".*?", keep.digits,
    "-",
    chromEnd="[0-9,]*", keep.digits)
  expected <- rbind(
    data.table(
      chrom="chr10", chromStart=213054000L, chromEnd=213055000L),
    data.table(
      chrom="chrM", chromStart=111000L, chromEnd=222000L),
    data.table(
      chrom=c("chr1", "chr2"),
      chromStart=as.integer(c("110", "220")),
      chromEnd=as.integer(c("111", "222"))))
  expect_identical(computed, expected)
})

test_engines("capture_all_str(type.convert=TRUE) returns one int column", {
  computed <- capture_all_str(
    "chr1:2-3,000 chr4:5-6,000",
    chrom="chr.*?",
    ":",
    chromStart=".*?",
    "-",
    chromEnd="[0-9,]*",
    type.convert=TRUE)
  expected <- data.table(
    chrom=c("chr1","chr4"),
    chromStart=c(2L,5L),
    chromEnd=c("3,000","6,000"))
  expect_identical(computed, expected)
})

test_engines("capture_all_str(type.convert=TRUE) returns two int columns", {
  computed <- capture_all_str(
    "chr1:2-3,000 chr4:5-6,000",
    chrom="chr.*?",
    ":",
    chromStart=".*?",
    "-",
    chromEnd="[0-9,]*", keep.digits,
    type.convert=TRUE)
  expected <- data.table(
    chrom=c("chr1","chr4"),
    chromStart=c(2L,5L),
    chromEnd=c(3000L,6000L))
  expect_identical(computed, expected)
})

test_engines("capture_all_str errors for one argument", {
  expect_error({
    capture_all_str("foo")
  }, "must have at least one named argument")
})

test_engines("capture_all_str errors for multi-dim patterns", {
  expect_error({
    capture_all_str("foo", sars=c("bar", "baz"))
  }, "patterns must be character vectors of length 1")
})

test_engines("capture_all_str errors for 0-length patterns", {
  expect_error({
    capture_all_str("foo", bar=character())
  }, "patterns must be character vectors of length 1")
})

test_engines("capture_all_str errors for non char/fun args", {
  expect_error({
    capture_all_str("foo", "bar", 1)
  }, "arguments must be", fixed=TRUE)
})

test_engines("capture_all_str errors for two funs in a row", {
  expect_error({
    capture_all_str("foo", g="bar", as.integer, as.numeric)
  },
  "too many functions; up to one function may follow each named pattern")
})

test_engines("capture_all_str errors for fun at start", {
  expect_error({
    capture_all_str("foo", as.numeric)
  },
  "too many functions; up to one function may follow each named pattern")
})

test_engines("capture_all_str errors for NA pattern", {
  expect_error({
    capture_all_str("foo", g="bar", NA_character_, "baz")
  }, "patterns must not be missing/NA")
})

test_engines("nested capture groups works", {
  trackDb.txt.gz <- system.file("extdata", "trackDb.txt.gz", package="nc")
  track.pattern <- list(
    cellType=".*?",
    "_",
    sampleName=list(as.factor,
                    "McGill",
                    sampleID="[0-9]+", as.integer),
    dataType="Coverage|Peaks",
    "|",
    "[^\n]+")
  match.dt <- capture_all_str(
    trackDb.txt.gz,
    "track ",
    track=track.pattern,
    "(?:\n[^\n]+)*",
    "\\s+bigDataUrl ",
    bigDataUrl="[^\n]+")
  expect_equal(nrow(match.dt), 78)
  expect_is(match.dt, "data.table")
  expect_identical(
    names(match.dt),
    c("track", "cellType", "sampleName", "sampleID",
      "dataType", "bigDataUrl"))
  expect_is(match.dt$sampleName, "factor")
  expect_is(match.dt$sampleID, "integer")
})

test_engines("error for capture all regex with literal groups, match", {
  expect_error({
    capture_all_str(
      c("chr1:100-200", "chr2:5-6"),
      chrom="chr.",
      ":",
      "([0-9]+)")
  }, "regex contains more groups than names; please remove literal groups (parentheses) from the regex pattern, and use named arguments in R code instead", fixed=TRUE)
})

test_engines("error for capture all regex with literal groups, no match", {
  expect_error({
    nc::capture_all_str("alias(es)", foo="alias(es)")
  }, "regex contains more groups than names; please remove literal groups (parentheses) from the regex pattern, and use named arguments in R code instead", fixed=TRUE)
})

test_engines("before_match markdown link, no before", {
  markdown_link <- list(
    "\\[",
    title="[^]]+?",
    "\\]\\(",
    url="http.*?",
    "\\)")
  markdown_subject <- "[foo](http)[bar text](http)"
  expected_dt <- rowwiseDT(
    before=, match=,             title=,      url=,
    "",      "[foo](http)",      "foo",       "http",
    "",      "[bar text](http)", "bar text",  "http")
  markdown_dt <- nc::capture_all_str(markdown_subject, markdown_link)
  expect_identical(markdown_dt, expected_dt[, .(title, url)])
  before_link <- nc::before_match(markdown_link)
  before_dt <- nc::capture_all_str(markdown_subject, before_link)
  expect_identical(before_dt, expected_dt)
  ch14.qmd <- system.file(package="nc", "extdata", "ch14.qmd", mustWork=TRUE)
  ch14.lines <- readLines(ch14.qmd)
  ch14.str <- paste(ch14.lines, collapse="\n")
  ch14_dt <- nc::capture_all_str(ch14.str, before_link)
  expect_equal(nrow(ch14_dt), 3)
  expect_identical(ch14_dt[, paste(paste0(before, match), collapse="")], ch14.str)
})

test_engines("before_match markdown link", {
  markdown_link <- list(
    "\\[",
    title="[^]]+?",
    "\\]\\(",
    url="http.*?",
    "\\)")
  markdown_subject <- "be\nfore [foo](http) be\ntween [bar text](http) af\nter\n"
  expected_dt <- rowwiseDT(
    before=,       match=,             title=,     url=,
    "be\nfore ",   "[foo](http)",      "foo",      "http",
    " be\ntween ", "[bar text](http)", "bar text", "http",
    " af\nter\n",  "",                 "",         "")
  markdown_dt <- nc::capture_all_str(markdown_subject, markdown_link)
  expect_identical(markdown_dt, expected_dt[1:2, .(title, url)])
  before_link <- nc::before_match(markdown_link)
  before_dt <- nc::capture_all_str(markdown_subject, before_link)
  expect_identical(before_dt, expected_dt)
  ch14.qmd <- system.file(package="nc", "extdata", "ch14.qmd", mustWork=TRUE)
  ch14.lines <- readLines(ch14.qmd)
  ch14.str <- paste(ch14.lines, collapse="\n")
  ch14_dt <- nc::capture_all_str(ch14.str, before_link)
  expect_equal(nrow(ch14_dt), 3)
  expect_identical(ch14_dt[, paste(paste0(before, match), collapse="")], ch14.str)
})

test_engines("before_match data.table Archive web page", {
  Archive.html <- system.file(
    package="nc", "extdata", "data.table.Archive.html", mustWork = TRUE)
  Archive.lines <- readLines(Archive.html)
  a.pattern <- list(
    '<a href="',
    href=".+?",
    '">',
    text=".+?",
    "</a>")
  adt <- nc::capture_all_str(
    Archive.lines,
    a.pattern)
  bdt <- nc::capture_all_str(
    Archive.lines,
    nc::before_match(a.pattern))
  expect_identical(bdt$text, c(adt$text, ""))
  expect_identical(names(bdt), c("before", "match", "href", "text"))
})

<!--
%\VignetteEngine{knitr::knitr}
%\VignetteIndexEntry{vignette 0: Overview}
-->

# Overview of nc functionality

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

Here is an index of topics which are explained in the different
vignettes, along with an overview of functionality using simple
examples.

## Capture first match in several subjects

[Capture first](v1-capture-first.html) is for the situation when your
input is a character vector (each element is a different subject), you
want find the first match of a regex to each subject, and your desired
output is a data table (one row per subject, one column per capture
group in the regex). 

```{r}
subject.vec <- c(
  "chr10:213054000-213,055,000",
  "chrM:111000",
  "chr1:110-111 chr2:220-222")
nc::capture_first_vec(
  subject.vec, chrom="chr.*?", ":", chromStart="[0-9,]+", as.integer)
```

A variant is doing the same thing, but with input
subjects coming from a data table/frame with character columns.

```{r}
subject.dt <- data.table::data.table(
  JobID = c("13937810_25", "14022192_1"),
  Elapsed = c("07:04:42", "07:04:49"))
int.pat <- list("[0-9]+", as.integer)
nc::capture_first_df(
  subject.dt,
  JobID=list(job=int.pat, "_", task=int.pat),
  Elapsed=list(hours=int.pat, ":", minutes=int.pat, ":", seconds=int.pat))
```

## Capture all matches in a single subject
  
[Capture all](v2-capture-all.html) is for the situation when your
input is a single character string or text file subject, you want to
find all matches of a regex to that subject, and your desired output
is a data table (one row per match, one column per capture group in
the regex).

```{r}
nc::capture_all_str(
  subject.vec, chrom="chr.*?", ":", chromStart="[0-9,]+", as.integer)
```

## Reshape a data table with regularly named columns 

[Capture melt](v3-capture-melt.html) is for the situation when your
input is a data table/frame that has regularly named columns, and your
desired output is a data table with those columns reshaped into a
taller/longer form. In that case you can use a regex to identify the
columns to reshape.

```{r}
(one.iris <- data.frame(iris[1,]))
nc::capture_melt_single  (one.iris, part  =".*", "[.]", dim   =".*")
nc::capture_melt_multiple(one.iris, column=".*", "[.]", dim   =".*")
nc::capture_melt_multiple(one.iris, part  =".*", "[.]", column=".*")
```

## Helper functions for defining complex pattterns
  
[Helpers](v5-helpers.html) describes various functions that simplify
the definition of complex regex patterns. For example `nc::field`
helps avoid repetition below,

```{r}
subject.vec <- c("sex_child1", "age_child1", "sex_child2")
pattern <- list(
  variable="age|sex", "_",
  nc::field("child", "", "[12]", as.integer))
nc::capture_first_vec(subject.vec, pattern)
```

It also explains how to define common sub-patterns which are used in
several different alternatives.

```{r}
subject.vec <- c("mar 17, 1983", "26 sep 2017", "17 mar 1984")
pattern <- nc::alternatives_with_shared_groups(
  month="[a-z]{3}", day="[0-9]{2}", year="[0-9]{4}",
  list(month, " ", day, ", ", year),
  list(day, " ", month, " ", year))
nc::capture_first_vec(subject.vec, pattern)
```

---
title: "Helper functions"
date: "`r Sys.Date()`"
output:
  rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Helper functions}
  %\VignetteEngine{knitr::rmarkdown}
  \usepackage[utf8]{inputenc}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

There are several "helper" functions which can simplify the definition
of complex patterns. First we define some functions that will help us
display the patterns:

```{r}
one.pattern <- function(pat){
  if(is.character(pat)){
    pat
  }else{
    nc::var_args_list(pat)[["pattern"]]
  }
}
show.patterns <- function(...){
  L <- list(...)
  str(lapply(L, one.pattern))
}
```

# `nc::field` for reducing repetition

The `nc::field` function can be used to avoid repetition when defining
patterns of the form `variable: value`. The example below shows three
(mostly) equivalent ways to write a regex that captures the text after
the colon and space; the captured text is stored in the `variable`
group or output column:

```{r}
show.patterns(
  "variable: (?<variable>.*)",      #repetitive regex string
  list("variable: ", variable=".*"),#repetitive nc R code
  nc::field("variable", ": ", ".*"))#helper function avoids repetition
```

Note that the first version above has a named capture group, whereas
the second and third patterns generated by nc have an un-named capture
group and some non-capturing groups (but they all match the same
pattern).

Another example:

```{r}
show.patterns(
  "Alignment (?<Alignment>[0-9]+)",
  list("Alignment ", Alignment="[0-9]+"),
  nc::field("Alignment", " ", "[0-9]+"))
```

Another example:

```{r}
show.patterns(
  "Chromosome:\t+(?<Chromosome>.*)",
  list("Chromosome:\t+", Chromosome=".*"),
  nc::field("Chromosome", ":\t+", ".*"))
```

# `nc::quantifier` for fewer parentheses

Another helper function is `nc::quantifier` which makes patterns
easier to read by reducing the number of parentheses required to
define sub-patterns with quantifiers. For example all three patterns
below create an optional non-capturing group which contains a named
capture group:

```{r}
show.patterns(
  "(?:-(?<chromEnd>[0-9]+))?",                #regex string
  list(list("-", chromEnd="[0-9]+"), "?"),    #nc pattern using lists
  nc::quantifier("-", chromEnd="[0-9]+", "?"))#quantifier helper function
```

Another example with a named capture group inside an optional
non-capturing group:

```{r}
show.patterns(
  "(?: (?<name>[^,}]+))?",
  list(list(" ", name="[^,}]+"), "?"),
  nc::quantifier(" ", name="[^,}]+", "?"))
```

# `nc::alternatives` for simplified alternation

We also provide a helper function for defining regex patterns with
[alternation](https://www.regular-expressions.info/alternation.html). The
following three lines are equivalent.

```{r}
show.patterns(
  "(?:(?<first>bar+)|(?<second>fo+))",
  list(first="bar+", "|", second="fo+"),
  nc::alternatives(first="bar+", second="fo+"))
```

# `nc::alternatives_with_shared_groups` for alternatives with identical named sub-pattern groups

Sometimes each alternative is just a re-arrangement of the same
sub-patterns. For example consider the following subjects, each of
which are dates, in one of two formats.

```{r}
subject.vec <- c("mar 17, 1983", "26 sep 2017", "17 mar 1984")
```

In each of the two formats, the month consists of three lower-case
letters, the day consists of two digits, and the year consists of four
digits. Is there a single pattern that can match each of these
subjects? Yes, such a pattern can be defined using the code below,

```{r}
pattern <- nc::alternatives_with_shared_groups(
  month="[a-z]{3}",
  day=list("[0-9]{2}", as.integer),
  year=list("[0-9]{4}", as.integer),
  list(american=list(month, " ", day, ", ", year)),
  list(european=list(day, " ", month, " ", year)))
```

In the code above, we used `nc::alternatives_with_shared_groups`,
which requires two kinds of arguments:

* named arguments (month, day, year) define sub-pattern groups that
  are used in each alternative.
* un-named arguments (last two) define alternative patterns, each
  which can use the sub-pattern group names (month, day, year).

The pattern can be used for matching, and the result is a data table
with one column for each unique name, 

```{r}
(match.dt <- nc::capture_first_vec(subject.vec, pattern))
```

After having parsed the dates into these three columns, we can add a
date column:

```{r}
Sys.setlocale(locale="C")#to recognize months in English.
match.dt[, date := data.table::as.IDate(
  paste(month, day, year), format="%b %d %Y")]
print(match.dt, class=TRUE)
```

Another example is parsing given and family names, in two different
formats:

```{r}
nc::capture_first_vec(
  c("Toby Dylan Hocking","Hocking, Toby Dylan"),
  nc::alternatives_with_shared_groups(
    family="[A-Z][a-z]+",
    given="[^,]+",
    list(given_first=list(given, " ", family)),
    list(family_first=list(family, ", ", given))
  )
)
```

---
title: "Capture first match"
date: "`r Sys.Date()`"
output:
  rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Capture first match}
  %\VignetteEngine{knitr::rmarkdown}
  \usepackage[utf8]{inputenc}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Capture first match in each character vector element

Consider the following vector which contains genome position strings,

```{r}
pos.vec <- c(
  "chr10:213,054,000-213,055,000",
  "chrM:111,000",
  "chr1:110-111 chr2:220-222") # two possible matches.
```

To capture the first genome position in each string, we use the following syntax.
The first argument is the subject character vector, 
and the other arguments are pasted together 
to make a capturing regular expression.
Each named argument generates a capture group; 
the R argument name is used for the column name of the result.

```{r}
(chr.dt <- nc::capture_first_vec(
  pos.vec, 
  chrom="chr.*?",
  ":",
  chromStart="[0-9,]+"))
str(chr.dt)
```

We can add type conversion functions on the same line as each named argument:

```{r}
keep.digits <- function(x)as.integer(gsub("[^0-9]", "", x))
(int.dt <- nc::capture_first_vec(
  pos.vec, 
  chrom="chr.*?",
  ":",
  chromStart="[0-9,]+", keep.digits))
str(int.dt)
```

Below we use list variables to create patterns which are re-usable,
and we use an un-named list to generate a non-capturing optional group:

```{r}
pos.pattern <- list("[0-9,]+", keep.digits)
range.pattern <- list(
  chrom="chr.*?",
  ":",
  chromStart=pos.pattern,
  list(
    "-",
    chromEnd=pos.pattern
  ), "?")
nc::capture_first_vec(pos.vec, range.pattern)
```

In summary, `nc::capture_first_vec` takes a variable number of arguments:

* The first argument is the subject character vector.
* The other arguments specify the pattern, via character strings,
  functions, and/or lists.
* If a pattern (character/list) is named, a capture group is generated in the regex.
* Each function is used to convert the text extracted by the previous
  named pattern argument.
* Lists may be used to avoid repetition in the definition of the pattern and type conversion functions.
* Each list generates a group in the regex (named list => named capture group, un-named list => non-capturing group).
* All patterns are pasted together in the order that they appear in
  the argument list.
  
## View generated regex

To see the generated regular expression pattern string, call
`nc::var_args_list` with the variable number of arguments that
specify the pattern:

```{r}
nc::var_args_list(range.pattern)
```

The generated regex is the `pattern` element of the resulting list
above. The other element `fun.list` indicates the names and type
conversion functions to use with the capture groups.

## Error/NA if any subjects do not match

The default is to stop with an error if any subject does not match:

```{r, error=TRUE, purl=FALSE}
bad.vec <- c(bad="does not match", pos.vec)
nc::capture_first_vec(bad.vec, range.pattern)
```

Sometimes you want to instead report a row of NA when a subject does
not match. In that case, use `nomatch.error=FALSE`:

```{r}
nc::capture_first_vec(bad.vec, range.pattern, nomatch.error=FALSE)
```

## Other regex engines

By default nc uses the PCRE regex engine. Other choices include ICU
and RE2. Each engine has different features, which are discussed in
[my R journal paper](https://github.com/tdhock/namedCapture-article).

The engine is configurable via the `engine` argument or the
`nc.engine` option:

```{r, error=TRUE, purl=FALSE}
u.subject <- "a\U0001F60E#"
u.pattern <- list(emoji="\\p{EMOJI_Presentation}")
old.opt <- options(nc.engine="ICU")
nc::capture_first_vec(u.subject, u.pattern)
nc::capture_first_vec(u.subject, u.pattern, engine="PCRE") 
nc::capture_first_vec(u.subject, u.pattern, engine="RE2")
options(old.opt)
```

For more details see the "engines" vignette, 

```r
vignette("v6-engines", package="nc")
```

## Capture first match from one or more character columns in a table

We also provide `nc::capture_first_df` which extracts text
from several columns of a data.frame, using a different 
regular expression for each column.

* It requires a `data.frame` (or `data.table`) as the first argument.
* It takes a variable number of other arguments, all of which must be
  named. For each other argument we call `nc::capture_first_vec` on one
  column of the input table.
* Each argument name specifies a column of the input table which will
  be used as the subject in `nc::capture_first_vec`.
* Each argument value specifies a pattern to be used with
  `nc::capture_first_vec`, in list/character/function format as
  explained in the previous section.
* The return value is a `data.table` with the same number of rows as
  the input, but with an additional column for each named capture
  group. New columns have the same names as the capture groups, so
  make sure they are unique.
* This function inputs `data.frame` and outputs `data.table` so can be
  used in a [pipe](https://r4ds.had.co.nz/pipes.html).
* For memory efficiency when the input is a `data.table`, it is
  modified to avoid copying the entire table.
  
This function can greatly simplify the code required to create numeric
data columns from character data columns. For example consider the
following data which was output from the
[sacct](https://slurm.schedmd.com/sacct.html) program.

```{r}
(sacct.df <- data.frame(
  Elapsed = c(
    "07:04:42", "07:04:42", "07:04:49",
    "00:00:00", "00:00:00"),
  JobID=c(
    "13937810_25",
    "13937810_25.batch",
    "13937810_25.extern",
    "14022192_[1-3]",
    "14022204_[4]"),
  stringsAsFactors=FALSE))
```

Say we want to filter by the total Elapsed time (which is reported as
hours:minutes:seconds), and base job id (which is the number before
the underscore in the JobID column). We could start by converting
those character columns to integers via:

```{r}
int.pattern <- list("[0-9]+", as.integer)
range.pattern <- list(
  "\\[",
  task1=int.pattern,
  list(
    "-",#begin optional end of range.
    taskN=int.pattern
  ), "?", #end is optional.
  "\\]")
nc::capture_first_df(sacct.df, JobID=range.pattern, nomatch.error=FALSE)
```
The result shown above is another data frame with an additional column
for each capture group. Next, we define another pattern that
matches either one task ID or the previously defined range pattern:

```{r}
task.pattern <- list(
  "_",
  list(
    task=int.pattern,
    "|",#either one task(above) or range(below)
    range.pattern))
nc::capture_first_df(sacct.df, JobID=task.pattern)
```

Below we match the complete JobID column:

```{r}
job.pattern <- list(
  job=int.pattern,
  task.pattern,
  list(
    "[.]",
    type=".*"
  ), "?")
nc::capture_first_df(sacct.df, JobID=job.pattern)
```

Below we match the Elapsed column with a different regex:

```{r}
elapsed.pattern <- list(
  hours=int.pattern,
  ":",
  minutes=int.pattern,
  ":",
  seconds=int.pattern)
nc::capture_first_df(sacct.df, JobID=job.pattern, Elapsed=elapsed.pattern)
```

Overall the result is another data table with an additional column for
each capture group. Note in the code below that the input `sacct.df`
was not modified, but if the input is `data.table` then it is modified:

```{r}
nc::capture_first_df(sacct.df, JobID=job.pattern)
sacct.df
(sacct.DT <- data.table::as.data.table(sacct.df))
nc::capture_first_df(sacct.DT, JobID=job.pattern)
sacct.DT
```

## Complex example: fsa file names

In this section we explain how to use various features to parse the
fsa file names in
[PROVEDIt](https://lftdi.camden.rutgers.edu/provedit/files/). Here are
a few representative examples:

```{r}
fsa.vec <- c(#control samples:
  "A01-Ladder-PP16-001.10sec.fsa",
  "D07_Ladder-IP_004.5sec.fsa",
  "A12_RB121514ADG_001.10sec.fsa", 
  "A10_RB102191515LEA-IP_001.10sec.fsa",
  ##single-source samples:
  "A02-RD12-0002-35-0.5PP16-001.10sec.fsa", 
  "G01_RD14-0003-35d3S30-0.01563P-Q10.0_003.10sec.fsa",
  "A06_RD14-0003-24d3a-0.0625IP-Q0.8_001.10sec.fsa", 
  "A08-RD12-0002-01d-0.125PP16-001.10sec.fsa",
  "A10-RD12-0002-04d1-0.0625PP16-001.10sec.fsa", 
  "C02_RD14-0003-15d2b-0.25IP-Q0.5_003.5sec.fsa",
  ##mixture samples:
  "A02-RD12-0002-1_2-1;9-0.125PP16-001.10sec.fsa", 
  "H07_RD14-0003-35_36_37_38_39-1;4;4;4;1-M2I35-0.75IP-QLAND_004.5sec.fsa")
```

The goal is to build a regex that can convert this character vector to
a data table with different columns for the different variables. The
structure of the file names is explained in [the supplementary
materials
PDF](https://lftdi.camden.rutgers.edu/wp-content/uploads/2019/12/PROVEDIt-Database-Naming-Convention-Laboratory-Methodsv1.pdf). We
will build a complex regex in terms of simpler sub-patterns. First
let's just match the start of each file name, which has the row/column
of the 96-well plate in which the sample was tested:

```{r}
well.pattern <- list(
  "^",
  well.letter="[A-H]",
  well.number="[0-9]+", as.integer)
nc::capture_first_vec(fsa.vec, well.pattern)
```

Now, let's match the end of each file name:

```{r}
end.pattern <- list(
  "[-_]",
  capillary="[0-9]+", as.integer,
  "[.]",
  seconds="[0-9]+", as.integer,
  "sec[.]fsa$")
nc::capture_first_vec(fsa.vec, end.pattern)
```

Now, let's take a look at what's in between:

```{r}
between.dt <- nc::capture_first_vec(
  fsa.vec, well.pattern, between=".*?", end.pattern)
between.dt$between
```

Notice that PP16/P/IP are the kit types. There may optionally be a DNA
template mass before, and optionally a Q score after. Let's match those:

```{r}
mass.pattern <- list(
  template.nanograms="[0-9.]*", as.numeric)
kit.pattern <- nc::quantifier(
  kit=nc::alternatives("PP16", "IP", "P"),
  "?")
q.pattern <- nc::quantifier(
  "-Q",
  Q.chr=nc::alternatives("LAND", "[0-9.]+"),
  "?")
old.opt <- options(width=100)
(before.dt <- nc::capture_first_vec(
  between.dt$between,
  "^",
  before=".*?",
  mass.pattern, kit.pattern, q.pattern, "$"))
options(old.opt)
```

Now to match the before column, we need some alternatives. The first
four rows are controls, and the other rows are samples which are
indicated by a project ID (prefix RD) and then a sample ID. The
mixture samples at the bottom contain semicolon-delimited mixture
proportions.

```{r}
project.pattern <- list(project="RD[0-9]+-[0-9]+", "-")
single.pattern <- list(id="[0-9]+", as.integer)
mixture.pattern <- list(
  ids.chr="[0-9_]+",
  "-",
  parts.chr="[0-9;]+",
  "-")
single.or.mixture <- list(
  project.pattern,
  nc::alternatives(mixture.pattern, single.pattern))
control.pattern <- list(control="[^-_]+")
single.mixture.control <- list("[-_]", nc::alternatives(
  single.or.mixture, control.pattern))
(rest.dt <- nc::capture_first_vec(
  before.dt$before, single.mixture.control, rest=".*"))
```

The rest may contain some information related to the dilution and
treatment, which we can parse as follows:

```{r}
dilution.pattern <- nc::quantifier(
  "[dM]",
  dilution.number="[0-9]?", as.integer,
  "?")
treatment.pattern <- nc::quantifier(
  treatment.letter="[a-zA-Z]",
  nc::quantifier(treatment.number="[0-9]+", as.integer, "?"),
  "?")
dilution.treatment <- list(
  dilution.pattern,
  treatment.pattern,
  "[_-]?")
nc::capture_first_vec(rest.dt$rest, "^", dilution.treatment, "$")
```

Now we just have to put all of those patterns together to get a full
match:

```{r}
fsa.pattern <- list(
  well.pattern,
  single.mixture.control,
  dilution.treatment,
  mass.pattern, kit.pattern, q.pattern,  
  end.pattern)
(match.dt <- nc::capture_first_vec(fsa.vec, fsa.pattern))
```

And to verify we parsed everything correctly we print the subjects
next to each row:

```{r}
disp.dt <- data.table::data.table(match.dt)
names(disp.dt) <- paste(1:ncol(disp.dt))
old.opt <- options("datatable.print.colnames"="none")
split(disp.dt, fsa.vec)
options(old.opt)
```

Incidentally, if you print the regex as a string it looks like
something that would be essentially impossible to write/maintain by
hand (without `nc` helper functions):

```{r}
nc::var_args_list(fsa.pattern)$pattern
```

In conclusion, we have shown how `nc` can help build complex regex
patterns from simple, understandable sub-patterns in R code. The
general workflow we have followed in this section can be generalized
to other projects. First, identify a small set of subjects to use for
testing (`fsa.vec` in the code above). Second, create some
sub-patterns for the start/end of the subjects, and create a
sub-pattern that will capture everything else in between. Third,
create some more sub-patterns to match the "everything else" in the
previous step (e.g. `before.dt$before`, `between.dt$between`,
`rest.dt$rest` above). Repeat the process until there is nothing left
to match, and then concatenate the sub-patterns to match the whole
subject.

<!--
%\VignetteEngine{knitr::knitr}
%\VignetteIndexEntry{vignette 1: capture first match}
-->

# Capture first match

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Capture first match in each character vector element

Consider the following vector which contains genome position strings,

```{r}
pos.vec <- c(
  "chr10:213,054,000-213,055,000",
  "chrM:111,000",
  "chr1:110-111 chr2:220-222") # two possible matches.
```

To capture the first genome position in each string, we use the following syntax.
The first argument is the subject character vector, 
and the other arguments are pasted together 
to make a capturing regular expression.
Each named argument generates a capture group; 
the R argument name is used for the column name of the result.

```{r}
(chr.dt <- nc::capture_first_vec(
  pos.vec, 
  chrom="chr.*?",
  ":",
  chromStart="[0-9,]+"))
str(chr.dt)
```

We can add type conversion functions on the same line as each named argument:

```{r}
keep.digits <- function(x)as.integer(gsub("[^0-9]", "", x))
(int.dt <- nc::capture_first_vec(
  pos.vec, 
  chrom="chr.*?",
  ":",
  chromStart="[0-9,]+", keep.digits))
str(int.dt)
```

Below we use list variables to create patterns which are re-usable,
and we use an un-named list to generate a non-capturing optional group:

```{r}
pos.pattern <- list("[0-9,]+", keep.digits)
range.pattern <- list(
  chrom="chr.*?",
  ":",
  chromStart=pos.pattern,
  list(
    "-",
    chromEnd=pos.pattern
  ), "?")
nc::capture_first_vec(pos.vec, range.pattern)
```

In summary, `nc::capture_first_vec` takes a variable number of arguments:

* The first argument is the subject character vector.
* The other arguments specify the pattern, via character strings,
  functions, and/or lists.
* If a pattern (character/list) is named, a capture group is generated in the regex.
* Each function is used to convert the text extracted by the previous
  named pattern argument.
* Lists may be used to avoid repetition in the definition of the pattern and type conversion functions.
* Each list generates a group in the regex (named list => named capture group, un-named list => non-capturing group).
* All patterns are pasted together in the order that they appear in
  the argument list.
  
## View generated regex

To see the generated regular expression pattern string, call
`nc::var_args_list` with the variable number of arguments that
specify the pattern:

```{r}
nc::var_args_list(range.pattern)
```

The generated regex is the `pattern` element of the resulting list
above. The other element `fun.list` indicates the names and type
conversion functions to use with the capture groups.

## Error/NA if any subjects do not match

The default is to stop with an error if any subject does not match:

```{r, error=TRUE, purl=FALSE}
bad.vec <- c(bad="does not match", pos.vec)
nc::capture_first_vec(bad.vec, range.pattern)
```

Sometimes you want to instead report a row of NA when a subject does
not match. In that case, use `nomatch.error=FALSE`:

```{r}
nc::capture_first_vec(bad.vec, range.pattern, nomatch.error=FALSE)
```

## Other regex engines

By default nc uses the PCRE regex engine. Other choices include ICU
and RE2. Each engine has different features, which are discussed in
[my R journal paper](https://github.com/tdhock/namedCapture-article).

The engine is configurable via the `engine` argument or the
`nc.engine` option:

```{r, error=TRUE, purl=FALSE}
u.subject <- "a\U0001F60E#"
u.pattern <- list(emoji="\\p{EMOJI_Presentation}")
old.opt <- options(nc.engine="ICU")
nc::capture_first_vec(u.subject, u.pattern)
nc::capture_first_vec(u.subject, u.pattern, engine="PCRE") 
nc::capture_first_vec(u.subject, u.pattern, engine="RE2")
options(old.opt)
```

## Capture first match from one or more data.frame character columns 

We also provide `nc::capture_first_df` which extracts text
from several columns of a data.frame, using a different 
regular expression for each column.

* It requires a data.frame (or data.table) as the first argument.
* It takes a variable number of other arguments, all of which must be
  named. For each other argument we call `nc::capture_first_vec` on one
  column of the input data.frame.
* Each argument name specifies a column of the data.frame which will
  be used as the subject in `nc::capture_first_vec`.
* Each argument value specifies a pattern to be used with
  `nc::capture_first_vec`, in list/character/function format as
  explained in the previous section.
* The return value is a data.table with the same number of rows as the
  input, but with an additional column for each named capture
  group. New columns are named using the R argument names, so make
  sure they are unique.
* This function inputs data.frame and outputs data.table so can be used in a
  [pipe](https://r4ds.had.co.nz/pipes.html).
  
This function can greatly simplify the code required to create numeric
data columns from character data columns. For example consider the
following data which was output from the [sacct](https://slurm.schedmd.com/sacct.html) program.

```{r}
(sacct.df <- data.frame(
  Elapsed = c(
    "07:04:42", "07:04:42", "07:04:49",
    "00:00:00", "00:00:00"),
  JobID=c(
    "13937810_25",
    "13937810_25.batch",
    "13937810_25.extern",
    "14022192_[1-3]",
    "14022204_[4]"),
  stringsAsFactors=FALSE))
```

Say we want to filter by the total Elapsed time (which is reported as
hours:minutes:seconds), and base job id (which is the number before
the underscore in the JobID column). We could start by converting
those character columns to integers via:

```{r}
int.pattern <- list("[0-9]+", as.integer)
range.pattern <- list(
  "\\[",
  task1=int.pattern,
  list(
    "-",#begin optional end of range.
    taskN=int.pattern
  ), "?", #end is optional.
  "\\]")
nc::capture_first_df(sacct.df, JobID=range.pattern, nomatch.error=FALSE)
```
The result shown above is another data frame with an additional column
for each capture group. Next, we define another pattern that
matches either one task ID or the previously defined range pattern:

```{r}
task.pattern <- list(
  "_",
  list(
    task=int.pattern,
    "|",#either one task(above) or range(below)
    range.pattern))
nc::capture_first_df(sacct.df, JobID=task.pattern)
```

Below we match the complete JobID column:

```{r}
job.pattern <- list(
  job=int.pattern,
  task.pattern,
  list(
    "[.]",
    type=".*"
  ), "?")
nc::capture_first_df(sacct.df, JobID=job.pattern)
```

Below we match the Elapsed column with a different regex:

```{r}
elapsed.pattern <- list(
  hours=int.pattern,
  ":",
  minutes=int.pattern,
  ":",
  seconds=int.pattern)
nc::capture_first_df(sacct.df, JobID=job.pattern, Elapsed=elapsed.pattern)
```

Overall the result is another data table with an additional column for
each capture group. 

---
title: "Overview of nc functionality"
date: "`r Sys.Date()`"
output:
  rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Overview of nc functionality}
  %\VignetteEngine{knitr::rmarkdown}
  \usepackage[utf8]{inputenc}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

`nc` is a package for named capture regular expressions (regex), which
are useful for parsing/converting text data to tabular data (one row
per match, one column per capture group). In the terminology of regex,
we attempt to match a regex/pattern to a subject, which is a string of
text data. The regex/pattern is typically defined using a single
string (in other frameworks/packages/languages), but in `nc` we use a
special syntax: one or more R arguments are concatenated to define a
regex/pattern, and named arguments are used as capture groups.  For
more info about regex in general see
[regular-expressions.info](https://www.regular-expressions.info/reference.html)
and/or the Friedl book. For more
info about the special `nc` syntax, see `help("nc",package="nc")`.

Below is an index of topics which are explained in the different
vignettes, along with an overview of functionality using simple
examples.

## Capture first match in several subjects

[Capture first](v1-capture-first.html) is for the situation when your
input is a character vector (each element is a different subject to
parse), you want find the first match of a regex to each subject, and
your desired output is a data table (one row per subject, one column
per capture group in the regex).

```{r}
subject.vec <- c(
  "chr10:213054000-213,055,000",
  "chrM:111000",
  "chr1:110-111 chr2:220-222")
nc::capture_first_vec(
  subject.vec, chrom="chr.*?", ":", chromStart="[0-9,]+", as.integer)
```

A variant is doing the same thing, but with input
subjects coming from a data table/frame with character columns.

```{r}
library(data.table)
subject.dt <- data.table(
  JobID = c("13937810_25", "14022192_1"),
  Elapsed = c("07:04:42", "07:04:49"))
int.pat <- list("[0-9]+", as.integer)
nc::capture_first_df(
  subject.dt,
  JobID=list(job=int.pat, "_", task=int.pat),
  Elapsed=list(hours=int.pat, ":", minutes=int.pat, ":", seconds=int.pat))
```

## Capture all matches in a single subject
  
[Capture all](v2-capture-all.html) is for the situation when your
input is a single character string or text file subject, you want to
find all matches of a regex to that subject, and your desired output
is a data table (one row per match, one column per capture group in
the regex).

```{r}
nc::capture_all_str(
  subject.vec, chrom="chr.*?", ":", chromStart="[0-9,]+", as.integer)
```

## Reshape a data table with regularly named columns 

[Capture melt](v3-capture-melt.html) is for the situation when your
input is a data table/frame that has regularly named columns, and your
desired output is a data table with those columns reshaped into a
taller/longer form. In that case you can use a regex to identify the
columns to reshape.

```{r}
(one.iris <- data.frame(iris[1,]))
nc::capture_melt_single  (one.iris, part  =".*", "[.]", dim   =".*")
nc::capture_melt_multiple(one.iris, column=".*", "[.]", dim   =".*")
nc::capture_melt_multiple(one.iris, part  =".*", "[.]", column=".*")
```

## Reading regularly named data files

[Capture glob](v7-capture-glob.html) is for the situation when you
have several data files on disk, with regular names that you can match
with a glob/regex. In the example below we first write one CSV file
for each iris Species,

```{r}
dir.create(iris.dir <- tempfile())
icsv <- function(sp)file.path(iris.dir, paste0(sp, ".csv"))
data.table(iris)[, fwrite(.SD, icsv(Species)), by=Species]
dir(iris.dir)
```

We then use a glob and a regex to read those files in the code below:

```{r}
nc::capture_first_glob(file.path(iris.dir,"*.csv"), Species="[^/]+", "[.]csv")
```

## Helper functions for defining complex pattterns
  
[Helpers](v5-helpers.html) describes various functions that simplify
the definition of complex regex patterns. For example `nc::field`
helps avoid repetition below,

```{r}
subject.vec <- c("sex_child1", "age_child1", "sex_child2")
pattern <- list(
  variable="age|sex", "_",
  nc::field("child", "", "[12]", as.integer))
nc::capture_first_vec(subject.vec, pattern)
```

It also explains how to define common sub-patterns which are used in
several different alternatives.

```{r}
subject.vec <- c("mar 17, 1983", "26 sep 2017", "17 mar 1984")
pattern <- nc::alternatives_with_shared_groups(
  month="[a-z]{3}", day="[0-9]{2}", year="[0-9]{4}",
  list(month, " ", day, ", ", year),
  list(day, " ", month, " ", year))
nc::capture_first_vec(subject.vec, pattern)
```

<!--
%\VignetteEngine{knitr::knitr}
%\VignetteIndexEntry{vignette 3: capture melt}
-->

# Capture melt

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

This vignette explains how to use functions for "melting" wide data
tables, i.e. converting to tall/long data tables. To clarify the
discussion we introduce the following three terms:

* Reshape columns contain the data which are present in equal amounts
  in the input and output, but in fewer columns in the output.
  These are `measure.vars` in reshape2/data.table.
* Copy columns contain data which are copied from a single row of the
  input into several rows of the output. These are `id.vars`
  in reshape2/data.table.
* Capture columns are only present in the output, and contain data
  extracted from the names of the input reshape columns,
  using a capturing regular expression.
  
The `nc` functions use `data.table::melt` internally:

* `nc::capture_melt_single` is useful when you have a wide data table
  with lots of columns, all of the same type, that you would like to
  melt/reshape into a single result/output column. This uses `melt`
  with a character vector for `measure.vars`.
* `nc::capture_melt_multiple` is useful when you have a wide data
  table with columns of possibly different types that you would like
  to melt/reshape into multiple result/output columns. This uses
  `melt` with a list for `measure.vars`.

Both are useful when you want to use a regular expression to specify
both (1) the set of input columns to reshape and (2) some information
to extract from those column names.

## Reshaping several input columns into a single output column

Sometimes you want to melt a "wide" data table which has several
distinct pieces of information encoded in each column name. One
example is the familiar iris data, which have flower part and
measurement dimension encoded in each of four column names:

```{r}
head(iris)
```

Those four reshape column names can be specified via a regex in
`nc::capture_melt_single`. The first argument is the input data table
to reshape, and the subsequent arguments are interpreted as a pattern
which is passed to `nc::capture_first_vec`. Any input column names
which match the specified regex will be passed as `measure.vars` to
`melt`:

```{r}
(iris.tall <- nc::capture_melt_single(
  iris,
  part=".*",
  "[.]",
  dim=".*",
  value.name="cm"))
```

Note the output above has one copy column (Species), two capture
columns (part, dim), and one reshape column (cm). Internally the
function joins the result of `nc::capture_first_vec` (on column names)
to the result of `melt` (on the data).

The reshaped data can be plotted with different parts on rows and
different dimensions on columns:

```{r, fig.width=10}

if(require(ggplot2)){
  ggplot()+
    theme_bw()+
    theme(panel.spacing=grid::unit(0, "lines"))+
    facet_grid(part ~ dim)+
    geom_bar(aes(cm, fill=Species), data=iris.tall)
}

```

## Reshaping several input columns into multiple output columns

We could instead use `capture_melt_multiple` to get multiple output
columns. Like `capture_melt_single`, the first argument of
`capture_melt_multiple` is the subject data table and the following
arguments form a pattern which is matched to the input data column
names. However the pattern must have at least two groups:

* One group must be named `column`. This group is used to capture the
  part of the input column names which will be used for the output
  column names. The output will contain one column for each unique
  value captured in this group.
* One or more other groups (with any names) specify an identifier for
  each observation.

```{r}
(iris.part.cols <- nc::capture_melt_multiple(
  iris,
  column=".*",
  "[.]",
  dim=".*"))
```

Note that the reshaped table above contains one copy column (Species),
one capture column (dim), and two reshape columns (Petal, Sepal). We
can plot these data to see whether or not sepals are bigger than
petals:

```{r, fig.width=10}

if(require(ggplot2)){
  ggplot()+
    theme_bw()+
    theme(panel.spacing=grid::unit(0, "lines"))+
    facet_grid(dim ~ Species)+
    coord_equal()+
    geom_abline(slope=1, intercept=0, color="grey")+
    geom_point(aes(
      Petal, Sepal),
      data=iris.part.cols)
}

```

It is clear from the plot above that sepals are indeed both longer and
wider than petals, on each measured plant.

## Melting WHO data with a more complex pattern

Another data set where it is useful to do column name pattern matching
followed by melting is the World Health Organization data:

```{r}

if(requireNamespace("tidyr")){
  data(who, package="tidyr")
}else{
  who <- data.frame(id=1, new_sp_m5564=2, newrel_f65=3)
}
names(who)

```

Each column which starts with `new` has three distinct pieces of
information encoded in its name: diagnosis type (e.g. sp or rel),
gender (m or f), and age range (e.g. 5564 or 1524). We would like to
use a regex to match these column names, then using the matching
columns as measure.vars in a melt, then join the two results. 

```{r}
new.diag.gender <- list(
  "new_?",
  diagnosis=".*",
  "_",
  gender=".")
nc::capture_melt_single(who, new.diag.gender, ages=".*")
```

Note the output includes the new reshape column called `value` by
default, as in `melt`. The input reshape column names which matched
the specified pattern, and there is a new column for each group in
that pattern. The following example shows how to rename the `value`
column and use numeric type conversion functions:

```{r}
years.pattern <- list(new.diag.gender, ages=list(
  min.years="0|[0-9]{2}", as.numeric,
  max.years="[0-9]{0,2}", function(x)ifelse(x=="", Inf, as.numeric(x))))
(who.typed <- nc::capture_melt_single(
  who, years.pattern,
  value.name="count"))
str(who.typed)
```

Note in the code/result above that non-character captured output
columns can be obtained by specifying type conversion functions in the
pattern.

## Melting monthly Eurostat data

Another example is [exchange rate data from
Eurostat](https://ec.europa.eu/eurostat/estat-navtree-portlet-prod/BulkDownloadListing?file=data/ert_eff_ic_m.tsv.gz).
We first use read the entire data set into R:

```{r}
ert.gz <- system.file(
  "extdata", "ert_eff_ic_m.tsv.gz", package="nc", mustWork=TRUE)
ert.all <- data.table::fread(ert.gz, na.strings=":")
ert.all[1:5, 1:5]
```

We see that the first column has some CSV data which we can parse via:

```{r}
ert.first <- ert.all[, 1]
csv.lines <- c(sub("\\\\.*", "", names(ert.first)), ert.first[[1]])
ert.first.dt <- data.table::fread(text=paste(csv.lines, collapse="\n"))
ert.wide <- data.table::data.table(ert.first.dt, ert.all[,-1])
ert.wide[1:5, 1:5]
```

The wide data table can then be melted:


```{r}
(ert.tall <- nc::capture_melt_single(
  ert.wide,
  year="[0-9]{4}", as.integer,
  "M",
  month="[0-9]{2}", as.integer))
```

After that we can create a time variable and plot via

```{r}

ert.tall[, month.IDate := data.table::as.IDate(
  sprintf("%d-%d-15", year, month))]
if(require("ggplot2")){
  ggplot()+
    geom_hline(aes(
      yintercept=value),
      color="grey",
      data=data.frame(value=100))+
    geom_line(aes(
      month.IDate, value, color=geo),
      data=ert.tall[geo %in% c("CA", "US", "JP", "FR")])+
    facet_grid(exch_rt ~ .)+
    theme_bw()+
    theme(panel.spacing=grid::unit(0, "lines"))
}

```

Another way to do it would be via

```{r}
nc::capture_melt_single(ert.wide, month.POSIXct="[0-9].*", function(x){
  as.POSIXct(strptime(paste0(x,"15"), "%YM%m%d"))
})
```

## Melting into multiple output columns with missing input columns

What if the input data set has "missing" input columns?

```{r}
iris.missing <- iris[, names(iris) != "Sepal.Length"]
head(iris.missing)
```

In that case melting into multiple columns is an error by default:

```{r, error=TRUE, purl=FALSE}
iris.pattern <- list(column=".*", "[.]", dim=".*")
nc::capture_melt_multiple(iris.missing, iris.pattern)
```

The error message explains that the number of input columns for each
value of `dim` must be the same, but there is one for `Length` and two
for `Width`. To ignore the error and fill the output with missing
values,

```{r}
nc::capture_melt_multiple(iris.missing, iris.pattern, fill=TRUE)
```

Note the missing values in the table above, which correspond to the
missing input column in the original/wide data set.

## Multiple output columns of different types

Some real-world data sets can be reshaped into output columns with
different types. An example data set from the
[PROVEDIt](https://lftdi.camden.rutgers.edu/repository/PROVEDIt_1-5-Person%20CSVs%20Filtered.zip)
benchmark in criminology:

```{r}
peaks.csv <- system.file(
  "extdata", "RD12-0002_PP16HS_5sec_GM_F_1P.csv",
  package="nc", mustWork=TRUE)
peaks.wide <- data.table::fread(peaks.csv)
print(data.table::data.table(
  names=names(peaks.wide),
  class=sapply(peaks.wide, class)),
  topn=10)
```

There are 303 columns, with info for 100 peaks. Each peak has three
features: Allele=character, Size=numeric, and Height=integer. The
ending peaks are class logical because they are all missing. These
data can be reshaped via

```{r}
peaks.tall <- nc::capture_melt_multiple(
  peaks.wide,
  column=".*",
  " ",
  peak="[0-9]+", as.integer,
  na.rm=TRUE)
options(width=90)
print(peaks.tall)
str(peaks.tall)
```

<!--
%\VignetteEngine{knitr::knitr}
%\VignetteIndexEntry{vignette 5: Helper functions}
-->

# Helper functions

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

There are several "helper" functions which can simplify the definition
of complex patterns. First we define some functions that will help us
display the patterns:

```{r}
one.pattern <- function(pat){
  if(is.character(pat)){
    pat
  }else{
    nc::var_args_list(pat)[["pattern"]]
  }
}
show.patterns <- function(...){
  L <- list(...)
  str(lapply(L, one.pattern))
}
```

# nc::field for reducing repetition

The `nc::field` function can be used to avoid repetition when defining
patterns of the form `variable: value`. The example below shows three
(mostly) equivalent ways to write a regex that captures the text after
the colon and space; the captured text is stored in the `variable`
group or output column:

```{r}
show.patterns(
  "variable: (?<variable>.*)",      #repetitive regex string
  list("variable: ", variable=".*"),#repetitive nc R code
  nc::field("variable", ": ", ".*"))#helper function avoids repetition
```

Note that the first version above has a named capture group, whereas
the second and third patterns generated by nc have an un-named capture
group and some non-capturing groups (but they all match the same
pattern).

Another example:

```{r}
show.patterns(
  "Alignment (?<Alignment>[0-9]+)",
  list("Alignment ", Alignment="[0-9]+"),
  nc::field("Alignment", " ", "[0-9]+"))
```

Another example:

```{r}
show.patterns(
  "Chromosome:\t+(?<Chromosome>.*)",
  list("Chromosome:\t+", Chromosome=".*"),
  nc::field("Chromosome", ":\t+", ".*"))
```

# nc::quantifier for fewer parentheses

Another helper function is =nc::quantifier= which makes patterns
easier to read by reducing the number of parentheses required to
define sub-patterns with quantifiers. For example all three patterns
below create an optional non-capturing group which contains a named
capture group:

```{r}
show.patterns(
  "(?:-(?<chromEnd>[0-9]+))?",                #regex string
  list(list("-", chromEnd="[0-9]+"), "?"),    #nc pattern using lists
  nc::quantifier("-", chromEnd="[0-9]+", "?"))#quantifier helper function
```

Another example with a named capture group inside an optional
non-capturing group:

```{r}
show.patterns(
  "(?: (?<name>[^,}]+))?",
  list(list(" ", name="[^,}]+"), "?"),
  nc::quantifier(" ", name="[^,}]+", "?"))
```

# nc::alternatives for simplified alternation

We also provide a helper function for defining regex patterns with
[alternation](https://www.regular-expressions.info/alternation.html). The
following three lines are equivalent.

```{r}
show.patterns(
  "(?:(?<first>bar+)|(?<second>fo+))",
  list(first="bar+", "|", second="fo+"),
  nc::alternatives(first="bar+", second="fo+"))
```

# `nc::alternatives_with_shared_groups` for alternatives with identical named sub-pattern groups

Sometimes each alternative is just a re-arrangement of the same
sub-patterns. For example consider the following subjects, each of
which are dates, in one of two formats.

```{r}
subject.vec <- c("mar 17, 1983", "26 sep 2017", "17 mar 1984")
```

In each of the two formats, the month consists of three lower-case
letters, the day consists of two digits, and the year consists of four
digits. Is there a single pattern that can match each of these
subjects? Yes, such a pattern can be defined using the code below,

```{r}
pattern <- nc::alternatives_with_shared_groups(
  month="[a-z]{3}",
  day=list("[0-9]{2}", as.integer),
  year=list("[0-9]{4}", as.integer),
  list(month, " ", day, ", ", year),
  list(day, " ", month, " ", year))
```

In the code above, we used `nc::alternatives_with_shared_groups`,
which requires two kinds of arguments:

* named arguments (month, day, year) define sub-pattern groups that
  are used in each alternative.
* un-named arguments (last two) define alternative patterns, each
  which can use the sub-pattern group names (month, day, year).

The pattern can be used for matching, and the result is a data table
with one column for each unique name, 

```{r}
(match.dt <- nc::capture_first_vec(subject.vec, pattern))
```

After having parsed the dates into these three columns, we can add a
date column:

```{r}
Sys.setlocale(locale="C")#to recognize months in English.
match.dt[, date := data.table::as.IDate(
  paste(month, day, year), format="%b %d %Y")]
print(match.dt, class=TRUE)
```

# `nc::altlist` for named alternatives

For most use cases, `nc::alternatives_with_shared_groups` is
sufficient, but one case where it does not work is when you want to
name each alternative (for example, to easily count how many matches
to each alternative there were). In that case you can instead use
`nc::altlist` as in the code below,

```{r}
shared.groups <- nc::altlist(
  month="[a-z]{3}",
  day=list("[0-9]{2}", as.integer),
  year=list("[0-9]{4}", as.integer))
alt.args <- with(shared.groups, list(
  american=list(month, " ", day, ", ", year),
  european=list(day, " ", month, " ", year)))
pattern <- do.call(nc::alternatives, alt.args)
(match.dt <- nc::capture_first_vec(subject.vec, pattern))
match.dt[, lapply(.SD, function(x)sum(x!="")), .SDcols=names(alt.args)]
```

<!--
%\VignetteEngine{knitr::knitr}
%\VignetteIndexEntry{vignette 6: regex engines}
-->

# Uniform interface to three regex engines

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

Several C libraries providing regular expression engines are available
in R. The standard R distribution has included the Perl-Compatible
Regular Expressions (PCRE) C library since 2002.  CRAN package re2r
provides the RE2 library, and stringi provides the ICU library. Each
of these regex engines has a unique feature set, and may be preferred
for different applications. For example, PCRE is installed by default,
RE2 guarantees matching in polynomial time, and ICU provides strong
unicode support. For a more detailed comparison of the relative
strengths of each regex library, we refer the reader to our previous
research paper, [Comparing namedCapture with other R packages for
regular
expressions](https://journal.r-project.org/archive/2019/RJ-2019-050/index.html).

Each regex engine has a different R interface, so switching from one
engine to another may require non-trivial modifications of user
code. In order to make switching between engines easier, the
namedCapture package provides a uniform interface for capturing text
using PCRE and RE2. The user may specify the desired engine via an
option; the namedCapture package provides the output in a uniform
format. However namedCapture requires the engine to support specifying
capture group names in regex pattern strings, and to support output of
the group names to R (which ICU does not support).

Our proposed nc package provides support for the ICU engine in
addition to PCRE and RE2. The nc package implements this functionality
using un-named capture groups, which are supported in all three regex
engines. In particular, a regular expression is constructed in R code
that uses named arguments to indicate capturing sub-patterns, which
are translated to un-named groups when passed to the regex engine. For
example, consider a user who wants to capture the two pieces of the
column names of the iris data, e.g., `Sepal.Length`. The user would
typically specify the capturing regular expression as a string
literal, e.g., `"(.*)[.](.*)"`.  Using nc the same pattern can be
applied to the iris data column names via

```{r}
nc::capture_first_vec(
  names(iris), 
  part = ".*", "[.]", dim = ".*", 
  engine = "ICU", nomatch.error = FALSE)
```

Above we see an example usage of `nc:capture_first_vec`, which
is for capturing the first match of a regex from each element of a
character vector subject (the first argument). There are a variable
number of other arguments (`...`) which are used to define the
regex pattern. In this case there are three pattern arguments:
`part = ".*", "[.]", dim = ".*"`. Each named R argument in the
pattern generates an un-named capture group by enclosing the specified
character string in parentheses, e.g., `(.*)` for both `part`
and `dim` arguments above. All of the sub-patterns are pasted
together in the sequence they appear in order to create the final
pattern that is used with the specified regex engine. The
`nomatch.error = FALSE` argument is given because the default is
to stop with an error if any subjects do not match the specified
pattern (the fifth subject `Species` does not match). Under the
hood, the following function is called to parse the pattern arguments:

```{r}
str(compiled <- nc::var_args_list(part = ".*", "[.]", dim = ".*"))
```

This function is intended mostly for internal use, but can be useful
for viewing the generated regex pattern (or using it as input to
another regex function). The return value is a named list of two
elements: `pattern` is the capturing regular expression which is
generated based on the input arguments, and `fun.list` is a named
list of type conversion functions. If the user does not specify a type
conversion function for a group (as in the example code above), then
the default is `base::identity`, which simply returns the
captured character strings. Group-specific type conversion functions
are useful for converting captured text into numeric output columns. 
Note that the order of elements in
`fun.list` corresponds to the order of capture groups in the
pattern (e.g., first capture group named `part`, second
`dim`). These data can be used with any regex engine that
supports un-named capture groups (including ICU) in order to get a
capture matrix with column names, e.g.

```{r}
m <- stringi::stri_match_first_regex(names(iris), compiled$pattern)
colnames(m) <- c("match", names(compiled$fun.list))
m
```

Again, this is not the recommended usage of nc, but here we give
these details in order to explain how it works. Note that the result
from stringi is a character matrix with three columns: first for
the entire match, and another column for each capture group. Using the
same pattern with `base::regexpr` (PCRE engine) or
`re2r::re2_match` (RE2 engine) yields output in varying formats.
The nc package takes care of converting these different results
into a standard data table format which makes it easy to switch regex
engines (by changing the value of the `engine` argument). 
Most of the time the different engines give similar results, 
but in some cases there are differences:

```{r, error=TRUE, purl=FALSE}
u.subject <- "a\U0001F60E#"
u.pattern <- list(
  emoji="\\p{EMOJI_Presentation}")#only supported in ICU.
old.opt <- options(nc.engine="ICU")
nc::capture_first_vec(u.subject, u.pattern)
nc::capture_first_vec(u.subject, u.pattern, engine="PCRE") 
nc::capture_first_vec(u.subject, u.pattern, engine="RE2")
options(old.opt)
```

Note that the standard output format used by nc, as shown above with
`nc::capture_first_vec`, is a data table (not a character matrix, as
in other regex packages). The main reason that data tables are always
output by nc is in order to support output columns of different types,
when type conversion functions are specified.

<!--
%\VignetteEngine{knitr::knitr}
%\VignetteIndexEntry{vignette 4: comparisons with other packages}
-->

# Comparisons with other packages

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

This vignette provides comparisons with other packages that provide
similar functionality. This is a work in progress -- for a more
detailed / complete / coherent comparison with other packages which
provide wide-to-tall data reshaping, see [my
paper](https://github.com/tdhock/nc-article#paper).

## Are sepals larger or smaller than petals in the iris data?

Sometimes you want to melt a "wide" data table which has several
distinct pieces of information encoded in each column name. One
example is the familiar iris data, which have flower part and
measurement dimension encoded in each of four column names:

```{r}

library(data.table)
data.table(iris)

```

The goal in this section will be to convert these data into a format
with a column for each flower part (`Sepal` and `Petal`) so we can
easily make a facetted scatterplot to visually examine whether or not
sepals or larger than petals. The easiest way to perform this
conversion is with packages which provide a function for melting into
multiple output columns:

```{r}

iris.parts <- list(
  nc=nc::capture_melt_multiple(
    iris,
    column=".*?",
    "[.]",
    dim=".*"),
  tidyr=if(requireNamespace("tidyr"))tidyr::pivot_longer(
    iris, 
    cols=1:4, 
    names_to=c(".value", "dim"),
    names_sep="[.]"),
  stats=stats::reshape(
    iris,
    direction="long",
    timevar="dim",
    varying=1:4,
    sep="."),
  "data.table::melt"=melt(
    data.table(iris),
    measure.vars=patterns(
      Sepal="^Sepal",
      Petal="^Petal")
  )[data.table(
    variable=factor(1:2), dim=c("Length", "Width")
  ), on=.(variable)],
  if(requireNamespace("cdata"))cdata::rowrecs_to_blocks(
    iris,
    controlTable=data.frame(
      dim=c("Length", "Width"),
      Petal=c("Petal.Length", "Petal.Width"),
      Sepal=c("Sepal.Length", "Sepal.Width"),
      stringsAsFactors=FALSE),
    columnsToCopy="Species"))
iris.parts$nc

```

It is clear from the code above that each package is capable of
the conversions. However the syntax and level of explicitness varies:

* `nc::capture_melt_multiple` requires a regular
  expression: (most implicit, least repetition)
  * any input column names that match the regex are melted.
  * values matched in the `column` group are used for the output column names.
* `tidyr::pivot_longer` and `stats::reshape` require specification of
  the input columns to melt along with a separator.
  * `stats::reshape` assumes the output columns names occur in the
    part of the input column name before the separator.
  * `tidyr::pivot_longer` assumes the output columns occur in the part
    which corresponds to the `.value` element of the `names_to`
    argument.
* `data.table::melt` requires a join to recover the `dim` output
  column.
* `cdata::rowrecs_to_blocks` requires explicit specification of a
  control table. (most explicit, most repetition)

Any of the results can be visualized via:

```{r}

if(require(ggplot2)){
  ggplot()+
    theme_bw()+
    theme(panel.spacing=grid::unit(0, "lines"))+
    facet_grid(dim ~ Species)+
    coord_equal()+
    geom_abline(slope=1, intercept=0, color="grey")+
    geom_point(aes(
      Petal, Sepal),
      data=iris.parts$nc)
}

```

It is clear from the plot above that sepals are larger than petals,
for every measured flower. 

## Comparing dimensions in iris data

What if we wanted to compare dimensions rather than parts? 

```{r}

iris.dims <- list(
  nc=nc::capture_melt_multiple(
    iris,
    part=".*?",
    "[.]",
    column=".*"),
  stats=stats::reshape(
    structure(iris, names=sub("(.*?)[.](.*)", "\\2.\\1", names(iris))),
    direction="long",
    timevar="part",
    varying=1:4,
    sep="."))
iris.dims$nc

```

The code above shows that the syntax is mostly the same for this
example. The biggest difference is for `stats::reshape` which assumes
that each input column name is composed of (1) the output column name,
(2) a delimiter, and (3) some additional information to be stored in
the output column given by `timevar`. Therefore we need to pre-process
column names using `sub` for it to work.

```{r}

if(require(ggplot2)){
  ggplot()+
    theme_bw()+
    theme(panel.spacing=grid::unit(0, "lines"))+
    facet_grid(part ~ Species)+
    coord_equal()+
    geom_abline(slope=1, intercept=0, color="grey")+
    geom_point(aes(
      Length, Width),
      data=iris.dims$nc)
}

```

It is clear from the plot above that `Length` is larger than `Width`
for every measured flower part.

## Do columns need to be sorted?

Consider the following wide data set:

```{r}

TC <- data.table::data.table(
  age.treatment=c(1, 5),
  sex.control=c("M", "M"),
  sex.treatment=c("F", "F"),
  age.control=c(10, 50))

```

It is clear from the column names how the data should be grouped when
they are converted to tall format. However the columns do not appear
in regular order (age is before sex for treatment, but age is after
sex for control), which causes a problem for stats and data.table:

```{r}

input.list <- list(
  "nc"=nc::capture_melt_multiple(
    TC,
    column=".*?",
    "[.]",
    group=".*"),
  "cdata"=if(requireNamespace("cdata"))cdata::rowrecs_to_blocks(
    TC,
    controlTable=data.frame(
      group=c("treatment", "control"),
      age=c("age.treatment", "age.control"),
      sex=c("sex.treatment", "sex.control"),
      stringsAsFactors=FALSE)),
  "data.table"=data.table::melt(TC, measure.vars=patterns(
    age="age",
    sex="sex")),
  "stats"=stats::reshape(
    TC,
    varying=1:4,
    direction="long"),
  "tidyr"=if(requireNamespace("tidyr"))tidyr::pivot_longer(
    TC, 1:4,
    names_to=c(".value", "group"),
    names_sep="[.]"))
output.list <- list()
for(pkg in names(input.list)){
  df.or.null <- input.list[[pkg]]
  if(is.data.frame(df.or.null)){
    output.list[[pkg]] <- data.table::data.table(df.or.null)[order(age)]
  }
}
output.list
sapply(output.list, function(DT)identical(DT$sex, c("F", "F", "M", "M")))

```

In conclusion, when the input column names to melt do not appear in
the same order across groups or output columns, then the correct tall
data can be computed using one of `nc::capture_melt_multiple`,
`tidyr::pivot_longer`, `cdata::rowrecs_to_blocks`.

## Melting into a single output column, who data

Another data set where it is useful to do column name pattern matching
followed by melting is the World Health Organization data:

```{r}

if(requireNamespace("tidyr")){
  data(who, package="tidyr")
}else{
  who <- data.frame(id=1, new_sp_m5564=2, newrel_f65=3)
}
names(who)

```

Each column which starts with `new` has three distinct pieces of
information encoded in its name: diagnosis type (e.g. sp or rel),
gender (m or f), and age range (e.g. 5564 or 1524). We would like to
use a regex to match these column names, then using the matching
columns as measure.vars in a melt, then join the two results. The most
convenient way to do that is via:

```{r}

who.chr.list <- list(
  nc=nc::capture_melt_single(
    who,
    "new_?",
    diagnosis=".*",
    "_",
    gender=".",
    ages=".*"),
  tidyr=if(requireNamespace("tidyr"))tidyr::pivot_longer(
    who,
    new_sp_m014:newrel_f65,
    names_to=c("diagnosis", "gender", "ages"),
    names_pattern="new_?(.*)_(.)(.*)"))

```

Note the result includes additional column `value` which contains the
melted data. There is also a column for each capture group in the
specified pattern. The following example shows how to rename the
`value` column, remove missing values, and use numeric type conversion
functions:

```{r}

who.pattern <- "new_?(.*)_(.)((0|[0-9]{2})([0-9]{0,2}))"
as.numeric.Inf <- function(y)ifelse(y=="", Inf, as.numeric(y))
who.typed.list <- list(
  nc=nc::capture_melt_single(
    who,
    "new_?",
    diagnosis=".*",
    "_",
    gender=".",
    ages=list(
      ymin.num="0|[0-9]{2}", as.numeric,
      ymax.num="[0-9]{0,2}", as.numeric.Inf),
    value.name="count",
    na.rm=TRUE),
  tidyr=if(requireNamespace("tidyr"))try(tidyr::pivot_longer(
    who,
    cols=grep(who.pattern, names(who)),
    names_transform=list(
      ymin.num=as.numeric,
      ymax.num=as.numeric.Inf),
    names_to=c("diagnosis", "gender", "ages", "ymin.num", "ymax.num"),
    names_pattern=who.pattern,
    values_drop_na=TRUE,
    values_to="count")))
str(who.typed.list)

```

The result above shows that `nc::capture_melt_single` (1) makes it
easier to define complex patterns (2) supports type conversion without
a post-processing step, and (3) reduces repetition in user code. There
are several sources of repetition in `tidyr` code:

* Syntax of type conversion:
  * capture group name `ymin.num` appears only once for `nc` but twice for `tidyr`.
  * capture group name `ymax.chr` appears only once for `nc` but three times for `tidyr`.
* Specification of input data and columns to melt:
  * input data table `who` appears only once for `nc` but twice for `tidyr`.
  * pattern for input columns to melt appears only once for `nc` but twice for `tidyr`.

Other packages for doing this include:

```{r}

if(requireNamespace("tidyr")){
  gather.result <- tidyr::gather(
    who,
    "variable",
    "count",
    grep(who.pattern, names(who)),
    na.rm=TRUE)
  extract.result <- tidyr::extract(
    gather.result,
    "variable",
    c("diagnosis", "gender", "ages", "ymin.int", "ymax.int"),
    who.pattern,
    convert=TRUE)
  transform.result <- base::transform(
    extract.result,
    ymin.num=as.numeric(ymin.int),
    ymax.num=ifelse(is.na(ymax.int), Inf, as.numeric(ymax.int)))
  str(transform.result)
}

```

Note that `tidyr::gather` requires two post-processing steps, which cause the
same two types of repetition as `tidyr::pivot_longer`:

* Syntax of type conversion:
  `base::transform` is used for converting age range variables to
  numeric, since default types are int with `convert=TRUE`.
* Specification of input data and columns to melt:
  `tidyr::extract` is used to convert the melted column into
  several output columns; this results in repetition in the code
  because the regex is also used to define the columns to melt/gather.

The `reshape2` package suffers from the same two issues:
  
```{r}

reshape2.result <- if(requireNamespace("reshape2")){
  reshape2:::melt.data.frame(
    who,
    measure.vars=grep(who.pattern, names(who)),
    na.rm=TRUE,
    value.name="count")
}

```

Interestingly, `data.table::patterns` can be used to avoid repeating
the data set name, `who`. However it supports neither type conversion
nor regex capture groups.

```{r}

dt.result <- data.table::melt.data.table(
  data.table(who),
  measure.vars=patterns(who.pattern),
  na.rm=TRUE,
  value.name="count")

```

Neither cdata nor stats provide an na.rm option:

```{r}

who.df <- data.frame(who)
is.varying <- grepl(who.pattern, names(who))
names(who.df)[is.varying] <- paste0("count.", names(who)[is.varying])
stats.result <- stats::reshape(
  who.df,
  direction="long",
  timevar="variable",
  varying=is.varying)

```

```{r}

if(requireNamespace("cdata")){
  cdata.result <- cdata::rowrecs_to_blocks(
    who, 
    cdata::build_unpivot_control(
      "variable",
      "count",
      grep(who.pattern, names(who), value=TRUE)),
    columnsToCopy=grep(who.pattern, names(who), value=TRUE, invert=TRUE))
}

```

## Melting a wider iris back to original

```{r}

## Example 1: melting a wider iris data back to original.
library(data.table)
iris.dt <- data.table(
  i=1:nrow(iris),
  iris[,1:4],
  Species=paste(iris$Species))
print(iris.dt)

## what if we had two observations on each row?
set.seed(1)
iris.rand <- iris.dt[sample(.N)]
iris.wide <- cbind(treatment=iris.rand[1:75], control=iris.rand[76:150])
print(iris.wide, topn=2, nrows=10)

## This is the usual data.table syntax for getting the original iris back.
iris.melted <- melt(iris.wide, value.factor=TRUE, measure.vars = patterns(
  i="i$",
  Sepal.Length="Sepal.Length$",
  Sepal.Width="Sepal.Width$",
  Petal.Length="Petal.Length$",
  Petal.Width="Petal.Width$",
  Species="Species$"))
identical(iris.melted[order(i), names(iris.dt), with=FALSE], iris.dt)

## nc can do the same thing -- you must define an R argument named
## column, and another named argument which identifies each group.
(nc.melted <- nc::capture_melt_multiple(
  iris.wide,
  group="[^.]+",
  "[.]",
  column=".*"))
identical(nc.melted[order(i), names(iris.dt), with=FALSE], iris.dt)

## This is how we do it using stats::reshape.
iris.wide.df <- data.frame(iris.wide)
names(iris.wide.df) <- sub("(.*?)[.](.*)", "\\2_\\1", names(iris.wide))
iris.reshaped <- stats::reshape(
  iris.wide.df,
  direction="long",
  timevar="group",
  varying=names(iris.wide.df),
  sep="_")
identical(data.table(iris.reshaped[, names(iris.dt)])[order(i)], iris.dt)

## get the parts columns and groups -- is there any difference
## between groups? of course not!
parts.wide <- nc::capture_melt_multiple(
  iris.wide,
  group=".*?",
  "[.]",
  column=".*?",
  "[.]",
  dim=".*")
if(require("ggplot2")){
  ggplot()+
    theme_bw()+
    theme(panel.spacing=grid::unit(0, "lines"))+
    facet_grid(dim ~ group)+
    coord_equal()+
    geom_abline(slope=1, intercept=0, color="grey")+
    geom_point(aes(
      Petal, Sepal),
      data=parts.wide)
}

```

## Lots of column types

```{r}

## Example 2. Lots of column types, from example(melt.data.table).
DT <- data.table(
  i_1 = c(1:5, NA),
  i_2 = c(NA,6:10),
  f_1 = factor(sample(c(letters[1:3], NA), 6, TRUE)),
  f_2 = factor(c("z", "a", "x", "c", "x", "x"), ordered=TRUE),
  c_1 = sample(c(letters[1:3], NA), 6, TRUE),
  d_1 = as.Date(c(1:3,NA,4:5), origin="2013-09-01"),
  d_2 = as.Date(6:1, origin="2012-01-01"))
## add a couple of list cols
DT[, l_1 := DT[, list(c=list(rep(i_1, sample(5,1)))), by = i_1]$c]
DT[, l_2 := DT[, list(c=list(rep(c_1, sample(5,1)))), by = i_1]$c]

## original DT syntax is quite repetitive.
melt(DT, measure=patterns(
  i="^i",
  f="^f",
  d="^d",
  l="^l"
))

## nc syntax uses a single regex rather than four.
nc::capture_melt_multiple(
  DT,
  column="^[^c]",
  "_",
  number="[12]")

## id.vars can be specified using original DT syntax.
melt(DT, id=1:2, measure=patterns(
  f="^f",
  l="^l"
))

nc::capture_melt_multiple(
  DT,
  column="^[fl]",
  "_",
  number="[12]")

## reshape does not support list columns.
reshape(
  DT,
  varying=grep("^[fid]", names(DT)),
  sep="_",
  direction="long",
  timevar="number")

## tidyr does, but errors for combining ordered and un-ordered factors.
if(requireNamespace("tidyr")){
  tidyr::pivot_longer(
    DT, grep("[cf]", names(DT), invert=TRUE),
    names_pattern="(.)_(.)",
    names_to=c(".value", "number"))
}

```

## family data

```{r}

## Example 3, three children, one family per row, from data.table
## vignette.
family.dt <- fread(text="
family_id age_mother dob_child1 dob_child2 dob_child3 gender_child1 gender_child2 gender_child3
1         30 1998-11-26 2000-01-29         NA             1             2            NA
2         27 1996-06-22         NA         NA             2            NA            NA
3         26 2002-07-11 2004-04-05 2007-09-02             2             2             1
4         32 2004-10-10 2009-08-27 2012-07-21             1             1             1
5         29 2000-12-05 2005-02-28         NA             2             1            NA")
(children.melt <- melt(family.dt, measure = patterns(
  dob="^dob", gender="^gender"
), na.rm=TRUE, variable.factor=FALSE))

## nc::field can be used to define group name and pattern at the
## same time, to avoid repetitive code.
(children.nc <- nc::capture_melt_multiple(
  family.dt,
  column="[^_]+",
  "_",
  nc::field("child", "", "[1-3]"),
  na.rm=TRUE))

## reshape works too.
stats::reshape(
  family.dt,
  varying=grep("child", names(family.dt)),
  direction="long",
  sep="_",
  timevar="child.str")

```

## histogram of iris

```{r}

## Comparison with base R. 1. mfrow means parts on rows, mfcol means
## parts on columns. 2. same number of lines of code. 3. nc/ggplot2
## code has more names and fewer numbers.
imat <- as.matrix(iris[, 1:4])
ylim <- range(table(imat))
xlim <- range(imat)
par(mfcol=c(2,2), mar=c(2,2,1,1))
for(col.i in 1:ncol(imat)){
  hist(
    imat[, col.i],
    breaks=seq(xlim[1], xlim[2], by=0.1),
    ylim=ylim,
    main=colnames(imat)[col.i])
}

```

### single output column simple type conversion

All four packages below can convert the input reshape column name to a
numeric output capture column.

```{r}

pen.peaks.wide <- data.table::data.table(
  data.set=c("foo", "bar"),
  "10.1"=c(5L, 10L),
  "0.3"=c(26L, 39L))
pen.peaks.gather <- if(requireNamespace("tidyr"))tidyr::gather(
  pen.peaks.wide,
  "penalty",
  "peaks",
  -1,
  convert=TRUE)
str(pen.peaks.gather)

pen.peaks.nc <- nc::capture_melt_single(
  pen.peaks.wide,
  penalty="^[0-9.]+", as.numeric,
  value.name="peaks")
str(pen.peaks.nc)

pen.peaks.pivot <- if(requireNamespace("tidyr"))try(tidyr::pivot_longer(
  pen.peaks.wide,
  -1,
  names_to="penalty",
  names_transform=list(penalty=as.numeric),
  values_to="peaks"))
str(pen.peaks.pivot)

varying <- 2:3
pen.peaks.reshape.times <- stats::reshape(
  pen.peaks.wide,
  direction="long",
  varying=varying,
  times=as.numeric(names(pen.peaks.wide)[varying]),
  v.names="peaks",
  timevar="penalty")
str(pen.peaks.reshape.times)

pen.peaks.renamed <- pen.peaks.wide
names(pen.peaks.renamed) <- paste0(ifelse(
  grepl("^[0-9]", names(pen.peaks.wide)),
  "peaks_", ""),
  names(pen.peaks.wide))
pen.peaks.reshape.sep <- stats::reshape(
  pen.peaks.renamed,
  direction="long",
  varying=varying,
  sep="_",
  timevar="penalty")
str(pen.peaks.reshape.sep)

```

### Multiple output columns of different types

```{r}
peaks.csv <- system.file(
  "extdata", "RD12-0002_PP16HS_5sec_GM_F_1P.csv",
  package="nc", mustWork=TRUE)
peaks.wide <- data.table::fread(peaks.csv)
tidyr.long <- tidyr::pivot_longer(
  peaks.wide,
  grep(" [0-9]", names(peaks.wide)),
  names_pattern = "(.*) ([0-9]+)",
  names_to = c(".value", "peak"),
  names_transform = list(peak=as.integer))
peaks.tall <- nc::capture_melt_multiple(
  peaks.wide,
  column=".*",
  " ",
  peak="[0-9]+", as.integer)
```

# `clickSelects` et `showSelected` nommés

<!-- paragraph -->

```{r setup, echo=FALSE}
knitr::opts_chunk$set(fig.path="ch14-figures/")
```

<!-- paragraph -->

Dans ce chapitre, nous expliquons comment utiliser les [variables `clickSelects` et `showSelected` nommées](/ch06#data-driven-selectors) pour créer des noms de sélecteurs à partir des données.
<!-- comment -->
Cette fonctionnalité facilite l’écriture du code `animint2` et accélère la compilation.

<!-- paragraph -->

Plan du chapitre :

<!-- paragraph -->

- Nous commençons par charger le jeu de données `PSJ` et calculer les données à tracer.
<!-- comment -->
- Nous présentons une méthode pour définir un `animint` avec de nombreux sélecteurs, en utilisant des boucles `for`.
<!-- comment -->
Cette approche est techniquement correcte, mais peu efficace.
<!-- comment -->
- Nous expliquons ensuite la méthode à privilégier pour définir un `animint` avec de nombreux sélecteurs, c’est-à-dire en utilisant des `clickSelects` et `showSelected` nommés.
<!-- comment -->
Cette méthode est plus efficace sur le plan computationnel et s’avère plus facile à coder.

<!-- paragraph -->

## Téléchargement du jeu de données {#download}

<!-- paragraph -->

Les données de cet exemple proviennent du [package PeakSegJoint](https://github.com/tdhock/PeakSegJoint) qui sert à détecter les pics dans des séquences de données génomiques.
<!-- comment -->
Le code ci-dessous télécharge le jeu de données.

<!-- paragraph -->

```{r}
if(!requireNamespace("animint2data"))
  remotes::install_github("animint/animint2data")
data(PSJ, package="animint2data")
sapply(PSJ, class)
```

<!-- paragraph -->

On peut voir, ci-dessus, que `PSJ` est une liste de plusieurs éléments (listes et tableaux de données).

<!-- paragraph -->

## Exploration des données `PSJ` à l’aide de ggplots statiques {#explore-data-with-static-ggplots}

<!-- paragraph -->

Nous commençons par créer un graphique montrant certaines données ChIP-seq génomiques, des données séquentielles dont les valeurs augmentent dans les zones actives (généralement lorsque des gènes sont transcrits).
<!-- comment -->
Dans le code ci-dessous, nous affichons chaque échantillon dans un panneau distinct :

<!-- paragraph -->

```{r}
library(animint2)
ann.colors <- c(
  noPeaks="#f6f4bf",
  peakStart="#ffafaf",
  peakEnd="#ff4c4c",
  peaks="#a445ee")
(gg.cov <- ggplot()+
  scale_y_continuous(
    "couverture des lectures alignées",
    breaks=function(limits){
      floor(limits[2])
    })+
  scale_x_continuous(
    "position sur chr11 (kilobases = kb)")+
  coord_cartesian(xlim=c(118167.406, 118238.833))+
  geom_tallrect(aes(
    xmin=chromStart/1e3, xmax=chromEnd/1e3,
    fill=annotation),
    alpha=0.5,
    color="grey",
    data=PSJ$filled.regions)+
  scale_fill_manual(values=ann.colors)+
  theme_bw()+
  theme(panel.margin=grid::unit(0, "cm"))+
  facet_grid(sample.id ~ ., labeller=function(df){
    df$sample.id <- sub("McGill0", "", sub(" ", "\n", df$sample.id))
    df
  }, scales="free")+
  geom_line(aes(
    base/1e3, count),
    data=PSJ$coverage,
    color="grey50"))
```

<!-- paragraph -->

La figure ci-dessus montre les données brutes bruitées sous forme d’un `geom_line()` gris.
<!-- comment -->
Les rectangles de couleur indiquent si le début ou la fin d’un pic doit être prédit dans une région et un échantillon donnés.
<!-- comment -->
Nous ajoutons ensuite un panneau pour les problèmes de segmentation.
Dans chacun d’eux, un algorithme cherche un pic commun à tous les échantillons
<!-- comment -->
et prédit les positions de début et de fin du pic de valeur élevée.
<!-- comment -->
Le code ci-dessous calcule un tableau de contingence avec une ligne par problème :

<!-- paragraph -->

```{r}
library(data.table)
(show.problems <- data.table(PSJ$problems)[
, y := problem.i/max(problem.i), by=bases.per.problem][])
```

<!-- paragraph -->

Dans le code ci-dessus, une colonne `y` est ajoutée pour l’affichage des problèmes.

<!-- paragraph -->

```{r}
(gg.cov.prob <- gg.cov+
  ggtitle("Sélection du problème")+
  geom_text(aes(
    chromStart/1e3, 0.9,
    label=sprintf(
      "%d taille moyenne des problèmes %.1f kb",
      problems, mean.bases/1e3)),
    showSelected="bases.per.problem",
    data=PSJ$problem.labels,
    hjust=0)+
  geom_segment(aes(
    problemStart/1e3, y,
    xend=problemEnd/1e3, yend=y),
    showSelected="bases.per.problem",
    clickSelects="problem.name",
    size=5,
    data=show.problems))
```

<!-- paragraph -->

Ci-dessus, on voit qu’un panneau inférieur intitulé `problems` a été ajouté au graphique précédent.
<!-- comment -->
Le graphique statique ci-dessus est surchargé ; la version interactive sera lisible parce qu’elle n’affichera qu’une seule valeur de `bases.per.problem` à la fois.
<!-- comment -->
Pour sélectionner les différentes valeurs de `bases.per.problem` (taille du problème), nous utiliserons un autre graphique, qui présentera le meilleur taux d’erreur pour chaque taille de problème, comme le montrent les données ci-dessous :

<!-- paragraph -->

```{r}
(res.error <- data.table(PSJ$error.total.chunk))
```

<!-- paragraph -->

Le tableau ci-dessus comporte une ligne par valeur de `bases.per.problem`, un paramètre de taille de fenêtre coulissante, que nous allons explorer à l’aide de l’interactivité.
<!-- comment -->
Nous utilisons ces données pour tracer le graphique ci-dessous :

<!-- paragraph -->

```{r}
(gg.res.error <- ggplot()+
  ggtitle("choisir la taille du problème")+
  ylab("nombre minimal d’erreurs de prédiction d’étiquettes")+
  geom_line(aes(
    bases.per.problem, errors),
    data=res.error)+
  geom_tallrect(aes(
    xmin=min.bases.per.problem,
    xmax=max.bases.per.problem),
    clickSelects="bases.per.problem",
    alpha=0.5,
    data=res.error)+
  scale_x_log10())
```

<!-- paragraph -->

La figure ci-dessus montre le nombre minimum d’erreurs d’étiquettes en fonction de la taille du problème.
<!-- comment -->
Des rectangles gris seront utilisés pour sélectionner la taille du problème.

<!-- paragraph -->

Un paramètre de pénalité contrôle le nombre d’échantillons ayant un pic commun ; il est défini dans le tableau de données de sélection de modèles dans le code ci-dessous :

<!-- paragraph -->

```{r}
pdot <- function(L){
  out_list <- list()
  for(i in seq_along(L)){
    out_list[[i]] <- data.table(
      problem.dot=names(L)[[i]], L[[i]])
  }
  rbindlist(out_list)
}
(all.modelSelection <- pdot(PSJ$modelSelection.by.problem))
```

<!-- paragraph -->

Le code ci-dessus utilise la fonction `pdot()` qui fait appel à la méthode [liste des tableaux de données](/ch99#list-of-data-tables) pour ajouter une colonne nommée `problem.dot`, laquelle servira ensuite à définir les sélecteurs dans la visualisation interactive.
<!-- comment -->
Nous traçons ci-dessous le nombre de pics et d’erreurs d’étiquettes, en fonction du paramètre de pénalité de l’algorithme :

<!-- paragraph -->

```{r}
long.modelSelection <- melt(
  data.table(all.modelSelection)[, errors := as.numeric(errors)],
  measure.vars=c("peaks","errors"))
log.lambda.range <- all.modelSelection[, c(
  min(max.log.lambda), max(min.log.lambda))]
modelSelection.labels <- unique(all.modelSelection[, data.table(
  problem.name,
  bases.per.problem,
  problemStart,
  problemEnd,
  log.lambda=mean(log.lambda.range),
  peaks=max(peaks)+0.5)])
(gg.model.selection <- ggplot()+
   scale_x_continuous("log(pénalité)")+
   geom_segment(aes(
     min.log.lambda, value,
     xend=max.log.lambda, yend=value),
     showSelected=c("bases.per.problem", "problem.name"),
     data=long.modelSelection,
     size=5)+
   geom_text(aes(
     log.lambda, peaks,
     label=sprintf(
       "%.1f kb dans le problème %s",
       (problemEnd-problemStart)/1e3, problem.name)),
     showSelected=c("bases.per.problem", "problem.name"),
     data=data.frame(modelSelection.labels, variable="peaks"))+
   ylab("")+
   facet_grid(variable ~ ., scales="free"))
```

<!-- paragraph -->

La figure ci-dessus montre `errors` (panneau supérieur) et `peaks` (panneau inférieur) en fonction de `log(penalty)`.
<!-- comment -->
Là encore, cette version statique est surchargée ; l’interactivité sera utilisée pour rendre la figure lisible (en n’affichant que le sous-ensemble de données correspondant aux valeurs sélectionnées de `bases.per.problem` et `problem.name`).

<!-- paragraph -->

## Visualisation interactive des données (incomplète) {#incomplete}

<!-- paragraph -->

Dans cette section, nous combinons les `ggplots` de la section précédente pour créer une visualisation avec interactivité qui relie les graphiques.
<!-- comment -->
Le code ci-dessous utilise `theme_animint()` pour associer certaines options d’affichage au graphique `coverage` précédent, et l’option `first` est ajoutée pour spécifier quels sous-ensembles de données doivent être affichés en premier.

<!-- paragraph -->

```{r}
timing.incomplete.construct <- system.time({
  coverage.counts <- table(PSJ$coverage$sample.id)
  facet.rows <- length(coverage.counts)+1
  vis.incomplete <- animint(
    out.dir="ch14incomplete",
    first=list(
      bases.per.problem=6516,
      problem.name="chr11:118174946-118177139"),
    coverage=gg.cov.prob+theme_animint(
      last_in_row=TRUE, colspan=2,
      width=800, height=facet.rows*100),
    resError=gg.res.error,
    modelSelection=gg.model.selection)
})
```

<!-- paragraph -->

La sortie temporelle ci-dessus montre que la définition initiale est rapide.
<!-- comment -->
Générer cette visualisation des données préliminaires (incomplète) dans le code ci-dessous est également rapide.

<!-- paragraph -->

```{r ch14incomplete}
before.incomplete <- Sys.time()
vis.incomplete
cat(elapsed.incomplete <- Sys.time()-before.incomplete, "secondes\n")
```

<!-- paragraph -->

Nous voyons ci-dessus une visualisation des données avec trois graphiques :

<!-- paragraph -->

- Dans la partie supérieure, quatre profils de données ChIP-seq sont affichés, ainsi qu’un panneau de problèmes qui divise l’axe `x` en sections dans lesquels l’algorithme de segmentation s’exécute.
<!-- comment -->
- En cliquant sur le graphique en bas à gauche, on sélectionne la taille du problème, ce qui met à jour les problèmes affichés en haut.
<!-- comment -->
- Le graphique en bas à droite montre le nombre de pics et d’erreurs en fonction de la pénalité (plus elle est élevée, moins il y a de pics).

<!-- paragraph -->

## Ajouter de l’interactivité à l’aide de boucles `for` {#define-using-for-loops}

<!-- paragraph -->

Dans cette section, nous ajoutons des couches aux `ggplots` précédents en utilisant une boucle `for`, une approche sous-optimale, que nous présentons pour la comparer à une meilleure approche (`clickSelects` et `showSelected` nommés), présentée dans la prochaine section.
<!-- comment -->
La visualisation de la section précédente est incomplète.
<!-- comment -->
Nous aimerions y ajouter :

<!-- paragraph -->

- des rectangles dans le graphique en bas à gauche pour permettre la sélection du nombre de pics prédits dans un problème donné ;
<!-- comment -->
- des segments et des rectangles dans le graphique du haut pour montrer les pics prédits et les erreurs d’étiquettes.

<!-- paragraph -->

Une façon (inefficace) d’ajouter ces éléments serait d’utiliser la boucle `for` codée ci-dessous.
<!-- comment -->
Pour chaque problème, on trouve un sélecteur (appelé `problem.dot`) pour le nombre de pics.
<!-- comment -->
Dans cette boucle `for`, nous ajoutons donc quelques couches aux graphiques `coverage` et `modelSelection` à l’aide de `clickSelects="problem.dot"` ou `showSelected="problem.dot"`.

<!-- paragraph -->

```{r}
vis.first <- vis.incomplete
vis.first$first <- c(vis.incomplete$first, PSJ$first)
vis.first$modelSelection <- vis.first$modelSelection+
  ggtitle("choisir le nombre d’échantillons ayant un pic commun")
print(timing.for.construct <- system.time({
  vis.for <- vis.first
  vis.for$title <- "PSJ avec boucles for"
  vis.for$out.dir <- "ch14for"
  for(problem.dot in names(PSJ$modelSelection.by.problem)){
    if(problem.dot %in% names(PSJ$peaks.by.problem)){
      peaks <- PSJ$peaks.by.problem[[problem.dot]]
      peaks[[problem.dot]] <- peaks$peaks
      prob.peaks.names <- c(
        "bases.per.problem", "problem.i", "problem.name",
        "chromStart", "chromEnd", problem.dot)
      prob.peaks <- unique(data.frame(peaks)[, prob.peaks.names])
      prob.peaks$sample.id <- "problems"
      vis.for$coverage <- vis.for$coverage +
        geom_segment(aes(
          chromStart/1e3, 0,
          xend=chromEnd/1e3, yend=0),
          clickSelects="problem.name",
          showSelected=c(problem.dot, "bases.per.problem"),
          data=peaks, size=7, color="deepskyblue")
    }
    modelSelection.dt <- PSJ$modelSelection.by.problem[[problem.dot]]
    modelSelection.dt[[problem.dot]] <- modelSelection.dt$peaks
    vis.for$modelSelection <- vis.for$modelSelection+
      geom_tallrect(aes(
        xmin=min.log.lambda, 
        xmax=max.log.lambda), 
        clickSelects=problem.dot,
        showSelected=c("problem.name", "bases.per.problem"),
        data=modelSelection.dt, alpha=0.5)
  }
}))
```

<!-- paragraph -->

Notez le temps d’exécution du code ci-dessus, qui ne comprend que l’évaluation du code R pour les ggplots (et non  son affichage).
<!-- comment -->
Ensuite, nous compilons la visualisation des données.

<!-- paragraph -->

```{r ch14for}
before.for <- Sys.time()
vis.for
cat(elapsed.for <- Sys.time()-before.for, "seconds\n")
```

<!-- paragraph -->

Notez que la compilation nécessite plusieurs secondes, car il y a de nombreux geoms (cliquez sur « Show download status table » pour les voir tous).
<!-- comment -->
Par rapport à la visualisation des données de la section précédente, celle-ci présente :

<!-- paragraph -->

- des segments bleus qui apparaissent dans le graphique des données de couverture (en haut), pour indiquer les pics prédits ;
<!-- comment -->
- des rectangles de sélection cliquables dans le graphique en bas à droite, pour modifier le nombre d’échantillons présentant un pic dans le problème sélectionné.

<!-- paragraph -->

Dans la prochaine section, nous allons créer la même visualisation, mais plus efficacement.

<!-- paragraph -->

## Définition de l’interactivité à l’aide de `clickSelects` et `showSelected` nommés {#define-using-named}

<!-- paragraph -->

Dans cette section, nous créons une version plus efficace de la précédente visualisation en utilisant des `clickSelects` et des `showSelected` nommés.
<!-- comment -->
En général, toute visualisation définie à l’aide de boucles `for` dans le code R peut être optimisée par cette méthode.
<!-- comment -->
Tout d’abord, nous définissons quelques données communes.

<!-- paragraph -->

```{r}
(sample.peaks <- pdot(PSJ$peaks.by.problem))
```

<!-- paragraph -->

La sortie ci-dessus montre un tableau avec une ligne par pic pouvant être affiché, pour différents échantillons, problèmes, et choix interactifs des paramètres `bases.per.problem` et `peaks`.
<!-- comment -->
Notez la colonne `problem.dot` : elle détermine le nom du sélecteur qui stockera le nombre de pics sélectionné pour le problème.

<!-- paragraph -->

L’idée principale du code ci-dessous est la suivante : pour chaque problème, un sélecteur est défini par la colonne `problem.dot` pour le nombre de pics dans ce problème.
<!-- comment -->
Nous utilisons `showSelected=c(problem.dot="peaks")` et `clickSelects=c(problem.dot="peaks")` pour indiquer que le nom du sélecteur se trouve dans la colonne `problem.dot`, et que la valeur de la sélection se trouve dans la colonne `peaks`.
<!-- comment -->
Le compilateur `animint2dir()` crée une variable de sélection pour chaque valeur unique de `problem.dot`, et utilise les valeurs correspondantes dans `peaks` pour définir ou mettre à jour la valeur sélectionnée et les geoms associés.

<!-- paragraph -->

```{r}
print(timing.named.construct <- system.time({
  vis.named <- vis.first
  vis.named$title <- "PSJ avec clickSelects et showSelected nommés"
  vis.named$out.dir <- "ch14named"
  vis.named$coverage <- vis.named$coverage+
    geom_segment(aes(
      chromStart/1e3, 0,
      xend=chromEnd/1e3, yend=0),
      clickSelects="problem.name",
      showSelected=c(problem.dot="peaks", "bases.per.problem"),
      data=sample.peaks, size=7, color="deepskyblue")
  vis.named$modelSelection <- vis.named$modelSelection+
    geom_tallrect(aes(
      xmin=min.log.lambda,
      xmax=max.log.lambda),
      clickSelects=c(problem.dot="peaks"),
      showSelected=c("problem.name", "bases.per.problem"),
      data=all.modelSelection, alpha=0.5)
}))
```

<!-- paragraph -->

De toute évidence, le code R ci-dessus, qui utilise les `clickSelects` et `showSelected` nommés, améliore la rapidité.
<!-- comment -->
Nous le compilons ci-dessous :

<!-- paragraph -->

```{r ch14named}
before.named <- Sys.time()
vis.named
cat(elapsed.named <- Sys.time()-before.named, "seconds\n")
```

<!-- paragraph -->

Le `animint` produit ci-dessus devrait être identique à la visualisation de la section précédente.
<!-- comment -->
Les temps d’exécution ci-dessus montrent que les `clickSelects` et `showSelected` nommés sont beaucoup plus rapides que les boucles `for`, tant pour la définition que pour la compilation.

<!-- paragraph -->

## Comparaison de l’utilisation du disque {#disk-usage}

<!-- paragraph -->

Dans cette section, nous calculons l’utilisation du disque pour les deux méthodes.

<!-- paragraph -->

```{r}
vis.dirs.vec <- c("ch14incomplete", "ch14for", "ch14named")
vis.dirs.text <- paste(vis.dirs.vec, collapse=" ")
(cmd <- paste("du -ks", vis.dirs.text))
(kb.dt <- fread(cmd=cmd, col.names=c("kilobytes", "path")))
```

<!-- paragraph -->

Le tableau ci-dessus montre que la visualisation définie à l’aide des boucles `for` occupe environ deux fois plus d’espace disque que la visualisation utilisant des `clickSelects` et `showSelected` nommés.

<!-- paragraph -->

## Résumé du chapitre et exercices {#ch14-exercises}

<!-- paragraph -->

Le tableau ci-dessous résume l’utilisation du disque et les temps d’exécution présentés dans ce chapitre :

<!-- paragraph -->

```{r}
data.table(
  kb.dt,
  construct.seconds=c(
    timing.incomplete.construct[["elapsed"]],
    timing.for.construct[["elapsed"]],
    timing.named.construct[["elapsed"]]),
  compile.seconds=as.numeric(c(
    elapsed.incomplete,
    elapsed.for,
    elapsed.named)))
```

<!-- paragraph -->

Il ressort clairement du tableau ci-dessus que les `clickSelects` et `showSelected` nommés sont plus efficaces à ces deux égards et devraient être privilégiés plutôt que les boucles `for`.

<!-- paragraph -->

Exercices :

<!-- paragraph -->

- Utilisez des `clickSelects` et des `showSelected` nommés pour créer une visualisation illustrant le surapprentissage et le sous-apprentissage, comme nous l'avons fait dans cette [visualisation du modèle linéaire et des plus proches voisins](https://tdhock.github.io/2023-12-04-degree-neighbors/).
<!-- comment -->
- Utilisez des `clickSelects` et des `showSelected` nommés pour créer une visualisation de données dans votre domaine d’expertise.

<!-- paragraph -->

Dans le [chapitre 15](/ch15), nous vous expliquerons comment visualiser les algorithmes de recherche de racines (« zero-finding »).

<!-- paragraph -->

---
title: "Capture all matches in a single subject string"
date: "`r Sys.Date()`"
output:
  rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Capture all matches in a single subject string}
  %\VignetteEngine{knitr::rmarkdown}
  \usepackage[utf8]{inputenc}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

The `nc::capture_all_str` function is for the common case of
extracting each match from a multi-line text file (a single large
subject string). In this section we demonstrate how to extract data
tables from such loosely structured text data. For example we consider
the following [track
hub](http://genome.cse.ucsc.edu/goldenPath/help/hgTrackHubHelp.html)
meta-data file:

```{r}
trackDb.txt.gz <- system.file(
  "extdata", "trackDb.txt.gz", package="nc")
trackDb.vec <- readLines(trackDb.txt.gz)
```

Some representative lines from that file are shown below.

```{r}
cat(trackDb.vec[78:107], sep="\n")
```

## Match all tracks in the text file

Each block of text begins with "track" and includes several lines of
data before the block ends with two consecutive newlines. That pattern
is coded below using a regex:

```{r}
tracks.dt <- nc::capture_all_str(
  trackDb.vec, 
  "track ",
  track="\\S+",
  fields="(?:\n[^\n]+)*",
  "\n")
str(tracks.dt)
```

The result is a data.table with one row for each track block that
matches the regex. There are two character columns: `track` is a
unique name, and `fields` is a string with the rest of the data
in that block:

```{r}
tracks.dt[, .(track, fields.start=substr(fields, 1, 30))]
```

## Match all fields in each track

Each block has a variable number of lines/fields. Each line starts
with a field name, followed by a space, followed by the field
value. That regex is coded below:

```{r}
(fields.dt <- tracks.dt[, nc::capture_all_str(
  fields,
  "\\s+",
  variable=".*?",
  " ",
  value="[^\n]+"),  
  by=track])
str(fields.dt)
```

Note that because `by=track` was specified, `nc::capture_all_str` is
called for each unique value of `track` (i.e. each row). The results
are combined into a single data.table with one row for each
field. This data.table can be easily queried, e.g.

```{r}
fields.dt[
  J("tcell_McGill0107Coverage", "bigDataUrl"),
  value,
  on=.(track, variable)]
fields.dt[, .(count=.N), by=variable][order(count)]
```

For more information about data.table syntax, read
`vignette("datatable-intro", package="data.table")`.

## Match all tracks and some fields with one regex

In the examples above we extracted all fields from all tracks (using
two regexes, one for the track, one for the field). In the example
below we extract only the track name, split into separate columns
(using a single regex for the track).

```{r}
cell.sample.type <- list(
  cellType="[^ ]*?",
  "_",
  sampleName=list(
    "McGill",
    sampleID="[0-9]+", as.integer),
  dataType="Coverage|Peaks")
nc::capture_all_str(trackDb.vec, cell.sample.type)
```

Note that the pattern above defines nested capture groups via named
lists (e.g. sampleID is a subset of sampleName). The pattern below
matches either the previously specified track pattern, or any other
type of track name:

```{r}
sample.or.anything <- list(
  cell.sample.type,
  "|",
  "[^\n]+")
track.pattern.old <- list(
  "track ",
  track=sample.or.anything)
nc::capture_all_str(trackDb.vec, track.pattern.old)
```

Notice the repetition of `track` in the pattern above. This can be
avoided by using the `nc::field` helper function, which takes three
arguments, that are pasted together to form a pattern:

* `field.name` is used as a pattern, and as the capture group
  (column) name for the pattern specified in the third argument.
* `between.pattern` is a pattern that matches between the other two patterns.
* `field.pattern` is the pattern that matches the text to be extracted
  in a capture group.
  
The example above can thus be re-written as below, avoiding the
repetition of `track` which was present above:

```{r}
track.pattern <- nc::field("track", " ", sample.or.anything)
nc::capture_all_str(trackDb.vec, track.pattern)
```

Finally we use `field` again to match the type column:

```{r}
any.lines.pattern <- "(?:\n[^\n]+)*"
nc::capture_all_str(
  trackDb.vec,
  track.pattern,
  any.lines.pattern,
  "\\s+",
  nc::field("type", " ", "[^\n]+"))
```

Exercise for the reader (easy): modify the above regex in order to capture
the bigDataUrl field, and three additional columns (red, green, blue)
from the color field. Assume that `bigDataUrl` occurs before `color`
in each track. Note that this is a limitation of the single regex
approach --- using two regex, as described in previous sections, could
extract any/all fields, even if they appear in different orders in
different tracks.

Exercise for the reader (hard): note that the last code block only
matches tracks which define the type field. How would you optionally
match the type field? Hint: the current `any.lines.pattern` can match
the type field.

# Parsing SweeD output files

Thanks to Marc Tollis for providing the example data used in this
section (from the SweeD bioinformatics program). Some representative
lines from one output file are shown below.

```{r}
info.txt.gz <- system.file(
  "extdata", "SweeD_Info.txt.gz", package="nc")
info.vec <- readLines(info.txt.gz)
info.vec[20:50]
```

The Alignment numbers must be matched with the numbers before slashes
in the other file,

```{r}
report.txt.gz <- system.file(
  "extdata", "SweeD_Report.txt.gz", package="nc")
report.vec <- readLines(report.txt.gz)
cat(report.vec[1:10], sep="\n")
cat(report.vec[1000:1010], sep="\n")
```

The goal is to produce a bed file, which has tab-separated values with
four columns: chrom, chromStart, chromEnd, Likelihood. The chrom
values appear in the info file (Chromosome) so we will need to join
the two files based on alignment ID.
First we capture all alignments in the info file:

```{r}
(info.dt <- nc::capture_all_str(
  info.vec,
  "Alignment ",
  alignment="[0-9]+",
  "\n\n\t\tChromosome:\t\t",
  chrom=".*",
  "\n"))
```

Then we capture all alignment/csv blocks in the report file:

```{r}
(report.dt <- nc::capture_all_str(
  report.vec,
  "//",
  alignment="[0-9]+",
  "\n",
  csv="[^/]+"
)[, {
  data.table::fread(text=csv)
}, by=alignment])
```

Note that because `by=alignment` was specified, `fread` is called for
each unique value of `alignment` (i.e. each row). The results are
combined into a single data.table with all of the csv data from the
original file, plus the additional `alignment` column. Next, we join
this table to the previous table in order to get the `chrom` column:

```{r}
(join.dt <- report.dt[info.dt, on=.(alignment)])
```

Finally the desired bed table can be created via

```{r}
join.dt[, .(
  chrom,
  chromStart=as.integer(Position-1),
  chromEnd=as.integer(Position),
  Likelihood)]
```

Exercise for the reader (easy): notice that the code above for
creating `info.dt` involves repetition in the pattern and group names
(`alignment`, `Alignment`, `chrom`, `Chromosome`). Re-write the
pattern using `nc::field` in order to eliminate that repetition.

Exercise for the reader (hard): notice that Chromosome is only the
first field -- how could you extract the other fields as well? Hint:
use `nc::field` in a helper function in order to avoid repetition.

<!--
%\VignetteEngine{knitr::knitr}
%\VignetteIndexEntry{vignette 1: capture first match}
-->

# Capture first match

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Capture first match in each character vector element

Consider the following vector which contains genome position strings,

```{r}
pos.vec <- c(
  "chr10:213,054,000-213,055,000",
  "chrM:111,000",
  "chr1:110-111 chr2:220-222") # two possible matches.
```

To capture the first genome position in each string, we use the following syntax.
The first argument is the subject character vector, 
and the other arguments are pasted together 
to make a capturing regular expression.
Each named argument generates a capture group; 
the R argument name is used for the column name of the result.

```{r}
(chr.dt <- nc::capture_first_vec(
  pos.vec, 
  chrom="chr.*?",
  ":",
  chromStart="[0-9,]+"))
str(chr.dt)
```

We can add type conversion functions on the same line as each named argument:

```{r}
keep.digits <- function(x)as.integer(gsub("[^0-9]", "", x))
(int.dt <- nc::capture_first_vec(
  pos.vec, 
  chrom="chr.*?",
  ":",
  chromStart="[0-9,]+", keep.digits))
str(int.dt)
```

Below we use list variables to create patterns which are re-usable,
and we use an un-named list to generate a non-capturing optional group:

```{r}
pos.pattern <- list("[0-9,]+", keep.digits)
range.pattern <- list(
  chrom="chr.*?",
  ":",
  chromStart=pos.pattern,
  list(
    "-",
    chromEnd=pos.pattern
  ), "?")
nc::capture_first_vec(pos.vec, range.pattern)
```

In summary, `nc::capture_first_vec` takes a variable number of arguments:

* The first argument is the subject character vector.
* The other arguments specify the pattern, via character strings,
  functions, and/or lists.
* If a pattern (character/list) is named, a capture group is generated in the regex.
* Each function is used to convert the text extracted by the previous
  named pattern argument.
* Lists may be used to avoid repetition in the definition of the pattern and type conversion functions.
* Each list generates a group in the regex (named list => named capture group, un-named list => non-capturing group).
* All patterns are pasted together in the order that they appear in
  the argument list.
  
## View generated regex

To see the generated regular expression pattern string, call
`nc::var_args_list` with the variable number of arguments that
specify the pattern:

```{r}
nc::var_args_list(range.pattern)
```

The generated regex is the `pattern` element of the resulting list
above. The other element `fun.list` indicates the names and type
conversion functions to use with the capture groups.

## Error/NA if any subjects do not match

The default is to stop with an error if any subject does not match:

```{r, error=TRUE, purl=FALSE}
bad.vec <- c(bad="does not match", pos.vec)
nc::capture_first_vec(bad.vec, range.pattern)
```

Sometimes you want to instead report a row of NA when a subject does
not match. In that case, use `nomatch.error=FALSE`:

```{r}
nc::capture_first_vec(bad.vec, range.pattern, nomatch.error=FALSE)
```

## Other regex engines

By default nc uses the PCRE regex engine. Other choices include ICU
and RE2. Each engine has different features, which are discussed in
[my R journal paper](https://github.com/tdhock/namedCapture-article).

The engine is configurable via the `engine` argument or the
`nc.engine` option:

```{r, error=TRUE, purl=FALSE}
u.subject <- "a\U0001F60E#"
u.pattern <- list(emoji="\\p{EMOJI_Presentation}")
old.opt <- options(nc.engine="ICU")
nc::capture_first_vec(u.subject, u.pattern)
nc::capture_first_vec(u.subject, u.pattern, engine="PCRE") 
nc::capture_first_vec(u.subject, u.pattern, engine="RE2")
options(old.opt)
```

For more details see the "engines" vignette, 

```r
vignette("v6-engines", package="nc")
```

## Capture first match from one or more character columns in a table

We also provide `nc::capture_first_df` which extracts text
from several columns of a data.frame, using a different 
regular expression for each column.

* It requires a `data.frame` (or `data.table`) as the first argument.
* It takes a variable number of other arguments, all of which must be
  named. For each other argument we call `nc::capture_first_vec` on one
  column of the input table.
* Each argument name specifies a column of the input table which will
  be used as the subject in `nc::capture_first_vec`.
* Each argument value specifies a pattern to be used with
  `nc::capture_first_vec`, in list/character/function format as
  explained in the previous section.
* The return value is a `data.table` with the same number of rows as
  the input, but with an additional column for each named capture
  group. New columns have the same names as the capture groups, so
  make sure they are unique.
* This function inputs `data.frame` and outputs `data.table` so can be
  used in a [pipe](https://r4ds.had.co.nz/pipes.html).
* For memory efficiency when the input is a `data.table`, it is
  modified to avoid copying the entire table.
  
This function can greatly simplify the code required to create numeric
data columns from character data columns. For example consider the
following data which was output from the
[sacct](https://slurm.schedmd.com/sacct.html) program.

```{r}
(sacct.df <- data.frame(
  Elapsed = c(
    "07:04:42", "07:04:42", "07:04:49",
    "00:00:00", "00:00:00"),
  JobID=c(
    "13937810_25",
    "13937810_25.batch",
    "13937810_25.extern",
    "14022192_[1-3]",
    "14022204_[4]"),
  stringsAsFactors=FALSE))
```

Say we want to filter by the total Elapsed time (which is reported as
hours:minutes:seconds), and base job id (which is the number before
the underscore in the JobID column). We could start by converting
those character columns to integers via:

```{r}
int.pattern <- list("[0-9]+", as.integer)
range.pattern <- list(
  "\\[",
  task1=int.pattern,
  list(
    "-",#begin optional end of range.
    taskN=int.pattern
  ), "?", #end is optional.
  "\\]")
nc::capture_first_df(sacct.df, JobID=range.pattern, nomatch.error=FALSE)
```
The result shown above is another data frame with an additional column
for each capture group. Next, we define another pattern that
matches either one task ID or the previously defined range pattern:

```{r}
task.pattern <- list(
  "_",
  list(
    task=int.pattern,
    "|",#either one task(above) or range(below)
    range.pattern))
nc::capture_first_df(sacct.df, JobID=task.pattern)
```

Below we match the complete JobID column:

```{r}
job.pattern <- list(
  job=int.pattern,
  task.pattern,
  list(
    "[.]",
    type=".*"
  ), "?")
nc::capture_first_df(sacct.df, JobID=job.pattern)
```

Below we match the Elapsed column with a different regex:

```{r}
elapsed.pattern <- list(
  hours=int.pattern,
  ":",
  minutes=int.pattern,
  ":",
  seconds=int.pattern)
nc::capture_first_df(sacct.df, JobID=job.pattern, Elapsed=elapsed.pattern)
```

Overall the result is another data table with an additional column for
each capture group. Note in the code below that the input `sacct.df`
was not modified, but if the input is `data.table` then it is modified:

```{r}
nc::capture_first_df(sacct.df, JobID=job.pattern)
sacct.df
(sacct.DT <- data.table::as.data.table(sacct.df))
nc::capture_first_df(sacct.DT, JobID=job.pattern)
sacct.DT
```

## Complex example: fsa file names

In this section we explain how to use various features to parse the
fsa file names in
[PROVEDIt](https://lftdi.camden.rutgers.edu/provedit/files/). Here are
a few representative examples:

```{r}
fsa.vec <- c(#control samples:
  "A01-Ladder-PP16-001.10sec.fsa",
  "D07_Ladder-IP_004.5sec.fsa",
  "A12_RB121514ADG_001.10sec.fsa", 
  "A10_RB102191515LEA-IP_001.10sec.fsa",
  ##single-source samples:
  "A02-RD12-0002-35-0.5PP16-001.10sec.fsa", 
  "G01_RD14-0003-35d3S30-0.01563P-Q10.0_003.10sec.fsa",
  "A06_RD14-0003-24d3a-0.0625IP-Q0.8_001.10sec.fsa", 
  "A08-RD12-0002-01d-0.125PP16-001.10sec.fsa",
  "A10-RD12-0002-04d1-0.0625PP16-001.10sec.fsa", 
  "C02_RD14-0003-15d2b-0.25IP-Q0.5_003.5sec.fsa",
  ##mixture samples:
  "A02-RD12-0002-1_2-1;9-0.125PP16-001.10sec.fsa", 
  "H07_RD14-0003-35_36_37_38_39-1;4;4;4;1-M2I35-0.75IP-QLAND_004.5sec.fsa")
```

The goal is to build a regex that can convert this character vector to
a data table with different columns for the different variables. The
structure of the file names is explained in [the supplementary
materials
PDF](https://lftdi.camden.rutgers.edu/wp-content/uploads/2019/12/PROVEDIt-Database-Naming-Convention-Laboratory-Methodsv1.pdf). We
will build a complex regex in terms of simpler sub-patterns. First
let's just match the start of each file name, which has the row/column
of the 96-well plate in which the sample was tested:

```{r}
well.pattern <- list(
  "^",
  well.letter="[A-H]",
  well.number="[0-9]+", as.integer)
nc::capture_first_vec(fsa.vec, well.pattern)
```

Now, let's match the end of each file name:

```{r}
end.pattern <- list(
  "[-_]",
  capillary="[0-9]+", as.integer,
  "[.]",
  seconds="[0-9]+", as.integer,
  "sec[.]fsa$")
nc::capture_first_vec(fsa.vec, end.pattern)
```

Now, let's take a look at what's in between:

```{r}
between.dt <- nc::capture_first_vec(
  fsa.vec, well.pattern, between=".*?", end.pattern)
between.dt$between
```

Notice that PP16/P/IP are the kit types. There may optionally be a DNA
template mass before, and optionally a Q score after. Let's match those:

```{r}
mass.pattern <- list(
  template.nanograms="[0-9.]*", as.numeric)
kit.pattern <- nc::quantifier(
  kit=nc::alternatives("PP16", "IP", "P"),
  "?")
q.pattern <- nc::quantifier(
  "-Q",
  Q.chr=nc::alternatives("LAND", "[0-9.]+"),
  "?")
old.opt <- options(width=100)
(before.dt <- nc::capture_first_vec(
  between.dt$between,
  "^",
  before=".*?",
  mass.pattern, kit.pattern, q.pattern, "$"))
options(old.opt)
```

Now to match the before column, we need some alternatives. The first
four rows are controls, and the other rows are samples which are
indicated by a project ID (prefix RD) and then a sample ID. The
mixture samples at the bottom contain semicolon-delimited mixture
proportions.

```{r}
project.pattern <- list(project="RD[0-9]+-[0-9]+", "-")
single.pattern <- list(id="[0-9]+", as.integer)
mixture.pattern <- list(
  ids.chr="[0-9_]+",
  "-",
  parts.chr="[0-9;]+",
  "-")
single.or.mixture <- list(
  project.pattern,
  nc::alternatives(mixture.pattern, single.pattern))
control.pattern <- list(control="[^-_]+")
single.mixture.control <- list("[-_]", nc::alternatives(
  single.or.mixture, control.pattern))
(rest.dt <- nc::capture_first_vec(
  before.dt$before, single.mixture.control, rest=".*"))
```

The rest may contain some information related to the dilution and
treatment, which we can parse as follows:

```{r}
dilution.pattern <- nc::quantifier(
  "[dM]",
  dilution.number="[0-9]?", as.integer,
  "?")
treatment.pattern <- nc::quantifier(
  treatment.letter="[a-zA-Z]",
  nc::quantifier(treatment.number="[0-9]+", as.integer, "?"),
  "?")
dilution.treatment <- list(
  dilution.pattern,
  treatment.pattern,
  "[_-]?")
nc::capture_first_vec(rest.dt$rest, "^", dilution.treatment, "$")
```

Now we just have to put all of those patterns together to get a full
match:

```{r}
fsa.pattern <- list(
  well.pattern,
  single.mixture.control,
  dilution.treatment,
  mass.pattern, kit.pattern, q.pattern,  
  end.pattern)
(match.dt <- nc::capture_first_vec(fsa.vec, fsa.pattern))
```

And to verify we parsed everything correctly we print the subjects
next to each row:

```{r}
disp.dt <- data.table::data.table(match.dt)
names(disp.dt) <- paste(1:ncol(disp.dt))
old.opt <- options("datatable.print.colnames"="none")
split(disp.dt, fsa.vec)
options(old.opt)
```

Incidentally, if you print the regex as a string it looks like
something that would be essentially impossible to write/maintain by
hand (without `nc` helper functions):

```{r}
nc::var_args_list(fsa.pattern)$pattern
```

In conclusion, we have shown how `nc` can help build complex regex
patterns from simple, understandable sub-patterns in R code. The
general workflow we have followed in this section can be generalized
to other projects. First, identify a small set of subjects to use for
testing (`fsa.vec` in the code above). Second, create some
sub-patterns for the start/end of the subjects, and create a
sub-pattern that will capture everything else in between. Third,
create some more sub-patterns to match the "everything else" in the
previous step (e.g. `before.dt$before`, `between.dt$between`,
`rest.dt$rest` above). Repeat the process until there is nothing left
to match, and then concatenate the sub-patterns to match the whole
subject.

---
title: "Capture melt"
date: "`r Sys.Date()`"
output:
  rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Capture melt}
  %\VignetteEngine{knitr::rmarkdown}
  \usepackage[utf8]{inputenc}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

This vignette explains how to use functions for "melting" wide data
tables, i.e. converting to tall/long data tables. To clarify the
discussion we introduce the following three terms:

* Reshape columns contain the data which are present in equal amounts
  in the input and output, but in fewer columns in the output.
  These are `measure.vars` in reshape2/data.table.
* Copy columns contain data which are copied from a single row of the
  input into several rows of the output. These are `id.vars`
  in reshape2/data.table.
* Capture columns are only present in the output, and contain data
  extracted from the names of the input reshape columns,
  using a capturing regular expression.
  
The `nc` functions use `data.table::melt` internally:

* `nc::capture_melt_single` is useful when you have a wide data table
  with lots of columns, all of the same type, that you would like to
  melt/reshape into a single result/output column. This uses `melt`
  with a character vector for `measure.vars`.
* `nc::capture_melt_multiple` is useful when you have a wide data
  table with columns of possibly different types that you would like
  to melt/reshape into multiple result/output columns. This uses
  `melt` with a list for `measure.vars`.

Both are useful when you want to use a regular expression to specify
both (1) the set of input columns to reshape and (2) some information
to extract from those column names.

## Reshaping several input columns into a single output column

Sometimes you want to melt a "wide" data table which has several
distinct pieces of information encoded in each column name. One
example is the familiar iris data, which have flower part and
measurement dimension encoded in each of four column names:

```{r}
head(iris)
```

Those four reshape column names can be specified via a regex in
`nc::capture_melt_single`. The first argument is the input data table
to reshape, and the subsequent arguments are interpreted as a pattern
which is passed to `nc::capture_first_vec`. Any input column names
which match the specified regex will be passed as `measure.vars` to
`melt`:

```{r}
(iris.tall <- nc::capture_melt_single(
  iris,
  part=".*",
  "[.]",
  dim=".*",
  value.name="cm"))
```

Note the output above has one copy column (Species), two capture
columns (part, dim), and one reshape column (cm). Internally the
function joins the result of `nc::capture_first_vec` (on column names)
to the result of `melt` (on the data).

The reshaped data can be plotted with different parts on rows and
different dimensions on columns:

```{r, fig.width=10}

if(require(ggplot2)){
  ggplot()+
    theme_bw()+
    theme(panel.spacing=grid::unit(0, "lines"))+
    facet_grid(part ~ dim)+
    geom_bar(aes(cm, fill=Species), data=iris.tall)
}

```

## Reshaping several input columns into multiple output columns

We could instead use `capture_melt_multiple` to get multiple output
columns. Like `capture_melt_single`, the first argument of
`capture_melt_multiple` is the subject data table and the following
arguments form a pattern which is matched to the input data column
names. However the pattern must have at least two groups:

* One group must be named `column`. This group is used to capture the
  part of the input column names which will be used for the output
  column names. The output will contain one column for each unique
  value captured in this group.
* One or more other groups (with any names) specify an identifier for
  each observation.

```{r}
(iris.part.cols <- nc::capture_melt_multiple(
  iris,
  column=".*",
  "[.]",
  dim=".*"))
```

Note that the reshaped table above contains one copy column (Species),
one capture column (dim), and two reshape columns (Petal, Sepal). We
can plot these data to see whether or not sepals are bigger than
petals:

```{r, fig.width=10}

if(require(ggplot2)){
  ggplot()+
    theme_bw()+
    theme(panel.spacing=grid::unit(0, "lines"))+
    facet_grid(dim ~ Species)+
    coord_equal()+
    geom_abline(slope=1, intercept=0, color="grey")+
    geom_point(aes(
      Petal, Sepal),
      data=iris.part.cols)
}

```

It is clear from the plot above that sepals are indeed both longer and
wider than petals, on each measured plant.

## Melting WHO data with a more complex pattern

Another data set where it is useful to do column name pattern matching
followed by melting is the World Health Organization data:

```{r}

if(requireNamespace("tidyr")){
  data(who, package="tidyr")
}else{
  who <- data.frame(id=1, new_sp_m5564=2, newrel_f65=3)
}
names(who)

```

Each column which starts with `new` has three distinct pieces of
information encoded in its name: diagnosis type (e.g. sp or rel),
gender (m or f), and age range (e.g. 5564 or 1524). We would like to
use a regex to match these column names, then using the matching
columns as measure.vars in a melt, then join the two results. 

```{r}
new.diag.gender <- list(
  "new_?",
  diagnosis=".*",
  "_",
  gender=".")
nc::capture_melt_single(who, new.diag.gender, ages=".*")
```

Note the output includes the new reshape column called `value` by
default, as in `melt`. The input reshape column names which matched
the specified pattern, and there is a new column for each group in
that pattern. The following example shows how to rename the `value`
column and use numeric type conversion functions:

```{r}
years.pattern <- list(new.diag.gender, ages=list(
  min.years="0|[0-9]{2}", as.numeric,
  max.years="[0-9]{0,2}", function(x)ifelse(x=="", Inf, as.numeric(x))))
(who.typed <- nc::capture_melt_single(
  who, years.pattern,
  value.name="count"))
str(who.typed)
```

Note in the code/result above that non-character captured output
columns can be obtained by specifying type conversion functions in the
pattern.

## Melting monthly Eurostat data

Another example is [exchange rate data from
Eurostat](https://ec.europa.eu/eurostat).
We first use read the entire data set into R:

```{r}
ert.gz <- system.file(
  "extdata", "ert_eff_ic_m.tsv.gz", package="nc", mustWork=TRUE)
ert.all <- data.table::fread(ert.gz, na.strings=":")
ert.all[1:5, 1:5]
```

We see that the first column has some CSV data which we can parse via:

```{r}
ert.first <- ert.all[, 1]
csv.lines <- c(sub("\\\\.*", "", names(ert.first)), ert.first[[1]])
ert.first.dt <- data.table::fread(text=paste(csv.lines, collapse="\n"))
ert.wide <- data.table::data.table(ert.first.dt, ert.all[,-1])
ert.wide[1:5, 1:5]
```

The wide data table can then be melted:


```{r}
(ert.tall <- nc::capture_melt_single(
  ert.wide,
  year="[0-9]{4}", as.integer,
  "M",
  month="[0-9]{2}", as.integer))
```

After that we can create a time variable and plot via

```{r}

ert.tall[, month.IDate := data.table::as.IDate(
  sprintf("%d-%d-15", year, month))]
if(require("ggplot2")){
  ggplot()+
    geom_hline(aes(
      yintercept=value),
      color="grey",
      data=data.frame(value=100))+
    geom_line(aes(
      month.IDate, value, color=geo),
      data=ert.tall[geo %in% c("CA", "US", "JP", "FR")])+
    facet_grid(exch_rt ~ .)+
    theme_bw()+
    theme(panel.spacing=grid::unit(0, "lines"))
}

```

Another way to do it would be via

```{r}
nc::capture_melt_single(ert.wide, month.POSIXct="[0-9].*", function(x){
  as.POSIXct(strptime(paste0(x,"15"), "%YM%m%d"))
})
```

## Melting into multiple output columns with missing input columns

What if the input data set has "missing" input columns?

```{r}
iris.missing <- iris[, names(iris) != "Sepal.Length"]
head(iris.missing)
```

In that case melting into multiple columns is an error by default:

```{r, error=TRUE, purl=FALSE}
iris.pattern <- list(column=".*", "[.]", dim=".*")
nc::capture_melt_multiple(iris.missing, iris.pattern)
```

The error message explains that the number of input columns for each
value of `dim` must be the same, but there is one for `Length` and two
for `Width`. To ignore the error and fill the output with missing
values,

```{r}
nc::capture_melt_multiple(iris.missing, iris.pattern, fill=TRUE)
```

Note the missing values in the table above, which correspond to the
missing input column in the original/wide data set.

## Multiple output columns of different types

Some real-world data sets can be reshaped into output columns with
different types. An example data set from the
[PROVEDIt](https://lftdi.camden.rutgers.edu/repository/PROVEDIt_1-5-Person%20CSVs%20Filtered.zip)
benchmark in criminology:

```{r}
peaks.csv <- system.file(
  "extdata", "RD12-0002_PP16HS_5sec_GM_F_1P.csv",
  package="nc", mustWork=TRUE)
peaks.wide <- data.table::fread(peaks.csv)
print(data.table::data.table(
  names=names(peaks.wide),
  class=sapply(peaks.wide, class)),
  topn=10)
```

There are 303 columns, with info for 100 peaks. Each peak has three
features: Allele=character, Size=numeric, and Height=integer. The
ending peaks are class logical because they are all missing. These
data can be reshaped via

```{r}
peaks.tall <- nc::capture_melt_multiple(
  peaks.wide,
  column=".*",
  " ",
  peak="[0-9]+", as.integer,
  na.rm=TRUE)
old.opt <- options(width=90)
print(peaks.tall)
str(peaks.tall)
options(old.opt)
```

%\VignetteIndexEntry{Examples}
%\VignetteEngine{knitr::knitr}
\documentclass[12pt]{article}
\usepackage[cm]{fullpage}
\usepackage{hyperref}
\usepackage{amssymb,amsmath}
\DeclareMathOperator*{\minimize}{minimize}

\begin{document}

\title{PeakSegDisk usage examples}
\author{Toby Dylan Hocking}
\maketitle

Welcome to PeakSegDisk, an R package for optimal peak
detection in very large count data sequences.

<<setup, echo=FALSE>>=
knitr::opts_chunk$set(
  collapse = TRUE,
  fig.width = 7,
  fig.height = 3,
  fig.align = "center",
  comment = "#>"
)
@

\section{Related work}

The PeakSeg R packages contain algorithms for inferring optimal
segmentation models subject to the constraint that up changes must be
followed by down changes, and vice versa. This ensures that the model
can be interpreted in terms of peaks (after up changes) and background
(after down changes).

\begin{description}
\item[PeakSegDP] the historically first PeakSeg package,
  \url{https://CRAN.R-project.org/package=PeakSegDP} provides a
  heuristic quadratic time algorithm for computing models from 1 to S
  segments for a single sample. This was the original algorithm
  described in our ICML'15 paper,
  \url{http://jmlr.org/proceedings/papers/v37/hocking15.html}, but it
  is neither fast nor optimal, so in practice we recommend to use our
  newer packages below instead.
\item[PeakSegOptimal]
  \url{https://CRAN.R-project.org/package=PeakSegOptimal} provides
  log-linear time algorithms for computing optimal models with
  multiple peaks for a single sample. The algorithms are faster and
  more accurate than PeakSegDP, \url{https://arxiv.org/abs/1703.03352}
\item[PeakSegDisk] \url{https://github.com/tdhock/PeakSegDisk}
  provides an on-disk implementation of optimal log-linear algorithms
  for computing multiple peaks in a single sample. Computes same
  models as PeakSegOptimal but works for much larger data sets because
  disk is used for storage instead of memory.
  \url{https://arxiv.org/abs/1810.00117}
\item[PeakSegJoint] \url{https://CRAN.R-project.org/package=PeakSegJoint}
  provides a fast heuristic algorithm for computing models with a single
  common peak in $0,...,S$
  samples. \url{https://arxiv.org/abs/1506.01286}
\item[PeakSegPipeline]
  \url{https://github.com/tdhock/PeakSegPipeline} provides a pipeline
  for genome-wide peak calling using the other PeakSeg packages.
\end{description}

The remainder of this vignette is dedicated to an explanation of how
to use PeakSegDisk.

\section{Simulate a noisy integer vector with changes}

The first example we will treat is detecting peaks in a vector of
integer data, with possibly the same values at adjacent
positions. This is an inefficient representation for large genomic
data, but it is the typical output from simulation functions like
\texttt{rpois}:

<<>>=
sim.seg <- function(seg.mean, size.mean=15){
  seg.size <- rpois(1, size.mean)
  rpois(seg.size, seg.mean)
}
set.seed(1)
seg.mean.vec <- c(1.5, 3.5, 0.5, 4.5, 2.5)
z.list <- lapply(seg.mean.vec, sim.seg)
(z.rep.vec <- unlist(z.list))
@

From the output above it is clear that these simulated data are
integers, with some identical values at adjacent positions.

Below we put these data into a data table in order to plot them along
with the model using ggplot2:

<<ggcount>>=
count.df <- data.frame(
  chrom="chrUnknown",
  chromStart=0:(length(z.rep.vec)-1),
  chromEnd=1:length(z.rep.vec),
  count=z.rep.vec)
library(ggplot2)
gg.count <- ggplot()+
  xlab("position")+
  geom_point(aes(
    chromEnd, count),
    shape=1,
    data=count.df)
gg.count
@

The true changepoints in the simulation are shown below.

<<>>=
n.segs <- length(seg.mean.vec)
seg.size.vec <- sapply(z.list, length)
seg.end.vec <- cumsum(seg.size.vec)
change.vec <- seg.end.vec[-n.segs]+0.5
change.df <- data.frame(
  changepoint=change.vec)
gg.change <- gg.count+
  geom_vline(aes(
    xintercept=changepoint, color=model),
    data=data.frame(change.df, model="simulation"))+
  scale_color_manual(
    values=c(
      simulation="black",
      fitted="green"))
gg.change
@

\section{Segment a vector of integers}

Let $z_1, \dots, z_n\in\mathbb Z_+$ be the sequence of $n$
non-negative count data in z.rep.vec, and let $w_1=\cdots=w_n=1$ be
weights which are all 1. The peak detection algorithm computes the
solution to the following optimization problem:

\begin{align*}
  \minimize_{
    \substack{
    \mathbf m\in\mathbb R^n,\ \mathbf s\in\{0, 1\}^n\\
\mathbf c\in\{-1, 0,1\}^{n-1}\\
}
    } &\ \
  \sum_{i=1}^n w_i \ell(m_i, z_i) + \lambda \sum_{i=1}^{n-1} I(c_i = 1) \\
  \text{subject to\ \ } &\ \text{no change: }c_i = 0 \Rightarrow m_i = m_{i+1}\text{ and }s_i=s_{i+1}
  \nonumber\\
&\ \text{go up: }c_i = 1 \Rightarrow m_i \leq m_{i+1}\text{ and }(s_i,s_{i+1})=(0,1),
  \nonumber\\
&\ \text{go down: } c_i = -1 \Rightarrow m_i \geq m_{i+1}\text{ and }(s_i,s_{i+1})=(1,0),\\
  & \ \text{start and end down: } s_1=s_n=0.\nonumber
\end{align*}
where $\ell(m, z)= m - z\log m$ is the Poisson loss. The optimization
variables are $m_i$ for the segment mean, $s_i$ for hidden state, and
$c_i$ for type of changepoint. The penalty term is proportional to the number of
changepoint variables $c_i$ which are equal to 1 (which is the same as
the number of peaks in the resulting model).

To run the peak detection algorithm a numeric penalty parameter
$\lambda\geq 0$ must be specified by the user. The smallest value is 0
which yields max peaks, and the largest value is Inf which yields no
peaks.  The code below runs the peak detection algorithm on this count
data vector, using the penalty parameter $\lambda = 10.5$:

<<>>=
fit <- list()
(fit$vec <- PeakSegDisk::PeakSegFPOP_vec(z.rep.vec, 10.5))
@


The model output list above includes \verb|segments|, a data table with
one row for each segment mean, and \verb|loss|, a data table with one
row that reports the model meta-data. Of interest are:

\begin{itemize}
\item \verb|penalty|, the user-provided penalty value,
\item \verb|segments|, the number of segments,
\item \verb|peaks|, the number of peaks (even-numbered segments),
\item \verb|bases|, the number of data points in repetitive form (not run-length encoding),
\item \verb|bedGraph.lines|, the number of data points in run-length encoding form,
\item \verb|mean.pen.cost|, the optimal mean loss plus penalty*peaks,
\item \verb|total.loss|, the optimal total Poisson loss over all data points,
\item \verb|equality.constraints|, the number of adjacent segment
  means that are equal in the optimal solution. Note that when this
  number is greater than 0, then there are some active equality
  constraints, and the optimal model is therefore not feasible for the
  strict inequality constraints, which implies that the optimum of the
  problem with strict inequality constraints is undefined, i.e. for
  any sub-optimal solution that satisfies the strict inequality
  constraints, we can find a lower cost solution that satifies the
  strict inequality constraints (but is still sub-optimal), by getting
  closer to the solution with active equality constraints.
\item \verb|megabytes|, the storage space on disk used by the solver,
\item \verb|seconds|, the amount of time used by the solver,
\item \verb|mean.intervals|, \verb|max.intervals|, statistics over all intervals
  (candidate changepoints) computed by the functional pruning
  algorithm, useful for analyzing computational complexity, which is
  linear in the number of intervals.
\end{itemize}
Note in particular that \verb|PeakSegFPOP_vec| internally uses \verb|rle| to
construct a run-length encoding, which is passed to the solver to save
time/storage. In this case the repetitive integer data vector contains
\Sexpr{fit$vec$loss$bases} elements but the coverage.bedGraph data file
contains only \Sexpr{fit$vec$loss$bedGraph.lines} lines. In real genomic data
sets the difference is typically much larger.

<<>>=
gg.change+
  geom_segment(aes(
    chromStart+0.5, mean, xend=chromEnd+0.5, yend=mean, color=model),
    data=data.frame(fit$vec$segments, model="fitted"))
@

It is clear from the plot above that the first three changepoints are
estimated exactly and the last one is a bit over-estimated.

Also note that a default plot method is defined for these objects:

<<>>=
plot(fit$vec)
@

\section{Segment a data frame}

Another interface that can be used on a data.frame with $n$ rows and
exactly 4 columns (chrom, chromStart, chromEnd, count) is
\verb|PeakSegFPOP_df|. For each row $i\in\{1,\dots, n\}$, let
$z_i\in\mathbb Z_+$ be the non-negative count data (count column), and
let $w_i>0$ be the weight (equal to the number of bases,
chromEnd-chromStart). The optimization problem we solve is the same as
before. Note that this function does not perform run-length encoding
for you:

<<>>=
(fit$df <- PeakSegDisk::PeakSegFPOP_df(count.df, 10.5))
@

Note how \verb|bedGraph.lines| is now the same size as \verb|bases|,
\Sexpr{fit$df$loss$bedGraph.lines}. The time/storage complexity is
log-linear in the number of \verb|bedGraph.lines|, so it is more efficient
to use the run-length encoding. This can be easily done in R:

<<>>=
z.rle.vec <- rle(z.rep.vec)
chromEnd <- cumsum(z.rle.vec$lengths)
rle.df <- data.frame(
  chrom="chrUnknown",
  chromStart=c(0L, chromEnd[-length(chromEnd)]),
  chromEnd,
  count=z.rle.vec$values)
gg.rle <- ggplot()+
  geom_segment(aes(
    chromStart+0.5, count, xend=chromEnd+0.5, yend=count),
    data=rle.df)+
  geom_point(aes(
    chromEnd, count),
    shape=1,
    data=rle.df)+
  geom_vline(aes(
    xintercept=changepoint, color=model),
    data=data.frame(change.df, model="simulation"))+
  scale_color_manual(
    values=c(
      simulation="black",
      fitted="green"))+
  xlab("position")
gg.rle
@

The plot above shows the run-length encoded data, with a \verb|geom_point|
for the last position in each run, and a \verb|geom_segment| extending left
to the first position. These data can be segmented as above:

<<>>=
(fit$rle <- PeakSegDisk::PeakSegFPOP_df(rle.df, 10.5))
gg.rle+
  geom_segment(aes(
    chromStart+0.5, mean, xend=chromEnd+0.5, yend=mean, color=model),
    data=data.frame(fit$rle$segments, model="fitted"))
@

\section{Write the file yourself}

The interfaces discussed in the previous sections are perhaps the most
intuitive for useRs, but they are also the least efficient, so they
are not recommended for large data.

In this section we introduce the most efficient way of using PeakSegDisk, which involves:

\begin{itemize}
\item creating a ``problem'' directory for each segmentation problem
  (sample and genome subset),
\item saving the data to \verb|coverage.bedGraph| in that directory,
\item and then running \verb|PeakSegFPOP_dir|.
\end{itemize}

The reason why this method is recommended for large data is because
\verb|PeakSegFPOP_dir| saves its results to the ``problem'' directory. So
if a certain result has already been computed, these result files are
used as a cache, and are read instead of doing computations, which
saves a lot of time.
The file system is used as the interface in
order to support very large data sets with very little memory usage.

To use \verb|PeakSegFPOP_dir| the data should be saved to a
chrXX-start-end/coverage.bedGraph file, where the problem directory
``chrXX-start-end'' should be named using a genome postion string:

\begin{itemize}
\item chrXX is the chromosome (which is irrelevant to the algorithm),
\item start is the 0-based first position of the region to segment (the smallest possible value is 0),
\item end is the 1-based end position (the smallest possible value is 1).
\end{itemize}

<<>>=
data.dir <- file.path(
  tempfile(),
  with(rle.df, sprintf(
    "%s-%d-%d", chrom[1], min(chromStart), max(chromEnd))))
dir.create(data.dir, showWarnings=FALSE, recursive=TRUE)
coverage.bedGraph <- file.path(data.dir, "coverage.bedGraph")
write.table(
  rle.df, coverage.bedGraph,
  sep="\t", row.names=FALSE, col.names=FALSE)
@

The next step is to run the main solver,

<<>>=
(fit$dir <- PeakSegDisk::PeakSegFPOP_dir(data.dir, 10.5))
@

The
underlying C++ code
creates penalty-specific files such as

\verb|chrXX-start-end/coverage.bedGraph_penalty=0.1_loss.tsv| which
are used to store/cache the results.  If the files already exist (and
are consistent) then \verb|PeakSegFPOP_dir| just reads them; otherwise
it runs the dynamic programming C++ code in order to create those
files, which are then read into R.

\section{Computing the model with a given number of peaks}

The \verb|sequentialSearch_dir| function can be used to compute the
optimal model with a certain number of peaks:

<<>>=
if(interactive() && requireNamespace("future"))future::plan("multiprocess")
(fit$search <- PeakSegDisk::sequentialSearch_dir(data.dir, 2L, verbose=1))
@

The algorithm must evaluate several penalty values to compute the
optimal model with a certain number of peaks. The \verb|others| component of
the model list above shows that

\begin{itemize}
\item the search starts with penalty values 0 and Inf, which result in models
  with \Sexpr{fit$search$others[penalty==0, peaks]} and 0 peaks, respectively.
\item the next penalty evaluated is \Sexpr{sprintf("%.2f", fit$search$others[iteration==2, penalty])},
  which results in \Sexpr{fit$search$others[iteration==2, peaks]} peaks.
\item the final penalty evaluated is \Sexpr{sprintf("%.2f", fit$search$others[iteration==3, penalty])},
  which results in \Sexpr{fit$search$others[iteration==3, peaks]} peaks.
\end{itemize}

At each step (except the first) the new penalties are computed based
on the loss values found in the previous step.
If present with a registered parallel future plan,
the \verb|future.apply| package is used to run the first step
(penalties $0,\infty$) in parallel.

Note how the number of peaks and \verb|total.loss| of this model is
the same as the other models computed above,

<<>>=
lossDF <- function(L)data.frame(L$loss)[, names(fit$dir$loss)]
do.call(rbind, lapply(fit, lossDF))
@

Finally we demonstrate how the filesystem caching is especially useful
for the sequential search. In the code below we ask the sequential
search algorithm to compute the optimal model with four peaks:

<<>>=
four.peaks <- PeakSegDisk::sequentialSearch_dir(data.dir, 4L)
four.peaks$others[, .(iteration, penalty, peaks)]
@

Looking at the output above, we see that the first three iterations of
the sequential search require computing models with 26, 0, 6, 2
peaks. Since all of these have been previously computed (and saved to
disk), the dynamic programming algorithm does not need to be re-run,
and instead the model results are simply read from the files. After
that the dynamic programming is run for the subsequent iterations
4-6. In this particular example the savings in computation time is not
extraordinary, but in real genomic data, this can result in
substantial speed-ups.

\end{document}

\name{check_df_names}
\alias{check_df_names}
\title{check df names}
\description{Check that first argument is a data frame and then call
\code{\link{check_names}} on its names.}
\usage{check_df_names(...)}
\arguments{
  \item{\dots}{data frame, regex pattern args.}
}



\author{Toby Hocking <toby.hocking@r-project.org> [aut, cre]}

\name{quantifier}
\alias{quantifier}
\title{quantifier}
\description{Create a \code{\link{group}} with a quantifier.}
\usage{quantifier(...)}
\arguments{
  \item{\dots}{Pattern(s) to be enclosed in a \code{\link{group}}, and a quantifier (last
argument). A quantifier is character string: "?" for zero or one,
"*?"  for non-greedy zero or more, "+" for greedy one or more,
etc.}
}

\value{A pattern list.}

\author{Toby Hocking <toby.hocking@r-project.org> [aut, cre]}




\examples{

## No need to use nc::quantifier when the pattern to be quantified
## is just a string literal.
digits <- "[0-9]+"

## nc::quantifier is useful when there is a sequence of patterns to
## be quantified, here an optional group with a dash (not captured)
## followed by some digits (captured in the chromEnd group).
str(optional.end <- nc::quantifier("-", chromEnd=digits, "?"))
str(optional.end <- list(list("-", chromEnd=digits), "?"))#same

## Use it as a sub-pattern for capturing genomic coordinates.
chr.pos.vec <- c(
  "chr10:213054000-213055000",
  "chrM:111000",
  "chr1:110-111 chr2:220-222") # two possible matches.
nc::capture_first_vec(
  chr.pos.vec,
  chrom="chr.*?",
  ":",
  chromStart=digits,
  optional.end)

## Another example which uses quantifier twice, for extracting code
## chunks from Rmd files.
vignette.Rmd <- system.file(
  "extdata", "vignette.Rmd", package="nc")
non.greedy.lines <- nc::quantifier(".*\\n", "*?")
optional.name <- nc::quantifier(" ", name="[^,}]+", "?")
Rmd.dt <- nc::capture_all_str(
  vignette.Rmd,
  before=non.greedy.lines,
  "```\\\\{r",
  optional.name,
  parameters=".*",
  "\\\\}\\n",
  code=non.greedy.lines,
  "```")
Rmd.dt[, chunk := 1:.N]
Rmd.dt[, .(chunk, name, parameters, some.code=substr(code, 1, 20))]

}

\name{try_or_stop_print_pattern}
\alias{try_or_stop_print_pattern}
\title{try or stop print pattern}
\description{Try to run a capture function. If it fails we wrap the error
message with a more informative message that also includes the
generated pattern.}
\usage{try_or_stop_print_pattern(expr, 
    pat, engine)}
\arguments{
  \item{expr}{expression to try.  }
  \item{pat}{regex pattern string.}
  \item{engine}{string: regex \code{engine}.}
}



\author{Toby Hocking <toby.hocking@r-project.org> [aut, cre]}

\name{stop_for_engine}
\alias{stop_for_engine}
\title{stop for engine}
\description{Stop if specified \code{engine} is not available.}
\usage{stop_for_engine(engine)}
\arguments{
  \item{engine}{character string: PCRE, RE2, or ICU.}
}

\value{character string.}

\author{Toby Hocking <toby.hocking@r-project.org> [aut, cre]}

\name{capture_longer_spec}
\alias{capture_longer_spec}
\title{capture longer spec}
\description{Create a spec \code{data} table for input to \code{\link[tidyr]{pivot_longer_spec}}.}
\usage{capture_longer_spec(data, 
    ..., values_to = "value")}
\arguments{
  \item{data}{Data table to reshape (actually the \code{data} are ignored, and only the
column names are used).}
  \item{\dots}{Regex and conversion as described in \code{\link{capture_first_vec}}. This is
processed by \code{\link{measure}} so if "column" is used as an argument name
then there will be multiple output columns in the reshaped \code{data}
(analogous to names_to=".value" in \code{\link[tidyr]{pivot_longer}}).}
  \item{values_to}{string to use for name of value column in output (only used if
there is a single output column in the reshaped \code{data}).}
}

\value{\code{data} table describing a reshape longer operation.}

\author{Toby Hocking <toby.hocking@r-project.org> [aut, cre]}




\examples{

(one.iris <- iris[1,])
(single.spec <- nc::capture_longer_spec(iris, part=".*", "[.]", dim=".*", values_to="cm"))
(multiple.spec <- nc::capture_longer_spec(iris, part=".*", "[.]", column=".*"))
if(requireNamespace("tidyr")){
  tidyr::pivot_longer_spec(one.iris, single.spec)
  tidyr::pivot_longer_spec(one.iris, multiple.spec)
}

}

\name{capture_first_df}
\alias{capture_first_df}
\title{Capture first match in columns of a data frame}
\description{Capture first matching text from one or more character columns of
a data frame, using a different regular expression for each
column.}
\usage{capture_first_df(..., 
    nomatch.error = getOption("nc.nomatch.error", 
        TRUE), existing.error = getOption("nc.existing.error", 
        TRUE), engine = getOption("nc.engine", 
        "PCRE"), type.convert = getOption("nc.type.convert", 
        FALSE))}
\arguments{
  \item{\dots}{subject data frame, colName1=list(groupName1=pattern1, fun1, etc),
colName2=list(etc), etc. First argument must be a data frame with
one or more character columns of subjects for matching. If the
first argument is a data table then it will be modified using
\code{\link[data.table]{set}} (for memory efficiency, to avoid copying the whole
data table); otherwise the input data frame will be copied to a
new data table. Each other argument must be named using a column
name of the subject data frame, e.g. colName1, colName2. Each
other argument value must be a list that specifies the
regex/conversion to use (in string/function/list format as
documented in \code{\link{capture_first_vec}}, which is used on each named
column), and possibly a column-specific \code{engine} to use.}
  \item{nomatch.error}{if TRUE (default), stop with an error if any subject does not
match; otherwise subjects that do not match are reported as
missing/NA rows of the result.}
  \item{existing.error}{if TRUE (default to avoid data loss), stop with an error if any
capture groups have the same name as an existing column of
subject.}
  \item{engine}{character string, one of PCRE, ICU, RE2. This \code{engine} will be used
for each column, unless another \code{engine} is specified for that
column in \code{...}}
  \item{type.convert}{Default conversion function, which will be used on each capture
\code{\link{group}}, unless a specific conversion is specified for that
\code{\link{group}}. If TRUE, use \code{\link[utils]{type.convert}}; if FALSE, use
\code{\link[base]{identity}}; otherwise must be a function of at least one
argument (character), returning an atomic vector of the same
length.}
}

\value{data.table with same number of rows as subject, with an additional
column for each named capture \code{\link{group}} specified in \code{...}}

\author{Toby Hocking <toby.hocking@r-project.org> [aut, cre]}




\examples{

## The JobID column can be match with a complicated regular
## expression, that we will build up from small sub-pattern list
## variables that are easy to understand independently.
(sacct.df <- data.frame(
  JobID = c(
    "13937810_25", "13937810_25.batch",
    "13937810_25.extern", "14022192_[1-3]", "14022204_[4]"),
  Elapsed = c(
    "07:04:42", "07:04:42", "07:04:49",
    "00:00:00", "00:00:00"),
  stringsAsFactors=FALSE))

## Just match the end of the range.
int.pattern <- list("[0-9]+", as.integer)
end.pattern <- list(
  "-",
  task.end=int.pattern)
nc::capture_first_df(sacct.df, JobID=list(
  end.pattern, nomatch.error=FALSE))

## Match the whole range inside square brackets.
range.pattern <- list(
  "[[]",
  task.start=int.pattern,
  end.pattern, "?", #end is optional.
  "[]]")
nc::capture_first_df(sacct.df, JobID=list(
  range.pattern, nomatch.error=FALSE))

## Match either a single task ID or a range, after an underscore.
task.pattern <- list(
  "_",
  list(
    task.id=int.pattern,
    "|",#either one task(above) or range(below)
    range.pattern))
nc::capture_first_df(sacct.df, JobID=task.pattern)

## Match type suffix alone.
type.pattern <- list(
  "[.]",
  type=".*")
nc::capture_first_df(sacct.df, JobID=list(
  type.pattern, nomatch.error=FALSE))

## Match task and optional type suffix.
task.type.pattern <- list(
  task.pattern,
  type.pattern, "?")
nc::capture_first_df(sacct.df, JobID=task.type.pattern)

## Match full JobID and Elapsed columns.
nc::capture_first_df(
  sacct.df,
  JobID=list(
    job=int.pattern,
    task.type.pattern),
  Elapsed=list(
    hours=int.pattern,
    ":",
    minutes=int.pattern,
    ":",
    seconds=int.pattern))

## If input is data table then it is modified for memory efficiency,
## to avoid copying entire table.
sacct.DT <- data.table::as.data.table(sacct.df)
nc::capture_first_df(sacct.df, JobID=task.pattern)
sacct.df #not modified.
nc::capture_first_df(sacct.DT, JobID=task.pattern)
sacct.DT #modified!

}

\name{check_names}
\alias{check_names}
\title{check names}
\description{Check that \code{subject} is a vector of unique names and then call
\code{\link{capture_first_vec}}.}
\usage{check_names(subject, 
    var.args)}
\arguments{
  \item{subject}{character vector, data frame column names.}
  \item{var.args}{regex pattern list.}
}



\author{Toby Hocking <toby.hocking@r-project.org> [aut, cre]}

\name{apply_type_funs}
\alias{apply_type_funs}
\title{apply type funs}
\description{Convert columns of \code{match.mat} using corresponding functions from
\code{fun.list}, then handle any duplicate capture \code{\link{group}} names.}
\usage{apply_type_funs(match.mat, 
    fun.list)}
\arguments{
  \item{match.mat}{Character matrix (matches X groups).}
  \item{fun.list}{Named list of functions to apply to captured groups. If there are
any duplicate names, they must be in \code{\link{alternatives}} (only one match
per unique \code{\link{group}} name, otherwise error).}
}

\value{data.table with columns defined by calling the functions in
\code{fun.list} on the corresponding column of \code{match.mat}. Even if
\code{fun.list} has duplicated names, the output data.table will have
unique column names (identically named capture groups in
\code{\link{alternatives}} will be combined into a single output column).}

\author{Toby Hocking <toby.hocking@r-project.org> [aut, cre]}

\name{var_args_list}
\alias{var_args_list}
\title{var args list}
\description{Parse the variable-length argument list used in \code{\link{capture_first_vec}},
\code{\link{capture_first_df}}, and \code{\link{capture_all_str}}. This function is mostly
intended for internal use, but is useful if you want to see the
regex pattern generated by the variable argument syntax.}
\usage{var_args_list(..., type.convert = getOption("nc.type.convert", 
    FALSE))}
\arguments{
  \item{\dots}{character vectors (for regex patterns) or functions (which specify
how to convert extracted character vectors to other types). All
patterns must be character vectors of length 1. If the pattern is
a named argument in R, it becomes a capture \code{\link{group}} in the regex
pattern. All patterns are pasted together to obtain the final
pattern used for matching. Each named pattern may be followed by
at most one function which is used to convert the previous named
pattern. Patterns may also be lists, which are parsed recursively
for convenience.}
  \item{type.convert}{Default conversion function, which will be used on each capture
\code{\link{group}}, unless a specific conversion is specified for that
\code{\link{group}}. If TRUE, use \code{\link[utils]{type.convert}}; if FALSE, use
\code{\link[base]{identity}}; otherwise must be a function of at least one
argument (character), returning an atomic vector of the same
length.}
}

\value{a list with two named elements
\item{fun.list}{list of conversion functions with names corresponding to capture group(s)}
\item{pattern}{regular expression string with capture group(s)}}

\author{Toby Hocking <toby.hocking@r-project.org> [aut, cre]}




\examples{

pos.pattern <- list("[0-9]+", as.integer)
nc::var_args_list(
  chrom="chr.*?",
  ":",
  chromStart=pos.pattern,
  list(
    "-",
    chromEnd=pos.pattern
  ), "?")

}

\name{alternatives}
\alias{alternatives}
\title{alternatives}
\description{Make a pattern that matches one of the specified alternatives. The
\code{\link{altlist}} function can be helpful for defining named sub-patterns
that are used in several alternatives.}
\usage{alternatives(...)}
\arguments{
  \item{\dots}{Each argument is a different alternative pattern.}
}

\value{Pattern list.}

\author{Toby Hocking <toby.hocking@r-project.org> [aut, cre]}




\examples{

## simple example.
subject <- c("foooo1", "barrr2")
str(foo.or.bar <- nc::alternatives(bar="bar+", foo="fo+"))
str(foo.or.bar <- list(bar="bar+", "|", foo="fo+"))#same
nc::capture_first_vec(subject, foo.or.bar, number="[12]")

## More complicated regular expression for matching the JobID column
## of SLURM sacct output.
JobID <- c(
  "13937810_25", "13937810_25.batch",
  "13937810_25.extern", "14022192_[1-3]", "14022204_[4]")
int.pattern <- list("[0-9]+", as.integer)
## Match the whole range inside square brackets.
range.pattern <- list(
  "[[]",
  task.start=int.pattern,
  nc::quantifier("-", task.end=int.pattern, "?"),
  "[]]")
nc::capture_first_vec(JobID, range.pattern, nomatch.error=FALSE)

## Match either a single task ID or a range, after an underscore.
task.pattern <- list(job="[0-9]+", "_", nc::alternatives(
  task.id=int.pattern,
  range.pattern))
nc::capture_first_vec(JobID, task.pattern)

}

\name{melt_list}
\alias{melt_list}
\title{melt list}
\description{Compute a list of arguments to pass to
\code{\link[data.table]{melt.data.table}}.}
\usage{melt_list(measure.fun, 
    dot.args, ...)}
\arguments{
  \item{measure.fun}{\code{\link{measure_single}} or \code{\link{measure_multiple}}.}
  \item{dot.args}{list of arguments for \code{\link{check_df_names}}.}
  \item{\dots}{passed to \code{measure.fun}.}
}



\author{Toby Hocking <toby.hocking@r-project.org> [aut, cre]}

\name{subject_var_args}
\alias{subject_var_args}
\title{subject var args}
\description{Parse the complete argument list including subject.}
\usage{subject_var_args(..., 
    type.convert = getOption("nc.type.convert", 
        FALSE))}
\arguments{
  \item{\dots}{subject, regex/conversion.}
  \item{type.convert}{passed to \code{\link{var_args_list}}.}
}

\value{Result of \code{\link{var_args_list}} plus subject.}

\author{Toby Hocking <toby.hocking@r-project.org> [aut, cre]}

\name{capture_first_vec}
\alias{capture_first_vec}
\alias{nc}
\title{Capture first match in each character vector element}
\description{Use a regular expression (regex) with capture groups to extract
the first matching text from each of several subject strings. For
all matches in one multi-line text file or string use
\code{\link{capture_all_str}}. For the first match in every row of a data.frame,
using a different regex for each column, use \code{\link{capture_first_df}}. For
reading regularly named files, use \code{\link{capture_first_glob}}. For
matching column names in a wide data frame and then
melting/reshaping those columns to a taller/longer data frame, see
\code{\link{capture_melt_single}} and \code{\link{capture_melt_multiple}}. To simplify the
definition of the regex you can use \code{\link{field}}, \code{\link{quantifier}}, and
\code{\link{alternatives}}.}
\usage{capture_first_vec(..., 
    nomatch.error = getOption("nc.nomatch.error", 
        TRUE), engine = getOption("nc.engine", 
        "PCRE"), type.convert = getOption("nc.type.convert", 
        FALSE))}
\arguments{
  \item{\dots}{subject, name1=pattern1, fun1, etc. The first argument must be a
character vector of length>0 (subject strings to parse with a
regex). Arguments after the first specify the regex/conversion and
must be string/list/function. All character strings are pasted
together to obtain the final regex used for matching. Each
string/list with a named argument in R becomes a capture \code{\link{group}} in
the regex, and the name is used for the corresponding column of
the output data table. Each function must be un-named, and is used
to convert the previous capture \code{\link{group}}. Each un-named list becomes
a non-capturing \code{\link{group}}. Elements in each list are parsed
recursively using these rules.}
  \item{nomatch.error}{if TRUE (default), stop with an error if any subject does not
match; otherwise subjects that do not match are reported as
missing/NA rows of the result.}
  \item{engine}{character string, one of PCRE, ICU, RE2}
  \item{type.convert}{Default conversion function, which will be used on each capture
\code{\link{group}}, unless a specific conversion is specified for that
\code{\link{group}}. If TRUE, use \code{\link[utils]{type.convert}}; if FALSE, use
\code{\link[base]{identity}}; otherwise must be a function of at least one
argument (character), returning an atomic vector of the same
length.}
}

\value{data.table with one row for each subject, and one column for each
capture \code{\link{group}}.}

\author{Toby Hocking <toby.hocking@r-project.org> [aut, cre]}




\examples{

chr.pos.vec <- c(
  "chr10:213,054,000-213,055,000",
  "chrM:111,000",
  "chr1:110-111 chr2:220-222") # two possible matches.
## Find the first match in each element of the subject character
## vector. Named argument values are used to create capture groups
## in the generated regex, and argument names become column names in
## the result.
(dt.chr.cols <- nc::capture_first_vec(
  chr.pos.vec,
  chrom="chr.*?",
  ":",
  chromStart="[0-9,]+"))

## Even when no type conversion functions are specified, the result
## is always a data.table:
str(dt.chr.cols)

## Conversion functions are used to convert the previously named
## group, and patterns may be saved in lists for re-use.
keep.digits <- function(x)as.integer(gsub("[^0-9]", "", x))
int.pattern <- list("[0-9,]+", keep.digits)
range.pattern <- list(
  chrom="chr.*?",
  ":",
  chromStart=int.pattern,
  list( # un-named list becomes non-capturing group.
    "-",
    chromEnd=int.pattern
  ), "?") # chromEnd is optional.
(dt.int.cols <- nc::capture_first_vec(
  chr.pos.vec, range.pattern))

## Conversion functions used to create non-char columns.
str(dt.int.cols)

## NA used to indicate no match or missing subject.
na.vec <- c(
  "this will not match",
  NA, # neither will this.
  chr.pos.vec)
nc::capture_first_vec(na.vec, range.pattern, nomatch.error=FALSE)

## another subject from https://adventofcode.com/2024/day/14
## type.convert=TRUE means to use utils::type.convert as default
## conversion function
pvxy.subject <- c("p=0,4 v=3,-3","p=6,3 v=-1,-3")
nc::capture_first_vec(
  pvxy.subject,
  "p=",
  px="[0-9]",
  ",",
  py="[0-9]",
  " v=",
  vx="[-0-9]+",
  ",",
  vy="[-0-9]+",
  type.convert=TRUE)

## to do the same as above but with less repetition:
g <- function(prefix,suffix)nc::group(
  name=paste0(prefix,suffix),
  "[-0-9]+")
xy <- function(prefix)list(
  prefix,
  "=",
  g(prefix,"x"),
  ",",
  g(prefix,"y"))
nc::capture_first_vec(
  pvxy.subject,
  xy("p"),
  " ",
  xy("v"),
  type.convert=TRUE)

## or use a sub-pattern list without type.convert arg:
ipat <- list("[-0-9]+", as.integer)
nc::capture_first_vec(
  pvxy.subject,
  "p=",
  px=ipat,
  ",",
  py=ipat,
  " v=",
  vx=ipat,
  ",",
  vy=ipat)

}

\name{field}
\alias{field}
\title{Capture a field}
\description{Capture a field with a pattern of the form
list("field.name", \code{between.pattern},
field.name=list(\code{...})) -- see examples.}
\usage{field(field.name, between.pattern, 
    ...)}
\arguments{
  \item{field.name}{Field name, used as a pattern and as a capture \code{\link{group}} (output
column) name.}
  \item{between.pattern}{Pattern to match after \code{field.name} but before the field value.}
  \item{\dots}{Pattern(s) for matching field value.}
}

\value{Pattern list which can be used in \code{\link{capture_first_vec}},
\code{\link{capture_first_df}}, or \code{\link{capture_all_str}}.}

\author{Toby Hocking <toby.hocking@r-project.org> [aut, cre]}




\examples{

## Two ways to create the same pattern.
str(list("Alignment ", Alignment="[0-9]+"))
## To avoid typing Alignment twice use:
str(nc::field("Alignment", " ", "[0-9]+"))

## An example with lots of different fields.
info.txt.gz <- system.file(
  "extdata", "SweeD_Info.txt.gz", package="nc")
info.vec <- readLines(info.txt.gz)
info.vec[24:40]
## For each Alignment there are many fields which have a similar
## pattern, and occur in the same order. One way to capture these
## fields is by coding a pattern that says to look for all of those
## fields in that order. Each field is coded using this helper
## function.
g <- function(name, fun=identity, suffix=list()){
  list(
    "\\t+",
    nc::field(name, ":\\t+", ".*"),
    fun,
    suffix,
    "\\n+")
}
nc::capture_all_str(
  info.vec,
  nc::field("Alignment", " ", "[0-9]+"),
  "\\n+",
  g("Chromosome"),
  g("Sequences", as.integer),
  g("Sites", as.integer),
  g("Discarded sites", as.integer),
  g("Processing", as.integer, " seconds"),
  g("Position", as.integer),
  g("Likelihood", as.numeric),
  g("Alpha", as.numeric))

## Another example where field is useful.
trackDb.txt.gz <- system.file(
  "extdata", "trackDb.txt.gz", package="nc")
trackDb.vec <- readLines(trackDb.txt.gz)
cat(trackDb.vec[101:115], sep="\\n")
int.pattern <- list("[0-9]+", as.integer)
 cell.sample.type <- list(
  cellType="[^ ]*?",
  "_",
  sampleName=list(
    "McGill",
    sampleID=int.pattern),
  dataType="Coverage|Peaks")
## Each block in the trackDb file begins with track, followed by a
## space, followed by the track name. That pattern is coded below,
## using field:
track.pattern <- nc::field(
  "track",
  " ",
  cell.sample.type,
  "|",
  "[^\\n]+")
nc::capture_all_str(trackDb.vec, track.pattern)

## Each line in a block has the same structure (field name, space,
## field value). Below we use the field function to extract the
## color line, along with columns for each of the three channels
## (red, green, blue).
any.lines.pattern <- "(?:\\n[^\\n]+)*"
nc::capture_all_str(
  trackDb.vec,
  track.pattern,
  any.lines.pattern,
  "\\\\s+",
  nc::field(
    "color", " ",
    red=int.pattern, ",",
    green=int.pattern, ",",
    blue=int.pattern))

}

\name{capture_first_glob}
\alias{capture_first_glob}
\title{capture first glob}
\description{Glob files, then use \code{\link{capture_first_vec}} to get meta-data from each
file name, and combine with contents of each file.}
\usage{capture_first_glob(glob, 
    ..., READ = fread)}
\arguments{
  \item{glob}{string: \code{glob} specifying files to read.}
  \item{\dots}{passed to \code{\link{capture_first_vec}}, should include pattern used on vector
of file names to get meta-data.}
  \item{READ}{function of one argument (file name) which returns a data table,
default \code{\link[data.table]{fread}}.}
}

\value{Data table with columns of meta-data specified by pattern, plus
contents of all files specified by \code{glob}.}

\author{Toby Hocking <toby.hocking@r-project.org> [aut, cre]}




\examples{

data.table::setDTthreads(1)

## Example 0: iris data, one file per species.
library(data.table)
dir.create(iris.dir <- tempfile())
icsv <- function(sp)file.path(iris.dir, paste0(sp, ".csv"))
data.table(iris)[, fwrite(.SD, icsv(Species)), by=Species]
dir(iris.dir)
data.table::fread(file.path(iris.dir,"setosa.csv"), nrows=2)
(iglob <- file.path(iris.dir,"*.csv"))
nc::capture_first_glob(iglob, Species="[^/]+", "[.]csv")

## Example 1: four files, two capture groups, custom read function.
db <- system.file("extdata/chip-seq-chunk-db", package="nc", mustWork=TRUE)
suffix <- if(interactive())"gz" else "head"
(glob <- paste0(db, "/*/*/counts/*", suffix))
Sys.glob(glob)
read.bedGraph <- function(f)data.table::fread(
  f, skip=1, col.names = c("chrom","start", "end", "count"))
data.chunk.pattern <- list(
  data="H.*?",
  "/",
  chunk="[0-9]+", as.integer)
(data.chunk.dt <- nc::capture_first_glob(glob, data.chunk.pattern, READ=read.bedGraph))

## Write same data set in Hive partition, then re-read.
if(requireNamespace("arrow") && arrow::arrow_with_dataset()){
  path <- tempfile()
  max_rows_per_file <- if(interactive())3 else 1000
  arrow::write_dataset(
    dataset=data.chunk.dt,
    path=path,
    format="csv",
    partitioning=c("data","chunk"),
    max_rows_per_file=max_rows_per_file)
  hive.glob <- file.path(path, "*", "*", "*.csv")
  hive.pattern <- list(
    nc::field("data","=",".*?"),
    "/",
    nc::field("chunk","=",".*?", as.integer),
    "/",
    nc::field("part","-","[0-9]+", as.integer))
  hive.dt <- nc::capture_first_glob(hive.glob, hive.pattern)
  hive.dt[, .(rows=.N), by=.(data,chunk,part)]
}

## Example 2: more complex pattern.
count.dt <- nc::capture_first_glob(
  glob,
  data.chunk.pattern,
  "/counts/", 
  name=list("McGill", id="[0-9]+", as.integer),
  READ=read.bedGraph)
count.dt[, .(count=.N), by=.(data, chunk, name, chrom)]

if(require(ggplot2)){
  ggplot()+
    facet_wrap(~data+chunk+name+chrom, labeller=label_both, scales="free")+
    geom_step(aes(
      start, count),
      data=count.dt)
}

## Example 3: parsing non-CSV data.
vignettes <- system.file("extdata/vignettes", package="nc", mustWork=TRUE)
non.greedy.lines <- list(
  list(".*\\n"), "*?")
optional.name <- list(
  list(" ", chunk_name="[^,}]+"), "?")
chunk.pattern <- list(
  before=non.greedy.lines,
  "```\\\\{r",
  optional.name,
  parameters=".*",
  "\\\\}\\n",
  code=non.greedy.lines,
  "```")
chunk.dt <- nc::capture_first_glob(
  paste0(vignettes, "/*.Rmd"),
  "/v",
  vignette_number="[0-9]", as.integer,
  "-",
  vignette_name=".*?",
  ".Rmd",
  READ=function(f)nc::capture_all_str(f, chunk.pattern))
chunk.dt[, chunk_number := seq_along(chunk_name), by=vignette_number]
chunk.dt[, .(
  vignette_number, vignette_name, chunk_number, chunk_name, 
  lines=nchar(code))]
cat(chunk.dt$code[2])
  
}

\name{group}
\alias{group}
\title{Capture group}
\description{Create a capture group (named column in output). In the vast
majority of patterns R arguments can/should be used to specify
names, e.g. list(name=pattern). This is a helper function which is
useful for programmatically creating group names (see example for
a typical use case).}
\usage{group(name, ...)}
\arguments{
  \item{name}{Column \code{name} in output.}
  \item{\dots}{Regex pattern(s).}
}

\value{Named list.}

\author{Toby Hocking <toby.hocking@r-project.org> [aut, cre]}




\examples{

## Three ways to create a group named data which matches zero or
## more non-newline characters.
str(list(data=".*"))
str(nc::group("data", ".*"))
g.name <- "data"
str(nc::group(g.name, ".*"))

## Data downloaded from
## https://en.wikipedia.org/wiki/Hindu\%E2\%80\%93Arabic_numeral_system
numerals <- system.file(
  "extdata", "Hindu-Arabic-numerals.txt.gz", package="nc")

## Use engine="ICU" for unicode character classes
## http://userguide.icu-project.org/strings/regexp e.g. match any
## character with a numeric value of 2 (including japanese etc).
if(requireNamespace("stringi"))
  nc::capture_all_str(
  numerals,
  " ",
  two="[\\\\p{numeric_value=2}]",
  " ",
  engine="ICU")

## Create a table of numerals with script names.
digits.pattern <- list()
for(digit in 0:9){
  digits.pattern[[length(digits.pattern)+1]] <- list(
    "[|]",
    nc::group(paste(digit), "[^{|]+"),
    "[|]")
}
nc::capture_all_str(
  numerals,
  "\\n",
  digits.pattern,
  "[|]",
  " *",
  "\\\\[\\\\[",
  name="[^\\\\]|]+")

}

\name{capture_melt_single}
\alias{capture_melt_single}
\title{Capture and melt into a single column}
\description{Match a regex to column names of a wide data frame (many
columns/few rows), then melt/reshape the matching columns into a
single result column in a taller/longer data table (fewer columns/more
rows). It is for the common case of melting several columns of
the same type in a "wide" input data table which has several
distinct pieces of information encoded in each column name. For
melting into several result columns of possibly different types,
see \code{\link{capture_melt_multiple}}.}
\usage{capture_melt_single(..., 
    value.name = "value", 
    na.rm = TRUE, verbose = getOption("datatable.verbose"))}
\arguments{
  \item{\dots}{First argument must be a data frame to melt/reshape; column names
of this data frame will be used as the subjects for regex
matching. Other arguments (regex/conversion/engine) are passed to
\code{\link{capture_first_vec}} along with nomatch.error=FALSE.}
  \item{value.name}{Name of the column in output which has values taken from
melted/reshaped column values of input (passed to
\code{\link[data.table]{melt.data.table}}).}
  \item{na.rm}{remove missing values from melted data? (passed to
\code{\link[data.table]{melt.data.table}})}
  \item{verbose}{Print \code{verbose} output messages? (passed to
\code{\link[data.table]{melt.data.table}})}
}

\value{Data table of reshaped/melted/tall/long data, with a new column
for each named argument in the pattern, and additionally
variable/value columns.}

\author{Toby Hocking <toby.hocking@r-project.org> [aut, cre]}




\examples{

data.table::setDTthreads(1)

## Example 1: melt iris data and barplot for each numeric variable.
(iris.tall <- nc::capture_melt_single(
  iris,
  part=".*",
  "[.]",
  dim=".*",
  value.name="cm"))
## Histogram of cm for each variable.
if(require("ggplot2")){
  ggplot()+
    theme_bw()+
    theme(panel.spacing=grid::unit(0, "lines"))+
    facet_grid(part ~ dim)+
    geom_bar(aes(cm), data=iris.tall)
}

## Example 2: melt who data and use type conversion functions for
## year limits (e.g. for censored regression).
if(requireNamespace("tidyr")){
  data(who, package="tidyr", envir=environment())
  ##2.1 just extract diagnosis and gender to chr columns.
  new.diag.gender <- list(#save pattern as list for re-use later.
    "new_?",
    diagnosis=".*",
    "_",
    gender=".")
  who.tall.chr <- nc::capture_melt_single(who, new.diag.gender, na.rm=TRUE)
  print(head(who.tall.chr))
  str(who.tall.chr)
  ##2.2 also extract ages and convert to numeric output columns.
  who.tall.num <- nc::capture_melt_single(
    who,
    new.diag.gender,#previous pattern for matching diagnosis and gender.
    ages=list(#new pattern for matching age range.
      min.years="0|[0-9]{2}", as.numeric,#in-line type conversion functions.
      max.years="[0-9]{0,2}", function(x)ifelse(x=="", Inf, as.numeric(x))),
    value.name="count",
    na.rm=TRUE)
  print(head(who.tall.num))
  str(who.tall.num)
  ##2.3 compute total count for each age range then display the
  ##subset with max.years lower than a threshold.
  who.age.counts <- who.tall.num[, .(
    total=sum(count)
  ), by=.(min.years, max.years)]
  print(who.age.counts[max.years < 50])
}

## Example 3: pepseq data.
if(requireNamespace("R.utils")){#for reading gz files with data.table
  pepseq.dt <- data.table::fread(
    system.file("extdata", "pepseq.txt", package="nc", mustWork=TRUE))
  u.pepseq <- pepseq.dt[, unique(names(pepseq.dt)), with=FALSE]
  nc::capture_melt_single(
    u.pepseq,
    "^",
    prefix=".*?",
    nc::field("D", "", ".*?"),
    "[.]",
    middle=".*?",
    "[.]",
    "[0-9]+",
    suffix=".*",
    "$")
}

}

\name{alternatives_with_shared_groups}
\alias{alternatives_with_shared_groups}
\title{alternatives with shared groups}
\description{Create a pattern which matches \code{\link{alternatives}} with common
sub-pattern groups.}
\usage{alternatives_with_shared_groups(...)}
\arguments{
  \item{\dots}{named arguments are sub-pattern groups, un-named arguments are
alternative patterns which can refer to argument names.}
}

\value{Pattern created by first using \code{\link{altlist}} on named arguments to
create a list of sub-patterns, then using \code{\link{alternatives}} on un-named
arguments (evaluated using the names defined in the sub-pattern
list).}

\author{Toby Hocking <toby.hocking@r-project.org> [aut, cre]}




\examples{

## Example 0: matching family and given names.
nc::capture_first_vec(
  c("Toby Dylan Hocking","Hocking, Toby Dylan"),
  nc::alternatives_with_shared_groups(
    family="[A-Z][a-z]+",
    given="[^,]+",
    list(given, " ", family),
    list(family, ", ", given)
  )
)

## Example 1: matching dates in different formats, but always same
## type in each alternative.
subject.vec <- c("mar 17, 1983", "26 sep 2017", "17 mar 1984")
## Another way is with alternative regular expressions.
pattern <- nc::alternatives_with_shared_groups(
  month="[a-z]{3}",
  day="[0-9]{2}",
  year="[0-9]{4}",
  list(american=list(month, " ", day, ", ", year)),
  list(european=list(day, " ", month, " ", year)))
(match.dt <- nc::capture_first_vec(subject.vec, pattern))
match.dt[, data.table::as.IDate(paste0(month, day, year), format="\%b\%d\%Y")]
## american and european columns can be searched to see which
## alternative matched.
match.dt[european!=""]

## Example 2: matching dates in different formats, but with
## different types in different alternatives.
subject.vec <- c("3/17/1983", "26 sep 2017")
month2int <- c(#this approach is locale-indepdendent.
  jan=1L, feb=2L, mar=3L, apr=4L,  may=5L,  jun=6L,
  jul=7L, aug=8L, sep=9L, oct=10L, nov=11L, dec=12L)
pattern <- nc::alternatives_with_shared_groups(
  day=list("[0-9]{2}", as.integer),
  year=list("[0-9]{4}", as.integer),
  list(month="[0-9]", as.integer, "/", day, "/", year),
  list(day, " ", month="[a-z]{3}", function(m)month2int[m], " ", year))
match.dt <- nc::capture_first_vec(subject.vec, pattern)
print(match.dt, class=TRUE)

## Example 3: three alternatives with four groups each.
subject.vec <- c(
  "EUR 5.00 Theft in delivery inserted in wire transfer 11/02/2021",
  "EUR 50.00 - Refund for theft in delivery - 30/07/2020",
  "EUR68.50 - Refund for theft in delivery 02/07/2020",
  "45.00 EUR 29/10/2020 Refund for theft in delivery",
  "53.00\\u20ac Refund for theft in delivery 24/09/2020")
sep <- function(x, y, ...){
  if(missing(y)){
    list("^", x, "$")
  }else{
    sep(list(x, list(" - | |"), y), ...)
  }
}
pattern <- nc::alternatives_with_shared_groups(
  currency="EUR|\\u20ac",
  amount=list("[0-9.]+", as.numeric),
  reason="[A-Za-z ]+?",
  date=list(
    "[0-9]{2}/[0-9]{2}/[0-9]{4}",
    function(d)data.table::as.IDate(d, format="\%d/\%m/\%Y")),
  sep(currency, amount, reason, date),
  sep(amount, currency, date, reason),
  sep(amount, currency, reason, date))
match.dt <- nc::capture_first_vec(subject.vec, pattern)
print(match.dt, class=TRUE)

}

\name{only_captures}
\alias{only_captures}
\title{only captures}
\description{Extract capture \code{\link{group}} columns from \code{match.mat} and assign optional
groups to "".}
\usage{only_captures(match.mat, 
    stop.fun)}
\arguments{
  \item{match.mat}{character matrix.}
  \item{stop.fun}{function to call on logical matrix of missing indicators.}
}



\author{Toby Hocking <toby.hocking@r-project.org> [aut, cre]}

\name{measure_single}
\alias{measure_single}
\title{measure single}
\description{Compute a measure.vars vector (indicating a single output column)
with variable_table attribute to pass to
\code{\link[data.table]{melt.data.table}}.}
\usage{measure_single(subject.names, 
    match.dt, no.match, 
    value.name = NULL)}
\arguments{
  \item{subject.names}{character vector of data frame column names.}
  \item{match.dt}{data table of matches.}
  \item{no.match}{logical vector.}
  \item{value.name}{string.}
}



\author{Toby Hocking <toby.hocking@r-project.org> [aut, cre]}

\name{capture_all_str}
\alias{capture_all_str}
\title{Capture all matches in a single subject string}
\description{Capture each match of a regex pattern from one multi-line subject
string or text file. It can be used to convert any regular text
file (web page, log, etc) to a data table, see examples.}
\usage{capture_all_str(..., 
    engine = getOption("nc.engine", 
        "PCRE"), collapse = "\\n", 
    type.convert = getOption("nc.type.convert", 
        FALSE))}
\arguments{
  \item{\dots}{subject, name1=pattern1, fun1, etc. The first argument must be a
subject character vector (or file name which is read via
\code{\link[base]{readLines}} to get a subject). After removing missing values,
we use \code{\link[base]{paste}} to \code{collapse} the subject (by default using
newline) and treat it as single character string to
search. Arguments after the first specify the regex/conversion and
must be string/function/list, as documented in \code{\link{capture_first_vec}}.}
  \item{engine}{character string, one of PCRE, ICU, RE2}
  \item{collapse}{separator string for combining elements of subject into a single
string, used as \code{collapse} argument of \code{\link[base]{paste}}.}
  \item{type.convert}{Default conversion function, which will be used on each capture
\code{\link{group}}, unless a specific conversion is specified for that
\code{\link{group}}. If TRUE, use \code{\link[utils]{type.convert}}; if FALSE, use
\code{\link[base]{identity}}; otherwise must be a function of at least one
argument (character), returning an atomic vector of the same
length.}
}

\value{data.table with one row for each match, and one column for each
capture \code{\link{group}}.}

\author{Toby Hocking <toby.hocking@r-project.org> [aut, cre]}




\examples{

data.table::setDTthreads(1)

chr.pos.vec <- c(
  "chr10:213,054,000-213,055,000",
  "chrM:111,000-222,000",
  "this will not match",
  NA, # neither will this.
  "chr1:110-111 chr2:220-222") # two possible matches.
keep.digits <- function(x)as.integer(gsub("[^0-9]", "", x))
## By default elements of subject are treated as separate lines (and
## NAs are removed). Named arguments are used to create capture
## groups, and conversion functions such as keep.digits are used to
## convert the previously named group.
int.pattern <- list("[0-9,]+", keep.digits)
(match.dt <- nc::capture_all_str(
  chr.pos.vec,
  chrom="chr.*?",
  ":",
  chromStart=int.pattern,
  "-",
  chromEnd=int.pattern))
str(match.dt)

## Extract all fields from each alignment block, using two regex
## patterns, then dcast.
info.txt.gz <- system.file(
  "extdata", "SweeD_Info.txt.gz", package="nc")
info.vec <- readLines(info.txt.gz)
info.vec[24:40]
info.dt <- nc::capture_all_str(
  sub("Alignment ", "//", info.vec),
  "//",
  alignment="[0-9]+",
  fields="[^/]+")
(fields.dt <- info.dt[, nc::capture_all_str(
  fields,
  "\\t+",
  variable="[^:]+",
  ":\\t*",
  value=".*"),
  by=alignment])
(fields.wide <- data.table::dcast(fields.dt, alignment ~ variable))

## Capture all csv tables in report -- the file name can be given as
## the subject to nc::capture_all_str, which calls readLines to get
## data to parse.
(report.txt.gz <- system.file(
  "extdata", "SweeD_Report.txt.gz", package="nc"))
(report.dt <- nc::capture_all_str(
  report.txt.gz,
  "//",
  alignment="[0-9]+",
  "\\n",
  csv="[^/]+"
)[, {
  data.table::fread(text=csv)
}, by=alignment])

## Join report with info fields.
report.dt[fields.wide, on=.(alignment)]

## parsing nbib citation file.
(pmc.nbib <- system.file(
  "extdata", "PMC3045577.nbib", package="nc"))
blank <- "\\n      "
pmc.dt <- nc::capture_all_str(
  pmc.nbib,
  Abbreviation="[A-Z]+",
  " *- ",
  value=list(
    ".*",
    list(blank, ".*"), "*"),
  function(x)sub(blank, "", x))
str(pmc.dt)

## What do the variable fields mean? It is explained on
## https://www.nlm.nih.gov/bsd/mms/medlineelements.html which has a
## local copy in this package (downloaded 18 Sep 2019).
fields.html <- system.file(
  "extdata", "MEDLINE_Fields.html", package="nc")
if(interactive())browseURL(fields.html)
fields.vec <- readLines(fields.html)

## It is pretty easy to capture fields and abbreviations if gsub
## used to remove some tags first.
no.strong <- gsub("</?strong>", "", fields.vec)
no.comments <- gsub("<!--.*?-->", "", no.strong)
## grep then capture_first_vec can be used if each desired row in
## the output comes from a single line of the input file.
(h3.vec <- grep("<h3", no.comments, value=TRUE))
h3.pattern <- list(
  nc::field("name", '="', '[^"]+'),
  '"></a>',
  fields.abbrevs="[^<]+")
first.fields.dt <- nc::capture_first_vec(
  h3.vec, h3.pattern)
field.abbrev.pattern <- list(
  Field=".*?",
  " \\\\(",
  Abbreviation="[^)]+",
  "\\\\)",
  "(?: and |$)?")
(first.each.field <- first.fields.dt[, nc::capture_all_str(
  fields.abbrevs, field.abbrev.pattern),
  by=fields.abbrevs])

## If we want to capture the information after the initial h3 line
## of the input, e.g. the rest column below which contains a
## description/example for each field, then capture_all_str can be
## used on the full input file.
h3.fields.dt <- nc::capture_all_str(
  no.comments,
  h3.pattern,
  '</h3>\\n',
  rest="(?:.*\\n)+?", #exercise: get the examples.
  "<hr />\\n")
(h3.each.field <- h3.fields.dt[, nc::capture_all_str(
  fields.abbrevs, field.abbrev.pattern),
  by=fields.abbrevs])

## Either method of capturing abbreviations gives the same result.
identical(first.each.field, h3.each.field)

## but the capture_all_str method returns the additional rest column
## which contains data after the initial h3 line.
names(first.fields.dt)
names(h3.fields.dt)
cat(h3.fields.dt[fields.abbrevs=="Volume (VI)", rest])

## There are 66 Field rows across three tables.
a.href <- list('<a href=[^>]+>')
(td.vec <- fields.vec[240:280])
fields.pattern <- list(
  "<td.*?>",
  a.href,
  Fields="[^()<]+",
  "</a></td>\\n")
(td.only.Fields <- nc::capture_all_str(fields.vec, fields.pattern))

## Extract Fields and Abbreviations. Careful: most fields have one
## abbreviation, but one field has none, and two fields have two.
(td.fields.dt <- nc::capture_all_str(
  fields.vec,
  fields.pattern,
  "<td[^>]*>",
  "(?:\\n<div>)?",
  a.href, "?",
  abbrevs=".*?",
  "</"))

## Get each individual abbreviation from the previously captured td
## data.
td.each.field <- td.fields.dt[, {
  f <- nc::capture_all_str(
    Fields,
    Field=".*?",
    "(?:$| and )")
  a <- nc::capture_all_str(
    abbrevs,
    "\\\\(",
    Abbreviation="[^)]+",
    "\\\\)")
  if(nrow(a)==0)list() else cbind(f, a)
}, by=Fields]
str(td.each.field)
td.each.field[td.fields.dt, .(
  count=.N
), on=.(Fields), by=.EACHI][order(count)]

## There is a typo in the data captured from the h3 headings.
td.each.field[!Field \%in\% h3.each.field$Field]
h3.each.field[!Field \%in\% td.each.field$Field]

## Abbreviations are consistent.
td.each.field[!Abbreviation \%in\% h3.each.field$Abbreviation]
h3.each.field[!Abbreviation \%in\% td.each.field$Abbreviation]

## There is a a table that provides a description of each comment
## type.
(comment.vec <- fields.vec[840:860])
comment.dt <- nc::capture_all_str(
  fields.vec,
  "<td><strong>",
  Field="[^<]+",
  "</strong></td>\\n",
  "<td><strong>\\\\(",
  Abbreviation="[^)]+",
  "\\\\)</strong></td>\\n",
  "<td>",
  description=".*",
  "</td>\\n")
str(comment.dt)

## Join to original PMC citation file in order to see what the
## abbreviations used in that file mean.
all.abbrevs <- rbind(
  td.each.field[, .(Field, Abbreviation)],
  comment.dt[, .(Field, Abbreviation)])
all.abbrevs[pmc.dt, .(
  Abbreviation,
  Field,
  value=substr(value, 1, 20)
), on=.(Abbreviation)]

## There is a listing of examples for each comment type.
(comment.ex.dt <- nc::capture_all_str(
  fields.vec[938],
  "br />\\\\s*",
  Abbreviation="[A-Z]+",
  "\\\\s*-\\\\s*",
  citation="[^<]+?",
  list(
    "[.] ",
    nc::field("PMID", ": ", "[0-9]+")
  ), "?",
  "<"))

## Join abbreviations to see what kind of comments.
all.abbrevs[comment.ex.dt, on=.(Abbreviation)]

## parsing bibtex file.
refs.bib <- system.file(
  "extdata", "namedCapture-refs.bib", package="nc")
refs.vec <- readLines(refs.bib)
at.lines <- grep("@", refs.vec, value=TRUE)
str(at.lines)
refs.dt <- nc::capture_all_str(
  refs.vec,
  "@",
  type="[^{]+",
  "[{]",
  ref="[^,]+",
  ",\\n",
  fields="(?:.*\\n)+?.*",
  "[}]\\\\s*(?:$|\\n)")
str(refs.dt)

## parsing each field of each entry.
eq.lines <- grep("=", refs.vec, value=TRUE)
str(eq.lines)
strip <- function(x)sub("^\\\\s*\\\\{*", "", sub("\\\\}*,?$", "", x))
refs.fields <- refs.dt[, nc::capture_all_str(
  fields,
  "\\\\s+",
  variable="\\\\S+",
  "\\\\s+=",
  value=".*", strip),
  by=.(type, ref)]
str(refs.fields)
with(refs.fields[ref=="HockingUseR2011"], structure(
  as.list(value), names=variable))
## the URL of my talk is now
## https://user2011.r-project.org/TalkSlides/Lightening/2-StatisticsAndProg_3-Hocking.pdf

if(!grepl("solaris", R.version$platform)){#To avoid CRAN check error on solaris
  ## Parsing wikimedia tables: each begins with {| and ends with |}.
  emoji.txt.gz <- system.file(
    "extdata", "wikipedia-emoji-text.txt.gz", package="nc")
  tables <- nc::capture_all_str(
    emoji.txt.gz,
    "\\n[{][|]",
    first=".*",
    '\\n[|][+] style="',
    nc::field("font-size", ":", '.*?'),
    '" [|] ',
    title=".*",
    lines="(?:\\n.*)*?",
    "\\n[|][}]")
  str(tables)
  ## Rows are separated by |-
  rows.dt <- tables[, {
    row.vec <- strsplit(lines, "|-", fixed=TRUE)[[1]][-1]
    .(row.i=seq_along(row.vec), row=row.vec)
  }, by=title]
  str(rows.dt)
  ## Try to parse columns from each row. Doesn't work for second table
  ## https://en.wikipedia.org/w/index.php?title=Emoji&oldid=920745513#Skin_color
  ## because some entries have rowspan=2.
  contents.dt <- rows.dt[, nc::capture_all_str(
    row,
    "[|] ",
    content=".*?",
    "(?: [|]|\\n|$)"),
    by=.(title, row.i)]
  contents.dt[, .(cols=.N), by=.(title, row.i)]
  ## Make data table from
  ## https://en.wikipedia.org/w/index.php?title=Emoji&oldid=920745513#Emoji_versus_text_presentation
  contents.dt[, col.i := 1:.N, by=.(title, row.i)]
  data.table::dcast(
    contents.dt[title=="Sample emoji variation sequences"],
    row.i ~ col.i,
    value.var="content")
}

## Simple way to extract code chunks from Rmd.
vignette.Rmd <- system.file(
  "extdata", "vignette.Rmd", package="nc")
non.greedy.lines <- list(
  list(".*\\n"), "*?")
optional.name <- list(
  list(" ", name="[^,}]+"), "?")
Rmd.dt <- nc::capture_all_str(
  vignette.Rmd,
  before=non.greedy.lines,
  "```\\\\{r",
  optional.name,
  parameters=".*",
  "\\\\}\\n",
  code=non.greedy.lines,
  "```")
Rmd.dt[, chunk := 1:.N]
Rmd.dt[, .(chunk, name, parameters, some.code=substr(code, 1, 20))]

## Extract individual parameter names and values.
Rmd.dt[, nc::capture_all_str(
  parameters,
  ", *",
  variable="[^= ]+",
  " *= *",
  value="[^ ,]+"),
  by=chunk]

## Simple way to extract code chunks from Rnw.
vignette.Rnw <- system.file(
  "extdata", "vignette.Rnw", package="nc")
Rnw.dt <- nc::capture_all_str(
  vignette.Rnw,
  before=non.greedy.lines,
  "<<",
  name="[^,>]*",
  parameters=".*",
  ">>=\\n",
  code=non.greedy.lines,
  "@")
Rnw.dt[, .(name, parameters, some.code=substr(code, 1, 20))]

## The next example involves timing some compression programs that
## were run on a 159 megabyte input/uncompressed text file. Here is
## how to get a data table from the time command line output.
times.out <- system.file(
  "extdata", "compress-times.out", package="nc", mustWork=TRUE)
times.dt <- nc::capture_all_str(
  times.out,
  "coverage.bedGraph ",
  program=".*?",
  " coverage.bedGraph.",
  suffix=".*",
  "\\n\\nreal\\t",
  minutes.only="[0-9]+", as.numeric,
  "m",
  seconds.only="[0-9.]+", as.numeric)
times.dt[, seconds := minutes.only*60+seconds.only]
times.dt

## join with output from du command line program.
sizes.out <- system.file(
  "extdata", "compress-sizes.out", package="nc", mustWork=TRUE)
sizes.dt <- data.table::fread(
  file=sizes.out,
  col.names=c("megabytes", "file"))
sizes.dt[, suffix := sub("coverage.bedGraph.?", "", file)]
join.dt <- times.dt[sizes.dt, on="suffix"][order(megabytes)]
join.dt[file=="coverage.bedGraph", seconds := 0]
join.dt

## visualize with ggplot2.
if(require(ggplot2)){
  ggplot(join.dt, aes(
    seconds, megabytes, label=suffix))+
    geom_text(vjust=-0.5)+
    geom_point()+
    scale_x_log10()+
    scale_y_log10()
}

}

\name{measure}
\alias{measure}
\title{measure}
\description{Computes a value to be used as measure.vars argument to
\code{\link[data.table]{melt.data.table}}. NOTE: only works on newer versions of
data.table that include the measure function.}
\usage{measure(..., cols)}
\arguments{
  \item{\dots}{Regular expression pattern list, passed to \code{\link{capture_first_vec}} with
\code{cols} as subject.}
  \item{cols}{Character vector, column names to match with regex.}
}
\details{\code{\link{measure_multiple}} is called if there is a capture \code{\link{group}}
named "column" and \code{\link{measure_single}} is called otherwise.}
\value{List or vector to use as measure.vars argument to
\code{\link[data.table]{melt.data.table}}.}

\author{Toby Hocking <toby.hocking@r-project.org> [aut, cre]}




\examples{

library(data.table)
iris.dt <- data.table(datasets::iris[c(1,150),])
melt(iris.dt, measure=nc::measure(part  =".*", "[.]", dim   =".*"))
melt(iris.dt, measure=nc::measure(column=".*", "[.]", dim   =".*"))
melt(iris.dt, measure=nc::measure(part  =".*", "[.]", column=".*"))

}

\name{stop_for_subject}
\alias{stop_for_subject}
\title{stop for subject}
\description{Error if \code{subject} incorrect type.}
\usage{stop_for_subject(subject)}
\arguments{
  \item{subject}{character vector.}
}



\author{Toby Hocking <toby.hocking@r-project.org> [aut, cre]}

\name{collapse_some}
\alias{collapse_some}
\title{collapse some}
\description{Create character string with some or all items.}
\usage{collapse_some(all.vec, 
    max.first.last = 5, 
    collapse = ",")}
\arguments{
  \item{all.vec}{Vector of all items.}
  \item{max.first.last}{Max number of items to show.}
  \item{collapse}{Passed to paste.}
}

\value{Character string formed by paste with \code{collapse} on some items of
\code{all.vec} (first/last few items used if length is greater than
max.first.last*2, otherwise all items).}

\author{Toby Hocking <toby.hocking@r-project.org> [aut, cre]}

\name{alevels}
\alias{alevels}
\title{Alternative levels}
\description{Create a pattern and a conversion function based on alternative
string literals.}
\usage{alevels(...)}
\arguments{
  \item{\dots}{Optional names are string literals to match; values are
corresponding factor levels.}
}

\value{List of pattern and conversion function that returns factor.}

\author{Toby Hocking <toby.hocking@r-project.org> [aut, cre]}




\examples{

## Example 0: melt iris data with literal alternatives -> chr columns.
ichr <- nc::capture_melt_single(
  iris[1,],
  part="Sepal|Petal",
  "[.]",
  dim="Length|Width")
factor(ichr$part)#default factor levels are alphabetical.

## Example 1: melt iris data with alevels() -> factor columns.
(ifac <- nc::capture_melt_single(
  iris[1,],
  part=nc::alevels("Sepal","Petal"),
  "[.]",
  dim=nc::alevels("Length","Width")))
ifac$part #factor with levels in same order as given in alevels().

## Example 2: alevels(literals_to_match="levels_to_use_in_output").
tv_wide <- data.frame(
  id=0,
  train.classif.logloss = 1, train.classif.ce = 2, 
  valid.classif.logloss = 3, valid.classif.ce = 4)
nc::capture_melt_single(
  tv_wide, 
  set=nc::alevels(valid="validation", train="subtrain"),
  "[.]classif[.]",
  measure=nc::alevels(ce="error_prop", auc="AUC", "logloss"))

## Example 3: additional groups which output character columns.
nc::capture_melt_single(
  tv_wide, 
  set_chr=list(set_fac=nc::alevels(valid="validation", train="subtrain")),
  "[.]classif[.]",
  measure_chr=list(measure_fac=nc::alevels(ce="error_prop", auc="AUC", "logloss")))

}

\name{altlist}
\alias{altlist}
\title{altlist}
\description{Create a named list containing named patterns, useful for creating
a named list of named sub-pattern groups to be used with
\code{\link{alternatives}}. This function is used to implement the more
user-friendly \code{\link{alternatives_with_shared_groups}}, which should be
preferred.}
\usage{altlist(...)}
\arguments{
  \item{\dots}{Named patterns which will be used as sub-pattern groups in \code{\link{alternatives}}.}
}

\value{Named list of patterns to be used for constructing \code{\link{alternatives}}
using \code{\link[base]{with}}, see examples.}

\author{Toby Hocking <toby.hocking@r-project.org> [aut, cre]}




\examples{

## Example 1: matching dates in different formats, but always same
## type in each alternative.
subject.vec <- c("mar 17, 1983", "26 sep 2017", "17 mar 1984")
pat.list <- nc::altlist(month="[a-z]{3}", day="[0-9]{2}", year="[0-9]{4}")
pattern <- with(pat.list, nc::alternatives(
  american=list(month, " ", day, ", ", year),
  european=list(day, " ", month, " ", year)))
match.dt <- nc::capture_first_vec(subject.vec, pattern)
print(match.dt, class=TRUE)
match.dt[, data.table::as.IDate(paste0(month, day, year), format="\%b\%d\%Y")]

## Example 2: matching dates in different formats, but with
## different types in different alternatives.
subject.vec <- c("3/17/1983", "26 sep 2017")
month2int <- c(#this approach is locale-indepdendent.
  jan=1L, feb=2L, mar=3L, apr=4L,  may=5L,  jun=6L,
  jul=7L, aug=8L, sep=9L, oct=10L, nov=11L, dec=12L)
pat.list <- nc::altlist(
  day=list("[0-9]{2}", as.integer),
  year=list("[0-9]{4}", as.integer))
pattern <- with(pat.list, nc::alternatives(
  american=list(month="[0-9]", as.integer, "/", day, "/", year),
  european=list(
    day, " ", month="[a-z]{3}", function(m)month2int[m], " ", year)))
match.dt <- nc::capture_first_vec(subject.vec, pattern)
print(match.dt, class=TRUE)

}

\name{before_match}
\alias{before_match}
\title{before match}
\description{Augment pattern so that it can be used to match an entire string. Useful for complex find and replace operations, when used with \code{\link{capture_all_str}}.}
\usage{before_match(...)}
\arguments{
  \item{\dots}{pattern as in \code{\link{capture_first_vec}}.}
}

\value{Pattern with two new groups, before (everything before \code{...}) and match (text matching \code{...} or blank for end of string).}

\author{Toby Hocking <toby.hocking@r-project.org> [aut, cre]}




\examples{

markdown_link <- list(
  "\\\\[",
  title=".*?",
  "\\\\]\\\\(",
  url="http.*?",
  "\\\\)")
markdown_subject <- "before [foo](http) between [bar text](http) after\\n"
nc::capture_all_str(markdown_subject, markdown_link)

before_link <- nc::before_match(markdown_link)
(all_dt <- nc::capture_all_str(markdown_subject, before_link))

## before + match = full subject.
identical(all_dt[, paste(paste0(before, match), collapse="")], markdown_subject)

## replace with org link.
all_dt[, paste(paste0(before, ifelse(
  match=="", "", sprintf("[[\%s][\%s]]", url, title)
)), collapse="")]

## also works with no extra text before/after match.
nc::capture_all_str("[foo](http) between [bar text](http)", before_link)

}

\name{stop_for_capture_same_as_id}
\alias{stop_for_capture_same_as_id}
\title{stop for capture same as id}
\description{Error if capture names same as \code{id.vars}.}
\usage{stop_for_capture_same_as_id(capture.vars, 
    id.vars)}
\arguments{
  \item{capture.vars}{character vector of capture column names.}
  \item{id.vars}{character vector of id column names.}
}



\author{Toby Hocking <toby.hocking@r-project.org> [aut, cre]}
